
'SuperStrict
'Import "tool.bmx"

Private

'TODO: this should realy be taken from settings, or from the current gui font.
?Win32
Const EDITOR_FONT_HEIGHT:Int = 16
Const ITEM_FONT_HEIGHT:Int = 14
?Linux
Const EDITOR_FONT_HEIGHT:Int = 16
Const ITEM_FONT_HEIGHT:Int = 14
?MacOS
Const EDITOR_FONT_HEIGHT:Int = 16
Const ITEM_FONT_HEIGHT:Int = 14
?




' marks selected item in autocomplete window, since the window is inactive the colors can be quite dim
Incbin "resources/autocompleteiconssmall.png"
Global AutoCompleteIconStrip:TIconStrip = LoadIconStrip( "incbin::resources/autocompleteiconssmall.png")

Public


Type TGUIAutoComplete Extends TGUIContainer
	Field _owner:TCodePanel
	Field _editor:TGadget
	Field _list:TGadget
	Field _hint:TGUIHintWindow
	Field IsOpen:Int
	Field FirstSpace:Int
	Field LastKeyword:String
	Field CurrentLine:Int
	Field AutocVal:String[], AutocInt:Int
	Field AutocTip:String[]
	Field options:TAutoCompleteOptions
	Field CaseSensitive:Int	
	Field AllowOneCharLookup:Int ' for indexing by prefix (might be slow?)

	Method Init( owner:TCodePanel, options:TAutoCompleteOptions)
		?Debug
			If Not owner Then Throw "TGUIAutoComplete.Init() needs a valid TCodePanel"
			If Not options Then Throw "TGUIAutoComplete.Init() needs valid TAutoCompleteOptions"
		?
		Self.options = options
		_owner = owner
		_editor = owner.Gadget()
		_gadget = CreateWindow( "AutoComplete", 0,0, options.winsize.w,options.winsize.h, _owner.host._gadget, WINDOW_RESIZABLE | WINDOW_HIDDEN)
		_list = CreateListBox( 0,0, ClientWidth(_gadget),ClientHeight(_gadget), _gadget)
		SetGadgetLayout _list, 1,1,1,1
		SetGadgetIconStrip( _list, AutoCompleteIconStrip)
		_hint = New TGUIHintWindow
		_hint.Init( Self)
	EndMethod

	Method Free() 
		FreeGadget _gadget
		_gadget = Null
		_hint.Free() 
		_owner = Null
		options = Null		
	EndMethod

	Method Poll(e:TEvent) 
		If Not IsOpen Then Return	
		If e.source = _list Then
			Select e.id
				' autocomplete on clicking an item
				Case EVENT_GADGETACTION , EVENT_GADGETSELECT				
					AutoComplete()
					Return
			EndSelect
		ElseIf e.source = _gadget Then
			' set options on window resize
			If e.ID = EVENT_WINDOWSIZE Then
				'FIXME: make the gadget size after its item height (its 14 on windows because of the icon)
				'Local h:Int = ??
				'SetGadgetShape GadgetX(_gadget),GadgetY(_gadget),GadgetWidth(_gadget), h
				options.SetWindowSize( GadgetWidth(_gadget), GadgetHeight(_gadget))
				Return
			EndIf
		Else
			' close window if any other gadget is active
			Select ActiveGadget()
				Case _editor, _gadget, _list
					' do nothing
				Default
					Hide()
					Return
			EndSelect
			' close window if the line has changed
			If _owner.cursorline <> CurrentLine Then
				Hide()
				Return
			EndIf
		EndIf
		PopulateList()
	EndMethod

	Method Show()
		If IsOpen Then Return
		FirstSpace = False
		LastKeyword = Null
		If Not PopulateList() Then Return
		IsOpen = True
		CurrentLine = _owner.cursorline
		Local pt:Int[2]
		?Win32
			Local rect:Int[4]
			GetWindowRect( _editor.Query(QUERY_HWND), rect)
			GetCaretPos_( pt)
			pt[0] :+ rect[0] + 2
			pt[1] :+ rect[1] + EDITOR_FONT_HEIGHT
		?
		?Not win32
			'TODO: add proper popup location for Linux & MacOSX
			' center the window for now
			pt[0] = (_owner.host._gadget.xpos / 2) + (_owner.host._gadget.width / 2)
			pt[1] = ((_owner.host._gadget.ypos / 2) + (_owner.host._gadget.height / 2)) + 128
		?
		SetGadgetShape( _gadget, pt[0],pt[1], _gadget.Width,_gadget.Height)
		' reselect to show hintbox properly
		SelectItem( SelectedGadgetItem(_list))
		ShowGadget _gadget
		ActivateGadget _editor
	EndMethod

	Method Hide()
		IsOpen = False
		_hint.Hide()
		HideGadget _gadget
		ActivateGadget _editor
	EndMethod

	Method SelectItem( index:Int)
		Function UnselectCurrentItem( gadget:TGadget)
			Local index:Int = SelectedGadgetItem( gadget)
			If (index >= 0) And (index < CountGadgetItems(gadget)) Then
				Local text:String = GadgetItemText( gadget, index)
				Local flags:Int = GadgetItemFlags( gadget, index)
				Local tip:String = gadget.itemTip( index)
				Local extra:Object = GadgetItemExtra( gadget, index)
				ModifyGadgetItem gadget, index, text, flags, 0, tip, extra
			EndIf
		EndFunction

		If (index >= 0) And (index < CountGadgetItems(_list)) Then
			UnselectCurrentItem( _list)
			Local text:String = GadgetItemText( _list, index)
			Local flags:Int = GadgetItemFlags( _list, index)
			Local tip:String = _list.itemTip( index)
			Local extra:Object = GadgetItemExtra( _list, index)
			ModifyGadgetItem _list, index, text, flags, 1, tip, extra
			SelectGadgetItem _list, index
			' show hint
			If index <> -1 Then
				Local y:Int = _gadget.ypos + (index * ITEM_FONT_HEIGHT)
				_hint.SetContent( tip)
				?Win32
				' we need to account for the window border
				y :+ 4
				?
				_hint.ShowPos( _gadget.xpos + _gadget.width, y)
			EndIf
		EndIf
	EndMethod

	' select last item in list
	Method LastItem()
		SelectItem( SelectedGadgetItem(_list) - 1)
	EndMethod

	' select next item in list
	Method NextItem()
		SelectItem( SelectedGadgetItem(_list) + 1)
	EndMethod

	' autocompletes with currently selected item
	Method AutoComplete()
		Local s:String = GetSelected()
		If s.Length > 0 Then _owner.AutoComplete(s)
		Hide()
	EndMethod

	Method GetSelected:String()
		Local index:Int = SelectedGadgetItem(_list)
		If (index >= 0) And (index < AutocVal.Length) Then Return AutocVal[index]
	EndMethod

	Method ClearList()
		AutocVal = Null
		AutocInt = 0
		ClearGadgetItems _list
	EndMethod

	' ripped from the original CreateAutoCompleteMenu() in editor.bmx
	' it now populates the listbox instead
	' Returns TRUE if there is any items in the list FALSE otherwise
	Method PopulateList:Int()
		' This function builds the auto-complete menu
		Local t:String, tempt:String, nbItem:Int
		Local i:Int, xPos:Int
		
		Local src:String
		If CaseSensitive Then
			src = _owner.cleansrc
		Else
			src = _owner.cleansrcl
		EndIf

' find beginning of keyword
		i = _owner.cursorpos
		xPos = Instr(Separ, Mid(src,i,1),1)
		While xPos = 0 And i > 0
			t = Mid(src,i,1) + t
			i :- 1
			xPos = Instr(Separ, Mid(src,i,1),1)
		Wend
		If t.Length = 0 Then
			ClearList()
			Return False
		EndIf
		If Not CaseSensitive Then t = t.ToLower()
		
		' for indexing by a prefix
		If t.Length = 1 And Not AllowOneCharLookup Then
			ClearList()
			Return False
		EndIf
		
		i = _owner.cursorpos
		While Mid(src,i,1) <> Chr(10) And i > 0
			tempt = Mid(src,i,1) + tempt
			i :- 1
		Wend

		' dont bother updating list if same keyword
		If LastKeyword.Length > 0 Then
			If CaseSensitive Then
				If t = LastKeyword Then Return False
			Else
				If t.ToLower() = LastKeyword.ToLower() Then Return False
			EndIf
		EndIf
		LastKeyword = t
		' reset keyword list
		ClearList()

' build keyword list / search for keyword ?
		t = BuildKeywordList( t)
		If t.Length <= 0 Then
			ClearList()
			Return False
		EndIf

		' if there's just one option, and it matches what's already written, forget it!
		If AutocVal.length = 1 Then
			If CaseSensitive Then
				If AutocVal[0] = t Then
					ClearList()
					Return False
				EndIf
			Else
				If AutocVal[0].ToLower() = t.ToLower() Then
					ClearList()
					Return False
				EndIf
			EndIf
		EndIf

		' populate the list
		If AutocVal.Length > 0 Then
			For i = 0 Until AutocVal.Length
				'TODO: add some extra data to the item ?
				AddGadgetItem _list, AutocVal[i], 0, 0, AutocTip[i]
			Next
			' select top item
			SelectItem(0)
		EndIf

		Return (CountGadgetItems( _list) > 1)
	EndMethod


	Method Refresh()
	EndMethod

	Method Invoke:Object( command:Int, argument:Object = Null)
	EndMethod

	Method Query:Int(property:Int)
	EndMethod

	Method BuildKeywordList:String( keyword:String) Abstract
EndType



Type TBMXAutoComplete Extends TGUIAutoComplete
	Method New()
		AllowOneCharLookup = True
	EndMethod

	' a bit messy having to use an abstract method like this, but it gets the job done.
	' takes in the keyword to "search" for,
	' should build the AutocVal & AutocTip arrays from whatever source, and SORT IT!!
	' and then return the keyword
	Method BuildKeywordList:String( keyword:String)
		' make sure owner is TBMXCodePanel
		Local bmxo:TBMXCodePanel = TBMXCodePanel(_owner)
		If Not bmxo Then Return Null
		Local bmxh:TBMXHost = TBMXHost(bmxo.host)
		If Not bmxh Then Return Null
		
		AutocInt = keyword.length
		
		' search blitzmax commands in quickhelp
		Local list:TList = New TList
		For Local tok:TToken = EachIn MapValues(bmxh.quickhelp.map)
			If tok.token.ToLower().Find(keyword) = 0
				list.AddLast( TKeyword.Create( tok.token, bmxh.quickhelp.help(tok.token)))
			EndIf
		Next
		
		' gets next datanode from link
		Function NextNode:TDataNode( link:TLink Var)
			Local val:TDataNode = TDataNode( link.Value())
			link = link.NextLink()
			Return val
		EndFunction

		' search code tree		
		Local nn:TLink = bmxo.codenode._kids.FirstLink()
		Local consts:TDataNode = NextNode(nn)	' 0-Constants
		Local globals:TDataNode = NextNode(nn)	' 1-Globals
		Local funcs:TDataNode = NextNode(nn)	' 2-Functions
		Local types:TDataNode = NextNode(nn)	' 3-Types
		
		' searches a codenode for partial matches and adds them to the list 
		' !! keyword must be upper case, as it uses TDataNode._sortname
		Function SearchCodeNodes( bmxo:TBMXCodePanel, list:TList, root:TDataNode, keyword:String)
			For Local n:TCodeNode = EachIn root._kids
				If n._sortname.Find(keyword) = 0 Then
					' get line from editor
					Local ln:Int = TextAreaLine( bmxo._textarea, n.pos)
					Local line:String = TextAreaText( bmxo._textarea, ln, 1, TEXTAREA_LINES)
					list.AddLast( TKeyword.Create( n._name, "In file ~q" + bmxo.codenode._name + "~q:~n~n" + line))
				EndIf
			Next
		EndFunction
		
		Local t:String = keyword.ToUpper()
		Local doclist:TList
		If maxide.options_code.dockeysfirst Then
			doclist = New TList
		Else
			doclist = list
		EndIf
		SearchCodeNodes( bmxo, doclist, consts, t)
		SearchCodeNodes( bmxo, doclist, globals, t)
		SearchCodeNodes( bmxo, doclist, funcs, t)
		SearchCodeNodes( bmxo, doclist, types, t)
		
		If maxide.options_code.dockeysfirst Then
			' put keywords local to the current document first
			If list.IsEmpty() And doclist.IsEmpty() Then Return Null
			list.Sort( True, TKeyword._Compare)
			If Not doclist.IsEmpty() Then
				doclist.Sort( True, TKeyword._Compare)
				Local nn:TLink = doclist.LastLink(), first:TLink = doclist.FirstLink()
				Repeat
					list.AddFirst( nn.Value())
					If nn = first Then Exit
					nn = nn.PrevLink()
				Forever
			EndIf
		Else
			' sort the list
			If list.IsEmpty() Then Return Null
			list.Sort( True, TKeyword._Compare)
		EndIf
		
		' build the value & tip arrays
		Local keys:TKeyword[] = TKeyword[] list.ToArray()
		AutocVal = New String[keys.Length]
		AutocTip = New String[keys.Length]
		For Local i:Int = 0 Until keys.Length
			AutocVal[i] = keys[i].Key
			AutocTip[i] = keys[i].Help
		Next

		Return keyword
	EndMethod
EndType


Private
Type TKeyword
	Field Key:String
	Field Help:String
	
	Function Create:TKeyword( key:String, help:String = Null)
			Local k:TKeyword = New TKeyword
			k.Key = key
			k.Help = help
			Return k
	EndFunction
	
	Function _CompareCase:Int( a:Object, b:Object)
		Return TKeyword(a).Key.Compare( TKeyword(b).Key)
	EndFunction
	
	Function _Compare:Int( a:Object, b:Object)
		Return TKeyword(a).Key.ToLower().Compare( TKeyword(b).Key.ToLower())
	EndFunction
EndType
Public


'
' generic keyword autocomplete
'
Type TKeywordAutoComplete Extends TGUIAutoComplete
	Field Keywords:TMap = New TMap
	
	Method Register( key:String, help:String = Null)
		Keywords.Insert( key, TKeyword.Create( key, help))
	EndMethod
	
	Method BuildKeywordList:String( keyword:String)
		' this is used to determine where on the line to insert text
		AutocInt = keyword.Length

		' search for matches
		Local list:TList = New TList
		If CaseSensitive Then
			For Local k:TKeyword = EachIn Keywords.Values()
				If k.Key.Find(keyword) = 0 Then list.AddLast( k)
			Next		
		Else
			For Local k:TKeyword = EachIn Keywords.Values()
				If k.Key.ToLower().Find(keyword) = 0 Then list.AddLast( k)
			Next
		EndIf
		
		If list.IsEmpty() Then Return Null
		
		' sort the list
		If CaseSensitive Then
			list.Sort( True, TKeyword._CompareCase)
		Else
			list.Sort( True, TKeyword._Compare)
		EndIf
		
		' build the value & tip arrays
		Local keys:TKeyword[] = TKeyword[] list.ToArray()
		AutocVal = New String[keys.Length]
		AutocTip = New String[keys.Length]
		For Local i:Int = 0 Until keys.Length
			AutocVal[i] = keys[i].Key
			AutocTip[i] = keys[i].Help
		Next

		Return keyword
	EndMethod
EndType



Type TCAutoComplete Extends TKeywordAutoComplete
	Method New()
		AllowOneCharLookup = True
		CaseSensitive = True
		
		'TODO: this should be loaded from a user configurable file
		'TODO: add more stuff!
		
		' types
		Register( "unsigned")
		Register( "signed")
		Register( "void")
		Register( "char")
		Register( "short")
		Register( "int")
		Register( "long")
		Register( "float")
		Register( "double")
		' reserved words
		Register( "typedef")
		Register( "struct")
		Register( "union")
		Register( "static")
		Register( "extern")
		Register( "register")
		Register( "if")
		Register( "else")
		Register( "switch")
		Register( "case")
		Register( "default")
		Register( "for")
		Register( "while")
		Register( "do")
		Register( "break")
		Register( "continue") 
		Register( "goto")
		Register( "return")
		Register( "sizeof", "returns the size of things")
		' pre-processing directives
		Register( "#define", "Pre-Processing Directive:~n~nDefine a macro")
		Register( "#include", "Pre-Processing Directive:~n~nInclude a file")
		Register( "#ifdef", "Pre-Processing Directive:")
		Register( "#ifndef", "Pre-Processing Directive:")
		Register( "#elseif", "Pre-Processing Directive:")
		Register( "#else", "Pre-Processing Directive:")
		Register( "#endif", "Pre-Processing Directive:")
	EndMethod
EndType

Type THTMLAutoComplete Extends TKeywordAutoComplete
	Method New()
		AllowOneCharLookup = True
		
		'TODO: this should be loaded from a user configurable file		
		'TODO: add more stuff!		
		
		' opening tags
		Register( "html>", "Element:")
		Register( "head>", "Element:")
		Register( "meta>", "Element:")
		Register( "link>", "Element:")
		Register( "style>", "Element:")
		Register( "script>", "Element:")
		Register( "title>", "Element:")
		Register( "body>", "Element:")
		Register( "a>", "Element:")
		Register( "b>", "Element:")
		Register( "i>", "Element:")
		Register( "u>", "Element:")
		Register( "pre>", "Element:")
		Register( "h1>", "Element:")
		Register( "h2>", "Element:")
		Register( "h3>", "Element:")
		Register( "h4>", "Element:")
		Register( "h5>", "Element:")
		Register( "h6>", "Element:")
		Register( "hr>", "Element:")
		Register( "table>", "Element:")
		Register( "th>", "Element:")
		Register( "td>", "Element:")
		Register( "tr>", "Element:")
		Register( "div>", "Element:")
		Register( "span>", "Element:")
		Register( "font>", "Element:")
		Register( "p>", "Element:")
		Register( "br>", "Element:")
		' attributes		
		Register( "class", "Attribute:~nAll Elements")		
		Register( "style", "Attribute:~nAll Elements")
		Register( "name", "Attribute:~nAll Elements")
		Register( "id", "Attribute:~nAll Elements")
		Register( "href", "Attribute:~n<A> <LINK>")
		Register( "src", "Attribute:~n<IMG>")
		Register( "width", "Attribute:")
		Register( "height", "Attribute:")
		Register( "type", "Attribute:")
		Register( "rel", "Attribute:~n<LINK>")
		Register( "colspan", "Attribute:~n<TABLE>")
		Register( "rowspan", "Attribute:~n<TABLE>")
		Register( "cellspacing", "Attribute:~n<TABLE>")
		Register( "language", "Attribute:~n<SCRIPT>")
	EndMethod
EndType
