'****************************************************************************
'* All items below this point appear to have stuff specific to
'* editing And compiling BMX code, e.g. the path.  This may not be entirely
'* correct, e.g. the HelpPanel!
'****************************************************************************
Global is_demo:Int
Global bmxpath$,bmkpath$

Incbin "resources/splash.png" ' splash screen

' TOOLMENU's are sections of menu gadgets
' they that can be enabled,disabled
' gadgets can allso be added to/from these sections
' and you can get their state
'
' number of toolmenu sections
Const TOOLMENU_COUNT:Int = 5
' toolmenu section indices
Const TOOLMENU_FILES:Int = 0
Const TOOLMENU_EDITOR:Int = 1
Const TOOLMENU_SELECTION:Int = 2 ' implies EDITOR
Const TOOLMENU_BUILD:Int = 3
Const TOOLMENU_DEBUG:Int = 4

'useful constants
Const TAB$=Chr(9)
Const Separ$ = "<>+-*\/= '[](),.:~q~t~r~n"
Const EOL$="~n"

Function CheckDemo:Int()
	If Not is_demo Return True
	Notify "@@message/demo_error"
	Return False
End Function


Type TBMXHost Extends TGUIHost

	Field	panels:TGUIPanel[]
	Field	helppanel:THelpPanel
	Field	output:TGUIOutput
	Field	tempfilecount:Int ' number of temporary files created

	' options
	Field	options_all:TOptionsGroup
	Field	options_host:THostOptions
	Field	options_editor:TEditorOptions
	Field	options_code:TCodeOptions
	Field	options_build:TBuildOptions
	Field	options_tools:TToolsOptions
	Field	options_sync:TSyncOptions
	Field	options_keys:TKeyOptions
	Field	options_help:THelpOptions
	Field	options_autoc:TAutoCompleteOptions
	Field	optionsgui:TOptionsRequester

	Field	cmdlinereq:TCmdLineRequester
	Field	cmdline$
	'Field	syncmodsreq:TBMXSyncModsRequester
	Field	buildmodsreq:TBMXBuildModsRequester
	Field	gotoreq:TGotoRequester
'	Field	progress:TProgressRequester
	Field	projectreq:TGUIProjectRequester
	Field	searchreq:TSearchRequester

	Field 	toolmenus:TList[TOOLMENU_COUNT]
	Field	menubar:Int,toolbar:TGadget,tabbar:TGadget
	Field	split:TSplitter
	Field	debugtree:TDebugTree

	Field	root:TDataNode
	Field	projects:TProjects
	Field	coderoot:TDataNode
	Field	projectroot:TDataNode
	Field	todoroot:TDataNode	'bregors todo-tree-list
	Field	navbar:TNavBar

	Field	Mode:Int
	Field	debugcode:TBMXCodePanel

	Field	quickenable:TGadget	'menu,state
	Field	debugenable:TGadget	'menu,state
	Field	threadenable:TGadget

	Field	menu_app_console:TGadget
	Field	menu_app_gui:TGadget
	Field	menu_app_wxgui:TGadget		
	
	Field	menu_target_win:TGadget
	Field	menu_target_linux:TGadget
	Field	menu_target_macos_native:TGadget
	Field	menu_target_macos_universal:TGadget	
		
	Field	quickhelp:TQuickHelp
	Field	recentmenu:TGadget
	Field	recentmenus:TGadget[]
	Field	tooly:Int,debugview:Int,navtab:Int
	Field	progress:Int

	Field	_client:TGUIGenericContainer

	Field activepanel:TGUIPanel

	Field hotkey_next:THotKey
	Field hotkey_prev:THotKey
	
	Field _iconstrip:TIconStrip

	' help related bits
	Field helproot:TDataNode

	Method Refresh()
		RefreshToolbar()
		RefreshToolMenus()
	End Method

	Method SelectActivePanel()
		SelectPanel activepanel
	End Method

	Method BuildPanel:TCodePanel()
		Local fn$ = options_build.buildfile
		Local t:TCodePanel
		For t = EachIn panels
			If t.path.ToLower() = fn Then Return t
		Next
		DebugLog "BuildPanel():: ~q" + fn + "~q not found!"
		Return Null
	End Method

	Method CheckVersion$()
		Local	process:TProcess
		Local	bytes:Byte[]
		Local	cmd$,version$

		cmd$=bmxpath+"/bin/bcc"
?win32
		cmd:+".exe"
?
		process=CreateProcess("~q"+cmd$+"~q",HIDECONSOLE)
		If process
			While True
				bytes=process.pipe.ReadPipe()
				If bytes
					version:+String.FromBytes(bytes,Len bytes)
				EndIf
				If Not process.status() Exit
				Delay 10
			Wend
		EndIf

		If version=""
			Notify "Unable to determine BlitzMax version.~n~nPlease reinstall BlitzMax to repair this problem."
			End
		EndIf

'		print "bcc version="+version

		If version.find("(expired)")>-1	'Demo timed out
			Notify "@@message/demo_expired"
			End
		EndIf

		If version.find("Demo Version")>-1
			is_demo=True
			Notify "@@message/demo_info|" + version
		EndIf

	End Method

	Method OpenProgress(message$)
'		progress.Open message
		DisableGadget _gadget
		SetStatus message
		progress=-1
	End Method

	Method CloseProgress()
'		progress.Hide
		SetStatus ""
		EnableGadget _gadget
		progress=0
	End Method

	Method UpdateProgress(message$,value:Int=0)		'returns false if cancelled
'		Return progress.Update(message,value)
		If progress
			If progress/5<>value/5
				SetStatus message+" "+value+"%"
				progress=value
'				Pollsystem
			EndIf
		EndIf
	End Method

	Method CloseAll:Int(warning$="",dontask:Int=False)	'returns true if successful
		Local	panel:TGUIPanel
		Local	mypanels:TGUIPanel[]
		Local	cancel:Int, count:Int

		For panel=EachIn panels
			If TCodePanel(panel) Then count :+ 1 
		Next

		If count=0 Then warning$ = ""
		If (warning$ = "") Or Confirm(warning$) Then
			mypanels=panels

			For panel=EachIn mypanels
				If panel.invoke(TOOLCLOSE,[dontask]) cancel=True
			Next

			Return Not cancel
		EndIf
	End Method

	Method Quit()
		SnapshotWindow()
		options_all.WriteConfig()
		If CloseAll() Then running=False
	End Method

	Method DebugExit()
		If debugcode
			debugcode.Edit		'restore cursor etc.
			debugcode=Null
		EndIf
		SetMode EDITMODE
	End Method

	Method DebugSource(path$,line:Int,column:Int)
		Local	code:TBMXCodePanel
		path=FullPath(path)
		code=TBMXCodePanel(OpenSource(path))
		If code=Null
			Notify("@@message/open_error|"+path)
			Return
		EndIf
		If debugcode And debugcode<>code debugcode.Edit		'restore cusor etc.
		debugcode=code
		debugcode.debug(line,column)
		ActivateWindow _gadget
		PollSystem
	End Method

	Method SetMode(m:Int)
		If Mode=m Return
		ActivateWindow _gadget
		Select m
		Case DEBUGMODE
			navtab=navbar.SelectedView()
			navbar.SelectView debugview
		Case EDITMODE
			navbar.SelectView navtab
		End Select
		Mode=m
		Refresh()
	End Method

	Method RefreshToolbar()
		Local	i:Int
' sourceedit buttons
		If TCodePanel(CurrentPanel)
			EnableGadgetItem toolbar , TOOLBAR_CLOSE
			EnableGadgetItem toolbar , TOOLBAR_SAVE
			For i=TOOLBAR_CUT To TOOLBAR_PASTE
				EnableGadgetItem toolbar,i
			Next
			EnableGadgetItem toolbar , TOOLBAR_UNDO
			EnableGadgetItem toolbar , TOOLBAR_REDO
			EnableGadgetItem toolbar , TOOLBAR_FIND
		Else
			If TGUIOutput(CurrentPanel) Then
				EnableGadgetItem toolbar , TOOLBAR_CLOSE
			Else
				DisableGadgetItem toolbar , TOOLBAR_CLOSE
			EndIf
			DisableGadgetItem toolbar , TOOLBAR_SAVE			
			For i=TOOLBAR_CUT To TOOLBAR_PASTE
				DisableGadgetItem toolbar,i
			Next
			DisableGadgetItem toolbar , TOOLBAR_UNDO
			DisableGadgetItem toolbar , TOOLBAR_REDO
			DisableGadgetItem toolbar , TOOLBAR_FIND
		EndIf
' debug buttons
		If Mode = DEBUGMODE And debugtree.cancontinue
			For i = TOOLBAR_BUILD To TOOLBAR_BUILDANDDEBUG
				DisableGadgetItem toolbar,i
			Next
			For i=TOOLBAR_STEP To TOOLBAR_STEPOUT
				EnableGadgetItem toolbar,i
			Next
		Else
			For i = TOOLBAR_BUILD To TOOLBAR_BUILDANDDEBUG
				EnableGadgetItem toolbar,i
			Next
			For i = TOOLBAR_STEP To TOOLBAR_STEPOUT
				DisableGadgetItem toolbar,i
			Next
		EndIf
' build buttons
		'if you change this dont forget to change TOOLMENU_BUILD test in RefreshToolMenus()
		If ((Mode <> DEBUGMODE) And (panels.Length > 1) And (TCodePanel(CurrentPanel) Or (TGUIOutput(CurrentPanel) And panels.Length > 2))) Or ..
			((THelpPanel(CurrentPanel) Or TGUIOutput(CurrentPanel)) And options_build.buildfile.Length)
			For i = TOOLBAR_BUILD To TOOLBAR_BUILDANDDEBUG
				EnableGadgetItem toolbar,i
			Next
		Else
			For i = TOOLBAR_BUILD To TOOLBAR_BUILDANDDEBUG
				DisableGadgetItem toolbar,i
			Next
		EndIf
' stop button
		If output And output.process
			EnableGadgetItem toolbar , TOOLBAR_STOP
			EnableGadgetItem toolbar , TOOLBAR_CONTINUE
		Else
			DisableGadgetItem toolbar , TOOLBAR_STOP
			DisableGadgetItem toolbar , TOOLBAR_CONTINUE
		EndIf
	End Method

' TOOLMENU related methods, for keeping track of various sections of menu item states
	Method RefreshToolMenus()
		EnableToolMenus( TOOLMENU_FILES, TCodePanel(CurrentPanel) Or TGUIOutput(CurrentPanel))
		EnableToolMenus( TOOLMENU_EDITOR, TCodePanel(CurrentPanel) <> Null)
		EnableToolMenus( TOOLMENU_SELECTION, TCodePanel(CurrentPanel) And CurrentPanel.Query(QUERY_HASSELECTION))
		'MARK: if you change this dont forget to change the same in RefreshToolbar() build buttons
		EnableToolMenus( TOOLMENU_BUILD, ((Mode <> DEBUGMODE) And (panels.Length > 1) And (TCodePanel(CurrentPanel) Or (TGUIOutput(CurrentPanel) And panels.Length > 2))) Or ..
			((THelpPanel(CurrentPanel) Or TGUIOutput(CurrentPanel)) And options_build.buildfile.Length))
		EnableToolMenus( TOOLMENU_DEBUG, (Mode = DEBUGMODE) And debugtree.cancontinue)
	EndMethod

	Method AddToolMenuItem( index:Int, value:Object)
		If (index >= 0) And (index < toolmenus.Length) Then
			If TGadget(value) Then
				' set same state as group
				If MenuEnabled( TGadget(toolmenus[index].First())) Then
					EnableMenu TGadget(value)
				Else
					DisableMenu TGadget(value)
				EndIf
				toolmenus[index].AddLast( value)
			ElseIf TList(value) And (Not TList(value).IsEmpty()) Then
				' set same state as group
				If MenuEnabled( TGadget(toolmenus[index].First())) Then
					For Local m:TGadget = EachIn TList(value)
						EnableMenu m
					Next
				Else
					For Local m:TGadget = EachIn TList(value)
						DisableMenu m
					Next
				EndIf
				toolmenus[index].AddLast( value)
			EndIf
		EndIf
	EndMethod

	Method RemoveToolMenuItem( index:Int, value:Object)
		If (index >= 0) And (index < toolmenus.Length) Then
			If TGadget(value) Then
				toolmenus[index].Remove( value)
			ElseIf TList(value) Then
				toolmenus[index].Remove( value)
			EndIf
		EndIf
	EndMethod

	Method EnableToolMenus( index:Int, enable:Int = True)
		If (index >= 0) And (index < toolmenus.Length) Then
			Local g:TGadget = TGadget(toolmenus[index].First() ) 
			If g=Null Then Return 
			If MenuEnabled( g ) = enable Then Return
			For Local o:Object = EachIn toolmenus[index]
				If TGadget(o) Then
					If enable Then
						EnableMenu TGadget(o)					
					Else
						DisableMenu TGadget(o)
					EndIf
				ElseIf TList(o) Then
					If enable Then
						For Local menuitem:TGadget = EachIn TList(o)
							EnableMenu menuitem
						Next
					Else
						For Local menuitem:TGadget = EachIn TList(o)
							DisableMenu menuitem
						Next
					EndIf
				EndIf
			Next
			' update main menu
			UpdateWindowMenu _gadget
		EndIf
	EndMethod

	Method ToolMenuEnabled:Int( index:Int)
		If (index >= 0) And (index < toolmenus.Length) Then
			Return MenuEnabled( TGadget(toolmenus[index].First()))
		EndIf
		Return False
	EndMethod

	Method IsSourceOpen:Int(path$)
		Local	panel:TGUIEditor
		Local	p$
		p$=FullPath(path)
		For panel=EachIn panels
			If panel.path=p Return True
		Next
	End Method

	Method MakePathTemp$(path$)
		Return FullPath("$BMXPATH/tmp/."+StripDir(path))
	End Method

	Method OpenSource:TCodePanel(path$,temp:Int=False)
		Local	code:TCodePanel
		Local	panel:TGUIEditor
		Local	ext$,p$
		Local	files:String[]
		Local	lastpanel:TGUIEditor
		Local	lastcode:TCodePanel

		path=path.Replace("\","/")

' only do the temp thing if we're in the bmx directory
		If temp And path[0..Len(bmxpath)] <> bmxpath Then temp = False

		If path.Length <= 0 Then
' handler for new
			' new files are allways of type BMX (can be changed by saving as different type)
			code = TBMXCodePanel.Create( Null , Self) 

			If temp Then
				code.SaveSource MakePathTemp(path)
			EndIf
			If code Then
				AddRecent code.path
				AddOpenFile code.path
			EndIf
			Return code
		ElseIf path = "." Then
' handler for open
DebugLog FILTER_Files
			files = RequestMultiFile( "@@main_window/open_title", FILTER_Files )
			If files.Length <= 0 Then Return Null
			' prepend path to all filenames
			If files.Length > 1 Then
				For Local i:Int = 1 Until files.Length
					files[i] = files[0] + files[i]
				Next
				files = files[1..] ' remove path
			EndIf
		Else
' handler for single file
			files = [ path ]
		EndIf

' check if files are already open
		For Local i:Int = 0 Until files.Length
			If files[i].Length <= 0 Then Continue
			p = FullPath(files[i]).ToLower()
			For panel = EachIn panels
				If panel.Path.ToLower() = p Then
					lastpanel = panel
					files[i] = Null ' so we dont try to reload it later
					Exit
				EndIf
			Next
		Next

' open multiple files
		For Local s:String = EachIn files
			If s.Length <= 0 Then Continue
			OpenProgress "Loading " + StripDir( s)
			code = TCodePanel.Create( s, Self)
			If code Then
				If temp Then code.SaveSource MakePathTemp(path)
				AddRecent code.path
				AddOpenFile code.path
				'FIX: fixes not updating highlighting
				code.ForceHighlight()
				lastcode = code
			EndIf
			CloseProgress()
		Next
		
' select/activate the last panel/loaded file
		If lastcode Then
			lastcode.Invoke(TOOLSHOW)
			Return lastcode
		ElseIf lastpanel Then
			SelectPanel lastpanel
			Return TCodePanel(lastpanel)
		EndIf
		Return Null
	End Method

	Method IsTempPath:Int(path$)
		Local s$ = FullPath("$BMXPATH/tmp/") 
		If path[..s.length]=s Then Return True
	End Method

	Method OpenSourceAt:TCodePanel(path$,line:Int,column:Int)
		Local	code:TCodePanel= OpenSource(path$)
		code.invoke TOOLNAVIGATE,[line]
	End Method

	Method AddOpenFile(path$)
		If IsTempPath(path$) Then Return
		If FileType(path) <> FILETYPE_FILE Then Return
		Local	f$
		For f$=EachIn options_host.openlist
			If f$=path$ Then Return
		Next
		options_host.openlist.AddLast(path$)
	End Method

	Method RemoveOpenFile(path$)
		Local	f$
		For f$=EachIn options_host.openlist
			If f$=path$ Then options_host.openlist.Remove(f$);Exit
		Next
	End Method

	Method AddRecent(path$)
		If IsTempPath(path$) Then Return
		If FileType(path) <> FILETYPE_FILE Then Return		
		Local	f$
		For f$ = EachIn options_host.recentlist
			If f$=path$ Then options_host.recentlist.Remove(f$);Exit
		Next
		options_host.recentlist.AddFirst(path$)
		RefreshRecentFiles
		UpdateWindowMenu _gadget
	End Method

	Method RefreshRecentFiles()
		Local	n:Int,f$
		Local	m:TGadget
		
		For m=EachIn recentmenus
			RemoveMenuItem m
		Next
		
		' check if files exist
		For f=EachIn options_host.recentlist
			If FileType(f) <> FILETYPE_FILE Then options_host.recentlist.Remove(f)
		Next		
		
		n=Min(options_host.recentlist.count(),16)
		recentmenus=New TGadget[n]
		n=0
		For f$=EachIn options_host.recentlist
			recentmenus[n]=AddMenuItem(f$,MENURECENT+n,recentmenu,0,0,PixmapFromIconStrip(_iconstrip,ICON_FILE))
			n:+1
			If n=16 Exit
		Next
	End Method

	Method ImportBB()
		Local	f$,cmd$,path$
		f$=RequestFile("@@main_window/import_bb_title","bb" )
		If Not f$ Return
		cmd$=quote(bmkpath$)
		cmd$:+" convertbb "
		cmd$:+quote(FullPath(f$))
		Execute cmd,"@@output_panel/cmd_convert_bb|"+StripExt(StripDir(f$))
		output.wait
		OpenSource(StripExt(f$)+".bmx")
	End Method

'TODO: check zapped version against any installed version and over-write only earlier modules (possibly add option).
	Method ImportMod()
		Local cmd$
		Local zapPath$ = RequestFile("@@main_window/import_zap_title", "@@zap:zap")
		If zapPath$ = Null Then Return

		cmd$ = Quote(bmkpath$)
		cmd$ :+ " unzapmod "
		cmd$ :+ Quote(FullPath(zapPath$))
		Execute cmd, "@@output_panel/cmd_import_mod"
		Execute Quote(bmxpath$ + "/bin/makedocs"), "@@output_panel/cmd_update_docs" ', MENUDOCMODS
	End Method

'TODO: add "Release mode" option to filter out everything but .a and .i files. Will require either, additions to bmk or mirroring of bmk functionality in the editor (not desirable).
	Method ExportMod()
		Local modName$, cmd$, basePath$ = (bmxpath$ + "/mod/").ToLower()
		Local modPath$ = RequestDir("@@main_window/export_mod_title", basePath$).ToLower()
		If modPath$ = Null Then Return

		modPath$ = FullPath(modPath$)

		If (ExtractExt(modPath$).ToLower() <> "mod")
			Notify("@@message/mod_dir_error|"+modPath$)
			Return
		Else
			modName$ = modPath$.Replace(basePath$, "")
			modName$ = modName$.Replace(".mod", "")
			modName$ = modName$.Replace("/", ".")

			' This block duplicates code from bmk. Not as scary as reading the output from bmk ;o)
			Local opt_arch$, cfg_platform$

			?MacOS
				cfg_platform$ = "macos"
				?MacOsPPC
					If is_pid_native(0) opt_arch$ = "ppc" Else opt_arch$ = "x86"
				?MacOsX86
					If is_pid_native(0) opt_arch$ = "x86" Else opt_arch$ = "ppc"
			?Win32
				opt_arch$ = "x86"
				cfg_platform$ = "win32"
			?Linux
				opt_arch$ = "x86"
				cfg_platform$ = "linux"
			?

			If FileType(ModuleInterface(modName$, "release." + cfg_platform + "." + opt_arch)) <> FILETYPE_FILE
				Notify("@@message/mod_found_error")
				Return
			EndIf

'			Print modName$
		EndIf

		Local zapPath$ = RequestFile("@@main_window/save_zap_title", "@@zap:zap", True)
		If zapPath$
			zapPath$ = FullPath(zapPath$)
			If ExtractExt(zapPath$).ToLower() <> "zap" Then zapPath$ = StripExt(zapPath$) + ".zap"
		Else
			Return
		EndIf

		cmd$ = Quote(bmkpath$)
		cmd$ :+ " zapmod "
		cmd$ :+ modName$ + " "
		cmd$ :+ Quote(zapPath$)
		Execute cmd, "@@output_panel/cmd_export_mod"

	End Method
'End additions

	Method GetCommandLine$()
		Return cmdline
	End Method

	Method SetCommandLine(text$)
		cmdline=text
	End Method

	Method Execute(cmd$,mess$="",post$="",home:Int=True,tool:TBase=Null)
		If Not output output=TGUIOutput.Create(Self,options_build,debugtree)
		output.execute cmd$,mess$,post$,home,tool
	End Method

	Method SelectError(path$,column:Int,line:Int)
		Local	panel:TGUIEditor,found:Int
		For panel=EachIn panels
			If panel.path=path found=True;Exit
		Next
		If Not found panel=OpenSource(path)
		If panel
			SelectPanel panel
			panel.Invoke TOOLGOTO,String(line)
			'TODO: maybe turn into a TOOLSELECTLINE ?
			TCodePanel(panel).Debug( line, column)
		EndIf
	End Method

	' this is a catch all error parser, normally the code panel would catch BMK/GCC errors.
	' but since the host does BuildMods and other stuff without codepanels, we must try to handle it here too.
	Method ParseError( cmd:String, err:String)
		Local mess$,file$,p:Int,q:Int
		Local line:Int,column:Int
		
' bcc error
		If err$[..13]="Compile Error"
			err=err[14..]
			p=err.find(EOL$)
			If p=-1 p=Len err
			mess=err[..p]
			err=err[p+1..]
			If err[..1]="["
				p=err.find("]")
				If p=-1 p=Len err
				file$=err[1..p]
				p=file.find(";")+1
				If p=0 p=Len err
				q=file.find(";",p)+1
				If q=0 q=Len err
				line=Int(file[p..q-1])
				column=Int(file[q..])
				file=FullPath(file[..p-1])
				SelectError file,column,line
			EndIf
			mess = TranslateMessage(mess)
			Local oldapptitle$ = AppTitle
			AppTitle = LanguageString("error_message/compile_error_title")
			Notify mess
			AppTitle = oldapptitle$
			SetStatus mess
			Return
		EndIf
		
' gcc error
		' this splits the errors into lines and then splits via ":",
		' it then loops over all elements until FILE MESSage and LINE are filled.		
		' if FILE and LINE are set, the file is opened and the line selected.
		' if MESS and LINE are set, a error message pops up
		Local lines:String[] = err.Split(EOL)
		For Local l:Int = 0 Until lines.Length

			Local emsg:String = lines[l]

			' make sure line contains an error header
			If Not (emsg.Find("error:")>-1) Then Continue
		
			emsg = emsg.Replace("::","^^")
			Local c:String[] = emsg.Trim().Split(":")
			?Win32
			' fix for DRIVE: syntax
			If (c.Length >= 2) And (c[0].Length = 1) Then
				c[1] = c[0] + ":" + c[1]
				c = c[1..]
			EndIf
			?
			If (c.Length >= 1) And Not file Then file = c[0]
			If (c.Length >= 2) And (c[1].Length > 0) And (c[1][0] = Asc(" ")) Then
				If Not mess Then mess = c[1][1..].Trim()
			ElseIf (c.Length >= 2) Then
				If Not line Then line = c[1].ToInt()
				If c.Length >= 4 And Not mess Then mess = c[3][1..]
			EndIf
			If file And line Then SelectError( file, 0,line)
			If mess And line Then
				mess = mess.Replace("^^","::")
				Local oldapptitle$ = AppTitle
				AppTitle = LanguageString("error_message/gcc_error_title")
				
				' if following lines contain notes then add them to the message
				Local nl:Int, pos:Int, bl:Int
				For nl=l+1 Until lines.length
					pos = lines[nl].find("note:")
					If pos<>-1 
						If Not bl Then mess:+"~n" ; bl=True
						mess:+"~n"+lines[nl][pos+6..].Replace("^^","::")
					Else 
						Exit ' no more notes for this error
					EndIf
				Next
				
				Notify mess
				AppTitle = oldapptitle$
			EndIf
			If file And line And mess Then Return
		Next
				
		' strip out archive creation (buildmods) and empty lines
		Local lns:String[]
		For Local s:String = EachIn lines
			s = s.Replace("~r","").Replace("~n","").Trim()
			If s And s.Find( "ar: creating") = -1 Then
				lns = lns[..lns.Length + 1]
				lns[lns.Length-1] = s
			EndIf
		Next
		lines = lns
		
' unknown error
		' return whatever is in the error (max 5 lines)
		If lines.Length = 0 Then Return
		If lines.Length = 1 Then
			mess = lines[0]
		Else
			mess = "~r~n".Join( lines[..4])
		EndIf
		If Not mess Then mess = "Unknown Error!"
		Local oldapptitle$ = AppTitle
		AppTitle = LanguageString("error_message/runtime_error_title")
		If cmd Then
			Notify "COMMAND::" + cmd + "~r~n" + mess
		Else
			Notify mess
		EndIf
		AppTitle = oldapptitle$
Rem		
		err=err.Replace(EOL+"   "," ")
		While err
			p=err.find(EOL)
			If p=-1 p=(Len err)-1
			mess=err[..p]
			err=err[p+1..]
			p=0
			While True
				p=mess.Find(":",p)+1
				If p=0 Exit
				?Win32
				' skip drive seperator
				If p = 2 Then p=mess.Find(":",p)+1
				?
				q=mess.Find(":",p)
				If q<>-1
					file=mess[..p-1]
					column=Int(mess[p..q])
					Local i:Int = mess.Find(":",q+1)
					If i<>-1 Then
						line=Int(mess[q+1..i])
						q = i
					EndIf
					If line
						mess=mess[q+1..]
						SelectError file,column,line
						Local oldapptitle$ = AppTitle
						AppTitle = LanguageString("error_message/gcc_error_title")
						Notify mess
						AppTitle = oldapptitle$
						Return
					EndIf
					p=q+1
				EndIf
			Wend
		Wend
EndRem
	End Method	

	Method AddPanel(tabpanel:TGUIPanel)
		Local	panel:TGadget , index:Int , icon:Int
DebugLog "AddPanel:Enter"				
		index=CountGadgetItems(tabbar)
		If panels.length <= index panels = panels[..index + 1]
		
		icon=-1
		If TCodePanel(tabpanel) Then icon = ICON_FILE
		If THelpPanel(tabpanel) Then icon = ICON_HELP
		If TGUIOutput(tabpanel) Then icon = ICON_OUTPUT
		
		AddGadgetItem(tabbar,tabpanel.name$,True,icon)

		panel = CreatePanel(0 , 0 , ClientWidth(tabbar) , ClientHeight(tabbar) , tabbar , 0)	'name
		HideGadget panel
		SetGadgetLayout panel,1,1,1,1
		tabpanel.panel=panel
		tabpanel.index=index
		panels[index] = tabpanel
		eventhandlers.addlast tabpanel
DebugLog "AddPanel:Exit"
	End Method

	Method RemovePanel(tabpanel:TGUIPanel)
		Local p:TGUIPanel[]
		Local index:Int
		eventhandlers.remove tabpanel
' unset debugcode
		If debugcode=tabpanel debugcode=Null
' activate next panel
		If tabpanel=activepanel activepanel=helppanel
		If tabpanel=BuildPanel() Then options_build.buildfile=""
		If TCodePanel(tabpanel) Then RemoveOpenFile(TCodePanel(tabpanel).path)
		If tabpanel=currentpanel
			index=tabpanel.index+1
			If index>=panels.length index=panels.length-2
			SelectPanel panels[index]
		EndIf
' remove from array
		p=panels
		panels=panels[..panels.length-1]
		For index=tabpanel.index To panels.length-1
			panels[index]=p[index+1]
			panels[index].index=index
		Next
' remove gadget
		FreeGadget tabpanel.panel
		RemoveGadgetItem tabbar,tabpanel.index
		tabpanel.panel=Null
' just in case SelectPanel() wasnt called
		Refresh()
	End Method

' used by editor when saving as new filetype
	Method SwapPanels( idx1:Int, idx2:Int)
		Function SwapGadgetItems( gadget:TGadget, idx1:Int, idx2:Int)
			If Not gadget Then Return
			Local text:String = gadget.ItemText(idx1)
			Local flags:Int = gadget.ItemFlags(idx1)
			Local icon:Int = gadget.ItemIcon(idx1)
			Local tip:String = gadget.ItemTip(idx1)
			Local extra:Object = gadget.ItemExtra(idx1)
			ModifyGadgetItem gadget, idx1, gadget.ItemText(idx2), gadget.ItemFlags(idx2), gadget.ItemIcon(idx2), gadget.ItemTip(idx2), gadget.ItemExtra(idx2)
			ModifyGadgetItem gadget, idx2, text, flags, icon, tip, extra
		EndFunction
		SwapGadgetItems( tabbar, idx1,idx2)
		' swap tab panels
		Local tmp:TGUIPanel = panels[idx1]
		panels[idx1] = panels[idx2]
		panels[idx2] = tmp
		panels[idx1].index = idx1
		panels[idx2].index = idx2
	EndMethod

	Method SelectPanel(panel:TGUIPanel)
		Local	Current:TGUIPanel
		If currentpanel = panel Then DebugLog "SelectPanel:I'm already active!";Return
		Current=currentpanel
		currentpanel = panel
		ShowGadget panel.panel
		SelectGadgetItem tabbar , panel.index
		DebugLog "SelectPanel:pre-Invoke"
		currentpanel.Invoke TOOLSHOW
		DebugLog "SelectPanel:post-Invoke"		
		If panel.active Then activepanel=panel
		If Current And Current <> currentpanel HideGadget Current.panel
		DebugLog "SelectPanel:pre-Refresh"
		Refresh() 
		DebugLog "SelectPanel:post-Refresh"
	End Method

	Method RefreshPanel(panel:TGUIPanel)	'call after a name change
		ModifyGadgetItem( tabbar,panel.index,panel.name )
	End Method

' bregors todo-tree-list
	Method SetTodoNode(todo:TDataNode)
		Local node:TDataNode
		If todoroot._kids.count() Then node = TDataNode(todoroot._kids.First() ) 
'If todo Then DebugLog "SetTodoNode:" + todo._name + "(" + todo._kids.count() + ")" Else DebugLog "SetTodoNode:Null"
		If node=todo Then Return
		If node Then node.Free
		If todo Then
			todoroot.Append todo
			todoroot.Refresh
			todoroot.Open
			todo.Open
		EndIf
	End Method

	Method SetCodeNode(code:TDataNode)
		Local node:TDataNode	
		If coderoot._kids.count() Then node = TDataNode(coderoot._kids.First() ) 
'If code Then DebugLog "SetCodeNode:" + code._name + "(" + code._kids.count() + ")" Else DebugLog "SetCodeNode:Null"
		If node=code Then Return
		If node Then node.Free
		If code Then
			coderoot.Append code
			coderoot.Refresh
			coderoot.Open
			code.Open
		EndIf
'DebugLog "SetCodeNode:Exit"
	End Method
		
	Method Initialize()
		Local	open:TCodePanel
		Local	i:Int,dir$,nomods:Int,p:Int
		Local	stream:TStream
		Local	splash:TGadget

		SetAppTitle("MaxIDE")

		splash=CreateWindow("Welcome to MaxIDE CE",200,200,400,140,Null,WINDOW_CLIENTCOORDS|WINDOW_HIDDEN|WINDOW_CENTER)
			Local splashpanel:TGadget = CreatePanel(0,0,ClientWidth(splash),ClientHeight(splash),splash,0)
			SetPanelColor splashpanel,255,255,255;SetPanelPixmap splashpanel, LoadPixmapPNG("incbin::resources/splash.png"), PANELPIXMAP_FIT2
			Local progress:TGadget = CreateProgBar(200,100,140,16,splashpanel)
			Local label:TGadget = CreateLabel("Initializing...",200,116,140,20,splashpanel,LABEL_CENTER);SetGadgetFont label,LookupGuiFont(GUIFONT_SYSTEM,8)
			ShowGadget splash;PollSystem

		SetPointer POINTER_APPSTARTING

		Try
			bmxpath = BlitzMaxPath()
			bmxpath = bmxpath.Replace("\" , "/")
		Catch err$
			Notify "Could not find BlitzMax compiler (bin/bcc), please Reinstall~nappdir="+AppDir,True
			End
		EndTry

		' store this such that host.FullPath can translate it where needed...
		AddPathKey "$BMXPATH" , bmxpath

		bmkpath = bmxpath + "/bin/bmk"
		bmkpath = bmkpath.Replace("\" , "/")
?win32
		bmkpath:+".exe"
?

		UpdateProgBar progress, 0.05;PollSystem ' 5% done - found blitzmax

		' construct all the options data structures that we care about
		' first we create an options group to hold all of our options
		options_all:TOptionsGroup = New TOptionsGroup

		' next we load the options that we want
		options_host=THostOptions.Create(options_all)
		options_editor=TEditorOptions.Create(options_all)
		options_code=TCodeOptions.Create(options_all)
		options_build=TBuildOptions.Create(options_all)
		options_tools=TToolsOptions.Create(options_all)
		options_sync=TSyncOptions.Create(options_all)
		options_keys=TKeyOptions.Create(options_all)
		options_help=THelpOptions.Create(options_all)
		options_autoc=TAutoCompleteOptions.Create(options_all)
		
		If FileType(bmxpath+"/cfg/ide-ce.ini")<>FILETYPE_NONE Then
			options_all.ReadConfig(bmxpath + "/cfg/ide-ce.ini")
		Else
?macos
			'Ranlibmods
			Local cmd$=Quote( bmxpath+"/bin/bmk" )+" ranlibdir "+Quote( bmxpath+"/mod" )
			system_ cmd
?
			options_all.readconfig(bmxpath + "/cfg/ide.ini")
			options_all.writeconfig(bmxpath + "/cfg/ide-ce.ini")
		EndIf
		
		UpdateProgBar progress, 0.1;PollSystem ' 10% done - loaded options
		
		' check for single instance mode
?Not debug
		If options_host.singleinstance Then
			If Not InitSingleInstance( "MaxIDE-Community-Edition-" + VERSION) Then
				SendOpenFileMessage( "BLITZMAX_WINDOW_CLASS", "MaxIDE", AppArgs[1..])
				End
			EndIf
		EndIf
?

		UpdateProgBar progress, 0.15;PollSystem ' 15% done - checked for existing instance if not debug

		dir$=bmxpath+"/mod"
		If FileType(dir)=0
			If Not CreateDir(dir)
				Notify "Failed to Create Module Directory "+dir,True
				End
			EndIf
			nomods=True
		EndIf
		dir$=bmxpath+"/tmp"
		If FileType(dir)=0
			If Not CreateDir(dir)
				Notify "Failed to Create Temp Directory "+dir,True
				End
			EndIf
		EndIf
		dir$=bmxpath+"/cfg"
		If FileType(dir)=0
			If Not CreateDir(dir)
				Notify "Failed to Create Config Directory "+dir,True
				End
			EndIf
		EndIf

		_iconstrip = LoadIconStrip("incbin::resources/icons_small.png") 
		If _iconstrip=Null Then DebugStop()

		UpdateProgBar progress, 0.2 ;SetGadgetText label,"Preparing Language Settings...";PollSystem' 20% done - found/create directories and loaded icons

		' load language file
		LanguageLoad options_host.languagefile
		LanguageLocaliseAppMenu() ' for MacOS application menu

		UpdateProgBar progress, 0.3 ;SetGadgetText label, "@@message/startup_gui";PollSystem' 30% done - loaded localisation

		'TODO: create the editor panels as a region
		'TODO: create the toolbar as a region

		' create requesters
		TCodePanel.Initialise(Self, options_host.autobackup,options_code.autotidymode)
		cmdlinereq=TCmdLineRequester.Create(Self)
		'syncmodsreq=TBMXSyncModsRequester.Create(Self)
		buildmodsreq=TBMXBuildModsRequester.Create(Self)
		gotoreq=TGotoRequester.Create(Self)
		optionsgui=TOptionsRequester.Create(Self,options_all)
'		progress=TProgressRequester.Create(Self)
		projectreq=TGUIProjectRequester.CreateProjectRequester(Self)
		searchreq=TSearchRequester.Create(Self)

		Local winsize:TRect=options_host.winsize
		If OutsideDesktop(winsize)
			winsize.set(20,20,760,540)
		EndIf
		'TODO: Should use WINDOW_CLIENTCOORDS
		_gadget=CreateWindow("MaxIDE",20,20,760,540,Null,WINDOW_TITLEBAR | WINDOW_RESIZABLE | WINDOW_MENU | WINDOW_STATUS|WINDOW_HIDDEN|WINDOW_ACCEPTFILES|WINDOW_CENTER)

		Local img:TPixmap=LoadPixmap("incbin::resources/main_icon.png")
		SetGadgetPixmap _gadget , img , GADGETPIXMAP_ICON
		SetGadgetShape _gadget , winsize.x , winsize.y , winsize.w , winsize.h		

		UpdateProgBar progress, 0.4;PollSystem ' 40% done - created all windows

		LanguageSetGroup "main_window"

' init toolbar
		toolbar = CreateToolBar("incbin::resources/icons.png" , 0 , 0 , 0 , 0 , _gadget) 

		SetToolbarToolTips() 

		If options_host.showtoolbar
			tooly=GadgetHeight(toolbar)
		Else
			HideGadget toolbar
			tooly=0
		EndIf
		
' init toolmenus
		For Local i:Int = 0 Until toolmenus.Length
			toolmenus[i] = New TList
			' add a dummy menu, so that we can keep track of Enable/Disable state even if there are no items
			toolmenus[i].AddLast CreateMenu("",0,Null)
		Next
	
		UpdateProgBar progress, 0.4;PollSystem ' 40% done - created menus & toolbars	

'TODO: MacOS specific hacks
?MacOS
'		client=window 'I would like to know why this is here... with it uncommented, refreshall moves the
					  'moves the window to the wrong location.
' cludge for window growing from adding of toolbar
		If Not options_host.maximised SetGadgetShape _gadget,winsize.x,winsize.y,winsize.w,winsize.h
?
		If Not _client
			'TODO: _Client doesnt really need to be a field on this type...
			_client = New TGUIGenericContainer
			_client.CreateGenericContainer("base panel",0,tooly,ClientWidth(_gadget),ClientHeight(_gadget)-tooly,Self)
			'CreatePanel(0,tooly,ClientWidth(_gadget),ClientHeight(_gadget)-tooly,_gadget,4)
			'SetGadgetLayout client,1,1,1,1
			_client.show()
			AddRegion _client
		EndIf

		'NAVIGATION BAR CODE - OLD
		'-------------------------

		tabbar = CreateTabber(0 , 0 , ClientWidth(_gadget) , ClientHeight(_gadget) - tooly , _client.gadget() ) 
		SetGadgetIconStrip tabbar , _iconstrip
		
		'MOT: for some reason, skid had replaced clientwidth/height above with zeroes, this screws something else up later on
		'     so have reinstated it.  I wonder if it's a Mac thing (see Mac specific bit above).

		debugtree = TDebugTree.CreateDebugTree(Self)
		
		root=TDataNode.CreateNode("@@help_tree/title")
		helproot=root.AddNode("@@help_tree/help")

		' create projects root
		projectroot=TDataNode.CreateNode("@@projects_tree/title")
		projects=TProjects.CreateProjects(Self)
		projectroot.Append projects
		projects.Open		

		' create the todo root
		todoroot=TDataNode.CreateNode("@@todo_tree/title")
		todoroot.Open

		' create the code tree root
		coderoot=TDataNode.CreateNode("@@code_tree/title")	'root.AddNode("Open")
		coderoot.Open

		navbar=TNavBar.Create(_client,options_tools, _iconstrip)

		navbar.AddView root,ICON_HELP
		navbar.AddView projectroot,ICON_PROJECTS
		navbar.AddView coderoot,ICON_CODESTRUCTURE
		navbar.AddView todoroot,ICON_TODO
		debugview=navbar.AddView(debugtree,ICON_DEBUG)
		' bregors todo-tree-list
		navbar.SelectView 0

		AddHandler navbar


		'NAVIGATION BAR CODE - NEW
		'-------------------------
		'TODO: Need to move the above old nav code into a new "region" based set of code

		'Local tpg:TGUITabbedPanelGroup = New TGUITabbedPanelGroup
		'tpg.InitTabbedPanelGroup(Self)
		'AddRegion tpg

		'SPLITTER BAR CODE
		'-----------------
		
		split=TSplitter.Create(TSplitter.VERTICAL,options_tools.splitpos,options_tools.navswap,tabbar,navbar.gadget())		'treeview)
		AddHandler split
		
		UpdateProgBar progress, 0.45;PollSystem ' 45% done - created all windows		
		
		' Initialise the menus
		InitMenu
		SetMode EDITMODE
DebugLog "6.3"
		' create the help & output panels and bring help to front
		helppanel=THelpPanel.Create(Self,bmxpath,options_help,helproot)
		helppanel.options_help = options_help
		output=TGUIOutput.Create(Self,options_Build,debugtree)
		activepanel = helppanel

DebugLog "6.4"	

		'Synchronize documentation
		UpdateProgBar progress, 0.55 ;SetGadgetText label,"@@message/startup_docs";PollSystem' 55% done - menus & helppanel done
		SyncDocs()
		UpdateProgBar progress, 0.65;PollSystem ' 65% done - docs sync'd
				
DebugLog "7"
		' check to see if we need to makedocs
		Local mkdocs:Int
		If FileType( bmxpath+"/docs/html/User Guide/index.html" )<>FILETYPE_FILE
			CreateDir bmxpath+"/docs/html"
			CreateFile bmxpath+"/docs/html/index.html"
			mkdocs=True
		EndIf
		
		' scan projects in projlist and add them
		For Local info$=EachIn options_host.projlist
			projects.AddProject info$
		Next
		UpdateProgBar progress, 0.7;PollSystem ' 70% done - loaded projects		
		
		' show the home panel
		helppanel.home
		selectPanel helppanel
DebugLog "8"
		UpdateProgBar progress, 0.8 ;SetGadgetText label,"@@message/startup_open";PollSystem' 80% done - just need to open any previous files

		' open files from .ini restorelist
		Local foundbuildfile%
		Local fcount:Float=0
		If options_host.restoreopenfiles
			For Local f$=EachIn options_host.openlist
				fcount=fcount+1
				If FileType(f$) = FILETYPE_FILE ' Stops empty code panel being created for unavailable files
					open=OpenSource(f$)
					If open And StripDir(f$)=options_build.buildfile Then
						open.Invoke(TOOLLOCK)
						foundbuildfile=True
					End If
				EndIf
				UpdateProgBar progress, 0.8+(0.15*fcount/options_host.openlist.count());PollSystem
			Next
		EndIf
		If Not foundbuildfile Then options_build.buildfile = ""
DebugLog "9"
		UpdateProgBar progress, 0.95;PollSystem ' 95% done - just need to open any specified files
		' open files specified in command line
		For i=1 To AppArgs.length-1
			open=OpenSource(AppArgs[i])
		Next
		
		UpdateProgBar progress, 1;PollSystem ' 100% done
		
		HideGadget splash
		FreeGadget splash
		PollSystem
DebugLog "10"
		If options_host.maximised Then MaximizeWindow(_gadget) 
		ShowGadget _gadget
		'TODO: Do we really need a refresh after initialisation?
		RefreshAll		
		PollSystem	'allow repaint
DebugLog "11"		
		running=True

		CheckVersion
		'If nomods syncmodsreq.Show
DebugLog "12"
		'build docs if not there		
		If mkdocs
			If Confirm( "Documentation not found - Rebuild documentation?" ) And CloseAll( False ) Then DocMods
		EndIf

		Refresh()
		SetPointer POINTER_DEFAULT
	End Method

	Method SetToolbarTooltips()
		Local str$[] = ConvertStringToArray(translate("@@main_window/toolbar_tips") , "|")
		For Local i% = 0 To Len(str) - 1
			Local id% = 0
			id = options_keys.FindKeyID(str[i]) 
			If id Then
				If options_keys.HotKeyName(id)<>"" Then str[i]:+ " (" + options_keys.HotKeyName(id) + ")"
			End If
		Next
		SetToolBarTips toolbar,str
	End Method

	Method DocMods()
		Local cmd$
		cmd$=quote(bmxpath+"/bin/makedocs")
		execute cmd,"Rebuilding Documentation...",MENUTRIGGERSYNCDOCS
	End Method
		
	Method SyncDocs()
		helppanel.SyncDocs(bmxpath)
		quickhelp = TBMXQuickHelp.Create(bmxpath+"/docs/html/Modules/commands.txt")
		'FIX: remove this call to helppanel.home - unneccessary and causes lots of random selectpanels on startup
		'helppanel.Home
	End Method

	Field	menuitems:TMap

	Method AddMenuItem:Tgadget(name$ , tag% , parent:TGadget , hotkey% = 0 , modifier% = 0,pixmap:TPixmap=Null,group%=-1) 
		Local id% = 0
		Local m:Tgadget
		
		If hotkey <> 0 Then
			id = options_keys.AddHotKeyDef(Translate(name$).Replace("&","").Replace("...","") , hotkey , modifier)
			hotkey = options_keys.GetKey(id) 
			modifier = options_keys.GetKeyMod(id)
			m = CreateMenu(name , tag , parent , hotkey , modifier) 
		Else
			m = CreateMenu(name , tag , parent)
		EndIf
		
		If pixmap Then SetGadgetPixmap m , pixmap
		
		MapInsert menuitems,m,String(id)
		If group<>-1 Then AddToolMenuItem group,m
		Return m
	End Method
	
	Method RemoveMenuItem(menu:TGadget)
		MapRemove menuitems , menu
		FreeMenu menu
	End Method

	Method DisableHotKeys()
		Local item:TGadget
		For item = EachIn MapKeys(menuitems)
			'TODO: SetGadgetHotKey with Nulls doesn't work on GTK - needs Brucey's love!
			SetGadgetHotKey item , Null , Null
		Next
		UpdateWindowMenu _gadget		
	End Method

	Method EnableHotKeys() 
		Local item:TGadget,id%,hotkey%,modifier%
		For item = EachIn MapKeys(menuitems) 
			id=Int(String(menuitems.valueforkey(item)))
			hotkey = options_keys.GetKey(id) 
			modifier = options_keys.GetKeyMod(id)	
			SetGadgetHotKey item,hotkey,modifier
		Next
		UpdateWindowMenu _gadget
	End Method

	Method AddMenu:TGadget(name$)
		Return CreateMenu(name,0,WindowMenu(_gadget))
	End Method
	
	Method AddMenuSep(parent:Tgadget) 
		CreateMenu "",0,parent
	End Method

	Method InitMenu()
		Local file:TGadget,edit:TGadget,program:TGadget,debug:TGadget,tools:TGadget,help:TGadget,buildoptions:TGadget
		Local buildmods:TGadget , buildallmods:TGadget , syncmods:TGadget , docmods:TGadget
	
		menuitems=CreateMap()
'---------------------FILE
		LanguageSetGroup "file_menu"
		file = AddMenu("@@file") 
		AddMenuItem "@@New" , MENUNEW , file , KEY_N , MODIFIER_COMMAND , PixmapFromIconStrip(_iconstrip,ICON_NEW)
		AddMenuItem "@@Open" , MENUOPEN , file , KEY_O , MODIFIER_COMMAND , PixmapFromIconStrip(_iconstrip,ICON_OPEN)

		recentmenu=AddMenuItem("@@open_recent",0,file,0,0,PixmapFromIconStrip(_iconstrip,ICON_OPEN))
?MacOS
		AddMenuItem "@@close",MENUCLOSE,file,KEY_W,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_CLOSE),TOOLMENU_FILES
		AddMenuItem "@@close_all", MENUCLOSEALL, file, KEY_W, MODIFIER_COMMAND| MODIFIER_SHIFT,PixmapFromIconStrip(_iconstrip,ICON_CLOSE)
?Win32
		AddMenuItem "@@close",MENUCLOSE,file,KEY_F4,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_CLOSE),TOOLMENU_FILES
		AddMenuItem "@@close_all",MENUCLOSEALL, file, KEY_F4, MODIFIER_COMMAND| MODIFIER_SHIFT,PixmapFromIconStrip(_iconstrip,ICON_CLOSE)
?Linux
		AddMenuItem "@@close",MENUCLOSE,file,KEY_W,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_CLOSE),TOOLMENU_FILES
		AddMenuItem "@@close_all", MENUCLOSEALL, file, KEY_W, MODIFIER_COMMAND| MODIFIER_SHIFT,PixmapFromIconStrip(_iconstrip,ICON_CLOSE)
?

		AddMenuSep file
		AddMenuItem "@@save",MENUSAVE,file,KEY_S,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_SAVE),TOOLMENU_FILES
		AddMenuItem "@@save_as",MENUSAVEAS,file,KEY_S,MODIFIER_COMMAND|MODIFIER_SHIFT,PixmapFromIconStrip(_iconstrip,ICON_SAVE),TOOLMENU_FILES
		AddMenuItem "@@save_all",MENUSAVEALL,file,,,PixmapFromIconStrip(_iconstrip,ICON_SAVE)

		AddMenuSep file
		AddMenuItem "@@Print" , MENUPRINT , file , 0 , 0 ,PixmapFromIconStrip(_iconstrip,ICON_PRINT), TOOLMENU_FILES
		
?Not macos
		AddMenuSep file
		AddMenuItem "@@exit",MENUQUIT,file,0,0,PixmapFromIconStrip(_iconstrip,ICON_EXIT)
?


'------------------------- EDIT
		LanguageSetGroup "edit_menu"

		edit=AddMenu("@@edit")
		AddMenuItem "@@undo",MENUUNDO,edit,KEY_Z,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_UNDO),TOOLMENU_EDITOR
?MacOS
		AddMenuItem "@@redo",MENUREDO,edit,KEY_Z,MODIFIER_COMMAND|MODIFIER_SHIFT,PixmapFromIconStrip(_iconstrip,ICON_REDO),TOOLMENU_EDITOR
?Win32
		AddMenuItem "@@redo",MENUREDO,edit,KEY_Y,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_REDO),TOOLMENU_EDITOR
?Linux
		AddMenuItem "@@redo",MENUREDO,edit,KEY_Y,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_REDO),TOOLMENU_EDITOR
?

		AddMenuSep edit
		AddMenuItem "@@cut",MENUCUT,edit,KEY_X,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_CUT),TOOLMENU_SELECTION
		AddMenuItem "@@copy",MENUCOPY,edit,KEY_C,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_COPY),TOOLMENU_SELECTION
		AddMenuItem "@@paste" , MENUPASTE , edit , KEY_V , MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_PASTE),TOOLMENU_EDITOR
		
		AddMenuSep edit
		AddMenuItem "@@select_all" , MENUSELECTALL , edit , KEY_A , MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_SELECTALL), TOOLMENU_EDITOR
		
		AddMenuSep edit
		AddMenuItem "@@block_indent",MENUINDENT,edit,KEY_CLOSEBRACKET,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_BLOCKINDENT),TOOLMENU_SELECTION
		AddMenuItem "@@block_outdent" , MENUOUTDENT , edit , KEY_OPENBRACKET , MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_BLOCKOUTDENT),TOOLMENU_SELECTION
		AddMenuItem "@@block_comment",MENUCOMMENT,edit,KEY_QUOTES,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_BLOCKCOMMENT),TOOLMENU_SELECTION
		AddMenuItem "@@block_uncomment" , MENUUNCOMMENT , edit , KEY_QUOTES , MODIFIER_COMMAND | MODIFIER_SHIFT,PixmapFromIconStrip(_iconstrip,ICON_BLOCKUNCOMMENT),TOOLMENU_SELECTION
		
		AddMenuSep edit
		AddMenuItem "@@find_replace",MENUFINDREPLACE,edit,KEY_F,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_FIND),TOOLMENU_EDITOR
?MacOS
		AddMenuItem "@@find_next",MENUFINDNEXT,edit,KEY_G,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_FIND),TOOLMENU_EDITOR
		AddMenuItem "@@goto_line",MENUGOTO,edit,KEY_L,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_FIND),TOOLMENU_EDITOR
?Win32
		AddMenuItem "@@find_next",MENUFINDNEXT,edit,KEY_F3,0,PixmapFromIconStrip(_iconstrip,ICON_FIND),TOOLMENU_EDITOR
		AddMenuItem "@@goto_line",MENUGOTO,edit,KEY_G,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_FIND),TOOLMENU_EDITOR
?Linux
		AddMenuItem "@@find_next",MENUFINDNEXT,edit,KEY_F3,0,PixmapFromIconStrip(_iconstrip,ICON_FIND),TOOLMENU_EDITOR
		AddMenuItem "@@goto_line",MENUGOTO,edit,KEY_G,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_FIND),TOOLMENU_EDITOR
?

		AddMenuSep edit
		AddMenuItem "@@toggle_bookmark",MENUTOGGLEMARK,edit,KEY_M,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_BOOKMARK),TOOLMENU_EDITOR
		AddMenuItem "@@next_bookmark",MENUNEXTMARK,edit,KEY_PERIOD,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_BOOKMARK_NEXT),TOOLMENU_EDITOR
		AddMenuItem "@@previous_bookmark",MENUPREVIOUSMARK,edit,KEY_COMMA,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_BOOKMARK_PREV),TOOLMENU_EDITOR


'-----------------------------Project
		LanguageSetGroup "project_menu"

		program=AddMenu("@@project")
		AddMenuItem "@@build",MENUBUILD,program,KEY_B,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_BUILD),TOOLMENU_BUILD
		AddMenuItem "@@build_and_run",MENURUN,program,KEY_R,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_BUILDANDRUN),TOOLMENU_BUILD
		AddMenuItem "@@build_and_debug",MENUDEBUG,program,KEY_R,MODIFIER_COMMAND| MODIFIER_SHIFT,PixmapFromIconStrip(_iconstrip,ICON_BUILDANDDEBUG),TOOLMENU_BUILD
		AddMenuItem "@@command_line" , MENUCOMMANDLINE , program,0,0,PixmapFromIconStrip(_iconstrip,ICON_OUTPUT)
		
		AddMenuSep program
		'TODO: add project properties to this menu
		'projprops=CreateMenu("@@properties",0,program)
		
		'the following items do NOT go through AddMenuItem on purpose as they are tickboxes
		buildoptions=AddMenuItem("@@build_options",0,program,0,0,PixmapFromIconStrip(_iconstrip,ICON_OPTIONS))
		quickenable=CreateMenu("@@quick_build",MENUQUICKENABLED,buildoptions)
		threadenable = CreateMenu("@@threaded_build" , MENUTHREADENABLED , buildoptions)
		
		Local appoptions:TGadget = AddMenuItem("@@app_options",0,program,0,0,PixmapFromIconStrip(_iconstrip,ICON_OPTIONS))
		menu_app_console = CreateMenu("@@app_console" , MENUAPPCONSOLE , appoptions)
		menu_app_gui = CreateMenu("@@app_gui" , MENUAPPGUI , appoptions)
		menu_app_wxgui = CreateMenu("@@app_wxgui" , MENUAPPWXGUI , appoptions)

		Local targetoptions:TGadget = AddMenuItem("@@target_platform" , 0 , program , 0 , 0 , PixmapFromIconStrip(_iconstrip , ICON_OPTIONS) )
		menu_target_win = CreateMenu("@@target_win32" , MENUTARGETWIN , targetoptions)
		menu_target_linux = CreateMenu("@@target_linux" , MENUTARGETLINUX , targetoptions)
		?x86
		menu_target_macos_native = CreateMenu("@@target_macos_intel" , MENUTARGETMACOS , targetoptions)
		?ppc
		menu_target_macos_native = CreateMenu("@@target_macos_ppc" , MENUTARGETMACOS , targetoptions)
		?
		menu_target_macos_universal = CreateMenu("@@target_macos_universal" , MENUTARGETMACOSUNI , targetoptions)			

		AddMenuSep program
		AddMenuItem "@@lock_build_file",MENULOCKBUILD,program,0,0,PixmapFromIconStrip(_iconstrip,ICON_LOCK),TOOLMENU_FILES
		AddMenuItem "@@unlock_build_file",MENUUNLOCKBUILD,program,0,0,PixmapFromIconStrip(_iconstrip,ICON_UNLOCK),TOOLMENU_FILES

		AddMenuSep program
		AddMenuItem "@@next_file",MENUNEXTFILE,program,KEY_TAB,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_NEXTFILE)
		AddMenuItem "@@prev_file",MENUPREVFILE,program,KEY_TAB,MODIFIER_COMMAND | MODIFIER_SHIFT,PixmapFromIconStrip(_iconstrip,ICON_PREVFILE)

'---------------------------------- DEBUG

		LanguageSetGroup "debug_menu"
		debug=AddMenu("@@debug")
?Not linux
		AddMenuItem "@@step" , MENUSTEP , debug, KEY_F9,0,PixmapFromIconStrip(_iconstrip,ICON_STEP),TOOLMENU_DEBUG
		AddMenuItem "@@step_in" , MENUSTEPIN , debug, KEY_F10,0,PixmapFromIconStrip(_iconstrip,ICON_STEPIN),TOOLMENU_DEBUG
		AddMenuItem "@@step_out",MENUSTEPOUT,debug,KEY_F11,0,PixmapFromIconStrip(_iconstrip,ICON_STEPOUT),TOOLMENU_DEBUG
?linux
' now consistent with Eclipse, and no clash on F10
		AddMenuItem "@@step" , MENUSTEP , debug, KEY_F6,0,PixmapFromIconStrip(_iconstrip,ICON_STEP),TOOLMENU_DEBUG
		AddMenuItem "@@step_in" , MENUSTEPIN , debug, KEY_F5,0,PixmapFromIconStrip(_iconstrip,ICON_STEPIN),TOOLMENU_DEBUG
		AddMenuItem "@@step_out",MENUSTEPOUT,debug,KEY_F7,0,PixmapFromIconStrip(_iconstrip,ICON_STEPOUT),TOOLMENU_DEBUG
?
		AddMenuItem "@@halt",MENUSTOP,debug,0,0,PixmapFromIconStrip(_iconstrip,ICON_STOP),TOOLMENU_DEBUG

'---------------------------------- TOOLS
		LanguageSetGroup "tools_menu"
		tools = AddMenu("@@tools" ) 

		AddMenuItem "@@ide_options" , MENUOPTIONS , tools ,0,0,PixmapFromIconStrip(_iconstrip,ICON_OPTIONS)
		
		AddMenuSep  tools
 		AddMenuItem "@@project_manager" , MENUPROJECTMANAGER , tools,0,0,PixmapFromIconStrip(_iconstrip,ICON_PROJECTS)

		AddMenuSep tools 
		AddMenuItem "@@import_bb_project",MENUIMPORTBB,tools ,0,0,PixmapFromIconStrip(_iconstrip,ICON_IMPORTBB)
		AddMenuItem "@@import_module", MENUIMPORTMOD, tools,0,0,PixmapFromIconStrip(_iconstrip,ICON_MODULEIMPORT)
		AddMenuItem "@@export_module" , MENUEXPORTMOD , tools,0,0,PixmapFromIconStrip(_iconstrip,ICON_MODULEEXPORT)
		
		AddMenuSep tools
		'syncmods=AddMenuItem ("@@synchronise_modules",MENUSYNCMODS,tools,0,0,PixmapFromIconStrip(_iconstrip,ICON_SYNC))
		buildmods=AddMenuItem ("@@build_modules",MENUBUILDMODULES,tools,KEY_D,MODIFIER_COMMAND,PixmapFromIconStrip(_iconstrip,ICON_MODULEBUILD))
		'buildallmods = AddMenuItem ("@@rebuild_all_modules" , MENUBUILDALLMODULES , tools,0,0,PixmapFromIconStrip(_iconstrip,ICON_MODULEBUILD))
		
		AddMenuSep tools
		docmods=AddMenuItem ("@@docmods",MENUDOCMODS,tools,0,0,PixmapFromIconStrip(_iconstrip,ICON_REBUILDDOCS))

'---------------------------------- HELP
		LanguageSetGroup "help_menu"

		help=AddMenu("@@help")
		AddMenuItem "@@home",MENUHOME,help,0,0,PixmapFromIconStrip(_iconstrip,ICON_HOME)
		'CreateMenu "@@back",MENUBACK,help
		'CreateMenu "@@forward",MENUFORWARD,help
		AddMenuItem "@@quick_help" , MENUQUICKHELP , help , KEY_F1 , 0 ,PixmapFromIconStrip(_iconstrip,ICON_HELP), TOOLMENU_EDITOR
		
		AddMenuSep help
		AddMenuItem "@@about_maxide",MENUABOUT,help,0,0,PixmapFromIconStrip(_iconstrip,ICON_ABOUT)

		'force refresh of all hotkeys
		EnableHotKeys()

		' UnckeckMenu forces the menuentry to show a checkbox (Linux)
		If options_build.quickenabled CheckMenu quickenable Else UncheckMenu quickenable
		If options_build.threadenabled CheckMenu threadenable Else UncheckMenu threadenable
		
		CheckMenuAppType(options_build.apptype)
		CheckMenuTarget(options_build.targetplatform)						
		
?win32
		Local mingw$=getenv_("MINGW")
		If Not mingw
			DisableMenu buildmods
			DisableMenu buildallmods
		EndIf
?
		If is_demo
			DisableMenu syncmods
		EndIf

		RefreshRecentFiles
		UpdateWindowMenu _gadget
	End Method

	Method RunCode()
		If Mode=DEBUGMODE And debugtree.cancontinue
			output.go
			Return
		EndIf
		output.stop()
		output.Open()
		SaveAll
		If options_build.buildfile<>""
			BuildPanel().invoke TOOLRUN
		Else
			activepanel.invoke TOOLRUN
		EndIf
	End Method

	Method RunCodeDebug()
		If Mode=DEBUGMODE And debugtree.cancontinue
			output.go
			Return
		EndIf
		output.stop
		output.open()
		SaveAll
		If options_build.buildfile<>""
			BuildPanel().invoke TOOLDEBUG
		Else
			activepanel.invoke TOOLDEBUG
		EndIf
	End Method

	Method BuildCode()
		output.Stop()
		output.Open()
		SaveAll
		If options_build.buildfile<>""
			BuildPanel().invoke TOOLBUILD
		Else
			activepanel.invoke TOOLBUILD
		EndIf
	End Method

	Method SaveAll()
		Local	panel:TGUIPanel
		For panel=EachIn panels
			panel.invoke TOOLQUICKSAVE
		Next
	End Method

	Method RefreshAll()
		'TODO: RefreshAll is using the _client - should it?
		Local	panel:TGUIPanel
' store current panel				
		Local	curr:TGUIPanel		
		If activepanel Then
			curr = activepanel
		Else
			curr = currentpanel
		EndIf
'change language
		LanguageLoad options_host.languagefile
		LanguageUpdateGadgets()		
		
		'update codepanel settings
		TCodePanel.Initialise(Self , options_host.autobackup , options_code.autotidymode)		
		
' flip navbar
		split.SetFlip options_tools.navswap
' hide/show toolbar
		If options_host.showtoolbar
			tooly=GadgetHeight(toolbar)
			SetGadgetShape _client.Gadget(),0,tooly,ClientWidth(_gadget),ClientHeight(_gadget)-tooly
'			SetGadgetShape tabbar,0,tooly,clientwidth(_gadget),clientheight(_gadget)-tooly
			SetToolBarToolTips()
'			ShowGadget toolbar
		Else
			tooly=0
			HideGadget toolbar
			SetGadgetShape _client.gadget(),0,tooly,ClientWidth(_gadget),ClientHeight(_gadget)-tooly
'			SetGadgetShape tabbar,0,tooly,clientwidth(_gadget),clientheight(_gadget)-tooly
		EndIf
' refresh panels
		For panel=EachIn panels
			panel.invoke TOOLREFRESH
		Next
' refresh navbar
		navbar.invoke TOOLREFRESH
' sort code stuff
		If options_tools.sorted Then coderoot.SortKids()
' restore current panel
		currentpanel = Null
		activepanel = Null
		SelectPanel(curr)
	End Method

	Method SnapshotWindow()
		options_tools.splitpos=split.pos
		If WindowMaximized(_gadget)
			options_host.setWindowSize(True)
		Else
			If Not WindowMinimized(_gadget)
				options_host.SetWindowSize(False,GadgetX(_gadget),GadgetY(_gadget),GadgetWidth(_gadget),GadgetHeight(_gadget))
			EndIf
		EndIf
		options_tools.SetDirty True
		options_host.SetDirty True		
	End Method

	Method OnMenu(menu:Int,extra:Object=Null)
		Local	index:Int
		Local	tool:TBase
		
		If menu=0 Then Return
		' First of all, check to see if the menu event came from a tool.  If so, pass it on.
		tool=TBase(extra)
		If tool And tool <> Self Then
			tool.Invoke(TranslateMenuToTool(menu),extra) 
			Return
		EndIf

		Select menu
			Case MENUNEW
				OpenSource ""
			Case MENUOPEN
				OpenSource "."
			Case MENUCLOSE
				currentpanel.invoke TOOLCLOSE
			Case MENUCLOSEALL
				CloseAll "@@message/close_proceed"
			Case MENUSAVE
				currentpanel.invoke TOOLSAVE
			Case MENUSAVEAS
				currentpanel.invoke TOOLSAVEAS
			Case MENUSAVEALL
				SaveAll

			Case MENUIMPORTMOD
				ImportMod()
			Case MENUEXPORTMOD
				ExportMod()
			Case MENUPRINT
				currentpanel.invoke TOOLPRINT
			Case MENUQUIT
				Quit

			Case MENUGOTO
				gotoreq.setTarget activepanel
				gotoreq.Show
			Case MENUFINDREPLACE
				currentpanel.Invoke TOOLFINDREPLACE
			Case MENUFINDNEXT
				currentpanel.invoke TOOLFINDNEXT

			Case MENUTOGGLEMARK
				currentpanel.invoke TOOLTOGGLEMARK
			Case MENUNEXTMARK
				currentpanel.invoke TOOLNEXTMARK
			Case MENUPREVIOUSMARK
				currentpanel.invoke TOOLPREVIOUSMARK

			Case MENUUNDO currentpanel.invoke TOOLUNDO
			Case MENUREDO currentpanel.invoke TOOLREDO

			Case MENUCUT currentpanel.invoke TOOLCUT
			Case MENUCOPY currentpanel.invoke TOOLCOPY
			Case MENUPASTE currentpanel.invoke TOOLPASTE
			Case MENUSELECTALL currentpanel.invoke TOOLSELECTALL

		' check TOOLMENU_BUILD since these can be called via a hotkey
			Case MENUBUILD
				If ToolMenuEnabled(TOOLMENU_BUILD) Then BuildCode
			Case MENURUN
				If ToolMenuEnabled(TOOLMENU_BUILD) Then RunCode
			Case MENUDEBUG
				If ToolMenuEnabled(TOOLMENU_BUILD) Then RunCodeDebug

			Case MENUBUILDMODULES
				If CheckDemo() Then buildmodsreq.Show()
				
'			Case MENUBUILDALLMODULES
'				If CheckDemo() BuildModules True

			Case MENUSTEP If output output.stepover
			Case MENUSTEPIN If output output.stepin
			Case MENUSTEPOUT If output output.stepout
			Case MENUSTOP If output output.stop

			Case MENULOCKBUILD
				activepanel.invoke TOOLLOCK
			Case MENUUNLOCKBUILD
				If options_build.buildfile<>"" Then BuildPanel().invoke TOOLUNLOCK ;options_build.buildfile=""

			Case MENUCOMMANDLINE cmdlinereq.Show

			Case MENUQUICKENABLED
				If options_build.quickenabled
					options_build.quickenabled=False
					UncheckMenu quickenable
				Else
					options_build.quickenabled=True
					CheckMenu quickenable
				EndIf
				UpdateWindowMenu _gadget

			Case MENUTHREADENABLED
				If options_build.threadenabled
					options_build.threadenabled=False
					UncheckMenu threadenable
				Else
					options_build.threadenabled=True
					CheckMenu threadenable
				EndIf
				UpdateWindowMenu _gadget

			Case MENUAPPGUI
				options_build.apptype = APPTYPE_GUI
				CheckMenuAppType(APPTYPE_GUI)

			Case MENUAPPCONSOLE
				options_build.apptype = APPTYPE_CONSOLE
				CheckMenuAppType(APPTYPE_CONSOLE)
				
			Case MENUAPPWXGUI
				options_build.apptype = APPTYPE_WXGUI
				CheckMenuAppType(APPTYPE_WXGUI)				

			Case MENUTARGETWIN
				options_build.targetplatform = TARGETPLATFORM_WIN32
				CheckMenuTarget(TARGETPLATFORM_WIN32)	

			Case MENUTARGETLINUX
				options_build.targetplatform = TARGETPLATFORM_LINUX
				CheckMenuTarget(TARGETPLATFORM_LINUX)	
				
			Case MENUTARGETMACOS
				options_build.targetplatform = TARGETPLATFORM_MACOS_NATIVE
				CheckMenuTarget(TARGETPLATFORM_MACOS_NATIVE)
				
			Case MENUTARGETMACOSUNI
				options_build.targetplatform = TARGETPLATFORM_MACOS_UNIVERSAL
				CheckMenuTarget(TARGETPLATFORM_MACOS_UNIVERSAL)
				
			Case MENUIMPORTBB
				ImportBB

			Case MENUFINDINFILES
				searchreq.Show
			Case MENUPROJECTMANAGER
				projectreq.Open projects
			Case MENUSHOWCONSOLE
				If output Then output.Open()

			Case MENUOPTIONS
				SnapshotWindow()
				optionsgui.Show

			Case MENUQUICKHELP
				currentpanel.invoke TOOLHELP

			Case MENUHOME
				helppanel.Home
				SelectPanel helppanel

			Case MENUABOUT
				Local typ$
				If is_demo HelpAbout(True) Else HelpAbout(False)
			Case MENUINDENT
				currentpanel.invoke TOOLINDENT
			Case MENUOUTDENT
				currentpanel.invoke TOOLOUTDENT
			Case MENUCOMMENT
				currentpanel.invoke TOOLCOMMENT
			Case MENUUNCOMMENT
				currentpanel.invoke TOOLUNCOMMENT

'			Case MENUSYNCMODS
'				If CheckDemo() And CloseAll("MaxIDE needs to close all documents before continuing") Then syncmodsreq.Show
			Case MENUDOCMODS
				If CheckDemo() And CloseAll("MaxIDE needs to close all documents before continuing") Then DocMods
			Case MENUTRIGGERDOCMODS
				DocMods
			Case MENUTRIGGERSYNCDOCS
				SyncDocs
				
			Case MENUNEXTFILE
				invoke TOOLNEXTFILE
			Case MENUPREVFILE
				invoke TOOLPREVFILE

			Default
				If menu>=MENURECENT
					Local f$=String(options_host.recentlist.ValueAtIndex(menu-MENURECENT))
					If f$ OpenSource f$
				EndIf
		End Select
		
	End Method

	Method CheckMenuAppType(apptype:Int)
		UncheckMenu menu_app_gui
		UncheckMenu menu_app_console
		UncheckMenu menu_app_wxgui
		
		Select apptype
			Case APPTYPE_GUI ; CheckMenu menu_app_gui
			Case APPTYPE_WXGUI ; CheckMenu menu_app_wxgui
			Case APPTYPE_CONSOLE ; CheckMenu menu_app_console
		End Select	

'disable menus not relevant to platform
?Not macos
		DisableMenu menu_app_wxgui
?macos
		If Not FileType( BlitzMaxpath()+"/mod/bah.mod/appstub.mod" ) Then DisableMenu menu_app_wxgui
?
		UpdateWindowMenu _gadget
			
	End Method

	Method CheckMenuTarget(target:Int)
		UncheckMenu menu_target_win
		UncheckMenu menu_target_linux
		UncheckMenu menu_target_macos_native
		UncheckMenu menu_target_macos_universal		
		
		Select target
			Case TARGETPLATFORM_WIN32 ; CheckMenu menu_target_win
			Case TARGETPLATFORM_LINUX ; CheckMenu menu_target_linux
			Case TARGETPLATFORM_MACOS_NATIVE ; CheckMenu menu_target_macos_native
			Case TARGETPLATFORM_MACOS_UNIVERSAL ; CheckMenu menu_target_macos_universal
		End Select	
		
'disable menus not relevant to platform		
?win32
		DisableMenu menu_target_linux
		DisableMenu menu_target_macos_native
		DisableMenu menu_target_macos_universal
?linux
		DisableMenu menu_target_macos_native
		DisableMenu menu_target_macos_universal
		If Not FileType( BlitzMaxpath()+"/bin/bcc.exe" )=FILETYPE_FILE Then DisableMenu menu_target_win
?macosppc
		DisableMenu menu_target_win32
		DisableMenu menu_target_linux
		DisableMenu menu_target_macos_universal
?macosx86
		DisableMenu menu_target_linux
		If FileType( BlitzMaxpath() + "/bin/bcc_ppc" ) <> FILETYPE_FILE Then DisableMenu menu_target_macos_universal
		If Not FileType( BlitzMaxpath()+"/bin/bcc.exe" )=FILETYPE_FILE Then DisableMenu menu_target_win	
?
		
		UpdateWindowMenu _gadget
			
	End Method

	Method Poll(e:TEvent=Null)
		Local	requester:TGUIToolWindow
		Local	src:TGadget,file$
		Local	index:Int
		Local	event:TEvent

		WaitEvent()
		event = CurrentEvent

		CheckHandlers(event)

		Select event.id 

			Case EVENT_GADGETACTION
				Select event.source
					Case toolbar
						Select event.data
							Case TOOLBAR_NEW OpenSource ""
							Case TOOLBAR_OPEN OpenSource "."
							Case TOOLBAR_CLOSE currentpanel.invoke TOOLCLOSE
							Case TOOLBAR_SAVE currentpanel.invoke TOOLSAVE
							Case TOOLBAR_CUT currentpanel.invoke TOOLCUT
							Case TOOLBAR_COPY currentpanel.invoke TOOLCOPY
							Case TOOLBAR_PASTE currentpanel.invoke TOOLPASTE
							Case TOOLBAR_UNDO currentpanel.invoke TOOLUNDO
							Case TOOLBAR_REDO currentpanel.invoke TOOLREDO
							Case TOOLBAR_FIND currentpanel.invoke TOOLFINDREPLACE
							Case TOOLBAR_BUILD BuildCode
							Case TOOLBAR_BUILDANDRUN RunCode
							Case TOOLBAR_BUILDANDDEBUG RunCodeDebug
							Case TOOLBAR_STEP If output output.stepover
							Case TOOLBAR_STEPIN If output output.stepin
							Case TOOLBAR_STEPOUT If output output.stepout
							Case TOOLBAR_STOP If output output.stop
							Case TOOLBAR_CONTINUE If output output.go
							Case TOOLBAR_HOME	helppanel.Home;SelectPanel helppanel
							Case TOOLBAR_BACK currentpanel.invoke TOOLBACK
							Case TOOLBAR_FORWARD	currentpanel.invoke TOOLFORWARD
						End Select

					Case tabbar
						index=event.data
						If index>=0 And index<panels.length
							SelectPanel panels[index]
						EndIf
				End Select

			Case EVENT_APPOPENFILE , EVENT_WINDOWACCEPT
				OpenSource(String(event.extra))

			Case EVENT_APPTERMINATE
				Quit

			Case EVENT_WINDOWACTIVATE
				If (event.source = _gadget) And _modal_tw.count()>0 Then
					TGUIToolWindow(_modal_tw.Last()).Activate
				ElseIf event.source = _gadget Then
					SelectPanel currentpanel
				EndIf

			Case EVENT_WINDOWCLOSE
				src=TGadget(event.source)
				For requester=EachIn eventhandlers
					If src=requester._gadget Then requester.Hide
				Next
				If src=_gadget Then Quit

			Case EVENT_WINDOWMOVE
				If event.source=_gadget Then SnapshotWindow

			Case EVENT_WINDOWSIZE
				If event.source=_gadget
					SnapshotWindow
					split.Resize
				EndIf

			Case EVENT_MENUACTION
				OnMenu event.data,event.extra

		End Select
	End Method

	Method WritePipe(text$)
		output.WritePipe(text$)
	End Method

	Method UpdateOptions()
		SnapshotWindow()
		options_host.SetProjectList projects.ProjectList()
	End Method
	
	Method Restart()
		If Confirm("MaxIDE would like to restart.")
			Quit
		EndIf
	End Method
	
	Field helpcmd$
	Method ShowHelp$(subject$)
		Local l$
		
		If Not quickhelp Return ""
		If subject$=helpcmd$
			l$=quickhelp.link(subject$)
			If l$
				helppanel.go bmxpath+l$
			EndIf
		Else
			helpcmd=subject$
			Return quickhelp.help(subject$)
		EndIf
		Return ""
	End Method

	Method Invoke:Object(command:Int , argument:Object = Null) 
		Local index:Int
		
		Select command		
			Case TOOLPARSEERROR
				ParseError( "", argument.ToString())
				Return "1"
				
			Case TOOLNEXTFILE
				If _modal_tw = Null Then Return "1" ' ignore if there's a modal window!
				If Not currentpanel Return "1"
				index=currentpanel.index+1
				If index=panels.length index=0
				SelectPanel panels[index]
			
			Case TOOLPREVFILE
				If _modal_tw = Null Then Return "1" ' ignore if there's a modal window!							' prev file / tab
				If Not currentpanel Return "1"
				index=currentpanel.index-1
				If index<0 index=panels.length-1
				SelectPanel panels[index]

			Default
				Return Super.Invoke( command, argument)
		EndSelect
	EndMethod
End Type




