'nodeview.bmx
'
' This file contains stuff support a tree of nodes.

SuperStrict

Import BRL.StandardIO
Import "tool.bmx"
Import "host.bmx"

	'BUG: Amending an existing node doesn't update it, it creates a new one anyway.  May be a bug in here, or in the calling code...
	'BUG: Applies to the code nodes and todo nodes...
	'TODO: data nodes need to be tidied up FULL-STOP!

Type TNewDataNode Extends TBase
	' A data node is a structure that holds named data items
	' Those data items can contain zero to many child nodes (kids).
	'
	' This Type should be *pure* data - nothing relating to GUI, etc
	
	'name
	Field	_name$
	Field	_sortname$

	'parent and children
	Field	_parent:TNewDataNode
	Field	_kids:TList = New TList
	
	Method New()
		_typename = "TNewDataNode"
	End Method
	
	Method ToString:String()
		Return "TNewDataNode(" + _name$ + ")"
	End Method

'	Method Compare:Int(o:Object)
'BUG: this compare method causes major slowdown, at least in debug
'		If o = Self Then Return 0
'		Local t:TNewDataNode = TNewDataNode(o)
'		If Not t Then Return 1 ' not a node!
'		If _sortname = t._sortname Then Return 0
'		If _sortname > t._sortname Then Return 1 Else Return -1		
'	End Method
	
	Method SetName(n$)
		_name = n.Trim()
		_sortname$=_name.ToUpper()
	End Method

	Method SetParent(v:TNewDataNode)
		_parent = v
	End Method

	Method Append(v:TNewDataNode)
		v._parent = Self
		_kids.AddLast v
	End Method

	Method InsertBefore(newnode:TNewDataNode , before:TNewDataNode)
		Local	link:TLink
		link=_kids.FindLink(before)
		_kids.InsertBeforeLink newnode, link
		newnode._parent = Self		
	End Method

	Method Query:Int(property:Int)
		' there are no properties to query - yet
	End Method

	Method SortKids( ascending:Int = True )
		Local term:TLink = _kids._head

		Repeat
			Local link:TLink=_kids._head._succ
			Local sorted:Int=True
			Repeat
				Local succ:TLink=link._succ
				If succ = term Exit

				Local cc:Int = TDataNode(link._value)._sortname.Compare( TDataNode(succ._value)._sortname )

				If (cc>0 And ascending) Or (cc<0 And Not ascending)
					Local link_pred:TLink=link._pred
					Local succ_succ:TLink=succ._succ
					link_pred._succ=succ
					succ._succ=link
					succ._pred=link_pred
					link._succ=succ_succ
					link._pred=succ
					succ_succ._pred=link
					sorted=False
				Else
					link=succ
				EndIf
			Forever
			If sorted Exit
			term=link
		Forever
		For Local temp:TNewDataNode = EachIn _kids
			temp.SortKids(ascending)
		Next
		Return
	End Method

	Method Detach() 
		If _parent Then
			' remove myself from my parent
			_parent._kids.remove Self
			' set my link to my parent to Null
			_parent = Null
		EndIf
	End Method

	Method FreeKids()
		Local	n:TNewDataNode
		' Loop through all my kids, and free them all in turn
		For n = EachIn _kids
			n.Free() 
			'NOTE: Detach now occurs in the Free, as does remove from _kids
		Next
		If Not _kids.IsEmpty() Then DebugStop()
	End Method

	Method Free()
		'Free anything that sits beneath this node in the structure		
		FreeKids()
		FreeGUIElements() ' must free gui elements *AFTER* the kids as kids may be freed by this!	
		'detach myself cleanly from my parent
		Detach()  
	End Method

	Method FindKid:TNewDataNode(name$)
		Local n:TNewDataNode
		For n=EachIn _kids
			If n._name=name Then Return n
		Next
		Return Null
	End Method

	Method NodeAfter:TNewDataNode(node:TNewDataNode)
		Local	link:TLink
		If node	Then Link=_kids.FindLink(node)
		If link Then Link=link.NextLink()
		If link Then Return TNewDataNode(link.Value())
	End Method

	Method SyncValues(snap:TNewDataNode) Abstract

	Method Sync(snap:TNewDataNode) 
		' synchronises two sets of data nodes - this routine assumes the lists are in the same sort order
		Local	snapkid:TNewDataNode
		Local	currentkid:TNewDataNode
		Local	kid:TNewDataNode
		Local	t:TNewDataNode

		' top level name must match, otherwise it's nowt to do with us!
		'DebugLog "TNewDataNode.Sync:comparing ~q"+snap._sortname+"~q to ~q"+_sortname+"~q"
		If snap._sortname <> _sortname Then Return
		
		'ensure I sync my own values
		SyncValues snap
		
		' start with the first kid in the current list (if there are any!)
		If Not _kids.IsEmpty() Then currentkid = TNewDataNode(_kids.First() ) 
		
		' loop through all of the kids in the NEW list
		For snapkid = EachIn snap._kids
			
			' Look for the same name in the existing list
			kid=currentkid
			While kid
				If kid._sortname = snapkid._sortname Exit
				kid=NodeAfter(kid)
			Wend
			' If kid is Null, it doesn't exist in the list
			
			' then remove entries from existing kids that are in front the matching entry
			If kid
				While currentkid <> kid
					t=currentkid
					currentkid = NodeAfter(currentkid) 
					'DebugLog "TNewDataNode.Sync:removed "+t._sortname					
					t.free() 
				Wend
			EndIf
			
			'By now:
			' * if we found it, kid is the one we're after
			' * elseif currentkid is null, we can add it the end
			' * else we insert before currentkid
			If kid Then
				'merge the new version of this kid with the one we just found in the existing list
				kid.Sync snapkid				
				' and move on to the next item
				currentkid=NodeAfter(kid)
			Else
				'By this stage, we now know this is a new item in the list
				snapkid.detach ' detach the NEW node from the NEW list
				
				' If we're still in the middle of the list, insert the new node
				If currentkid Then
					InsertBefore snapkid, currentkid
				Else ' otherwise, just append it to the end
					Append snapkid
				EndIf
				
			EndIf
		
		Next ' move on to the next item in the NEW list
		
		' Any entries left over by the time we get here must be old and should be removed
		While currentkid
			t=currentkid
			currentkid=NodeAfter(currentkid)
			t.free()
		Wend
		
		' trigger a Refresh of any views that link to this node
		Refresh()
		
	End Method

	Method _getindex:Int()
		Local	node:TNewDataNode
		Local	i:Int = 0

		If _parent
			For node=EachIn _parent._kids
				If node=Self Then Return i
				i:+1
			Next
		EndIf
	End Method
	
	Method Refresh() Abstract
	Method RefreshMe() Abstract
	Method FreeGUIElements() Abstract

End Type

'---------------------------------------
'TODO: here on down is GUI.  Should split this into true data stuff (TNewDataNode) and the rest...

Const	NODEVIEW_HIDESTATE:Int=0
Const	NODEVIEW_CLOSEDSTATE:Int=1
Const	NODEVIEW_OPENSTATE:Int = 2

Type TDataNodeView Extends TBase
	Field	_node:TGadget
	Field	_state:Int
	Field	host:TGUIHost
	
	Method New() 
		_typename = "TDataNodeView"
	End Method

	Method Hide()
		If _node Then FreeGadget _node
		_node = Null
		_state = NODEVIEW_HIDESTATE
	End Method
	
	Method IsHidden:Int() 
		If _state = NODEVIEW_HIDESTATE Or _node = Null Then
DebugLog "NV.IsHidden=True ("+_state+")"
			Return True
		Else
DebugLog "NV.IsHidden=False ("+_state+")"	
			Return False
		EndIf
		'NOTE: the above used to just compare to OPENSTATE, but that doesn't seem to make sense...
	End Method

	Method IsClosed:Int() 
		If _state <> NODEVIEW_OPENSTATE Or _node = Null Then
DebugLog "NV.IsClosed=True"		
			Return True
		Else
DebugLog "NV.IsClosed=False"		
			Return False
		EndIf
	End Method

	Method SelectNode()
		If _node Then SelectTreeViewNode _node
	End Method	
	
	Method Free() 
		Hide()
		host = Null
	End Method	
	
	Method Gadget:TGadget() 
		Return _node
	End Method
	
	Method Open() 
		_state=NODEVIEW_OPENSTATE 
		If Not _node Then Return'DebugStop() ; Return
		ExpandTreeViewNode _node
	End Method
		
	Method SetNode(gadget:TGadget) 
		If _node Then FreeGadget _node
		_node = gadget
	End Method	
	
	Method GetState:Int() 
		Return _state
	End Method	
	
	Method Close() 
		If _state<>NODEVIEW_CLOSEDSTATE
			_state=NODEVIEW_CLOSEDSTATE
			If _node CollapseTreeViewNode _node
		EndIf
	End Method
			
	Method SetName(name$)
		'TODO: This is sometimes called on the root node, which can never change.
		If _node Then ModifyTreeViewNode _node , name$
		If _state = NODEVIEW_OPENSTATE Then Open()				
	End Method

	Method Delete() 
		If _node Then DebugStop() 
	End Method
	
	Method Invoke:Object(command:Int , argument:Object = Null);End Method
	Method Query:Int(property:Int) ; End Method
	
End Type

Type TDataNode Extends TNewDataNode
	' Each data item can have an ACTIVATE and SELECT "target", "action" and "argument" attached to it.
	' If set, the target is invoked with the required action and argument when the node is invoked.
	'
	' Each data item is linked to one or more "views" which are gadgets that represent the node

	'invocation target, action and argument
	Field	_cmd:Int[]
	Field	_target:TBase[]
	Field	_action:Int[]
	Field	_argument:Object[]
	
	'where is this node accessible from?
	Field	_views:TDataNodeView[]

	Field	host:TGUIHost

	Method New()
		_cmd = New Int[1]
		_target = New TBase[1]
		_action = New Int[1]
		_argument = New Object[1]
		_typename = "TDataNode"
		_views = New TDataNodeView[1]
	End Method

	Function CreateNode:TDataNode(name$)
		Local	node:TDataNode
		node=New TDataNode
		node.SetName(name)
		Return node
	End Function
	
	Method AddNode:TDataNode(name$)
		Local	v:TDataNode
		v=TDataNode.CreateNode(name)
		Append v
		Return v
	End Method

	Method SyncValues(snap:TNewDataNode)
		Local n:TDataNode
		n = TDataNode(snap)
		If Not n Then DebugStop() 'ooops!
		setname n._name
		_cmd = n._cmd
		_target = n._target
		_action = n._action
		_argument = n._argument
	End Method

	Method SetAction(cmd:Int , tool:TBase , act:Int , arg:Object = Null)
		Local i:Int
		Local ni:Int = -1

		' find the command if it already exists
		'NOTE: In future, we might want to allow multiples of the same type?
		For i = 0 To _cmd.length-1
			If _cmd[i]=cmd Then ni=i
		Next

		' need to extend the array for the new action
		If ni = - 1 Then
			ni = _cmd.length
			_cmd = _cmd[..ni + 1]
			_target = _target[..ni + 1]
			_action = _action[..ni + 1]
			_argument = _argument[..ni + 1]
		EndIf

		' update / set the details for this node
		_cmd[ni]=cmd
		_target[ni]=tool
		_action[ni]=act
		_argument[ni]=arg
	End Method

	Method Invoke:Object(command:Int , arg:Object = Null)
		Local i:Int

		For i = 0 To _cmd.length - 1
			If _cmd[i] = command Then
				_target[i].invoke(_action[i],_argument[i])
			EndIf
		Next
	End Method

	Method Query:Int(property:Int);	End Method
	
	Method FindNodeByArgument:TDataNode(arg:Object) 
		'Locates a node by comparing a passed in argument and looking for that argument in the nodes
		'NOTE: This ONLY works with strings at the moment, as it's the only use of it to date!
		Local	r:TDataNode , find$ 

		' Extract the string we are looking for		
		find$ = (String(arg) ).ToLower() 
		If Not find Then Return Null
			
		'Look for it in my own arguments
		For Local a$ = EachIn _argument
			If a.ToLower()=find Then Return Self
		Next
		
		'If still not found, tell my kids to look for it
		For Local n:TDataNode=EachIn _kids
			r=n.FindNodeByArgument(arg)
			If r Return r
		Next
		
		'Didn't find it.  Sorry!
		Return Null
	End Method

	Method IsHidden:Int()
		Local	v:TDataNodeView
		
		'if I have no parent, I must be root, and therefore not hidden
		If Not _parent Return False

		'now I'll check my parents, to see if they are all not open (either hidden or closed)
		For v = EachIn TDataNode(_parent)._views
			If Not v.IsClosed() Then Return False 'NOTE: IsClosed checks for Not Open, i.b. closed & hidden covered by single call
		Next		

		'NOTE: Deliberately don't check my own views, otherwise things get buggy (might be worth investigating?)
		
		Return True
	End Method

	Method Hide()
		Local	n:TDataNode

		' hide the nodes beneath this one
		For n=EachIn _kids
			n.hide
		Next
		
		' then hide this node in the relevant view(s)
		For Local v:TDataNodeView=EachIn _views
			v.hide()
		Next
	End Method

	Method FreeGUIElements()
		Local view:Int
		For view=0 To _views.length-1
			If _views[view] Then _views[view].free()
			_views[view]=Null
		Next
		_views = _views[0..0]
		_cmd = New Int[1]
		_target = New TBase[1]
		_action = New Int[1]
		_argument = New Object[1]
		host = Null
	End Method

	Method FindNode:TDataNode(gadget:TGadget,view:Int)
		Local	n:TDataNode,r:TDataNode
		Local	v:TDataNodeView

		v=getview(view)
		If v And v.Gadget()=gadget Return Self
		For n=EachIn _kids
			r=n.FindNode(gadget,view)
			If r Return r
		Next
	End Method

	Method SetNode(gadget:TGadget , view:Int)
		' this method tells the datanode what gadget it is to attach itself to, e.g. a treeviewnode root
		'TODO: Check that if this is resetting the node, then should make sure it frees pre-existing...
		Local	v:TDataNodeView

		v = getview(view) 
		v.SetNode(gadget)
		open view
	End Method

	Method HighLight(view:Int=-1)
		Local	v:TDataNodeView
		If view = - 1
			For view = 0 To _views.length - 1
				v=GetView(view)
				v.SelectNode()
			Next
		Else
			v=GetView(view)
			v.selectnode() 
		EndIf
	End Method

	Method Open(view:Int = - 1) 
		Local v:TDataNodeView
		If view=-1
			For view = 0 To _views.length - 1
				v=GetView(view)
				v.Open()
			Next
		Else
			v=GetView(view)
			If v.GetState()<>NODEVIEW_OPENSTATE Then RefreshView view
			v.Open() 
		EndIf
	End Method

	Method Close(view:Int = - 1) 
		If view = - 1 Then
			For view=0 To _views.length-1
				_views[view].Close() 
			Next
		Else
			_views[view].Close() 
		EndIf
	End Method

	Method GetView:TDataNodeView(view:Int) 
		If view >= _views.length
			_views=_views[..view+1]
		EndIf
		If _views[view]=Null Then _views[view]=New TDataNodeView
		Return _views[view]
	End Method

	Method Refresh()
		For Local i:Int=0 To _views.length-1
			RefreshView i
		Next
	End Method

	Method RefreshMe() 
		For Local i:Int=0 To _views.length-1
			_views[i].setname _name
		Next		
	End Method

	Method RefreshView(view:Int)
		Local	n:TDataNode,quick:Int
		Local	v:TDataNodeView,parent_view:TDataNodeView
		Local	p:TDataNode
		
'TODO: This function is possibly getting called twice on a refresh according to markcw
'BUG: Memory is being lost whenever this routine is called and it definitely relates to creating treeview nodes

		p = TDataNode(_parent) 				' get the parent node
		If p Then parent_view=p.getview(view)	' get the link to the parents gadgetitem in this view
		v = GetView(view) 					' get the link to my gadgetitem in this view
				
		v.SetName(_name)

		If p Then
		
			v = getview(view)
	
			If v._node
				ModifyTreeViewNode v._node , _name 'BUG: can't modify root nodes!
				If v._state = NODEVIEW_OPENSTATE Then v.Open()
			Else		
				If _name
					parent_view=p.getview(view)
					If parent_view._node
						?MacOS ' MacOS doesn't allow "adding" TreeView nodes by "inserting" them after the last node
							If _GetIndex() > CountTreeViewNodes(parent_view._Node) Then
								If v._node Then FreeGadget v._node
								v._node=AddTreeViewNode(_name,parent_view._node)
							EndIf
						?Not macos
							If v._node Then FreeGadget v._node
							v._node = InsertTreeViewNode(_getindex() , _name , parent_view._node) 						
						?
						If v._state=NODEVIEW_HIDESTATE Then v._state=NODEVIEW_CLOSEDSTATE
						If parent_view._state=NODEVIEW_OPENSTATE Then parent_view.Open()
						quick=False
					EndIf
				EndIf
			EndIf
		EndIf
		
'TODO: Reinstate the "quick" mode that prevents children from being updated unless necessary...
		
		'If there is no parent, or if the parent gadget is open, then we also need to refresh kids...
		'If ( Not parent_view And _kids) Or (_kids And parent_view And parent_view.getstate()=NODEVIEW_OPENSTATE) Then
			For n=EachIn _kids
				n.RefreshView view
			Next
		'EndIf
	End Method

End Type


