' debugtree.bmx
'
' Contains all stuff relating to the debug tree

SuperStrict

Import "datanodes.bmx"
Import "host.bmx"

Type TObj
	Field	addr$,sync_count%,refs%,syncnext%
	
	Method ShouldSync( pDebugTree:TDebugTree )
		If sync_count < pDebugTree.sync_count Then pDebugTree.QueueSync( Self )
	EndMethod
	Method HasSynced( pSync% )
		sync_count = pSync;syncnext = False
	EndMethod
End Type

Type TVar Extends TDataNode

	Field	owner:TDebugTree
	Field	obj:Object

	Method Free()
		If obj owner.RemoveObj obj
		obj=Null
		Super.Free()
	End Method

	Method SetVarName(n$)
		Local	p:Int
		_name=n
' if object ref set addr$ field
		If _name.find("$=")=-1 And _name.find( ":String=" )=-1 And _name.find(")=$")=-1
			p=_name.find("=$")
			If p<>-1
				If TObj(obj) Then
					If TObj(obj).addr <> _name[p+2..] Then 
						TDebugTree.RemoveObj obj
					Else
						TObj(obj).refs:-1
					EndIf	
				EndIf				
				obj=TDebugTree.AddObj(_name[p+2..])
				'Request object dump if we are visible now that
				'we have updated our own object pointer.
				If Not IsHidden() Then Request()
				Return
			EndIf
			p=_name.find("=Null")
			If p<>-1
				FreeKids
				TDebugTree.RemoveObj obj
				obj=Null
			EndIf
		EndIf
	End Method

	Method AddVar(name$)
		Local	v:TVar
		v=New TVar
		v.owner=owner
		Append v
		v.setvarname name
	End Method

	Method SetValue(val:TVar)
		Local	v:TVar,w:TVar,i:Int
' if this is a reference to same object refresh values
		If obj And obj=val.obj
'			sync=val.sync
			If _kids.IsEmpty()
				For v=EachIn val._kids
					AddVar v._name
				Next
			Else
				For v=EachIn val._kids
					If i<_kids.count()
						w=TVar(_kids.ValueAtIndex(i))
						If w w.SetVarName v._name
					Else
						AddVar v._name
					EndIf
					i:+1
				Next
			EndIf
			Refresh
		EndIf
' recurse so all references are updated
If IsHidden() Then DebugLog "SetValue:" + _name + " IsHidden=True" Else DebugLog "SetValue:" + _name + " IsHidden=False"
		If IsHidden() Return				'parent And parent.state=CLOSEDSTATE Return
		For v = EachIn _kids
			'If v.obj <> obj Then
			v.SetValue val
		Next
	End Method

	Method Open(open%=-1)
		For Local kid:TVar = EachIn _kids
			kid.Request()
		Next
		Super.Open(open)
	EndMethod
	
	Method Request()
		If TObj(obj) Then TObj(obj).ShouldSync(owner)
	EndMethod

End Type

Type TScope Extends TVar
	Field	tree:TDebugTree
	Field	file$,line:Int,column:Int

	Method Invoke:Object(command:Int,argument:Object=Null)
		Select command
			Case TOOLACTIVATE
				tree.SelectScope Self,True
		End Select
	End Method

	Method SetScope(s:TScope)
		Local	v:TVar
		file=s.file
		line=s.line
		column=s.column
		s.obj=Self
		SetValue s
	End Method

	Method SetFile(debugtree:TDebugTree,f$)
		Local	p:Int,q:Int,r:Int
		tree=debugtree
		p=f.find("<",0)+1
		q=f.find(">",0)+1
		r=f.find(",",0)+1
		If p And q And r
			file=f[..p-1]
			line=Int(f[p..r-1])
			column=Int(f[r..q-1])
		EndIf
		obj=Self
	End Method
	
	Method Request()
		For Local kid:TVar = EachIn _kids
			kid.Request()
		Next
	EndMethod
	
End Type

Type TDebugTree Extends TVar
	Global	sync_count:Int
	Global	objlist:TList
	Field	host:TGUIHost
	Field	instack:TList
	Field	inscope:TScope
	Field	invar:TVar
	Field	infile$
	Field	inexception$
	Field	firststop:Int
	Field	cancontinue:Int

	Method Reset()
		objlist=New TList
		instack=Null
		inscope=Null
		invar=Null
		infile=""
		inexception=""
		sync_count=0
		firststop=True
		cancontinue = False
		freekids()
		SetStack New TList		
		refresh()
	End Method

	Function AddObj:TObj(addr$)
		Local	o:TObj
		For o=EachIn objlist
			If o.addr=addr
				o.refs:+1
				Return o
			EndIf
		Next
		o=New TObj
		o.addr=addr
		o.refs=1
		objlist.addlast o
		Return o
	End Function

	Function FindObj:TObj(addr$)
		Local	o:TObj
		For o=EachIn objlist
			If o.addr=addr Return o
		Next
	End Function

	Function RemoveObj(obj:Object)		':TObj
		Local	o:TObj
		o=TObj(obj)
		If Not o Return
		o.refs:-1
		If o.refs=0
			objlist.Remove o
		EndIf
	End Function

	Method SyncVars()
		sync_count:+1
		For Local tmpVar:TVar = EachIn _kids
			tmpVar.Request()
		Next
	End Method
	
	Method QueueSync( pObj:TObj )
		If Not pObj Then Return
		'Sync as soon as the debug pipe is clear
		'(see TGUIOutput.SendDumpRequests()).
		pObj.syncnext = True
	EndMethod

	Method SetStack(list:TList)
		Local	scope:TScope
		Local	openscope:TScope
		Local	s:TScope
		Local	count:Int,i:Int

		count=_kids.count()			'root.varlist.count()
		For scope=EachIn list
			If i>=count
				Append scope		'root.Append scope
				s=scope
			Else
				s=TScope(_kids.ValueAtIndex(i))
'MARK: Added code from BRL IDE 1.20F to fix synchronization bug? - Yan
				If s._name=scope._name
					s.SetScope scope
					scope.Free
				Else
					While _kids.count()>i
						s=TScope(_kids.ValueAtIndex(i))
						s.free
					Wend

					Append scope
					s=scope
					count=i+1
				EndIf
'End of added code
			EndIf

			If firststop
				If host.IsSourceOpen(s.file) openscope=s
			Else
				openscope=s
			EndIf

			i:+1
		Next

		While _kids.count()>i
			scope=TScope(_kids.ValueAtIndex(i))
			scope.free
		Wend

		If list.isempty() Return
		If Not openscope openscope=TScope(list.First())
		If openscope SelectScope openscope,True

		Refresh
		firststop=False
	End Method

	Method SelectScope(scope:TScope,open:Int)
		If Not scope Return
		host.SetMode host.DEBUGMODE
		If scope.file host.DebugSource scope.file,scope.line,scope.column
		scope.Open
	End Method

	Method ProcessError$(line$)
		Local	p:Int

		While line[..1]=">"
			line=line[1..]
		Wend
		If Not line Return ""

		If line[..2]<>"~~>" Return line

		line = line[2..]
		
		If invar
			If line="}"
				SetValue invar		'root
				invar.Free
				invar=Null
			Else
'				If Not invar.name
'					invar.name=line
'				Else
					invar.AddVar line
'				EndIf
			EndIf
			Return ""
		EndIf

		If instack
			If line="}"
				sync_count:+1
				SetStack instack
				instack=Null
				inscope=Null
				SyncVars
				If inexception
					inexception = TranslateMessage(inexception)
					Local oldapptitle$ = AppTitle
					AppTitle = LanguageString("error_message/runtime_error_title")
					Notify inexception
					AppTitle = oldapptitle$
					host.SetStatus inexception

					inexception=""
				EndIf
				Return ""
			EndIf

			If infile	
				If line="Local <local>"
				Else
					inscope=New TScope
					inscope._name=line
					inscope.owner=Self
					instack.AddLast inscope
				EndIf
				If inscope inscope.setfile Self,infile
				infile=""
				Return ""
			EndIf

			If line.find("@",0)=0 And line.find("<",0)>-1
				infile=line[1..]
			Else
				If inscope inscope.AddVar line
			EndIf

			Return ""
		EndIf

		If line[..20] = "Unhandled Exception:"
			inexception=line
			host.WritePipe "t"
			cancontinue=False
			Return ""
		EndIf

		If line = "StackTrace{"
			instack=New TList
			Return ""
		EndIf

		If line = "Debug:" Or line = "DebugStop:"
			host.WritePipe "t"
			cancontinue=True
			Return ""
		EndIf

		If line[..11] = "ObjectDump@"
			p=line.find("{")
			If p=-1 Return line
			line=line[11..p]
			invar=New TVar
'			invar.obj=AddObj(line)
			invar.obj=FindObj(line)	'simon come here
			invar.owner=Self
			Return ""
		EndIf

	End Method

	Function CreateDebugTree:TDebugTree(host:TGUIHost)
		Local	d:TDebugTree
		objlist=New TList
		d=New TDebugTree
		d.owner=d
		d.SetName "@@debug_tree/title"
		d.host=host
		d.Open
		Return d
	End Function
End Type

