'code editor
'
'TODO: This should be an imported Module
'SuperStrict
'Import "host.bmx"
'Import "icons.bmx"
'Import "diff.bmx"
'Import "requesters.bmx"		' provides all of the MaxIDE specific requesters

Const FILTER_Files$ = "@@bb:bmx,bb,bbdoc;" + ..
				"@@c:c,cpp,cxx,h,hpp,m;" + ..
				"@@web:css,html,htm,xml,css,js,vbs;" + ..
				"@@doc:txt,doc,ini,cmd,bat,sh;" + ..
				"@@code:s,fs,vs,vp,plist,p,fp,glsl,rc;" + .. ' p= pascal. s=source. plist=macos xml appinfo, fp=foxpro?
				"@@zap:zap;" + ..
				"@@bak:bak;" + ..
				"@@all:*"

'Used by auto-completion
'TODO: Win32 specific code for cursor pos in textarea
?Win32
	Extern "Win32"
		Function GetCaretPos_(lpPoint:Byte Ptr) = "GetCaretPos@4"
	End Extern
?


Type TCodePanel Extends TGUIEditor Abstract' possibly rename as TGUITextEditor

' This type provides core text editing functionality.  It can then be extended
' to provide further functionality, e.g. the BMX parser, and potentially others,
' e.g. C++, HTML, SQL, etc.

	Global CodeTypeExt:String[][]   
	Global CodeTypeCreate:Byte Ptr[]
	Global _autobackup:Int
	Global _autotidymode:Int
	Global _keywords:TMap
	Global _casesensitivekeywords:Int = False
	
	Method New() 
		If CodeTypeCreate.length = 0 Then
			CodeTypeCreate = [Byte Ptr TUnknownCodePanel.Create]
			CodeTypeExt = [ [""] ]
		EndIf
	End Method
	
	Function Initialise(host:TGUIHost,autobackup:Int,autotidymode:Int)
		If Not findrepreq Then findrepreq = TFindReplaceRequester.Create(host) 
		_autobackup = autobackup
		_autotidymode = autotidymode
		_keywords = New TMap
	End Function
	
	Function Register:Int(id:Int , ext$[] , createfunc:Byte Ptr) 
		' if  the id is the highest so far, extend the array and carry on
		If id+1 > CodeTypeCreate.length Then
			CodeTypeCreate = CodeTypeCreate[..id+1]
			CodeTypeExt = CodeTypeExt[..id+1]
		' if the id already has a createfunc set in the array, error!
		ElseIf CodeTypeCreate[id] Then
			Return False
		EndIf
		
		' next check if any of the extensions are already in use
		For Local a$ = EachIn ext$
			If CodeTypeFromExt(a$) Then Return False
		Next
		
		' we're clear to register it now
		CodeTypeExt[id] = ext
		CodeTypeCreate[id] = createfunc
		Return True
	End Function

	Function CodeTypeFromExt:Int( ext:String)
		For Local i:Int = 0 Until CodeTypeExt.Length
			For Local s:String = EachIn CodeTypeExt[i]
				If ext = s Then Return i
			Next
		Next
		Return 0
	EndFunction

	Method AddKeyword(command$)
		' this should be over-ridden if for some reason the structure of commands doesn't fit
		Local longcommand$ = command
		If command.find("(") >= 0 Then command = command[..command.find("(")]
		If command.find(" ") >= 0 Then command = command[..command.find(" ")]
		_keywords.Insert command$,longcommand$
	End Method
	
	Method IsKeyword:Int(command$)
		If Not _casesensitivekeywords Then command$=Lower(command$)
		Return _keywords.Contains(command$)
	End Method

	Global	findrepreq:TFindReplaceRequester ' find and replace requester used by all editors
	
	Field	codetype:Int=CODETYPE_UNKNOWN ' extending types should change this. The Field itself must not be overidden.

	'the host, the textarea, the editmenu, and the options
	Field	host:TBMXHost 'TODO: host should be TGUIHost
	Field	options_code:TCodeOptions ' code specific options
	Field	options_editor:TEditorOptions ' generic editor options
	Field	options_keys:TKeyOptions ' holds all configurable key codes
	Field	_textarea:TGadget		' possibly move this to TGUIEditor?
	Field	editmenu:TGadget
	Field	statustext$
	Field	dirty:Int = True

	Field	filesrc$		' the version of the source code in the file
	Field	cleansrc$		' the version of the source up to the last edit?
	Field	cleansrcl$	' as above, but lower case

	Field	helpcmd$
	Field	seek$

	Field	current:TDiff
	Field	deferpos:Int

	Field	codenode:TDataNode
	Field	todonode:TDataNode
	Field	dirtynode:Int

	'undo functionality
	Field	undolist:TList
	Field	redolist:TList

	'cursor stuff
	Field	cursorpos:Int,cursorlen:Int,cursorline:Int
	Field	oldpos:Int,oldlen:Int

	'codepanel goto history
	Field	backstack:IntStack
	Field	forwardstack:IntStack

	' auto complete
	Field	Autoc:TGUIAutoComplete

	' edit menu lists
	Field	editmenulist:TList = New TList
	Field	editmenusellist:TList = New TList

	Field	KEY_AUTOC% ' the index of the auto-complete hotkey
	
	Method Gadget:TGadget()
		Return _textarea
	End Method

	Function IsNotAlpha:Int(c:Int)
		If c<48 Return True
		If c>=58 And c<65 Return True
		If c>=91 And c<95 Return True
		If c>=96 And c<97 Return True
		If c>=123 Return True
	End Function

	Function IsNotAlphaNumeric:Int(c:Int)		'lowercase test only
		If c<47 Return True
		If c>=58 And c<95 Return True
		If c=96 Return True
		If c>=123 Return True
	End Function

	Function IsntAlphaNumericOrQuote:Int(c:Int)		'lowercase test only
		If c=34 Return False
		If c<47 Return True
		If c>=58 And c<95 Return True
		If c=96 Return True
		If c>=123 Return True
	End Function

	Function WordAtPos$(a$,p:Int)
		Local	c:Int,q:Int,r:Int,n:Int

		If a.Length - p < 0 Then Return Null ' removed equality test as seemed excessive and prevented help work on clean file
		If p=0 Then Return Null

	' string literal
		q=a.findlast(EOL$,a.length-p)
		If q=-1 q=0
		For q=q To p-1
			If a$[q]=34 n=Not n;r=q
		Next
		If n
			q=a.Find(Chr(34),r+1)+1
			If q=0 q=a.length
			Return a[r..q]
		EndIf
	' alphanumeric
		p=Min(p,a.length-1)	'simon was here - crash when checking at last char
		For p=p Until 0 Step -1	'simon was here unto->to
			If IsNotAlpha(a$[p]) Continue
			Exit
		Next
		For q=p-1 To 0 Step -1
			If IsNotAlpha(a$[q]) Exit
		Next
		For r=p To a.length-1
			If IsNotAlpha(a$[r]) Exit
		Next
		Return a$[q+1..r]
	End Function

	Function IsFirstCharOnLine:Int(src$,pos:Int)
		Local	i:Int,c:Int
		For i=1 To pos
			c=src[pos-i]
			If c=10 Or c=13 Return True
			If c>32 Return False
		Next
		Return True
	End Function

	Method AutoComplete(s:String)
		If Autoc Then
			If s.Length <= 0 Then s = Autoc.GetSelected()
			If s.Length > 0 Then
				SetTextAreaText _textarea, s, cursorpos-Autoc.AutocInt, Autoc.AutocInt
				SelectTextAreaText _textarea,cursorpos+s.Length-Autoc.AutocInt,0
				UpdateCursor
				UpdateCode
			EndIf
		EndIf
	End Method

	Method ForceHighlight() 
		If _textarea Then
			SetCode( TextAreaText(_textarea) )
		EndIf
	End Method

	Method Debug(line:Int,column:Int)
		SelectTextAreaText _textarea,line-1,0,TEXTAREA_LINES
		SelectTextAreaText _textarea,line-1,1,TEXTAREA_LINES
	End Method

	Method Edit()
		SelectTextAreaText _textarea,cursorpos,0,TEXTAREA_CHARS
	End Method

	Method Refresh()
		Local	rgb:TColor
		Local	src$

		' store the source
		src = cleansrc

		' update the textarea
		LockTextArea _textarea
		SetTextAreaTabs _textarea,options_code.tabsize
		'BUG: need to temporarily set the text to "" to force this to work due to a BMX bug
		SetTextAreaFont _textarea,options_editor.editstyle.font
		rgb=options_editor.editstyle.bg
		SetTextAreaColor _textarea , rgb.red , rgb.green , rgb.blue , True
		'BUG: the following two lines cause the update not to work, due to a BMX bug
		rgb=options_editor.styles[options_editor.NORMAL].color
		SetTextAreaColor _textarea , rgb.red , rgb.green , rgb.blue , False
		UnlockTextArea _textarea

		' redo the textarea
		cleansrc=""
		cleansrcl=""
		cursorpos = 0
		SetCode(src)
	End Method

	Method FindAndReplace()
		Local	p:Int,l$,a$
		p=TextAreaCursor(_textarea,TEXTAREA_CHARS)
		a$ = WordAtPos(cleansrc , p)
		findrepreq.SetTarget( Self)
		findrepreq.ShowFind( a)
	End Method

	Method FindNext:Int(s$, p:Int=0, showerror:Int=True)
		Local	l$ , notfound:Int
		Local	wrap:Int
		Global	ignore:Int
		If s seek=s Else s=seek
		'FIXME: hack to allow a kind of goto (so we dont have to recurse FindNext())
		p=TextAreaCursor(_textarea,TEXTAREA_CHARS)
		p:+ TextAreaSelLen(_textarea , TEXTAREA_CHARS) 
		wrap = ButtonState(findrepreq._wraparound)
		#continue_find
		While (notfound < 3) 			
			If p >= cleansrc.Length Then
				If Not wrap Then
					ignore = 0
					If showerror Then Notify "@@message/find_error|"+s$
					Return False
				Else
					ignore = 0
				EndIf
			ElseIf p < ignore Then
				ignore = 0
			EndIf
			If ButtonState(findrepreq._matchcase) Then
				' case sensetive find
				p=cleansrc.Find(s$,p+1)
				If p=-1 And wrap Then p=cleansrc.Find(s$)
			Else
				' case in-sensetive find
				l$=s.toLower()
				p=cleansrcl.Find(l$,p)
				If p=-1 And wrap Then p=cleansrcl.Find(l$)
			EndIf
			If p=-1 Then
				ignore = 0
				If showerror Then Notify "@@message/find_error|"+s$
				Return False
			Else	'** match found! now satisfy some more options

				' WHOLE WORD
				If ButtonState(findrepreq._matchwords) Then
					' match whole words
					Local a:Int,b:Int
					If p-1 > 0 Then
						a = Separ.Find( Chr(cleansrc[p-1])) <> -1
					ElseIf p = 0 Then
						a = True
					EndIf
					If p+s.Length < cleansrc.Length Then
						b = Separ.Find( Chr(cleansrc[p+s.Length])) <> -1
					ElseIf p+s.Length = cleansrc.Length Then
						b = True
					EndIf
					If (Not a) Or (Not b) Then
						' continue searching
						p :+ s.Length
						'notfound :+ 1
						Continue continue_find
					EndIf
				EndIf
'TODO: Find Ignore-Strings doesnt work, probably needs a reworking of FindNext()
Rem
				' IGNORE STRINGS
				If ButtonState(host.findrepreq._istrings) Then
					If p-1 > 0 Then
						' trace backwards
						Local i:Int = p - 1
						Local c:Byte
						While i > ignore
							c = cleansrc[i]
							If c = Asc("~q") Then
								ignore = i
								Local ii:Int = cleansrc.Find("~q", i+1)
								Local eol:Int
								If ii > p + s.Length Then
									eol = cleansrc.Find("~n",i+1)
									If eol<>-1 And eol < ii Then Exit
									ignore = ii
									p = ii+1
									'notfound :+ 1
									Continue continue_find
								EndIf
								Exit
							ElseIf c = Asc("~r") Or c = Asc("~n") Then
								Exit
							EndIf
							i :- 1
						Wend
					EndIf
				EndIf
EndRem
				' IGNORE COMMENTS
				If ButtonState(findrepreq._icomments) Then
					If p-1 > 0 Then
						' trace backwards for single line comments
						Local i:Int = p - 1
						Local c:Byte, goteol:Int = False
						While i >= 0
							c = cleansrcl[i]
							If (Not goteol) And (c = Asc("'")) Then
								' inside comment, so continue searching
								i = cleansrcl.Find("~n",i+1)
								If i >= 0 Then
									p = i+1
								Else
									p :+ s.Length
								EndIf
								'notfound :+ 1
								Continue continue_find
							ElseIf c = Asc("r") And cleansrcl.Find( "rem", i)=i Then
								' check for eol before Rem
								If (cleansrcl.Find("~n",i-1)=i-1 Or cleansrcl.Find("~r",i-1)=i-1) Then
									' search for EndRem
									i = cleansrcl.Find( "endrem", p + s.Length + 1)
									If i > 0 Then
										' check for eol before EndRem
										If (cleansrcl.Find("~n",i-1)=i-1 Or cleansrcl.Find("~r",i-1)=i-1) Then
											' inside multi comment, so continue searching
											' skip Rem-EndRem block
											i = cleansrcl.Find("~n",i+6)
											If i >= 0 Then
												p = i+1
											Else
												p = i + 6
											EndIf
											'notfound :+ 1
											Continue continue_find
										EndIf
									EndIf
								EndIf
								Exit
							ElseIf c = Asc("~r") Or c = Asc("~n") Then
								goteol = True
							EndIf
							i :- 1
						Wend
					EndIf
				EndIf
				SelectTextAreaText _textarea,p,s.Length,TEXTAREA_CHARS
				UpdateCursor
				Return True
			EndIf
			' make sure we always exit from the loop
			Exit
		Wend
		'ignore = 0
		Return False
	End Method
	
	Method ReplaceAll:Int(s$, r$, inSelection%=False)
		Local	p:Int , p_o:Int , p_n:Int , c:Int , cl:Int , l$ , count:Int, ss:Int, sl:Int
		
		LockTextArea _textarea
		'store cursor
		c = TextAreaCursor(_textarea , TEXTAREA_CHARS)
		
		If inSelection
			sl = TextAreaSelLen(_textarea)
			
			If Not sl
				UnlockTextArea _textarea
				Return 0
			EndIf
			
			ss = c
		Else
			sl = TEXTAREA_ALL
		EndIf
						
		' not case sensetive ?
		If ButtonState(findrepreq._matchcase) Then
			' case sensetive
			l = TextAreaText(_textarea, ss, sl)		
		Else
			' case in-sensetive
			l = TextAreaText(_textarea, ss, sl).ToLower()		
			s = s.ToLower() 
		EndIf			
		
		Repeat
			'p is current position in the text area
			'p_o is the position in the original text l
			'p_n is the position in l where s has just been found
'			If ButtonState(host.findrepreq._matchcase) Then
'				' case sensetive
'				l = TextAreaText(_textarea, ss, sl)
'			Else
'				' case in-sensetive
'				l = TextAreaText(_textarea, ss, sl).ToLower()
'			EndIf
		
			If (p_o + s.Length) >= l.Length Then Exit
			
			' look for the string in the original source
			p_n = l.Find(s$,p_o)
			If p_n = - 1 Then Exit ' didn't find it
			
			' work out the position in the textarea by adding p_n - p_o chars to our position
			p = p + p_n - p_o
			' then set p_o (our next search char) to be the length of s$ on from where we just found s
			p_o = p_n + s.length
			
			If ButtonState(findrepreq._matchwords) Then
				' match whole words
				Local a:Int,b:Int
			
				If p_n-1 > 0 Then
					a = Separ.Find( Chr(l[p_n-1])) <> -1
				ElseIf p_n = 0 Then
					a = True
				EndIf
				
				If p_n+s.Length < l.Length Then
					b = Separ.Find( Chr(l[p_n+s.Length])) <> -1
				ElseIf p_n+s.Length = l.Length Then
					b = True
				EndIf
				
				If (Not a) Or (Not b) Then
					' continue searching
					'p :+ s.Length
					Continue
				EndIf
			EndIf

			SelectTextAreaText _textarea, ss + p, s.length
			UpdateCursor 'DO NOT UPDATECODE HERE! it messes up the undo process, and is unecessary as it will happen later on!
			SetTextAreaText _textarea, r, ss + p, s.length
			
			If p < c Then c = c + (r.length - s.length)
			
			If inSelection
				ss = c
				sl :+ (r.length - s.length)
			EndIf 
			
			p = p + r.length
			SelectTextAreaText _textarea, ss + p, 0
			UpdateCursor
			UpdateCode
			count :+ 1
		Forever

		'restore cursor
		SelectTextAreaText _textarea,c,0
		UpdateCursor
		UnlockTextArea _textarea
		Return count
	End Method

	Method FindReplace:String(r$)
		Local	p:Int , n:Int

		p=TextAreaCursor(_textarea,TEXTAREA_CHARS)
		n=TextAreaSelLen(_textarea,TEXTAREA_CHARS)
		If n=0 Return Null
		SetTextAreaText _textarea,r$,p,n
		SelectTextAreaText _textarea,p+r.Length,0
		UpdateCursor
		UpdateCode
		Return r$
	End Method

	Method ReadSource(stream:TStream)
		Local	src$

		src=LoadString(stream)
		'LoadText doesn't rewind the stream after reading the unicode-tag when loading a non-unicode file.
		If (src[..2] = Chr($FE) + Chr($FF)) .. ' big endian UTF-16
		Or (src[..2] = Chr($FF) + Chr($FE)) .. ' little endian UTF-16
		Or (src[..3] = Chr($EF) + Chr($BB) + Chr($BF)) Then ' UTF-8
			If stream.Seek(0) = -1 Then ' rewind the stream
				Notify "Unicode files can not be loaded from a non-seekable source.", True
			Else
				src = LoadText(stream)
			EndIf
		EndIf
		src=src.Replace(Chr(13),"")
		src=src.Replace(Chr(11),"")
		LockTextArea _textarea
		SetTextAreaText _textarea,src
		UnlockTextArea _textarea
		filesrc=TextAreaText(_textarea)
		cleansrc=""
		cleansrcl=""
		ActivateGadget _textarea
	End Method

	Method SaveSource:Int(file$)
		Local	stream:TStream , src$

		If _autobackup And FileType(file) = FILETYPE_FILE
			DeleteFile file+".bak"
			RenameFile file,file+".bak"
		EndIf
		stream=WriteFile(file)
		If Not stream
			Notify "@@message/save_error|" + file
			Return False
		EndIf
		src=TextAreaText(_textarea)
		filesrc=src
		src=src.Replace$(Chr(13),Chr(10))
		src=src.Replace$(Chr(11),"")
		Local txt$=src.Replace$(Chr(10),Chr(13)+Chr(10))
		SaveText txt, stream
		CloseFile stream
		path=host.FullPath$(file)
		dirty=True
		SetDirty False
		host.AddRecent(path$)
		host.AddOpenFile(path$)
		Return True
	End Method

	Method Save:Int()
		If dirty Then SaveSource(path)
		Return True
	End Method

	Method UpdateCode(makeundo:Int=True)
		Local	cpos:Int,src$
		Local	d:TDiff
'DebugLog "UpdateCode:Enter"
		src=TextAreaText(_textarea)
		d=CalcDiff(src)
		If d Then
			If makeundo Then
				AddToUndoList d
				redolist.Clear
			EndIf
			SetCode src,d
			dirtynode=True
		Else
			'TODO: DEFERPOS : remove usage of deferpos (only changed in TBMXCodePanel)
			If deferpos Then SetCode src
		EndIf
		RefreshDependents()
'DebugLog "UpdateCode:Exit"
	End Method

	Method AddToUndoList(d:TDiff) 
		'if this is just another keypress, add it to the most recent one
		If undolist.count()>0 And TDiff(undolist.Last()).AppendIfMinor(d) Then Return
		
		' if we are applying a limit to the undo size, remove old items
		If options_code.undosize > - 1 Then 
			While undolist.count() > options_code.undosize 
				undolist.RemoveFirst
			Wend
		End If
		
		' and add the item to the list
		undolist.AddLast d
	End Method

	Method HighlightRegion(src$ , p:Int , p1:Int) Abstract

	Method SetCode:Int(src$ , diff:TDiff = Null)
		Local result:Int=False

		CheckDirty src
		'TODO: DEFERPOS : remove usage of deferpos (only changed in TBMXCodePanel)
		If src = cleansrc And Not deferpos And Not diff Then
			result=True
		ElseIf Not host.quickhelp Or Not options_code.syntaxhighlight Then
			'TODO: remove these checks as not apparently necessary, and should probably always update cleansrc, etc
			result=True
		EndIf
		If Not result Then
			Local p:Int , p1:Int
			Local cpos:Int , clen:Int
			
			CalcHighlightRegion(src.tolower() , diff , p , p1)
			
			' lock the textarea
			cpos = TextAreaCursor(_textarea , TEXTAREA_CHARS)
			clen = TextAreaSelLen(_textarea) 
			LockTextArea _textarea
			
			' Highlight it
			HighlightRegion(src , p , p1) 
			
			' unlock and reset where we are
			UnlockTextArea _textarea
			SelectTextAreaText _textarea,cpos,clen,TEXTAREA_CHARS
		EndIf
		cleansrc=src
		cleansrcl = src.ToLower() 		
		Return result
	End Method

	Method RefreshDependents() 
		'TODO: Need a neater way of refreshing dependent TTools
'DebugLog "RefreshDependents:Enter("+dirtynode+")"
		If dirtynode Then
			GetNode().Refresh
			' bregors todo-tree-list
			GetTodoNode().Refresh
			dirtynode=False
			'TODO: remove usage of deferpos (only changed in TBMXCodePanel)
			' can't see that this should be happening here, and it causes a recursive loop atm!
			'If deferpos Then UpdateCode	'SetCode cleansrc
		EndIf
'DebugLog "RefreshDependents:Exit"		
	End Method

	Method Help()
		Local	p:Int,a$
		
		p=TextAreaCursor(_textarea,TEXTAREA_CHARS)
		a$ = WordAtPos(cleansrc , p) 
		a$ = host.ShowHelp(a$) 
		If a$<>"" Then statustext=a$ ; UpdateStatus
	End Method

	Method Undo()
		Local	d:TDiff

		If Not undolist.isempty() Then

			' take the diff out of the undo list, and put it in the redo list
			d=TDiff(undolist.RemoveLast())
			redolist.AddLast d

			' replace the added text (d.add) at the correct position (d.pos1) with the deleted text (d.del)
			SetTextAreaText _textarea , d.del , d.pos1 , d.add.length

			' move the cursor, and select what was selected before
			SelectTextAreaText(_textarea , d.pos , d.count)

			' update our view of the view
			dirtynode = True			
			SetCode TextAreaText(_textarea) , d		
			UpdateCursor
			RefreshDependents			

		EndIf

	End Method

	Method Redo()
		Local	d:TDiff

		If Not redolist.isempty() Then

			' take the diff out of the redo list, and put it in the undo list
			d=TDiff(redolist.RemoveLast())
			AddToUndoList d

			' replace the ORIGINALLY deleted text (d.del) at the correct position (d.pos1) with the ORIGINALLY added text (d.add)
			SetTextAreaText _textarea , d.add , d.pos1 , d.del.length

			' move the cursor, and select what was selected before
			SelectTextAreaText(_textarea , d.pos0 , d.count0)

			' update our view of the view
			dirtynode=True			
			UpdateCursor
			SetCode TextAreaText(_textarea) , d
			RefreshDependents

		EndIf
	End Method

	Function FirstDiff:Int(s0$,s1$)
		Local	n:Int,i:Int
		n=Min(s0.length,s1.length)
		For i=0 Until n
			If s0[i]<>s1[i] Return i
		Next
		Return n
	End Function

	Function LastDiff:Int(s0$,s1$)
		Local	n:Int,i:Int,j:Int
		n=Min(s0.length,s1.length)
		i=s0.length-1
		j=s1.length-1
		While n>0
			If s0[i]<>s1[j] Exit
			i:-1;j:-1;n:-1
		Wend
		Return i+1
	End Function

	Method CalcDiff:TDiff(src$)
		'TODO: This would be better if moved into TDiff, although we'd need to pass a whole bunch of data cross...

		Local	d:TDiff

		d = New TDiff

		' store the new cursor position
		d.pos0 = cursorpos
		d.count0 = cursorlen

		' and the previous cursor position
		d.pos = oldpos
		d.count = oldlen

		' the following works in MOST instances:
		d.del = cleansrc[oldpos..oldpos + oldlen]		' what was the old text
		d.add = src[oldpos..cursorpos + cursorlen]		' what is the new text
		d.pos1 = oldpos							' where in the text these changes relate to

		' Now we check for various special scenarios

		' Firstly, if the overall length is completely unchanged, then some kind of
		' substitution / overwrite has occurred
		If src.length = cleansrc.length Then

			If cursorpos > oldpos Then
				d.del = cleansrc[oldpos..cursorpos]
				d.add = src[oldpos..cursorpos]
				d.pos1 = oldpos

			ElseIf cursorpos < oldpos Then
				' this scenario probably should never happen in editing, but might in code...
				d.del = cleansrc[cursorpos..oldpos]
				d.add = src[cursorpos..oldpos]
				d.pos1 = oldpos

			Else
				' if the cursorpos and oldpos are the same, then the "normal" routine will work.
			EndIf

		' Is it a deletion of text using delete key?  (could also be backspace on a selected area of text)
		ElseIf cursorlen=0 And cursorpos = oldpos Then
			d.del = cleansrc[cursorpos..cursorpos + cleansrc.length - src.length]
			d.add = "" ' ensure no chars added!
			' don't need to change d.pos1 as chars should get inserted at oldpos

		' Is it a deletion of text using backspace?
		ElseIf cursorlen=0 And cursorpos < oldpos Then 'cursorpos<=oldpos and cursorlen<=oldlen
			d.del = cleansrc[cursorpos..cursorpos + cleansrc.length - src.length]
			d.add = ""
			d.pos1 = cursorpos	' this is INTENTIONALLY different! Changes all relate to where the cursor ENDS UP, not where it WAS!
			
		EndIf

		' and if we end up with no change, well, just forget it!
		If d.del.length=0 And d.add.length=0 Then d=Null

		Return d
	End Method

	Method CalcHighlightRegion:Int(src$ , diff:TDiff = Null , p:Int Var , p1:Int Var)
		Assert src=src.tolower() ' src is assumed to be lowercase on input!
		Local r0:Int , r1:Int , q:Int
				
		' calculate highlight region
		If deferpos
			Assert False
			'TODO: DEFERPOS : this is quite possibly never used!		
			p=deferpos
			p1=p
			deferpos=0
		Else
			If diff
				If diff.add.length Then
					' if it's an addition, then we consider just the area that has been changed
					p = diff.pos
					p1 = p + diff.add.length
				ElseIf diff.del.length Then
					' if it's a deletion, then we consider where the cursor has ended up
					p = diff.pos0
					p1 = p
				Else
					Assert False
					Return False
				EndIf
			Else
				' Possibly used only when opening a file for the first time? (in which case, no need to "calc" the region!)
				p = firstdiff(src , cleansrc) 
				p1 = lastdiff(src , cleansrc)
			EndIf
			
			'Next we do a couple of checks that shouldn't be unecessary, but just in case...
			' if our end char is less than our start char, do a swap
			If p1 < p Then
				q = p1
				p1 = p
				p = q
			EndIf
						
			' if detected number of chars is less than those between p and p1, then force p,p1 bigger
			q = src.length - cleansrc.length
			If p1-p<q Then p1=p+q
			
		EndIf
		
DebugLog "Change is from " + p + " to " + p1
	
' round to whole lines for highlighting purposes
		' find the EOL before p
		q = src.findlast(EOL , src.length - p)
		If q > - 1 Then p = q Else p = 0
		
		' find the EOL after p1
		q = src.find(EOL , p1) 
		If q > - 1 Then p1 = q Else p1 = src.length
		
' Next we look for any multi-line rem blocks nearby that could cause us problems...	

		' check to see if we are inside a rem block, if so:
		'  - make our start position = start of rem block
		'  - make our end position = end of rem block / end of file
		q = IsRemmed(src , p)
		If q = - 1 Then q = IsRemmed(src , p1)
		If q > - 1 Then		
			If q < p Then p = q
			q = FindEndRem(src , p1 , True) ' returns last char pos if no endrem
			If q > p1 Then p1 = q
		EndIf
		
		' check to see if we WERE inside a rem block before, if so:
		'  - make our start position = start of OLD rem block
		'  - make our end position = end of OLD rem block / end of file
		q = IsRemmed(cleansrcl , p)
		If q = - 1 Then q = IsRemmed(cleansrcl , p1) 
		If q > - 1 Then
			If q < p Then p = q
			r0 = FindEndRem(src , p1 , True) ' returns last char pos if no endrem
			r1 = FindRem(src , p1) ' returns last char pos if no rem
			q = Min(r0 , r1) 
			If q > p1 Then p1 = q
			If p1 > src.length Then p1 = src.length
		EndIf

		' check to see if there's an EndRem in our highlighting area, if so:
		'  - make our end position either the next Rem, or the end of the file
		q = FindEndRem(src , p) 
		If q < p1 Then
			q = FindRem(src , q) 
			If q > p1 Then p1 = q
		EndIf

		' check to see if there's a Rem in our highlighting area, if so:
		'  - make our end position either the next Rem, or the end of the file
		q = FindPrevRem(src , p1)
		If q >= p Then
			q = FindEndRem(src , q , True) 
			If q > p1 Then p1 = q
		EndIf
		
		Return True
	End Method

	Method SetLocked( bool:Int )
		Local locked:TCodePanel
		locked=host.BuildPanel()
		If locked And locked<>Self Then locked.SetLocked False
		name=StripDir(path)
		If (bool)
			host.options_build.buildfile=path.ToLower()
			name=LanguageString("main_window/build_lock",name)
		EndIf
		If (dirty) name:+"*"
		host.RefreshPanel Self
	End Method

	Method CheckDirty(src$)
		Local	bool:Int
		bool = True
		If src = filesrc Then bool = False
		' MOT removed check for And undolist.isempty() in the above
		' caused error in refreshing the dirty flag, e.g. after changing IDE options
		' and doesn't make any sense (undolist is not cleared on saving a file, only on Undo)
		' and also is a bug in the original IDE.
		SetDirty bool
	End Method

	Method SetDirty( bool:Int )
		If dirty<>bool Then
			dirty=bool
			name=StripDir(path)
			If (host.options_build.buildfile=name) Then name = LanguageString("main_window/build_lock",name)
			If (dirty) name:+"*"
			host.RefreshPanel Self
			PollSystem
		EndIf
	End Method

	Method UpdateCursor()
		Local	l:Int
'DebugLog "UpdateCursor:Enter"		
		oldpos=cursorpos
		oldlen=cursorlen
		cursorpos=TextAreaCursor(_textarea,TEXTAREA_CHARS)
		cursorlen=TextAreaSelLen(_textarea,TEXTAREA_CHARS)
		If cursorpos<>oldpos Or cursorlen<>oldlen
			l=TextAreaLine(_textarea,cursorpos)+1
			If l<>cursorline Then dirtynode=True 'RefreshDependents() - removed from here and added to UpdateCode
			cursorline=l
			UpdateStatus() 
			' can't see a good reason for having this PollSystem here....
			'PollSystem
		EndIf
'DebugLog "UpdateCursor:Exit"		
	End Method

	Method UpdateStatus(text$="")
		Local	c:Int
		c = cursorpos + cursorlen - TextAreaChar(_textarea , cursorline - 1)
		If text$<>"" Then statustext=text
		'double tabs seem to not work with win32... on mine it would be centered and the
		'first letter would be cut off
?MacOS
		host.SetStatus statustext+"~t~t"+"Line: "+(cursorline)+" Char: "+(c+1)+"     "
?Linux
		host.SetStatus statustext+"~t~t"+"Line: "+(cursorline)+" Char: "+(c+1)+"     "
?Win32
		host.SetStatus statustext+"~t~t"+"Line: "+(cursorline)+" Char: "+(c+1)+"     "
?
	End Method

	Method IndentCode()
		Local	cpos:Int,p0:Int,p1:Int,i:Int,a$
		' blockindent
		' grab current cursor position and length
		p0 = TextAreaCursor(_textarea , TEXTAREA_LINES)
		p1 = TextAreaSelLen(_textarea , TEXTAREA_LINES)

		' make sure the entire block is selected (start cursor pos may be in the middle of the line)
		SelectTextAreaText _textarea , p0 , p1 , TEXTAREA_LINES
		UpdateCursor

		For i=0 Until p1
			a$="~t"+TextAreaText(_textarea,p0+i,1,TEXTAREA_LINES)
			SetTextAreaText _textarea,a$,p0+i,1,TEXTAREA_LINES
		Next
		SelectTextAreaText _textarea,p0,p1,TEXTAREA_LINES
		UpdateCursor
		UpdateCode
	End Method

	Method OutdentCode()
		Local	cpos:Int,p0:Int,p1:Int,i:Int,a$,modified:Int
		' blockoutdent
		' grab current cursor position and length
		p0 = TextAreaCursor(_textarea , TEXTAREA_LINES)
		p1 = TextAreaSelLen(_textarea , TEXTAREA_LINES)

		' make sure the entire block is selected (start cursor pos may in the middle of the line)
		SelectTextAreaText _textarea , p0 , p1 , TEXTAREA_LINES
		UpdateCursor

		For i=0 Until p1
			a$=TextAreaText(_textarea,p0+i,1,TEXTAREA_LINES)
			If a[0]=9 a$=a$[1..];modified=True
			SetTextAreaText _textarea,a$,p0+i,1,TEXTAREA_LINES
		Next
		If Not modified
			For i=0 Until p1
				a$=TextAreaText(_textarea,p0+i,1,TEXTAREA_LINES)
				If a[0]=32 a$=a$[1..]
				SetTextAreaText _textarea,a$,p0+i,1,TEXTAREA_LINES
			Next
		EndIf
		SelectTextAreaText _textarea,p0,p1,TEXTAREA_LINES
		UpdateCursor
		UpdateCode
	End Method

	' this is comment uncomment instead of a toggle becauseyou want to be able
	' to comment comments so when you uncomment you don't uncomment the
	' original comment -dmaz
	Method CommentCode(makecomment:Int=True)
		Local	cpos:Int,p0:Int,p1:Int,i:Int,a$

		' grab current cursor position and length
		p0 = TextAreaCursor(_textarea , TEXTAREA_LINES)
		p1 = TextAreaSelLen(_textarea , TEXTAREA_LINES)

		'select complete lines, ensuring whole of current line
		SelectTextAreaText _textarea, p0, p1, TEXTAREA_LINES
		UpdateCursor

		For i=0 Until p1 Or 1
			If makecomment = True
				SetTextAreaText _textarea,"'",p0+i,0,TEXTAREA_LINES
			Else
				a$=TextAreaText(_textarea,p0+i,1,TEXTAREA_LINES)
				If a[0]=Asc("'") a$=a$[1..]
				SetTextAreaText _textarea,a$,p0+i,1,TEXTAREA_LINES
			EndIf
		Next

		SelectTextAreaText _textarea, p0, p1, TEXTAREA_LINES
		UpdateCursor
		UpdateCode
	End Method

	' This will insert a new line with "'MARK:" at the begining if "'MARK" is not at the
	' begining already.  Otherwise it will delete the whole line.
	Method ToggleBookmark()
		Local	line:Int
		Local	a:String

		UpdateCursor
		' grab current cursor position
		line = TextAreaCursor(_textarea, TEXTAREA_LINES)
		' position the cursor at the begining of the line
		SelectTextAreaText _textarea, line, 0, TEXTAREA_LINES

		a$=TextAreaText(_textarea,line,1,TEXTAREA_LINES)
		If Lower(a[..6])="'mark:"
			'toggle mark off
			a$=""
			SetTextAreaText _textarea, a$, line, 1, TEXTAREA_LINES
			SelectTextAreaText _textarea, line, 0, TEXTAREA_LINES
		Else
			'toggle mark on
			a$="'MARK:~n"
			SetTextAreaText _textarea, a$, line, 0, TEXTAREA_LINES
			SelectTextAreaText _textarea,TextAreaCursor(_textarea,TEXTAREA_CHARS)+6,0,TEXTAREA_CHARS
		EndIf
		UpdateCursor
		UpdateCode
	End Method

	Method _ShowAtPos(pos:Int=0)
		host.SelectPanel Self
		backstack.Push TextAreaCursor(_textarea,TEXTAREA_CHARS)
		forwardstack.Clear

		'FIX: hack until GetScrollPos,SetScrollPos (sets the line at the top)
		SelectTextAreaText _textarea,TextAreaLine(_textarea,pos)+999,0,TEXTAREA_LINES
'		SelectTextAreaText _textarea,TextAreaLine(_textarea,pos)-5,0,TEXTAREA_LINES
		'
		SelectTextAreaText _textarea,TextAreaLine(_textarea,pos),0,TEXTAREA_LINES
		SelectTextAreaText _textarea,TextAreaLine(_textarea,pos),1,TEXTAREA_LINES
		UpdateCursor
		ActivateGadget _textarea
	End Method

	Method Back()
		Local pos:Int
		If backstack.NotEmpty()
			forwardstack.Push TextAreaCursor(_textarea)
			pos = backstack.Pop()

			SelectTextAreaText _textarea,TextAreaLine(_textarea,pos)+999,0,TEXTAREA_LINES 'FIX: hack until GetScrollPos,SetScrollPos
			SelectTextAreaText _textarea,TextAreaLine(_textarea,pos),0,TEXTAREA_LINES
			SelectTextAreaText _textarea,TextAreaLine(_textarea,pos),1,TEXTAREA_LINES
			UpdateCursor
			ActivateGadget _textarea
		EndIf
	End Method

	Method Forward()
		Local pos:Int
		If forwardstack.NotEmpty()
			backstack.Push TextAreaCursor(_textarea)
			pos = forwardstack.Pop()

			SelectTextAreaText _textarea,TextAreaLine(_textarea,pos)+999,0,TEXTAREA_LINES 'FIX: hack until GetScrollPos,SetScrollPos
			SelectTextAreaText _textarea,TextAreaLine(_textarea,pos),0,TEXTAREA_LINES
			SelectTextAreaText _textarea,TextAreaLine(_textarea,pos),1,TEXTAREA_LINES
			UpdateCursor
			ActivateGadget _textarea
		EndIf
	End Method

	Method NextBookMark(todonode:TDataNode)
		For Local n:TTodoNode = EachIn todonode._kids
			If n._name = "MARK"
				If n._kids.count()
					Local gotonode:TTodoNode
					For Local kidnode:TTodoNode = EachIn n._kids
						If kidnode.pos > TextAreaCursor(_textarea,TEXTAREA_CHARS)
							gotonode=kidnode
							Exit
						EndIf
					Next
					If Not gotonode gotonode=TTodoNode(n._kids.First())
					gotonode.invoke(TOOLACTIVATE)
				EndIf
			EndIf
		Next
	End Method

	Method PreviousBookMark(todonode:TDataNode)
		For Local n:TTodoNode = EachIn todonode._kids
			If n._name = "MARK"
				If n._kids.count()
					Local gotonode:TTodoNode
					For Local kidnode:TTodoNode = EachIn n._kids
						If kidnode.pos < TextAreaCursor(_textarea,TEXTAREA_CHARS)
							gotonode=kidnode
						EndIf
					Next
					If Not gotonode gotonode=TTodoNode(n._kids.Last())
					gotonode.invoke(TOOLACTIVATE)
				EndIf
			EndIf
		Next
	End Method

	Method Query:Int(property:Int)
		If property = QUERY_HASSELECTION Then
			If _TextArea Then Return (TextAreaSelLen(_TextArea) > 0)
		EndIf
	EndMethod

' common command interface
	Method Invoke:Object(command:Int,argument:Object=Null)
		Local	file$,ex$
		Local	p:Int

		Select command
			Case TOOLSHOW
				If host Then				
					'DebugLog "TCodePanel.Invoke(TOOLSHOW):Start"
					If _textarea Then ActivateGadget _textarea
					'TODO: are these two lines needed?
					GetNode() 
					GetTodoNode() ' bregors todo-tree-list			
					' code tree
					If codenode=Null Then GetNode()
					host.SetCodeNode codenode
					
					' Todo list
					If todonode=Null Then GetTodoNode()
					host.SetTodoNode todonode

					host.SetTitle path
					'DebugLog "TCodePanel.Invoke(TOOLSHOW):Done"
				Else
					DebugLog "Tried to activate a code panel that isn't yet attached to it's host!?!"
				EndIf

			Case TOOLNAVIGATE
				Local tmp:Int = Int[](argument)[0]
'DebugLog "TCodePanel.Invoke(TOOLNAVIGATE,"+tmp+")"				
				_ShowAtPos(tmp)

			Case TOOLCLOSE
				Local dontask:Int=False
				If argument<>Null Then
					dontask=Int[](argument)[0]
				EndIf		
				If dirty And dontask=False
					p=Proceed("@@message/save_proceed|"+name)	'the current file?
					If p=-1 Return [Self]
					If p=1
						If Not Save() Return [Self]
					EndIf
				EndIf
				
				'BUG: something in here isn't being cleared out properly
				If codenode Then codenode.free; host.setcodenode Null ; codenode = Null
				If todonode Then todonode.free; host.settodonode Null ;todonode = Null
				' remove items from host list before freeing them
				host.RemoveToolMenuItem( TOOLMENU_EDITOR, editmenulist)
				host.RemoveToolMenuItem( TOOLMENU_EDITOR, editmenusellist)
				editmenulist.Clear()
				editmenusellist.Clear() 
				FreeGadget _textarea
				_textarea=Null
				FreeGadget editmenu ' language.bmx: localised gadgets must be freed manually.
				editmenu=Null
				If Autoc Then
					host.RemoveHandler Autoc
					Autoc.Free()
					Autoc = Null
				EndIf
				host.RemovePanel Self
				undolist = Null
				redolist = Null
				host = Null			

			Case TOOLSAVE
				Save

			Case TOOLQUICKSAVE
				file=path
				If dirty SaveSource(file)

			Case TOOLSAVEAS
				Local newcodetype:Int = codetype

				file = path
				If host.IsTempPath(file) Then file$=""				
				file=RequestFile("@@main_window/save_as_title",FILTER_Files,True,file)
				If file="" Return Null
				ex$=ExtractExt(file)
				If Not ex$
					' set default codetype extension
					If codetype <> CODETYPE_UNKNOWN Then
						file = file + "." + CodeTypeExt[codetype][0]
					EndIf
				Else
					newcodetype = CodeTypeFromExt( ex.ToLower())
				EndIf
				SaveSource(file$)
				If codetype = newcodetype Then
					' same type as before save
					Refresh
					GetNode().Refresh
					GetTodoNode().Refresh
					setdirty False
					host.SetTitle path
				Else
					' not same type as before save, close panel and reopen it					
					'HACK: add some garbage to the name so we can open the same file
					path :+ "::" 
					Local code:TCodePanel = host.OpenSource( file)
					path = path[..path.Length-2]
					If code Then
						host.SwapPanels( Index, code.Index)
						host.RemovePanel(Self)
						host.SelectPanel(code)
						'HACK: force the right tab to be selected
						SelectGadgetItem host.tabbar , code.index
					EndIf
				EndIf

			Case TOOLGOTO
				Local line:Int=Int(String(argument))
				SelectTextAreaText _textarea,line-1,0,TEXTAREA_LINES
				UpdateCursor
				ActivateGadget _textarea

			Case TOOLFINDREPLACE
				FindAndReplace()

			Case TOOLFINDNEXT
				Local res:Int=FindNext(String(argument))
				ActivateGadget _textarea
				If res Then
					Return String(res) ' does it matter if we don't return anything?  Used to return res = true / false result
				Else
					Return Null
				EndIf

			Case TOOLREPLACE
				Return String( FindReplace( String(argument) ) )

			Case TOOLREPLACEALL
				Local a:String = argument.ToString()
				Local i:Int = a.Find( Chr(0))
				If i >= 0 Then
					Return String.FromInt( ReplaceAll( a[..i], a[i+1..]))
				Else
					Return String.FromInt( ReplaceAll( a[..i], Null))
				EndIf
			
			Case TOOLREPLACEALLINSELECTION
				Local a:String = argument.ToString()
				Local i:Int = a.Find( Chr(0))
				If i >= 0 Then
					Return String.FromInt(ReplaceAll(a[..i], a[i+1..], True))
				Else
					Return String.FromInt(ReplaceAll(a[..i], Null, True))
				EndIf

			Case TOOLUNDO
				Undo
			Case TOOLREDO
				Redo

			Case TOOLREFRESH
				Refresh

			Case TOOLCUT
				GadgetCut _textarea
			Case TOOLCOPY
				GadgetCopy _textarea
			Case TOOLPASTE
				GadgetPaste _textarea
			Case TOOLSELECTALL
				SelectTextAreaText _textarea

			Case TOOLINDENT
				IndentCode
			Case TOOLOUTDENT
				OutdentCode
			Case TOOLPRINT
				GadgetPrint _textarea

'MARK: should be implemented by subclasses if needed
'			Case TOOLHELP
'				If argument And argument.ToString() Then UpdateStatus(argument.ToString()) Else Help()

			Case TOOLAUTOCOMPLETE
				AutoComplete( argument.ToString())

			Case TOOLBUILD
				BuildSource host.options_build.quickenabled , False , host.options_build.threadenabled , host.options_build.apptype , host.options_build.targetplatform  , False
				
			Case TOOLRUN
				BuildSource host.options_build.quickenabled , False , host.options_build.threadenabled , host.options_build.apptype , host.options_build.targetplatform , True
				
			Case TOOLDEBUG
				BuildSource host.options_build.quickenabled,True,host.options_build.threadenabled,host.options_build.apptype , host.options_build.targetplatform,True
				
			Case TOOLPARSEERROR
				Return String.FromInt( ParseError( argument.ToString()))

			Case TOOLLOCK
				SetLocked True
			Case TOOLUNLOCK
				SetLocked False

			Case TOOLCOMMENT
				CommentCode(True)
			Case TOOLUNCOMMENT
				CommentCode(False)

			Case TOOLBACK
				Back()
			Case TOOLFORWARD
				Forward()

			Case TOOLTOGGLEMARK
				ToggleBookMark()
			Case TOOLNEXTMARK
				NextBookMark(todonode)
			Case TOOLPREVIOUSMARK
				PreviousBookMark(todonode)
			Default
				'NEVER put anything here!
		End Select
	End Method

'BUG: potential for memory leak, only 1 copy per TCodePanel !!!!
'its the TLists editmenulist & editmenusellist that would fill up, as they are only cleared & removed on TOOLCLOSE
	Method CreateEditMenu:TGadget()
		Local	edit:TGadget, e:TGadget
		Local	is:TIconStrip
		
		is = LoadIconStrip("incbin::resources/icons_small.png") 
		If is = Null Then DebugStop() 
		
		LanguageSetGroup "edit_context_menu"

		edit = CreateMenu("@@edit" , 0 , Null) 
		
		e = CreateMenu( "@@cut" , MENUCUT , edit)
		editmenusellist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is,ICON_CUT)
		
		e = CreateMenu( "@@copy" , MENUCOPY , edit)
		editmenusellist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is,ICON_COPY)	
		
		e = CreateMenu( "@@paste" , MENUPASTE , edit)
		editmenulist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is,ICON_PASTE)	
		
		'------------------
		CreateMenu "" , 0 , edit
		e=CreateMenu( "@@select_all",MENUSELECTALL,edit)
		editmenulist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is , ICON_SELECTALL)	
				
		'------------------
		CreateMenu "" , 0 , edit
		e=CreateMenu( "@@block_indent",MENUINDENT,edit)
		editmenusellist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is , ICON_BLOCKINDENT)	
		
		e=CreateMenu( "@@block_outdent" , MENUOUTDENT , edit)
		editmenusellist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is , ICON_BLOCKOUTDENT)	
		
		e = CreateMenu( "@@block_comment" , MENUCOMMENT , edit) 
		editmenusellist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is , ICON_BLOCKCOMMENT)	
		
		e = CreateMenu( "@@block_uncomment" , MENUUNCOMMENT , edit)
		editmenusellist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is,ICON_BLOCKUNCOMMENT)	
		
		'------------------
		CreateMenu "",0,edit
		e = CreateMenu( "@@find_replace" , MENUFINDREPLACE , edit)
		editmenulist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is , ICON_FIND) 
		
		e = CreateMenu( "@@find_next" , MENUFINDNEXT , edit) 
		editmenulist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is,ICON_FIND)
		
		'------------------
		CreateMenu "" , 0 , edit
		e = CreateMenu( "@@toggle_bookmark" , MENUTOGGLEMARK , edit) 
		editmenulist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is,ICON_BOOKMARK)
		
		e = CreateMenu( "@@next_bookmark" , MENUNEXTMARK , edit) 
		editmenulist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is,ICON_BOOKMARK_NEXT)
		
		e=CreateMenu( "@@previous_bookmark",MENUPREVIOUSMARK,edit)
		editmenulist.AddLast e
		SetGadgetPixmap e , PixmapFromIconStrip(is,ICON_BOOKMARK_PREV)	

		' add them to the host list
		host.AddToolMenuItem( TOOLMENU_EDITOR, editmenulist)
		host.AddToolMenuItem( TOOLMENU_SELECTION, editmenusellist)
		'host.EnableToolMenus( TOOLMENU_SELECTION, False)

		Return edit
	End Method


'----- for test harness
	Method SimulateKeyPress(key:Int,mods:Int=0,count:Int=1)
		Local m$=KeyToChar(key,mods)	
		Local e:TEvent = New TEvent
		Local p:Int,l:Int
		Local t$
		
		For Local i:Int=1 To count
			e.source = _textarea
			e.data = key
			e.mods = mods
			e.extra = m
			e.id = EVENT_KEYDOWN
			If FilterKey(e,Self) Then	
				e.id = EVENT_KEYCHAR
				If m$<>"" Then
					e.data = Asc(m)			
					If FilterKey(e , Self) Then
						p = TextAreaCursor(_textarea)
						l = TextAreaSelLen(_textarea)
						SetTextAreaText _textarea,m,p,l
						SelectTextAreaText _textarea, p+1,0
						e.id = EVENT_GADGETSELECT 
						Poll(e)		
						e.id = EVENT_GADGETACTION 
						Poll(e)			
					EndIf
					
				Else 'deal with unprintables
					Select key
						Case KEY_LEFT
							p = TextAreaCursor(_textarea)
							l = TextAreaSelLen(_textarea)
							If mods = MODIFIER_SHIFT Then
								If p > 0 Then p:- 1 Else Return ' at start of all text so can't go left
								l:+ 1
							ElseIf l = 0
								If p > 0 Then p:- 1 Else Return ' at start of all text so can't go left							
								l = 0
							Else
								l = 0
							EndIf
							SelectTextAreaText _textarea, p,l
							e.id = EVENT_GADGETSELECT 
							Poll(e)
							
						Case KEY_RIGHT
							p = TextAreaCursor(_textarea)
							l = TextAreaSelLen(_textarea)
							If mods = MODIFIER_SHIFT Then
								If p + l >= Len(TextAreaText(_textarea) ) Then Return ' at end of all text so can't go right						
								l:+ 1
							ElseIf l = 0
								If p >= Len(TextAreaText(_textarea) ) Then Return ' at end of all text so can't go right						
								p:+ 1
							Else
								p = p + l
								l = 0
							EndIf
							SelectTextAreaText _textarea, p,l
							e.id = EVENT_GADGETSELECT 
							Poll(e)
											
						Case KEY_BACKSPACE ' need to remove char before cursor
							If FilterKey(e,Self) Then
								p = TextAreaCursor(_textarea)
								l = TextAreaSelLen(_textarea)
								If l = 0 Then
									If p > 1 Then p:- 1 Else Return ' at start of all text so can't delete!
									SetTextAreaText _textarea,"",p,1
									SelectTextAreaText _textarea , p , 0
								Else
									SetTextAreaText _textarea,"",p,l
									SelectTextAreaText _textarea , p , 0
								EndIf
								e.id = EVENT_GADGETSELECT 
								Poll(e)		
								e.id = EVENT_GADGETACTION 
								Poll(e)							
							EndIf	
												
						Default ; Throw "Can't simulate keypress for unprintable keycode "+key
						
					End Select
							
				EndIf
			EndIf
		Next
	End Method
	
	Method SimulateTyping(text:String)
		'NOTE: This only works on a UK keyboard!
		For Local i:Int=0 To Len(text)-1
			Local a:Int=text[i]
			If a>=65 And a<=90 Then										' A-Z
				SimulateKeyPress(a,MODIFIER_SHIFT)
			ElseIf a>=97 And a<=122 Then									' a-z
				SimulateKeyPress(a-32)
			ElseIf a>=KEY_0 And a<=KEY_9 Then								' 0-9
				SimulateKeyPress(a)			
			Else
				Select a
					Case 9	; SimulateKeyPress(KEY_TAB)
					Case 10	; SimulateKeyPress(KEY_RETURN)					
					Case 13	; SimulateKeyPress(KEY_RETURN)
					Case 32	; SimulateKeyPress(KEY_SPACE)
					Case 33	; SimulateKeyPress(KEY_1,MODIFIER_SHIFT)		'	!
					Case 34	; SimulateKeyPress(KEY_2,MODIFIER_SHIFT)		'	"
					Case 35	; SimulateKeyPress(KEY_QUOTES)				'	#
					Case 36	; SimulateKeyPress(KEY_4,MODIFIER_SHIFT)		'	$
					Case 37	; SimulateKeyPress(KEY_5,MODIFIER_SHIFT)		'	%
					Case 38	; SimulateKeyPress(KEY_7,MODIFIER_SHIFT)		'	&
					Case 39	; SimulateKeyPress(KEY_TILDE)					'	'
					Case 40	; SimulateKeyPress(KEY_9,MODIFIER_SHIFT)		'	(
					Case 41	; SimulateKeyPress(KEY_0,MODIFIER_SHIFT)		'	)
					Case 42	; SimulateKeyPress(KEY_8,MODIFIER_SHIFT)
					Case 43	; SimulateKeyPress(KEY_EQUALS,MODIFIER_SHIFT)	'	=
					Case 44	; SimulateKeyPress(KEY_COMMA)					'	,
					Case 45	; SimulateKeyPress(KEY_MINUS)					'	-
					Case 46	; SimulateKeyPress(KEY_PERIOD)				'	.
					Case 47	; SimulateKeyPress(KEY_SLASH)	
					Case 58	; SimulateKeyPress(KEY_SEMICOLON,MODIFIER_SHIFT)
					Case 59	; SimulateKeyPress(KEY_SEMICOLON)
					Case 60	; SimulateKeyPress(KEY_COMMA,MODIFIER_SHIFT)
					Case 61	; SimulateKeyPress(KEY_EQUALS)
					Case 62	; SimulateKeyPress(KEY_PERIOD,MODIFIER_SHIFT)
					Case 63	; SimulateKeyPress(KEY_SLASH,MODIFIER_SHIFT)		'	?
					Case 64	; SimulateKeyPress(KEY_TILDE,MODIFIER_SHIFT)		'	@
					Case 91	; SimulateKeyPress(KEY_OPENBRACKET)
					Case 92	; SimulateKeyPress(KEY_BACKSLASH)				'	\
					Case 93	; SimulateKeyPress(KEY_CLOSEBRACKET)		
					Case 94	; SimulateKeyPress(KEY_6,MODIFIER_SHIFT)		
					Case 95	; SimulateKeyPress(KEY_MINUS,MODIFIER_SHIFT)		'	_		
					Case 123	; SimulateKeyPress(KEY_OPENBRACKET,MODIFIER_SHIFT)
					Case 124	; SimulateKeyPress(KEY_BACKSLASH,MODIFIER_SHIFT)	'	|
					Case 125	; SimulateKeyPress(KEY_CLOSEBRACKET,MODIFIER_SHIFT)					
					Case 126	; SimulateKeyPress(KEY_QUOTES,MODIFIER_SHIFT)	'	~
					Case 163	; SimulateKeyPress(KEY_3,MODIFIER_SHIFT)		'						

					Default ;Throw "Can't simulate typing ~q"+Chr(a)+"~q (ascii code "+a+")"
				End Select
			EndIf
		Next
	End Method
	
	Method Text:String()
		Return TextAreaText(_textarea)
	End Method
	
'----- end test harness code

	Method Poll(e:TEvent)
 		Select e.Source
			Case _textarea
				Select e.id
					Case EVENT_GADGETMENU
'DebugLog "==> TCodePanel.Poll(TextArea,GADGETMENU)"						
						host.EnableToolMenus( TOOLMENU_SELECTION, Query(QUERY_HASSELECTION))
						UpdateWindowMenu editmenu
						host.PopupMenu editmenu,Self
					Case EVENT_GADGETACTION
'DebugLog "==> TCodePanel.Poll(TextArea,GADGETACTION)"					
						UpdateCode
					Case EVENT_GADGETSELECT
'DebugLog "==> TCodePanel.Poll(TextArea,GADGETSELECT)"					
						UpdateCursor
						host.EnableToolMenus( TOOLMENU_SELECTION , Query(QUERY_HASSELECTION) ) 
					Default
'DebugLog "==> TCodePanel.Poll(TextArea,?)"					
				End Select
		End Select
	End Method

	Function FilterKey:Int(event:TEvent,context:Object)
		Local	this:TCodePanel
		Local	id:Int,key:Int,mods:Int
		Local	PT: POINT = New POINT

'		If event.id<>EVENT_KEYCHAR Return 1
		id=event.id
		key=event.data
		mods=event.mods
		this=TCodePanel(context)
?MacOS
		If key=25 And mods=MODIFIER_SHIFT Then key=KEY_TAB
?

		If this Then
			Select id
				Case EVENT_KEYDOWN
'DebugLog "TCodePanel.FilterKey(KEYDOWN,"+key+")"
					key:& $FF ' Crop key-value to one byte as linux seems to add some extra data.

					'Autocomplete
					If this.Autoc And this.options_keys.CheckKey(this.KEY_AUTOC,key,mods) And this.options_code.autocomplete Then
						this.Autoc.Show()
						 ' Just one option, so use it!
						If this.Autoc.AutocVal.Length = 1 Then
							this.AutoComplete( this.Autoc.AutocVal[0])
							this.Autoc.Hide()
						EndIf
						Return 0
					EndIf

					' more Autocomplete
					If this.Autoc And this.Autoc.IsOpen Then
						Select key
							Case KEY_DOWN ' next item in list
								this.Autoc.NextItem()
								Return False
							Case KEY_UP ' last item in list
								this.Autoc.LastItem()
								Return False
							Case KEY_RETURN ' replace with item & close
								this.Autoc.AutoComplete()
								Return False
							Case KEY_TAB, KEY_SPACE ' replace with item & close + allow key into editor
								this.Autoc.AutoComplete()
								Return True
							Case KEY_ESCAPE ' close
								this.Autoc.Hide()
								Return False
						EndSelect
					EndIf

					'auto-indent on return
					If key=KEY_RETURN And (this.options_code.autoindent Or this.options_code.autotidy) Then
						If this.AutoCodeChange() Then Return 0
						Return 1 ' never called as above always returns true
					EndIf

					' home to move to beginning of text (not line) if no modifiers
					If key = KEY_HOME And mods = 0 Then
						Local ta:TGadget = this.gadget()
						Local cursor:Int = TextAreaCursor(ta)
						Local linestart:Int = TextAreaChar(ta, TextAreaLine(ta, cursor))
						Local lineend:Int = TextAreaChar(ta, TextAreaLine(ta, cursor) + 1)
						Local line:String = TextAreaText(ta, linestart, lineend-linestart)
						Local codestart:Int = linestart
						If line.Trim() <> "" Then
							For Local a:Int = 0 To line.length-1 ' find first non-space character
								If line[a] > 32 Then
									codestart = linestart + a
									Exit
								EndIf
							Next
						EndIf
						If cursor = codestart Then
							SelectTextAreaText ta, linestart, 0
						Else
							SelectTextAreaText ta, codestart, 0
						EndIf
						this.UpdateCursor
						Return 0
					EndIf

				Case EVENT_KEYCHAR
'DebugLog "TCodePanel.FilterKey(KEYCHAR,"+key+")"
					Local ta:TGadget = this.gadget()
					If this.Autoc And this.Autoc.IsOpen Then
						' prevent the space in CTRL+SPACE to enter editor, for autocomplete
						If Not this.Autoc.FirstSpace And this.options_keys.CheckKey(this.KEY_AUTOC,key,mods) Then
							this.Autoc.FirstSpace = True
							Return False
						EndIf
					EndIf
					If key=KEY_TAB And TextAreaSelLen( ta,TEXTAREA_CHARS )
						Select mods
							Case 0
								this.IndentCode
							Case 1
								this.OutdentCode
						End Select
						Return 0
					EndIf
			End Select
			' autocomplete, just in case we missed a character
			If this.Autoc And this.Autoc.IsOpen Then this.Autoc.PopulateList()
		EndIf

		Return 1
	End Function

	Method Init( path:String, host:TBMXHost)
		Local	stream:TStream
		Local	isnew:Int
		
		undolist = New TList
		redolist = New TList
		backstack = New IntStack
		backstack.maxsize = 1000 ' set a maximum, otherwise it will just fill forever.  4 Bytes per item!
		forwardstack = New IntStack
		
		If path
			stream=ReadFile(path)
			If Not stream
'				Notify "Open could not read from "+path
				Return
			EndIf
		Else
			host.tempfilecount:+1
			path=host.FullPath("$BMXPATH/tmp/untitled"+host.tempfilecount+".bmx")
			isnew=True
		EndIf
		Self.host=host
		options_editor=host.options_editor
		options_code=host.options_code
		options_keys = host.options_keys
		KEY_AUTOC = options_keys.AddHotKeyDef("Auto-Complete",KEY_SPACE,MODIFIER_CONTROL)
		active=True
		Self.path = host.FullPath(path) 
		editmenu = CreateEditMenu() 
		host.addpanel(Self) 
		_textarea=CreateTextArea(0,0,ClientWidth(panel),ClientHeight(panel),panel,0)
		SetGadgetFilter _textarea,FilterKey,Self
		SetTextAreaText _textarea,"" ' was setting to ~n
		SetGadgetLayout _textarea,1,1,1,1

'TODO: !!! Should this refresh really be sat here? Could change to simply apply formatting as that's all its really doing
		Refresh()
		
		If isnew Then SaveSource path
		If stream
			host.UpdateProgress "Reading Stream"
			ReadSource(stream)
			CloseFile stream
		Else
			filesrc="~n"
			ActivateGadget _textarea
		EndIf

'TODO: !!! Theres another early refresh in UpdateCode to be removed		
		UpdateCode False
		filesrc=TextAreaText(_textarea)
		'TODO: !!! Should be removed in favour of only selecting when needed
		host.SelectPanel(Self)
	EndMethod

	Function AutoTidyLine$(oldtext$) 
		Return oldtext$
	End Function
	
	Function IsCodeBlockStart%(src$) 
		Return False
	End Function

	Function FindPrevRem%(src$,pos:Int) Abstract
	Function FindEndRem%(src$,pos:Int,returnlast:Int=False) Abstract
	Function FindRem%(src$ , pos:Int) Abstract
		
	Method IsRemmed:Int(src$ , pos:Int)
		' INPUTS: src = source code to check , pos = position to check
		' OUTPUTS: position of start of rem block, -1 if not remmed
		Local pr:Int, er:Int , nr:Int
		
		Assert src=src.tolower() ' src is assumed to be lowercase on input!
		
		'TODO: change IsRemmed such that it can deal with recursive rem blocks
		'suggest start at 0 and work forward until we get to pos, then say whether we are in a rem block or not?

		pr = FindPrevRem(src$ , pos )
		If pr = - 1 Then Return -1
		
		er = FindEndRem(src$ , pr , True)
		If er = - 1 Or pos < er Then Return pr	
		
		Return -1
	End Method

'MARK: TCodePanel non implemented functionality (for subclasses)
	Method AutoCodeChange:Int()
		'this does both auto-indent and auto-tidy

		Local	p_start:Int		' start of region to highlight, always one char on from an EOL or 1
		Local	p_leadingtab:Int	' the end of the tab characters at the start of this line
		Local	p_cursor:Int		' position of the cursor, PRIOR to new key being typed, but updated as we tidy
		Local	p_cursor_orig:Int	' position of the cursor, PRIOR to new key being typed, NOT updated as we tidy 
		Local 	curslen:Int		' current selection length
		Local	src$		' current contents of the textarea
		Local	srcl$		' lower case version of textarea contents
		Local	newtext$	' orevised line contents
		Local	oldtext$	' original line contents

		'NOTE: this routine is entered BEFORE the changes have been made to the text, or the cursor has moved
		'      In theory, we can use cleansrc & cleansrcl, as they should reflect the code in the textarea
		'      For safety, this code grabs the very latest from the textarea and uses that.

		' store the new source code (to compare to cleansrc)
		src = TextAreaText(_textarea) 
		srcl = src.tolower()
		
		' and store the current (new) cursor position and selection length
		'NOTE: p_cursor=0 is before the first character.  The typed character will become src[curspos]
		p_cursor = TextAreaCursor(_textarea , TEXTAREA_CHARS) 
		p_cursor_orig = p_cursor
		curslen = TextAreaSelLen(_textarea , TEXTAREA_CHARS) 
		
		'NOTE: If there's a region selected (curslen>0), then we've actually just "removed" 
		'      a bunch of code (inside curslen) and replaced it with an EOL.
		If curslen>0 Then DebugLog "!!AutoCodeChange: Called with selection in place"		

		' Find the start of the current line by looking for the first EOL prior to the cursor position (-1)
		' and then moving to the next character (+1)
		'NOTE: FindLast returns the position of the char in the string, working back from  
		'      the char before curspos (including src[curspos-1]).
		'NOTE: FindLast needs the number of chars from the *END* of the string to start searching back from
		p_start = src.FindLast( EOL$ , src.length - (p_cursor - 1)) + 1
		' and if we didn't find one, p_start=0, i.e. we must be at the start of the code!
		
		' Next find the position of the last leading tab on this line
		p_leadingtab = p_start
		While p_leadingtab<p_cursor And src[p_leadingtab]=KEY_TAB ' keep going while tabs, but stop at cursor!
			p_leadingtab:+1
		Wend
		Assert p_leadingtab<=p_cursor ' Should never happen, but there used to be a check for it...

		' As we don't want to auto-tidy inside multi-line Rem / EndRem blocks, find out if we are in one!
		Local inrem:Int = True
		If IsRemmed(srcl , p_cursor) = - 1 Then inrem = False

		' Let's dump a bunch of debug info here...
		DebugLog "p_cursor        = " + p_cursor 
		DebugLog "curslen         = " + curslen
		DebugLog "p_start         = " + p_start
		DebugLog "p_leadingtab    = " + p_leadingtab
		DebugLog "inrem?          = " + inrem
		DebugLog "considered text = " + src[p_start..p_cursor ] + "|"

		' Auto-Tidy
		' =========
		' This section ensures that there are spaces around all ,;=+-*/ And between (( And ))

		' Grab the current text (excl. tabs) from the source and store as oldtext
		oldtext$ = src[p_leadingtab..p_cursor] 'TODO: should we add curslen here?
		
		' and get AutoTidyLine to tidy it up
		If options_code.autotidy And Not(inrem) Then 
			newtext$ = AutoTidyLine(oldtext$) 
			
			' if it's any different, then we'd better deal with it...
			If newtext<>oldtext Then
				' make sure the entire block is selected (start cursor pos may in the middle of the line)
				SelectTextAreaText _textarea , p_start , curslen + (p_cursor - p_start) , TEXTAREA_CHARS
				' and trick the editor into thinking we edited this whole chunk of code by calling UpdateCursor!
				UpdateCursor
				' and then amend the code by overwriting in the text area
				SetTextAreaText _textarea , newtext$ , p_leadingtab , oldtext.length
				' and finally ensure that our apparent cursor position gets updated to reflect the extra text added
				p_cursor = p_cursor + newtext.length - oldtext.length
			End If
			
			' Dump some more debug info to show what we've done...
			DebugLog "oldtext         = " + oldtext
			DebugLog "newtext         = " + newtext
		End If

		' Auto-indent
		' ===========
		' Ensures that there are the same number of tabs on the New line as on the current line
		'TODO: AUTO-INDENT: Should also increase the tab count on an opening branch (If, ElseIf, Select, Case)
		'TODO: AUTO-INDENT: Should also decrease (or align) cloding branches (Endif, End Method/Function/Type, End Select, Wend, Forever, Until, Next)
		If options_code.autoindent And Not(inrem) Then
			' number of tabs on current line
			Local numtabs:Int = p_leadingtab - p_start ' if same, there are no tabs (0)
			
			' is this line the start of a code block?
			Local n:Int = IsCodeBlockStart(oldtext)
			If n>0 Then
				' insert the EOL, copy the tabs from the preceding line onto the new line, and add one more...
				Local t$
				For Local i:Int = 1 To n+numtabs
					t$:+ TAB$
				Next
				SetTextAreaText _textarea , EOL$ + t$ , p_cursor , curslen
				' udpate the cursor position - add the current number of tabs, plus EOL, plus extra tab
				p_cursor = p_cursor + numtabs + n + 1
			Else
				' just insert the EOL plus tabs from the preceding line onto the new line
				SetTextAreaText _textarea , EOL$ + src[p_start..p_leadingtab] , p_cursor , curslen
				' update the cursor position - add the current number of tabs plus EOL
				p_cursor =p_cursor + numtabs + 1
			End If
		Else
			' just insert the EOL, and increment cursor position by 1 for the EOL
			SetTextAreaText _textarea , EOL$ , p_cursor , curslen
			p_cursor =p_cursor + 1
		End If

		' ensure we have the correct cursor pos / selection
		SelectTextAreaText _textarea , p_cursor , 0

		'update everything
		UpdateCursor
		UpdateCode

		' if we changed the text using auto-tidy, then we need to tidy up the latest entry in the undolist
		' this is a bit hacky, but it works! (Mark Tiffany)
		If newtext <> oldtext Then
			Local d:TDiff = TDiff(undolist.last())
			d.count = d.count + d.pos - p_cursor_orig ' was curslen, but that seems a long way wrong...
			d.pos = p_cursor_orig			
		EndIf

		Return True
	End Method

	Method BuildSource(quick:Int,debug:Int,mt:Int,apptype:Int,targetplatform:Int,run:Int)
	EndMethod

	Method ParseError:Int( err:String)
		Return False
	EndMethod

	Method GetNode:TDataNode()
		Function CreateNode:TDataNode()
			Local node:TDataNode = New TDataNode
			node.setname( "Code functionality not present")
			Return node
		EndFunction
		Global node:TDataNode = CreateNode()
		Return node
	EndMethod

	Method GetTodoNode:TDataNode()
		Function CreateTodoNode:TDataNode()
			Local node:TDataNode = New TDataNode
			node.setname("Todo functionality not present")
			Return node
		EndFunction
		Global node:TDataNode = CreateTodoNode()
		Return node
	EndMethod

	Function Create:TCodePanel( path:String , host:TBMXHost)			
		If FileType(path)<>FILETYPE_FILE
			Return Null
		EndIf
		Local ext:String = ExtractExt(path).ToLower()
		For Local i:Int = 0 Until CodeTypeExt.Length
			For Local e:String = EachIn CodeTypeExt[i]				
				If e = ext Then
					Local createfunc:TCodePanel( path:String , host:TGUIHost) = CodeTypeCreate[i]								
					Return createfunc( path, host)
				EndIf
			Next
		Next
		Return TUnknownCodePanel.Create( path,host)
	EndFunction
End Type

Const CODETYPE_BLITZMAX:Int	 = 	1
TCodePanel.Register(CODETYPE_BLITZMAX , ["bmx"] , Byte Ptr TBMXCodePanel.Create)

Type TBMXCodePanel Extends TCodePanel
	
	Method Invoke:Object(command:Int,argument:Object=Null)
		Select command
			Case TOOLHELP
				If argument And argument.ToString() Then UpdateStatus(argument.ToString()) Else Help()
			Default
				Return Super.Invoke( command, argument)
		EndSelect
	EndMethod

	Method Save:Int()
		Local	file$
		file=path
		If host.IsTempPath(file)
			Invoke(TOOLSAVEAS)
			Return True
		EndIf		
		Return Super.Save()
	End Method

' bregors todo-tree-list
	Method GetTodoNode:TDataNode()
		Local	root:TTodoNode,n:TTodoNode
		Local tdn:TTodoNode
		Local bgn:TTodoNode
		Local fxn:TTodoNode
		Local bmn:TTodoNode
		Local	src$,csrc$,line:Int,col:Int
		Local	p:Int,p1:Int,r:Int,t:Int,m:Int,f:Int,l:Int,e:Int
		Local tok$ , lins$ , ccnt:Int = 0
		Local pos_comment:Int
		
		'DebugLog "GetTodoNode:Enter"
		n=New TTodoNode
		n.setname(StripDir(path))
		n.count=src.length
		root=n

		tdn = root.AddTodoNode(Self,"TODO",-1,-1)
		bgn = root.AddTodoNode(Self,"BUG",-1,-1)
		fxn = root.AddTodoNode(Self,"FIX",-1,-1)
		bmn = root.AddTodoNode(Self,"MARK",-1,-1)

' these don't appear to serve any purpose inside here
'		root.Refresh
'		root.Open

'update rem,type,method,function,label pointers
		src=cleansrcl+"~n"
		csrc=cleansrc+"~n"
		p1=src.length
		p=0
		r=-1;t=-1

		While p<p1			'src.length

			e=src.find(EOL,p)
			lins$ = Trim$(csrc$[p..e])
			
			pos_comment = Instr(lins$ , "'")
			If pos_comment < 1 Then p = e + 1 ; Continue

			'TODO: Need to make sure it isn't embedded inside quotes

			' we have found our comment, so let's strip out the rest
			lins$=Trim(lins$[pos_comment..])

			tok$ = Lower$(Trim$(lins$)[..5])
			If tok$="todo:" Then
				n=tdn.AddTodoNode(Self,lins$[5..],p,e)
' these don't appear to serve any purpose inside here
'				n.Refresh
'				n.Open
			End If

			tok$ = Lower$(Trim$(lins$)[..4])
			If tok$="bug:" Then
				n=bgn.AddTodoNode(Self,lins$[4..],p,e)
			End If

			tok$ = Lower$(Trim$(lins$)[..4])
			If tok$="fix:" Then
				n=fxn.AddTodoNode(Self,lins$[4..],p,e)
			End If

			tok$ = Lower$(Trim$(lins$)[..5])
			If tok$="mark:" Then
				If lins$[6..] > ""
					n=bmn.AddTodoNode(Self,lins$[5..],p,e)
				Else
					n=bmn.AddTodoNode(Self,"Bookmark",p,e)
				EndIf
			End If

			p=e+1

		Wend
		
		If todonode
			todonode.Sync(root)
			root.Free
		Else
			todonode=root
		EndIf
		'DebugLog "GetTodoNode:Exit"
		Return todonode
	End Method

	Method ParseBMX(n:TCodeNode)
	'Parses the file and generates the code tree.
	'Returns the root of the tree, usually a file name (*.bmx)

		Local	temp:TCodeNode
		Local	fn:TCodeNode[], mn:TCodeNode, tn:TCodeNode, gn:TCodeNode[], cn:TCodeNode[], ln:TCodeNode, dn:TCodeNode
		Local	kn:TCodeNode, inn:TCodeNode, imn:TCodeNode, icn:TCodeNode, dbg:TCodeNode, asr:TCodeNode
		Local	src$,line:Int,col:Int,tempstr$
		Local	p:Int,p1:Int,r:Int,t:Int,m:Int,f:Int,l:Int,e:Int,g:Int,c:Int,a:Int,ex:Int,d:Int,k:Int,im:Int,in:Int,ic:Int,db:Int,as:Int

		'nodes to divide the code up
		'things that exist globally and in types need 2 place holder nodes
		cn = New TCodeNode[2]
		gn = New TCodeNode[2]
		fn = New TCodeNode[2]
		cn[0] = n.AddCodeNode(Self,"Consts",0,n.count)
		gn[0] = n.AddCodeNode(Self,"Globals",0,n.count)
		fn[0] = n.AddCodeNode(Self,"Functions",0,n.count)
		'types
		tn = n.AddCodeNode(Self,"Types",0,n.count)
		'labels
		ln = n.AddCodeNode(Self,"Labels",0,n.count)
		'Frameworks and Files
		kn = n.AddCodeNode(Self,"Framework",0,n.count)
		inn = n.AddCodeNode(Self,"Include",0,n.count)
		imn = n.AddCodeNode(Self,"Import",0,n.count)
		icn = n.AddCodeNode(Self,"Incbin",0,n.count)
		dbg = n.AddCodeNode(Self,"Debug",0,n.count)
		asr = n.AddCodeNode(Self,"Assert",0,n.count)

		'fields aren't initialized because there's no globals
		'update rem,type,method,function,label,field,const, global pointers
		src=cleansrcl
		p1=src.length
		'locations of the next tokens
		p=-1;r=-1;t=-1;m=-1;f=-1;l=-1;g=-1;c=-1;d=-1;k=-1;im=-1;in=-1;ic=-1;db=-1;as=-1
		While p<p1			'src.length
			While r<=p
				r=FindToken("rem",src,r+1)
			Wend
			While t<=p
				t=FindToken("type",src,t+1)
			Wend
			While m<=p
				m=FindToken("method",src,m+1)
			Wend
			While f<=p
				f=FindToken("function",src,f+1)
			Wend
			While l<=p
				l=FindLabel(src,l+1)
			Wend
			While d<=p
				d=FindToken("field",src,d+1)
			Wend
			While g<=p
				g=FindToken("global",src,g+1)
			Wend			
			While c<=p
				c=FindToken("const",src,c+1)
			Wend		
			'---
			While k<=p
				k=FindToken("framework",src,k+1)
			Wend
			While in<=p
				in=FindToken("include",src,in+1)
			Wend
			While im<=p
				im=FindToken("import",src,im+1)
			Wend
			While ic<=p
				ic=FindToken("incbin",src,ic+1)
			Wend
			While db<=p
				Local tmp:Int
				tmp=FindToken("debuglog",src,db+1)
				db = FindToken("debugstop" , src , db + 1) 
				If tmp>=0 And tmp<db Then db=tmp
			Wend				
			While as<=p
				as=FindToken("assert",src,as+1)
			Wend				
' ** Rem
			If r<t And r<m And r<f And r<l And r<d And r<g And r<c And r<k And r<in And r<im And r<ic And r<db And r<as Then
				p=FindEndToken("rem",src,r+1,True)
				Continue
			EndIf
			p=Min(t,Min(m,Min(f,Min(l,Min(d,Min(g,Min(c,Min(k,Min(in,Min(im,Min(ic,Min(db,as))))))))))))
			While (n And n._parent And p>n.pos+n.count)
				For Local temp:TCodeNode = EachIn(n._kids)
					If temp._kids.IsEmpty() And temp._name.find("Extends",0) = -1 Then temp.Detach()
				Next
				n=TCodeNode(n._parent)
			Wend
			If p = src.length Then Exit
' ** Type
			If t<m And t<f And t<l And t<d And t<g And t<c And t<k And t<in And t<im And t<ic And t<db And t<as Then
				e=src.find("abstract",t)
				If (e > -1 And e < src.find(EOL,t)) Then tempstr = "Abstract" Else tempstr = ""
				e = src.find(EOL,t)
				ex = src.find("extends",t)
				If ex > e Or ex = -1 Then
					n = tn.AddCodeNode(Self,cleansrc[t+5..e],t,FindEndToken("type",src,t,True))
				Else
					n = tn.AddCodeNode(Self,cleansrc[t+5..ex]+tempstr,t,FindEndToken("type",src,t,True))
					If tempstr = "" Then
						For Local temp:TCodeNode = EachIn(tn._kids)
							If cleansrc[ex+8..e] = temp._name Then
								n.AddCodeNode(Self,"Extends " + cleansrc[ex+8..e],temp.pos,temp.count+temp.pos)
								Exit
							Else If temp = tn._kids.Last()
								n.AddCodeNode(Self,"Extends " + cleansrc[ex+8..e],ex,e)
							EndIf
						Next
					Else
						e = src.find("abstract",ex)
						For Local temp:TCodeNode = EachIn(tn._kids)
							If cleansrc[ex+8..e] = temp._name Then
								n.AddCodeNode(Self,"Extends " + cleansrc[ex+8..e],temp.pos,temp.pos + temp.count)
								Exit
							ElseIf temp = tn._kids.Last()
								n.AddCodeNode(Self,"Extends " + cleansrc[ex+8..e],ex,e)
							EndIf
						Next
					EndIf
				EndIf
				cn[1] = n.AddCodeNode(Self,"Consts",t,e)
				gn[1] = n.AddCodeNode(Self,"Globals",t,e)
				dn = n.AddCodeNode(Self,"Fields",t,e)
				mn = n.AddCodeNode(Self,"Methods",t,e)
				fn[1] = n.AddCodeNode(Self,"Functions",t,e)
				p=t+1
				Continue
			EndIf
' ** Method
			If m<f And m<l And m<d And m<g And m<c And m<k And m<in And m<im And m<ic And m<db And m<as Then
				If mn<>Null Then ' check if someone put a Method outside of a Type!
					e=src.find(EOL,m)
					mn.AddCodeNode(Self,cleansrc[m+7..e],m,e)
				EndIf
				p=m+1
				Continue
			EndIf	
' ** Function
			If f<l And f<d And f<g And f<c And f<k And f<in And f<im And f<ic And f<db And f<as Then
				e=src.find(EOL,f)
				If (n._Parent And n._Parent._Parent) Then a = 1 Else a = 0
				fn[a].AddCodeNode(Self,cleansrc[f+9..e],f,e)				
				p=f+1
				Continue
			EndIf
' ** Label
			If l<d And l<g And l<c And l<k And l<in And l<im And l<ic And l<db And l<as Then
				e=src.find(EOL,l)
				n.AddCodeNode(Self,cleansrc[l..e],l,e, True)
				p=l+1
				Continue
			EndIf						
' ** Field
			If d<g And d<c And d<k And d<in And d<im And d<ic And d<db And d<as Then
				If dn <> Null Then ' check if someone put a Field outside of a Type!
					Local line_end% = src.find(EOL , d) 
					e = d + 6
					Local oe:Int=e					
					While e < line_end
						Select src[e]
							Case Asc("~q")
								e = src.find("~q" , e + 1) 
								If e = - 1 Then e=line_end;Exit
							Case Asc("'")
								Exit
							Case Asc(",")
								dn.AddCodeNode(Self , cleansrc[oe..e] , oe , e) 
								oe=e+1
						End Select
						e:+1	
					Wend
					dn.AddCodeNode(Self , cleansrc[oe..e] , oe , e) 					
				EndIf
				p=d+1
				Continue
			EndIf		
' ** Global
			If g<c And g<k And g<in And g<im And g<ic And g<db And g<as Then
				If (n._Parent And n._Parent._Parent) Then a = 1 Else a = 0
				Local line_end% = src.find(EOL , g) 
				e = g + 7
				Local oe:Int=e					
				While e < line_end
					Select src[e]
						Case Asc("~q")
							e = src.find("~q" , e + 1) 
							If e = - 1 Then e=line_end;Exit
						Case Asc("'")
							Exit
						Case Asc(",")
							gn[a].AddCodeNode(Self , cleansrc[oe..e] , oe , e) 
							oe=e+1
					End Select
					e:+1	
				Wend
				gn[a].AddCodeNode(Self , cleansrc[oe..e] , oe , e) 				
				p=g+1
				Continue
			EndIf			
' ** Const
			If c<k And c<in And c<im And c<ic And c<db And c<as Then
				If (n._Parent And n._Parent._Parent) Then a = 1 Else a = 0
				Local line_end% = src.find(EOL , c) 
				e = c + 6
				Local oe:Int=e					
				While e < line_end
					Select src[e]
						Case Asc("~q")
							e = src.find("~q" , e + 1) 
							If e = - 1 Then e=line_end;Exit
						Case Asc("'")
							Exit
						Case Asc(",")
							cn[a].AddCodeNode(Self , cleansrc[oe..e] , oe , e) 
							oe=e+1
					End Select
					e:+1	
				Wend
				cn[a].AddCodeNode(Self , cleansrc[oe..e] , oe , e) 			
				p=c+1
				Continue
			EndIf
' ** Framework
			If k<in And k<im And k<ic And k<db And k<as Then
				e=src.find(EOL,k)
				kn.AddCodeNode(Self,cleansrc[k+10..e],k,e, True)
				p=k+1
				Continue
			EndIf			
' ** Include
			If in<im And in<ic And in<db And in<as Then
				e=src.find(EOL,in)
				inn.AddCodeNode(Self,cleansrc[in+8..e],in,e, True)
				p=in+1
				Continue
			EndIf
' ** Import
			If im<ic And im<db And im<as Then
				e=src.find(EOL,im)
				imn.AddCodeNode(Self,cleansrc[im+7..e],im,e, True)
				p=im+1
				Continue
			EndIf
' ** Incbin
			If ic<db And ic<as Then
				e=src.find(EOL,ic)
				icn.AddCodeNode(Self,cleansrc[ic+7..e],ic,e, True)
				p=ic+1
				Continue
			EndIf
' ** DebugLog / Stop
			If db < as Then
				e=src.find(EOL,db)
				dbg.AddCodeNode(Self,cleansrc[db..e],db,e, True)
				p=db+1
				Continue
			Else
' ** Assert
				e=src.find(EOL,as)
				asr.AddCodeNode(Self,cleansrc[as+6..e],as,e, True)
				p=as+1
				Continue
			EndIf
		Wend
		
	End Method
	
	Function Strim$(s$)
		'removes quotes and comments from a string
		Local q:Int,c:Int
		q = s.find("~q") 
		c = s.find("'")
		While q > - 1 Or c > - 1
			If q > - 1 And (q < c Or c = - 1) ' quote
				c = s.find("~q" , q + 1)
				If c = - 1 Then
					s = s[..q]
				Else
					s = s[..q] + s[c + 1..]
				EndIf
			Else ' comment 
				s=s[..c]			
			EndIf
			q = s.find("~q") 
			c = s.find("'")			
		Wend
		Return s
	End Function
	
	Method GetNode:TDataNode()
		Local	root:TCodeNode
		'DebugLog "GetNode:Enter"
		root=New TCodeNode
		root.setname(StripDir(path))
		root.count=cleansrc.length
		parsebmx(root) 
		If host.options_tools.sorted Then root.sortkids
		If codenode
			codenode.Sync(root)	
			root.Free
		Else
			codenode=root
		EndIf
		'DebugLog "GetNode:Exit"
		Return codenode
	End Method

	Function IsCode:Int(src$,p:Int)
		Local	l:Int,q:Int,n:Int
		l=src.FindLast(EOL$,src.length-p)
		If l=-1 l=0
		q=src.Find("'",l)
		If q<>-1 And q<p Return False
		q=l
		While q<p
			q=src.Find(Chr(34),q)+1 ' looking for "
			If q=0 Exit
			If q<=p n:+1
		Wend
		If n&1 Return False
		Return True
	End Function

	Function FindLabel:Int(src$,pos:Int)
		Local	p:Int , q:Int , c:Int
		
		Assert src=src.tolower() ' src is assumed to be lowercase on input!		
		
		While pos>=0
			p=src.Find("#",pos)
			If p=-1 Exit
			q=p
			While q>0
				q:-1
				c=src[q]
				If c=13 Return p
				If c=10 Return p
				If c=32 Or c=9 Continue
				Exit
			Wend
			If q=0 Return p
			pos=p+1
		Wend
		Return src.length
	End Function

	Function FindToken:Int(token$,src$,pos:Int)	'lowercase src only!
		Local	p:Int,c:Int,n:Int
		n=token.length
		While pos>=0
			p=src.Find(token,pos)
			If p=-1 Exit
			c=10
			If p>0 c=src[p-1]
			If IsNotAlphanumeric(c)
				If p+n<src.length c=src[p+n]
				If isnotalphanumeric(c)
					If iscode(src,p)
						If p<4 Or src[p-4..p]<>"end " Return p
					EndIf
				EndIf
			EndIf
			pos=p+1
		Wend
		Return src.length
	End Function

	Function FindEndToken:Int(token$,src$,pos:Int,returnlast:Int=False)	'if true returns first character after endtoken
		Local	p:Int,q:Int,e$,n:Int

		Assert src=src.tolower() ' src is assumed to be lowercase on input!

		p=pos
		e$="end"+token
		n=e.length
		While p<=src.length
			p=src.Find(e$,p)
			If p=-1 Exit
			If p+n=src.length Or isnotalphanumeric(src[p+n])
				If iscode(src,p)
					If p=0 Or isnotalphanumeric(src[p-1]) Exit
				EndIf
			EndIf
			p=p+n
		Wend
		If p=-1 p=src.length Else If returnlast p:+n
		q=pos
		e$="end "+token
		n=e.length
		While q<=src.length
			q=src.Find(e$,q)
			If q=-1 Exit
			If q+n=src.length Or isnotalphanumeric(src[q+n])
				If iscode(src,q)
					If q=0 Or isnotalphanumeric(src[q-1]) Exit
				EndIf
			EndIf
			q=q+n
		Wend
		If q=-1 q=src.length Else If returnlast q:+n
		Return Min(p,q)
	End Function

' rem and endrem must be first nonwhitespace on line - following funcs are for lowercase src only

	Function FindRem:Int(src$ , pos:Int) 
		Assert src=src.tolower() ' src is assumed to be lowercase on input!		
		While pos<=src.length
			pos=FindToken("rem",src,pos)
			If pos=src.length Exit
			If IsFirstCharOnLine(src,pos) Return pos
			pos:+1
		Wend
		Return pos
	End Function

	Function FindEndRem:Int(src$,pos:Int,returnlast:Int=False)
		Local	i:Int , c:Int
		
		Assert src=src.tolower() ' src is assumed to be lowercase on input!
		
		While pos<=src.length
			pos=FindEndToken("rem",src,pos)
			If pos=src.length Exit
			If IsFirstCharOnLine(src,pos)
				If returnlast
					If src[pos+5]=Asc("e") Or src[pos+5]=Asc("E") pos:+1
					pos:+6
					If pos<src.length Then
						Local qq:Int=src.find(EOL,pos)
						If qq>pos Then pos=qq
					End If

				EndIf
				Return pos
			EndIf
			pos:+1
		Wend
		Return src.length
	End Function

	Function FindPrevRem:Int(src$,pos:Int)	'lowercase src only!
		Local	p:Int , c:Int
		
		Assert src=src.tolower() ' src is assumed to be lowercase on input!		
		pos:+3 ' allow for the 3 chars of Rem
		If pos>=src.length Then pos=src.length
		While pos>=0
			p=src.FindLast("rem",src.length-pos) ' second param is chars from end to work back from
			If p=-1 Exit
			If (p=src.length-3 Or isnotalphanumeric(src[p+3])) And IsFirstCharOnLine(src,p) Return p
			pos=p-1
		Wend
		Return -1
	End Function

	Method HighlightRegion(src$ , p:Int , p1:Int)
		Local lsrc$
		Local autocap:Int
		Local style:TTextStyle[4] , s:TTextStyle
		Local t$ , h$ , a:Int
		Local q:Int , r:Int
		Local cpos:Int
		
DebugLog "Highlighting from "+p+" to "+p1		
		
		' let's just grab a few things for convenience...
		style = options_editor.styles
		autocap=options_code.autocapitalize		
		lsrc = src.ToLower()
		cpos = TextAreaCursor(_textarea,TEXTAREA_CHARS)
				
		' now lets get on with highlighting code
		s=style[options_editor.NORMAL]
		If p1 > p Then s.Format(_textarea , p , p1 - p) 
		While p<p1
			host.UpdateProgress("Highlighting Code",(p*100)/p1)
			a = src[p]

' quoted strings
			If a=34
				q=p1
				r=src.find(Chr(34),p+1)
				If r>-1 And r<q q=r+1
				r=src.find(EOL,p+1)
				If r>-1 And r<q q=r
				s=style[options_editor.QUOTED]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf
			
' reflection attributes
			If a=Asc("{")
				q=p1
				r=src.find("}",p+1)
				If r>-1 And r<q q=r+1
				r=src.find(EOL,p+1)
				If r>-1 And r<q q=r
				s=style[options_editor.ATTRIBUTE]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf	
			
' single line comments
			If a=39
				q=p1
				r=src.find(EOL,p+1)
				If r>-1 And r<q q=r
				s=style[options_editor.COMMENT]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf

' decimal numbers
			If 48 <= a And a <= 57
				q=p+1
				While q<p1
					a=src[q]
					If a < 48 Or 57 < a Then Exit
					q:+1
				Wend
				s=style[options_editor.NUMERAL]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf

' hexadecimal numbers
			If a = 36
				q = p + 1

				While q < p1
					a = src[q]
					If a < 48 Or (a > 57 And ((a & $DF) < 65 Or (a & $DF) > 70)) Then Exit
					q :+ 1
				Wend

				If q > p + 1
					s = style[options_editor.NUMERAL]
					s.Format(_textarea, p, q - p)
				EndIf

				p = q
				Continue
			EndIf

' binary numbers
			If a = 37
				q = p + 1

				While q < p1
					a = src[q]
					If a < 48 Or a > 49 Then Exit
					q :+ 1
				Wend

				If q > p + 1
					s = style[options_editor.NUMERAL]
					s.Format(_textarea, p, q - p)
				EndIf

				p = q
				Continue
			EndIf

' tokens
			If (a>=65 And a<91) Or (a>=97 And a<123) Or (a=95)
				q=p+1
				While q<p1
					a=src[q]
					If a<48 Exit
					If a>=58 And a<65 Exit
					If a>=91 And a<95 Exit
					If a=96 Exit
					If a>122 Exit
					q:+1
				Wend
				t$=src[p..q]
				h$=host.quickhelp.token(t$)
				If h$
					If h$<>t$ And autocap
						If cpos<p Or cpos>q ' needs avoid cpos - otherwise we end up over-aggressive in capitalising & highlighting!
							SetTextAreaText _textarea,h$,p,Len h
						Else
							'TODO: deferpos - what is it for?
							'deferpos=q
						EndIf
					EndIf
					s=style[options_editor.KEYWORD]
					If h$="Rem" And IsFirstCharOnLine(lsrc,p)	' Not (p>4 And lsrc[p-4..p]="end ")
						If q <> p 'TODO: Should this be returned to being cpos too?
							q=FindEndRem(lsrc,p,True)
							s=style[options_editor.COMMENT]
						Else
							deferpos=q
						EndIf
					EndIf
					s.Format(_textarea,p,q-p)
				EndIf
				p=q
				Continue
			EndIf
			p:+1
		Wend		
		style[options_editor.NORMAL].Format(_textarea , p1 , 0) 
	End Method

	Function AutoTidyLine$(oldtext$)
		Local 	pos%		' current working position (between q and c) of processing
		Local 	char$, charbef$, charaft$	' temp vars holding characters being considered
		Local	inquote%		'are we inside a text literal?
		Local	newtext$
		Local	seen_a_char%
		
		newtext$=oldtext$
		pos=0
		inquote=False
		While pos<newtext.length
			If pos>0 Then charbef=Chr(newtext[pos-1]) Else charbef=""
			If pos < newtext.length - 1 Then charaft = Chr(newtext[pos + 1]) 
			
			'don't want to auto-tidy code inside quotes!
			char = Chr(newtext[pos])
			If char = Chr(34) Then inquote = Not(inquote) 
			If char = "'" Then Exit ' ignore comments				
			If Not inquote Then
				If Not seen_a_char And char <> " " Then seen_a_char = True
				Select char
					Case " "
						' only scrunch spaces together if not at start of line
						If seen_a_char And charaft=" " Then char=""
					Case "("
						If charaft="(" Then char="( "
					Case ")"
						If charaft=")" Then char=") "
					Case "+","-","*","/","^","&","|","~~"
						If charbef<>" " And charbef<>":" Then char=" "+char
						If charaft<>" " Then char=char+" "
					Case "=","<",">"
						If charbef<>" " And charbef<>"=" And charbef<>"<" And charbef<>">" Then char=" "+char
						If charaft<>" " And charaft<>"=" And charaft<>"<" And charaft<>">"Then char=char+" "
					Case ";",","
						If _autotidymode & $1 Then ' space before seperators
							If charbef <> " " And charbef <> char Then char = " " + char
						EndIf
						If charaft <> " " And charaft <> char Then char = char + " "
				End Select
			EndIf
			newtext=newtext[..pos]+char+newtext[pos+1..]
			pos=pos+char.length
		Wend
		'remove any trailing spaces
		While newtext<>"" And newtext[Len(newtext)-1]=32
			newtext=newtext[..Len(newtext)-1]
		Wend
		Return newtext$
	End Function

	Function IsCodeBlockStart%(src$)
		' used by auto-indent to add an extra tab, and potentially also by code folding in the future
		'TODO: Improve detection of if's on same line

		Local p%,word$
		'tests the code segment to see if it is the start of a block
		src=Trim(src)
		'find the first word
		p=src.find(" ")
		If p>0 Then word=src[..p] Else word=src
		word=word.tolower()
		Select word
		  	Case "function","method","type"
				Return 1
			Case "while"
				src=src.tolower()
				If findtoken("wend",src,p)=src.length;Return 1
			Case "repeat"
				src=src.tolower()
				If findtoken("until",src,p)=src.length;Return 1
			Case "for"
				src=src.tolower()
				If findtoken("next",src,p)=src.length;Return 1

Rem	these ones are all a bit dificult and / or contentious	for automatically indenting
			Case "if"
				' yuck - how to spot if's that are all on one line?
				Return True
			Case "select"
				Return True
			Case "else","elseif"
				Return True
EndRem
		End Select
		Return 0
	End Function

	Method BuildSource(quick:Int,debug:Int,mt:Int,apptype:Int,targetplatform:Int,run:Int)
		'TODO: Should this really be in the code panel, or should it be in the Host?
		Local cmd$,out$,arg$
		cmd$=quote(bmkpath)
		cmd$:+" makeapp"
		If run Then cmd$:+" -x"
		If debug Then cmd$:+" -d" Else cmd$:+" -r"	'-v
		If mt Then cmd$:+ " -h"
		
		' handle different apptypes
		If apptype = APPTYPE_GUI Then cmd$:+ " -t gui"
		If apptype = APPTYPE_WXGUI Then
			If targetplatform = TARGETPLATFORM_MACOS_NATIVE Or targetplatform = TARGETPLATFORM_MACOS_UNIVERSAL Then
				cmd$:+ " -b bah.appstub -t gui"
			Else
				Notify "@@error_message/msg_target_error";Return
			EndIf
		EndIf
		
		' handle different target platforms
		Select targetplatform
			?win32
				Case TARGETPLATFORM_WIN32 ; cmd$:+ ""
				Case TARGETPLATFORM_LINUX ; Notify "@@error_message/msg_target_win_to_linux";Return
				Case TARGETPLATFORM_MACOS_NATIVE  ; Notify "@@error_message/msg_target_win_to_macos";Return
				Case TARGETPLATFORM_MACOS_UNIVERSAL  ; Notify "@@error_message/msg_target_win_to_macos";Return
			?macosx86
				Case TARGETPLATFORM_WIN32 ; cmd$:+ " -l win32"
				Case TARGETPLATFORM_LINUX ; Notify "@@error_message/msg_target_macos_to_linux";Return
				Case TARGETPLATFORM_MACOS_NATIVE  ; cmd$:+ ""
				Case TARGETPLATFORM_MACOS_UNIVERSAL  ; cmd$:+ " -i"
			?macosppc
				Case TARGETPLATFORM_WIN32 ; Notify "@@error_message/msg_target_macosppc_to_win";Return
				Case TARGETPLATFORM_LINUX ; Notify "@@error_message/msg_target_macos_to_linux";Return
				Case TARGETPLATFORM_MACOS_NATIVE  ; cmd$:+ ""
				Case TARGETPLATFORM_MACOS_UNIVERSAL  ; Notify "@@error_message/msg_target_macosppc_to_universal";Return
			?linux
				Case TARGETPLATFORM_WIN32 ; cmd$:+ " -l win32"
				Case TARGETPLATFORM_LINUX ; cmd$:+ ""
				Case TARGETPLATFORM_MACOS_NATIVE ; Notify "@@error_message/msg_target_linux_to_macos";Return
				Case TARGETPLATFORM_MACOS_UNIVERSAL ; Notify "@@error_message/msg_target_linux_to_macos";Return
			?			
			Default
				Notify "@@error_message/msg_target_error";Return
		End Select
		
		If Not quick Then cmd$:+" -a"
		If debug
			out$=StripExt(host.FullPath(path))+".debug"
			cmd$:+" -o "+quote(out$)+" "			'QUOTES+out$+QUOTES+" "
		EndIf
		
		cmd$:+" "+quote(host.FullPath(path))
		If run
			arg$=host.GetCommandLine()
			If arg cmd$:+" "+arg
		EndIf
		host.execute cmd,"@@output_panel/cmd_build|"+StripExt(StripDir(path)),"", True, Self	',exe$
	End Method
	
	Method ParseError:Int( err:String)
		DebugLog "TBMXCodePanel.ParseError()"
		Local mess$,file$,p:Int,q:Int
		Local line:Int,column:Int				
		If err$[..13]="Compile Error"
			err=err[14..]
			p=err.find(EOL$)
			If p=-1 p=Len err
			mess=err[..p]
			err=err[p+1..]
			If err[..1]="["
				p=err.find("]")
				If p=-1 p=Len err
				file$=err[1..p]
				p=file.find(";")+1
				If p=0 p=Len err
				q=file.find(";",p)+1
				If q=0 q=Len err
				line=Int(file[p..q-1])
				column=Int(file[q..])
				file=host.FullPath(file[..p-1])
				host.SelectError file,column,line
			EndIf
			mess = TranslateMessage(mess) 
			Print "BMXParseError: pre-notify"
			Local oldapptitle$ = AppTitle
			AppTitle = LanguageString("error_message/compile_error_title")
			Notify mess
			Print "BMXParseError: post-notify"
			AppTitle = oldapptitle$
			host.SetStatus mess
			Print "BMXParseError: exit"
			Return True
		EndIf
		Return False
	EndMethod

	Method Init( path:String, host:TBMXHost)
		Super.Init( path, host)
		codetype=CODETYPE_BLITZMAX
		' setup autocomplete window
		Autoc = New TBMXAutoComplete
		Autoc.Init( Self , host.options_autoc)
		host.AddHandler( Autoc)
	EndMethod
	
	Function Create:TBMXCodePanel(path$ , host:TBMXHost) 
'DebugLog "TBMXCodePanel.Create("+path$+")"		
		Local code:TBMXCodePanel = New TBMXCodePanel
		code.codetype = CODETYPE_BLITZMAX
		code.Init( path , host)
'DebugLog "TBMXCodePanel.Create:Exit"		
		Return code
	End Function
End Type


'MARK: TUnknownCodePanel stub
Const CODETYPE_UNKNOWN:Int	=	0
Type TUnknownCodePanel Extends TCodePanel
	Function Create:TUnknownCodePanel(path$,host:TBMXHost)
		Local code:TUnknownCodePanel = New TUnknownCodePanel
		code.codetype = CODETYPE_UNKNOWN
		code.Init( path, host)
		Return code
	End Function

	Method BuildSource( quick:Int,debug:Int,mt:Int,apptype:Int,targetplatform:Int,run:Int)
		Local cmd$
		'see what the system shell thinks of the file
		Local cd$=CurrentDir()
		ChangeDir ExtractDir(path)
		cmd=StripDir(path)
		host.execute cmd,"@@output_panel/cmd_build|"+cmd
		ChangeDir cd
	End Method
	
	Method HighlightRegion(src$,p:Int,p1:Int)
		' no code highlighting in unknown files!
	EndMethod	
	
	Function FindPrevRem%(src$,pos:Int);Return -1;End Function
	Function FindEndRem%(src$,pos:Int,returnlast:Int=False);Return -1;End Function
	Function FindRem:Int(src$,pos:Int);Return -1;End Function			

EndType

'MARK: TCCodePanel stub
Const CODETYPE_C:Int		 = 	2
TCodePanel.Register(CODETYPE_C , ["c" , "h"] , Byte Ptr TCCodePanel.Create) 

Type TCCodePanel Extends TCodePanel
	Function Create:TCCodePanel(path$,host:TBMXHost)
		Local code:TCCodePanel = New TCCodePanel
		code.codetype = CODETYPE_C
		code.Init( path, host)
		Return code
	End Function

	Function IsCode:Int(src$,p:Int)
		Local	l:Int,q:Int,n:Int
		l=src.FindLast(EOL$,src.length-p)
		If l=-1 l=0
		q=src.Find("//",l)
		If q<>-1 And q<p Return False
		q=l
		While q<p
			q=src.Find(Chr(34),q)+1 ' looking for "
			If q=0 Exit
			If q<=p n:+1
		Wend
		If n&1 Return False
		Return True
	End Function

	Function IsCodeBlockStart%(src$)
		' used by auto-indent to add an extra tab, and potentially also by code folding in the future
		Local n%
		Local bracecount:Int
		If src.length=0 Then Return 0
		For n% = 0 To src.length-1
			If src[n] = Asc("{") Then bracecount:+ 1
			If src[n] = Asc("}") Then bracecount:- 1
		Next
		If bracecount<0 Then bracecount=0
		Return bracecount
	End Function

	Function FindRem:Int(src$ , pos:Int) 
		pos = src.find("/*" , pos) 
		If pos=-1 Then pos=src.length
		Return pos
	End Function

	Function FindEndRem:Int(src$,pos:Int,returnlast:Int=False)
		pos = src.find("*/" , pos)
		If returnlast Then pos:+2
		If pos=-1 Then pos=src.length
		Return pos
	End Function
	
	Function FindPrevRem:Int(src$,pos:Int)
		Local	p:Int , c:Int
		pos:+2
		If pos > src.length Then pos = src.length
		p = src.FindLast("/*" , src.length - pos) ' second param is chars from end to work back from
		Return p
	End Function

	Function FindToken:Int(token$ , src$ , pos:Int)	
		'TODO: Can FindToken be made generic?
		Local	p:Int,c:Int,n:Int
		n=token.length
		While pos>=0
			p=src.Find(token,pos)
			If p=-1 Exit
			c=10
			If p>0 c=src[p-1]
			If IsNotAlphanumeric(c)
				If p+n<src.length c=src[p+n]
				If isnotalphanumeric(c)
					If iscode(src,p)Then Return p
				EndIf
			EndIf
			pos=p+1
		Wend
		Return src.length
	End Function	

	Method BuildSource( quick:Int,debug:Int,mt:Int,apptype:Int,targetplatform:Int,run:Int)
		'TODO: Should this really be in the code panel, or should it be in the Host?
		Local cmd$,out$,arg$
		cmd$=quote(bmkpath)
		cmd$:+" makeapp"
		If run cmd$:+" -x"
		If debug cmd$:+" -d" Else cmd$:+" -r"	'-v
		If mt cmd$:+" -h"
	
		' handle different apptypes
		If apptype = APPTYPE_GUI Then cmd$:+ " -t gui"
		If apptype = APPTYPE_WXGUI Then
			If targetplatform = TARGETPLATFORM_MACOS_NATIVE Or targetplatform = TARGETPLATFORM_MACOS_UNIVERSAL Then
				cmd$:+ " -b bah.appstub -t gui"
			Else
				cmd$:+ " -t gui"
			EndIf
		EndIf
		
		' handle different target platforms
		Select targetplatform
			?win32
				Case TARGETPLATFORM_WIN32 ; cmd$:+ ""
				Case TARGETPLATFORM_LINUX ; Notify "Cannot compile Linux binaries on Windows!";Return
				Case TARGETPLATFORM_MACOS_NATIVE  ; Notify "Cannot compile MacOS binaries on Windows!";Return
				Case TARGETPLATFORM_MACOS_UNIVERSAL  ; Notify "Cannot compile MacOS binaries on Windows!";Return
			?macosx86
				Case TARGETPLATFORM_WIN32 ; cmd$:+ " -l win32"
				Case TARGETPLATFORM_LINUX ; Notify "Cannot compile Linux binaries on MacOS!";Return
				Case TARGETPLATFORM_MACOS_NATIVE  ; cmd$:+ ""
				Case TARGETPLATFORM_MACOS_UNIVERSAL  ; cmd$:+ " -i"
			?macosppc
				Case TARGETPLATFORM_WIN32 ; Notify "Cannot compile Windows binaries on PowerPC MacOS!";Return
				Case TARGETPLATFORM_LINUX ; Notify "Cannot compile Linux binaries on MacOS!";Return
				Case TARGETPLATFORM_MACOS_NATIVE  ; cmd$:+ ""
				Case TARGETPLATFORM_MACOS_UNIVERSAL  ; Notify "Cannot compile Universal binaries on PowerPC MacOS!";Return
			?linux
				Case TARGETPLATFORM_WIN32 ; cmd$:+ " -l win32"
				Case TARGETPLATFORM_LINUX ; cmd$:+ ""
				Case TARGETPLATFORM_MACOS_NATIVE ; Notify "Cannot compile MacOS binaries on Linux!";Return
				Case TARGETPLATFORM_MACOS_UNIVERSAL ; Notify "Cannot compile MacOS binaries on Linux!";Return
			?			
			Default
				Notify "Error in building bmk string!";Return
		End Select	
	
		If Not quick cmd$:+" -a"
		If debug
			out$=StripExt(host.FullPath(path))+".debug"
			cmd$:+" -o "+quote(out$)+" "			'QUOTES+out$+QUOTES+" "
		EndIf
		cmd$:+" "+quote(host.FullPath(path))
		If run
			arg$=host.GetCommandLine()
			If arg cmd$:+" "+arg
		EndIf
		host.execute cmd,"@@output_panel/cmd_build|"+StripExt(StripDir(path)), "", True, Self ',exe$
	EndMethod
	
	Method ParseError:Int( err:String)
		DebugLog "TCCodePanel.ParseError()"
		Local file:String, mess:String, line:Int
		Local founderror:Int = False
		Local e:String[] = err.Split(EOL)
		' skip first line if there are more
		If e.Length > 1 Then e = e[1..]
		For Local l:Int = 0 Until e.Length
			Local emsg:String = e[l]

			' skips include file error header
			If emsg.Find("In file included from") <> -1 Then Continue
			' skip Function error header
			If emsg.Find("In function `") <> -1 Then Continue
			
			Local c:String[] = emsg.Trim().Split(":")
			?Win32
			' fix for DRIVE: syntax
			If (c.Length >= 2) And (c[0].Length = 1) Then
				c[1] = c[0] + ":" + c[1]
				c = c[1..]
			EndIf
			?
			If c.Length >= 5 Then Continue
			If (c.Length >= 1) And Not file Then file = c[0]
			If (c.Length >= 2) And (c[1].Length > 0) And (c[1][0] = Asc(" ")) Then
				If Not mess Then mess = c[1][1..].Trim()
			ElseIf (c.Length >= 2) Then
				If Not line Then line = c[1].ToInt()
				If c.Length >= 4 And Not mess Then mess = c[3][1..].Trim()
			EndIf
			If file And line Then
				host.SelectError( file, 0,line)
				founderror = True
			EndIf
			If mess And line Then		
				Local oldapptitle$ = AppTitle
				AppTitle = LanguageString("error_message/gcc_error_title")
				Notify mess
				AppTitle = oldapptitle$
				founderror = True
			EndIf
			If file And line And mess Then Exit
		Next
		Return founderror
	EndMethod

	Method Init( path:String, host:TBMXHost)
		Super.Init( path, host)
		codetype = CODETYPE_C
		' set up keywords
		AddKeyword "unsigned <var> : designates a variable as being unsigned"
		AddKeyword "signed <var> : designates a variable as being signed"
		AddKeyword "void(a$='') : tyest"
		AddKeyword "char"
		AddKeyword "short"
		AddKeyword "int"
		AddKeyword "long"
		AddKeyword "float"
		AddKeyword "double"
		AddKeyword "typedef"
		AddKeyword "struct"
		AddKeyword "union"
		AddKeyword "static"
		AddKeyword "extern"
		AddKeyword "register"
		AddKeyword "const"
		AddKeyword "volatile"
		AddKeyword "enum"
		AddKeyword "auto"
		AddKeyword "if"
		AddKeyword "else"
		AddKeyword "switch"
		AddKeyword "case"
		AddKeyword "default"
		AddKeyword "for"
		AddKeyword "while"
		AddKeyword "do"
		AddKeyword "break"
		AddKeyword "continue"
		AddKeyword "goto"
		AddKeyword "return"
		AddKeyword "sizeof"
		_caseSensitivekeywords = true
		
		' setup autocomplete window
		Autoc = New TCAutoComplete
		Autoc.Init( Self , host.options_autoc) 
		host.AddHandler( Autoc)
	EndMethod
		
	Method HighlightRegion(src$ , p:Int , p1:Int)
		Local lsrc$
		Local style:TTextStyle[4] , s:TTextStyle
		Local t$ , h$ , a:Int
		Local q:Int,r:Int
		
DebugLog "Highlighting from "+p+" to "+p1		
		
		' let's just grab a few things for convenience...
		style = options_editor.styles
		lsrc = src.ToLower()
				
		' now lets get on with highlighting code
		s = style[options_editor.NORMAL]
		If p1 > p Then s.Format(_textarea , p , p1 - p) 
		While p<p1
			host.UpdateProgress("Highlighting Code",(p*100)/p1)
			a = src[p]	
			
' quoted strings
			If a=Asc("~q") Then
				q=p1
				r=src.find( "~q",p+1)
				If r>-1 And r<q q=r+1
				r=src.find(EOL,p+1)
				If r>-1 And r<q q=r
				s=style[options_editor.QUOTED]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf
' quoted chars
			If a=Asc("'") Then
				q=p1
				r=src.find( "'",p+1)
				If r>-1 And r<q q=r+1
				r=src.find(EOL,p+1)
				If r>-1 And r<q q=r
				s=style[options_editor.QUOTED]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf
			
' macroes
			If a=Asc("#") Then
				'TODO: (TCCodePanel) syntax highlight multiline macroes
				q=p1
				r=src.find(EOL,p+1)
				If r>-1 And r<q q=r
				s=style[options_editor.ATTRIBUTE]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf
			
' comments
			If a = Asc("/") Then
				q = p + 1
				If q >= src.length Then p:+ 1 ; Continue
				Select src[q]
					Case Asc("/") ' single line comments
						q :+ 1
						r=src.find( EOL,q)
						If r <> -1 Then q=r
						s=style[options_editor.COMMENT]
						s.Format(_textarea,p,q-p)
						p=q
						Continue
						
					Case Asc("*") ' multiline comments
						q :+ 1
						r = FindEndRem( src, q)
						If r <> -1 Then
							q=r+2
							s=style[options_editor.COMMENT]
							s.Format(_textarea,p,q-p)
							p=q
							Continue
						EndIf
				EndSelect
			EndIf

' hexadecimal numbers
			If a = Asc("0") And src[p+1] = Asc("x") Then
				q = p + 2

				While q < p1
					a = src[q]
					If a < 48 Or (a > 57 And ((a & $DF) < 65 Or (a & $DF) > 70)) Then Exit
					q :+ 1
				Wend

				If q > p + 1
					s = style[options_editor.NUMERAL]
					s.Format(_textarea, p, q - p)
				EndIf

				p = q
				Continue
			EndIf			

' decimal numbers
			If 48 <= a And a <= 57
				q=p+1
				While q<p1
					a=src[q]
					If a < 48 Or 57 < a Then Exit
					q:+1
				Wend
				s=style[options_editor.NUMERAL]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf

' tokens
			If (a>=65 And a<91) Or (a>=97 And a<123) Or (a=95)
				q=p+1
				While q<p1
					a=src[q]
					If a<48 Exit
					If a>=58 And a<65 Exit
					If a>=91 And a<95 Exit
					If a=96 Exit
					If a>122 Exit
					q:+1
				Wend
				t$ = src[p..q]
				If IsKeyword(t$) Then					
					s=style[options_editor.KEYWORD]
					s.Format(_textarea , p , q - p) 							
				EndIf
				p=q
				Continue
			EndIf
			p:+1
		Wend
	EndMethod
	
	Method GetTodoNode:TDataNode()
		Local	root:TTodoNode,n:TTodoNode
		Local tdn:TTodoNode
		Local bgn:TTodoNode
		Local fxn:TTodoNode
		Local bmn:TTodoNode
		Local	src$,csrc$,line:Int,col:Int
		Local	p:Int,p1:Int,r:Int,t:Int,m:Int,f:Int,l:Int,e:Int
		Local tok$ , lins$ , ccnt:Int = 0
'DebugLog "GetTodoNode:Enter"
		n=New TTodoNode
		n.setname(StripDir(path))
		n.count=src.length
		root=n

		tdn = root.AddTodoNode(Self,"TODO",-1,-1)
		bgn = root.AddTodoNode(Self,"BUG",-1,-1)
		fxn = root.AddTodoNode(Self,"FIX",-1,-1)
		bmn = root.AddTodoNode(Self,"MARK",-1,-1)

' extract todo info
		src=cleansrcl+"~n"
		csrc=cleansrc+"~n"
		p1=src.length
		p=0
		r=-1;t=-1

		While p<p1			'src.length

			e = src.Find(EOL,p)
			Local sz:Int = (e - p)+1
			lins = src[p..e].Trim() ' lower case

	' C++ style comments
			If lins.StartsWith("//") Then
				lins = lins[2..].Trim()
				
				If lins.StartsWith("todo:") Then
					n = tdn.AddTodoNode( Self, lins[5..], p,e)
				ElseIf lins.StartsWith("bug:") Then
					n = bgn.AddTodoNode( Self, lins[4..], p,e)
				ElseIf lins.StartsWith("fix:") Then
					n = fxn.AddTodoNode( Self, lins[4..], p,e)
				ElseIf lins.StartsWith("mark:") Then
					Local s:String = lins[5..].Trim()
					If s Then
						n = bmn.AddTodoNode( Self, s, p,e)
					Else
						n = bmn.AddTodoNode( Self, "Bookmark", p,e)
					EndIf
				EndIf
				
				p :+ sz
				
	' C style comments
			ElseIf lins.StartsWith("/*") Then
				' allow multiple lines (but still only 1 tag per comment)
				p :+ 2
				Local i:Int = src.Find( "*/", p)
				If i = -1 Then Continue ' no end found, skip it
				lins = src[p..i]
				Local sz:Int = lins.Length
				lins = lins.Trim()
				
				If lins.StartsWith("todo:") Then
					n = tdn.AddTodoNode( Self, lins[5..], p,e)
				ElseIf lins.StartsWith("bug:") Then
					n = bgn.AddTodoNode( Self, lins[4..], p,e)
				ElseIf lins.StartsWith("fix:") Then
					n = fxn.AddTodoNode( Self, lins[4..], p,e)
				ElseIf lins.StartsWith("mark:") Then
					Local s:String = lins[5..].Trim()
					If s Then
						n = bmn.AddTodoNode( Self, s, p,e)
					Else
						n = bmn.AddTodoNode( Self, "Bookmark", p,e)
					EndIf
				EndIf
				
				p :+ sz + 2
				
	' nothing found, skip line
			Else
				p :+ sz
			EndIf
		Wend
		
'DebugLog "GetTodoNode:New node=" + root._name + "(" + root._kids.count() + ")"
'If todonode<>Null Then DebugLog "GetTodoNode:Old node=" + todonode._name + "(" + todonode._kids.count() + ")"		
		If todonode
			todonode.Sync(root)
			root.Free
		Else
			todonode=root
		EndIf
'DebugLog "GetTodoNode:PostSync=" + todonode._name + "(" + todonode._kids.count() + ")"	
'		DebugLog "GetTodoNode:Exit"
		Return todonode
	EndMethod	
EndType

'MARK: TCPPCodePanel stub
Const CODETYPE_CPP:Int		=	3
TCodePanel.Register(CODETYPE_CPP , ["cpp" , "cxx" , "hpp"] , Byte Ptr TCPPCodePanel.Create)

Type TCPPCodePanel Extends TCCodePanel
	Function Create:TCPPCodePanel(path$,host:TBMXHost)
		Local code:TCPPCodePanel = New TCPPCodePanel
		code.codetype = CODETYPE_CPP
		code.Init( path, host)
		Return code
	End Function

	Method Init( path:String, host:TBMXHost)
		Super.Init( path , host) 
		AddKeyword "asm"
		AddKeyword "bool"
		AddKeyword "catch"
		AddKeyword "class"
		AddKeyword "const_cast"
		AddKeyword "delete"
		AddKeyword "dynamic_cast"
		AddKeyword "explicit"
		AddKeyword "export"
		AddKeyword "false"
		AddKeyword "friend"
		AddKeyword "inline"
		AddKeyword "mutable"
		AddKeyword "namespace"
		AddKeyword "new"
		AddKeyword "operator"
		AddKeyword "private"
		AddKeyword "protected"
		AddKeyword "public"
		AddKeyword "reinterpret_cast"
		AddKeyword "static_cast"
		AddKeyword "template"
		AddKeyword "this"
		AddKeyword "throw"
		AddKeyword "true"
		AddKeyword "try"
		AddKeyword "typeid"
		AddKeyword "typename"
		AddKeyword "using"
		AddKeyword "virtual"
		AddKeyword "wchar_t"
		codetype = CODETYPE_CPP
	EndMethod

'TODO: (TCPPCodePanel) implement own syntax highlighting for C++ ?
'	Method SetCode:Int(src$,diff:TDiff=Null)
'	EndMethod
EndType

'MARK: TObjectiveCPCodePanel stub
Const CODETYPE_OBJC:Int		=	4
TCodePanel.Register(CODETYPE_OBJC, ["m"] , Byte Ptr TObjectiveCCodePanel.Create)

Type TObjectiveCCodePanel Extends TCCodePanel
	Function Create:TObjectiveCCodePanel(path$,host:TBMXHost)
		Local code:TObjectiveCCodePanel = New TObjectiveCCodePanel
		code.codetype = CODETYPE_OBJC
		code.Init( path, host)
		Return code
	End Function

	Method Init( path:String, host:TBMXHost)
		Super.Init( path, host)
		codetype = CODETYPE_OBJC
	EndMethod
	
'TODO: (TObjectiveCCodePanel) implement own syntax highlighting for Objective-C ?
'	Method SetCode:Int(src$,diff:TDiff=Null)
'	EndMethod	
EndType

'MARK: TMarkupCodePanel stub (html xml ...)
Type TMarkupCodePanel Extends TCodePanel
	Const SUBLANG_NONE:Int = 0
	Const SUBLANG_JS:Int = 1
	Const SUBLANG_CSS:Int = 2
	
	Field sublang:Int = SUBLANG_NONE
	Field UseSublang:Int = False
	
	Method BuildSource(quick:Int , debug:Int ,mt:Int, apptype:Int ,targetplatform:Int, run:Int) 
		DebugLog path
		host.helppanel.Go "file://"+path
	End Method

	Function FindPrevRem%(src$ , pos:Int);Return -1;End Function
	Function FindEndRem%(src$,pos:Int,returnlast:Int=False);Return -1;End Function
	Function FindRem:Int(src$,pos:Int);Return -1;End Function	

	Method HighlightRegion(src$ , p:Int , p1:Int)
		Local lsrc$
		Local style:TTextStyle[4] , s:TTextStyle
		Local t$ , h$ , a:Int
		Local q:Int,r:Int
		Local intag:Int = False
		Local isendtag:Int = False
	
DebugLog "Highlighting from "+p+" to "+p1		
		
		' let's just grab a few things for convenience...
		style = options_editor.styles
		lsrc = src.ToLower()
				
		' now lets get on with highlighting code
		s=style[options_editor.NORMAL]	
		If p1 > p Then s.Format(_textarea , p , p1 - p) 
		While p<p1
			host.UpdateProgress("Highlighting Code",(p*100)/p1)
			a = src[p]	
			
' elements
			If a=Asc("<") Then
				intag = True
				isendtag = False
				q = p + 1
				' comment
				If (src[q] = Asc("!")) And (src.Find( "--", q+1) = q+1) Then
					q :+ 3
					r = src.find( "-->", q)
					If r <> -1 Then
						q=r+3
						If sublang = SUBLANG_NONE Then
							s=style[options_editor.COMMENT]
							s.Format( _textarea,p,q-p)
						EndIf
						p=q
						Continue
					EndIf
				Else
					' skip whitespace and /
					While q < src.length
						Select src[q]
							Case Asc("/")
								q :+ 1
								isendtag = True
								Continue
							Case Asc(" "),Asc("~t"),Asc("~r"),Asc("~n")
								q :+ 1
								Continue
						EndSelect
						Exit
					Wend
					' trace element name, stop at whitespace and > /
					r = q
					While q < src.length
						Select src[q]
							Case Asc(" "),Asc("~t"),Asc("~r"),Asc("~n"),Asc(">"),Asc("/")
								Exit
						EndSelect
						q :+ 1
					Wend
					t = src[r..q].ToLower()
					If UseSublang Then
						If t = "script" And Not isendtag Then
							sublang = SUBLANG_JS
						ElseIf t = "style" And Not isendtag Then
							sublang = SUBLANG_CSS
						Else
							sublang = SUBLANG_NONE
						EndIf
					EndIf
					If isendtag Then intag = False
					If q < src.Length Then
						s=style[options_editor.KEYWORD]
						s.Format( _textarea,r,q-r)
						p=q
						Continue
					EndIf
				EndIf
				p = q
			EndIf
			If a = Asc(">") Then intag = False
			If Not intag And (sublang <> SUBLANG_NONE) Then 
				q :+ 1
				p = q
				Continue
			EndIf
			
' inside element
			If intag Then
			
' double quoted strings
				If a=Asc("~q") Then
					q=p1
					r=src.find( "~q",p+1)
					If r>-1 And r<q q=r+1
					r=src.find(EOL,p+1)
					If r>-1 And r<q q=r
					s=style[options_editor.QUOTED]
					s.Format(_textarea,p,q-p)
					p=q
					Continue
				EndIf
' single quoted strings
				If a=Asc("'") Then
					q=p1
					r=src.find( "'",p+1)
					If r>-1 And r<q q=r+1
					r=src.find(EOL,p+1)
					If r>-1 And r<q q=r
					s=style[options_editor.QUOTED]
					s.Format(_textarea,p,q-p)
					p=q
					Continue
				EndIf			
				
				
' hexadecimal numbers
				If a = Asc("#") Then
					q = p + 1
	
					While q < p1
						a = src[q]
						If a < 48 Or (a > 57 And ((a & $DF) < 65 Or (a & $DF) > 70)) Then Exit
						q :+ 1
					Wend
	
					If q > p + 1
						s = style[options_editor.NUMERAL]
						s.Format(_textarea, p, q - p)
					EndIf
	
					p = q
					Continue
				EndIf			
	
' decimal numbers
				If 48 <= a And a <= 57
					q=p+1
					While q<p1
						a=src[q]
						If a < 48 Or 57 < a Then Exit
						q:+1
					Wend
					s=style[options_editor.NUMERAL]
					s.Format(_textarea,p,q-p)
					p=q
					Continue
				EndIf

' properties						
				If (a>=65 And a<91) Or (a>=97 And a<123) Or (a=95)
					q=p+1
					While q<p1
						a=src[q]
						If a<48 Exit
						If a>=58 And a<65 Exit
						If a>=91 And a<95 Exit
						If a=96 Exit
						If a>122 Exit
						q:+1
					Wend
					s=style[options_editor.ATTRIBUTE]
					s.Format(_textarea,p,q-p)							
					p=q
					Continue
				EndIf
			EndIf
			p:+1
		Wend
	EndMethod
EndType


'MARK: THTMLCodePanel stub
Const CODETYPE_HTML:Int	 = 	5
TCodePanel.Register(CODETYPE_HTML , ["htm" , "html"] , Byte Ptr THTMLCodePanel.Create)

Type THTMLCodePanel Extends TMarkupCodePanel
	Method New()
		UseSublang = True
	EndMethod
		
	Function Create:THTMLCodePanel(path$,host:TBMXHost)
		Local code:THTMLCodePanel = New THTMLCodePanel
		code.codetype = CODETYPE_HTML
		code.Init( path, host)
		Return code
	EndFunction

	Method Init( path:String, host:TBMXHost)
		Super.Init( path, host)
		codetype = CODETYPE_HTML
		' setup autocomplete window
		Autoc = New THTMLAutoComplete 
		Autoc.Init( Self , host.options_autoc) 
		host.AddHandler( Autoc)	
	EndMethod	
	
EndType

'MARK: TXMLCodePanel stub
Const CODETYPE_XML:Int	 = 	6
TCodePanel.Register(CODETYPE_XML, ["xml"] , Byte Ptr TXMLCodePanel.Create)

Type TXMLCodePanel Extends TMarkupCodePanel
	Method New()
		UseSublang = False
	EndMethod
	
	Function Create:TXMLCodePanel(path$,host:TBMXHost)
		Local code:TXMLCodePanel = New TXMLCodePanel
		code.codetype = CODETYPE_XML
		code.Init( path, host)
		Return code
	EndFunction

	Method Init( path:String, host:TBMXHost)
		Super.Init( path, host)
		codetype = CODETYPE_XML
	EndMethod
EndType

'MARK: TConfigCodePanel stub
Const CODETYPE_CONFIG:Int		 = 	7
TConfigCodePanel.Register(CODETYPE_CONFIG, ["ini" , "cfg", "conf", "config"] , Byte Ptr TConfigCodePanel.Create) 

Type TConfigCodePanel Extends TCodePanel
	Function Create:TConfigCodePanel(path$,host:TBMXHost)
		Local code:TConfigCodePanel = New TConfigCodePanel
		code.codetype = CODETYPE_CONFIG
		code.Init( path, host)
		Return code
	End Function

	Method Init( path:String, host:TBMXHost)
		Super.Init( path, host)
		codetype = CODETYPE_CONFIG
	EndMethod

	Function FindPrevRem%(src$,pos:Int);Return -1;End Function
	Function FindEndRem%(src$,pos:Int,returnlast:Int=False);Return -1;End Function
	Function FindRem:Int(src$,pos:Int);Return -1;End Function	

	Method HighlightRegion(src$ , p:Int , p1:Int)
		Local lsrc$
		Local autocap:Int
		Local style:TTextStyle[4] , s:TTextStyle
		Local t$ , h$ , a:Int
		Local q:Int,r:Int
		
DebugLog "Highlighting from "+p+" to "+p1		
		
		' let's just grab a few things for convenience...
		style = options_editor.styles
		autocap=options_code.autocapitalize		
		lsrc = src.ToLower()
				
		' now lets get on with highlighting code
		s = style[options_editor.NORMAL]

		If p1 > p Then s.Format(_textarea , p , p1 - p) 
		While p<p1
			host.UpdateProgress("Highlighting Code",(p*100)/p1)
			a = src[p]
			
' quoted strings
			If a=Asc("~q") Then
				q=p1
				r=src.find( "~q",p+1)
				If r>-1 And r<q q=r+1
				r=src.find(EOL,p+1)
				If r>-1 And r<q q=r
				s=style[options_editor.QUOTED]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf
' quoted chars
			If a=Asc("'") Then
				q=p1
				r=src.find( "'",p+1)
				If r>-1 And r<q q=r+1
				r=src.find(EOL,p+1)
				If r>-1 And r<q q=r
				s=style[options_editor.QUOTED]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf
' comments
			If a=Asc("#") Or a=Asc(";") Then
				q=p1
				r=src.find(EOL,p+1)
				If r>-1 And r<q q=r
				s=style[options_editor.COMMENT]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf
' sections
			If a = Asc("[") Then
				q=p1
				style[options_editor.KEYWORD].Format( _textarea, p, 1)
				p :+ 1
				q :+ 1
				
				r=src.find( "]",p+1)
				If r>-1 And r<q q=r+1
				r=src.find(EOL,p+1)
				If r>-1 And r<q q=r
				style[options_editor.QUOTED].Format(_textarea,p,q-p)
				p=q
				
				style[options_editor.KEYWORD].Format( _textarea, q-1, 1)
				Continue
			EndIf
' keys
			If a = Asc("=") Then
				q = p + 1
				Local e:Int = src.Find( EOL, p)
				Local b:Int = src.FindLast( EOL, src.Length-p)
				If e > b Then
					style[options_editor.ATTRIBUTE].Format( _textarea, b+1, p-b)
					style[options_editor.KEYWORD].Format( _textarea, p, 1)
				EndIf
				p=q
				Continue
			EndIf
			
' hexadecimal numbers
			If a = Asc("0") And src[p+1] = Asc("x") Then
				q = p + 2

				While q < p1
					a = src[q]
					If a < 48 Or (a > 57 And ((a & $DF) < 65 Or (a & $DF) > 70)) Then Exit
					q :+ 1
				Wend

				If q > p + 1
					s = style[options_editor.NUMERAL]
					s.Format(_textarea, p, q - p)
				EndIf

				p = q
				Continue
			EndIf

' decimal numbers
			If 48 <= a And a <= 57
				q=p+1
				While q<p1
					a=src[q]
					If a < 48 Or 57 < a Then Exit
					q:+1
				Wend
				s=style[options_editor.NUMERAL]
				s.Format(_textarea,p,q-p)
				p=q
				Continue
			EndIf
						
			p:+1
		Wend

	EndMethod
EndType
