'help.bmx

SuperStrict

Import BRL.Retro
Import BRL.MaxUtil
Import "datanodes.bmx"
Import "host.bmx" 
Import "projects.bmx"
Import "options.bmx"
Import "version.bmx"		' the versioning file
Import maxgui.proxygadgets

Incbin "about.html"
Incbin "version_history.html"

Type TToken
	Field token$
	Field help$
	Field ref$	
	
	Method Create:TToken(t$, h$, r$)
		token=t
		help=h
		ref=r
		
		Return Self
	End Method
End Type

Type TQuickHelp
	Field map:TMap=New TMap	'key=lower(token) value=token:TToken
		
	Method AddCommand:TQuickHelp(t$,l$,a$)
		map.Insert Lower(t$),New TToken.Create(t$,l$,a$)
	End Method
	
	Method Token$(cmd$)
		Local t:TToken
		t=TToken(map.ValueForKey(cmd.toLower()))
		If t Return t.token
	End Method
	
	Method Help$(cmd$)
		Local t:TToken
		t=TToken(map.ValueForKey(cmd.toLower()))
		If t Return t.help
	End Method
	
	Method Link$(cmd$)
		Local t:TToken
		t=TToken(map.ValueForKey(cmd.toLower()))
		If t Return t.ref
	End Method
	
	Function Create:TQuickHelp(file$) Abstract
	
End Type

Type TBMXQuickHelp Extends TQuickHelp
	Function Create:TQuickHelp(file$)
		Local	stream:TStream
		Local	qh:TQuickHelp
		Local	l$,c$,p%,q%
		Local	token$,help$,anchor$
		
		stream=ReadFile(file)
		If stream=Null Return Null
		qh=New TBMXQuickHelp
		While Not Eof(stream)
			l$=ReadLine(stream)
			Local i%
			For i=0 Until l.length
				Local c% = l[i]
				If c=Asc("_") Continue
				If c>=Asc("0") And c<=Asc("9") Continue
				If c>=Asc("a") And c<=Asc("z") Continue
				If c>=Asc("A") And c<=Asc("Z") Continue
				Exit
			Next
			token$=l[..i]
			help$=""
			anchor$=""
			q=l.findlast("|")
			If q
				help=l[..q]
				anchor=l[q+1..]
			EndIf			
			qh.AddCommand token,help,anchor
		Wend
		stream.Close
		Return qh
	End Function	
End Type

Type THelpPanel Extends TGUIPanel

	Field	host:TGUIHost
	Field	htmlview:TGadget
	Field	homepage$
	Field	options_help:THelpOptions
	Field	helproot:TDataNode
	Field	helpuser:TDataNode
	Field	helplang:TDataNode
	Field	helpmods:TDataNode
	Field	helpmods_tp:TDataNode
	Field	helpindex:TDataNode	

	Method Refresh()
		HtmlViewGo htmlview,HtmlViewCurrentURL(htmlview)
	End Method

	Method AddLink:TDataNode(parent:TDataNode,name$,href$)
		Local	n:TDataNode
		If parent 
			n=parent.AddNode(name)
		Else
			n=helproot
		EndIf
		If href href=RealPath(href)
		n.SetAction(TOOLACTIVATE,Self,TOOLNAVIGATE,href)
		Return n
	End Method

	Method FindLink:TNewDataNode(parent:TDataNode,name$)
		If parent = Null Then Return Null
		Return parent.FindKid(name)
	End Method
	
	Method ImportLinks( node:TDataNode,path$ )

		Local t$=path+"/index.html"
		If FileType( t )<>FILETYPE_FILE Return
		
'		Select StripDir(path$).ToLower()
'			Case "language", "licenses", "tutorials"
'				node = TDataNode(FindLink(node, "User Guide"))
'		End Select
		
		Local existingNode:TDataNode = TDataNode(FindLink(node, StripDir(path$)))
		If existingNode
			node = existingNode
			node.SetAction(TOOLACTIVATE, Self, TOOLNAVIGATE, t$)
		Else
			node=AddLink(node, StripDir(path$), t$) 
		EndIf 
		
		Local map:TMap=New TMap
		
		'scan for html files
		For Local e$=EachIn LoadDir( path )
			If e="index.html" Continue
			Local p$=path+"/"+e
			Select FileType( p )
			Case FILETYPE_DIR
				ImportLinks node,p
			Case FILETYPE_FILE
				If Not e.StartsWith( "_" ) And ExtractExt( e ).Tolower()="html"
					map.Insert StripExt( e ),p
				EndIf
			End Select
		Next

		'scan for anchors in index.html...
		'
		'note: anchors must be quote enclosed and of simple form <a name="blah">
		Local c$=LoadText( t ), i%
		Repeat
			i=c.Find( "<a name=~q",i )
			If i=-1 Exit
			i:+9
			Local i2% = c.Find( "~q>",i )
			If i2=-1 Exit
			Local q$ = c[i..i2]
			If q.StartsWith( "_" ) Continue
			map.Insert q,t+"#"+q
			i=i2+1
		Forever

		For Local kv:TKeyValue=EachIn map	
			AddLink node,String( kv.Key() ),String( kv.Value() )
		Next

	End Method
	
	Method SyncDocs(path$)
		'clear down any old entries
		helproot.FreeKids
		helpmods = Null

		'create the new top level entries
		helpuser = helproot.AddNode("User Guide")
		helplang = AddLink(helproot , "Language" , "")
		helpmods =AddLink(helproot , "Modules" , "" )
		helpmods_tp = AddLink(helproot , "Third Party Modules" , "" )
		helpindex = AddLink(helproot , "@@help_tree/index" , "")	
		helproot.AddNode("Tutorials")
		helproot.append THelpSamples.CreateHelpSamples(host ,path$ ,Self) ' Import the samples
		helproot.AddNode("Licenses")
		helproot.append THelpSource.CreateHelpSource(host ,path$ ,Self) ' Import the source code links
		
		'Official Modules		
		ImportLinks Null , path$ + "/docs/html"
		
		' Third Party Modules
		For Local m$ = EachIn EnumModules()
			If m.StartsWith( "brl." ) Or m.StartsWith( "pub." ) Continue
			Local p$=ModulePath( m )+"/doc/commands.html"
			If FileType( p )<>FILETYPE_FILE Continue
			Addlink helpmods_tp,m,p
		Next

		helpmods.Refresh
		helpmods_tp.refresh
				
		' Import the alphabetic Index
		If FileType(path$ + "/docs/html/Modules/commands.txt" ) = FILETYPE_FILE
			Local comm$=LoadText(path$ + "/docs/html/Modules/commands.txt")
			Local splitLine$[] = comm.Split( "~n" ), bits$[]
			Local tokens$[][], count%
			tokens = tokens[..splitLine.length]
			
			'Fixed module index sorting. Overkill?
			For Local line$ = EachIn splitLine$
				bits$ = line.Split( "|" )
				If bits.length<>2 Continue
				Local i% = bits[0].Find( " : " )
				If i<>-1 bits[0]=bits[0][..i]
				Local c%
				For c%=0 Until bits[0].length
					Select bits[0].ToLower()[c]
						Case 40, 58, 61; Exit
					End Select
				Next
				
				tokens[count] = tokens[count][..2] 
				tokens[count][0] = bits[0][..c]
				tokens[count][1] = path$ + bits[1]
				count :+ 1
			Next
			
			tokens = tokens[..count]
			QuickSortTokens(tokens, 0, tokens.length - 1)
			
			For Local c%=0 Until tokens.length
				AddLink helpindex, tokens[c][0], tokens[c][1]
			Next
		EndIf
		
		' Import the User Guide
		CopyFile path + "/ReleaseNotes.doc" , path + "/ReleaseNotes.txt"
		CopyFile path + "/Versions.doc" , path + "/Versions.txt"
		CopyFile "incbin::about.html" , path + "/about.html"
		CopyFile "incbin::version_history.html",path+"/version_history.html"		
		AddLink(helpuser , "BlitzMax IDE (CE)" , path + "/about.html") 
		AddLink(helpuser , "BlitzMax IDE (CE) Version History" , path + "/version_history.html")		
		AddLink(helpuser , "BlitzMax Release Notes" , path + "/ReleaseNotes.txt")
		AddLink(helpuser , "BlitzMax Version History" , path + "/Versions.txt")
		
		helproot.open()		
		helproot.Refresh
		
	End Method

	Function QuickSortTokens(array$[][] Var, low%, high%)
    	Local l% = low, h% = high
    	Local x$ = array[(low + high) / 2][0].ToLower()

    	'partition
    	While l < h
        While array[l][0].ToLower() < x
					l :+ 1
				Wend
				 
        While array[h][0].ToLower() > x 
					h :- 1
				Wend
				
        If l <= h
        	Local s$[] = array[l]
					array[l] = array[h]
					array[h] = s
          l :+ 1 ; h :- 1
        EndIf
    	Wend

    	'recursion
		If low < h Then QuickSortTokens(array, low, h)
		If l < high Then QuickSortTokens(array, l, high)
	End Function

	Method Invoke:Object(command:Int,argument:Object=Null)
		Local	href$
		If Not htmlview Return Null
		Select command
			Case TOOLCUT
				GadgetCut htmlview
			Case TOOLCOPY
				GadgetCopy htmlview
			Case TOOLPASTE
				GadgetPaste htmlview
			Case TOOLSHOW
				ActivateGadget htmlview
				host.SetTitle
			Case TOOLNAVIGATE
				href$=String(argument)
				If href Go href
			Case TOOLPRINT
				GadgetPrint htmlview
			Case TOOLBACK
				HtmlViewBack htmlview
			Case TOOLFORWARD
				HtmlViewForward htmlview

		End Select
	End Method

	Method Query:Int(property:Int)
		' there are no properties to query
	End Method

	Method Poll(e:TEvent)
		Local	url$,p:Int,t$
		If e.source = htmlview
			Select e.id
				Case EVENT_GADGETACTION				'NAVIGATEREQUEST
					url$=String( e.extra )
					If url[..5]="http:"
						OpenURL url
					Else
						url=Replace(url,"%20"," ")
						p=url.findlast(".")
						If p>-1
							t$=url[p..].tolower()
							If t$=".bmx"
								If url.Find( "file://" )=0
									url=url[7..]
									?Win32
									url=url[1..]
									?
								EndIf
								host.OpenSource(url,options_help.tempsourcelinks)
							Else
								url=url.Replace("\","/")
								url=url.Replace("user/index","user/welcome")
								url=url.Replace("lang/index","lang/welcome")
								url=url.Replace("mods/index","mods/welcome")
								Go url$
							EndIf
						EndIf
					EndIf
			End Select
		EndIf
	End Method

	Method Go(url$, internal:Int = False)
		Local	node:TDataNode
		
		If Not url Then Return
		
		If options_help.externalhelp And Not internal
			PollSystem
			OpenURL url
			MinimizeWindow host.gadget()
			PollSystem
			Return
		EndIf		
		
		If url.Find("mailto:") = 0 Then ' start system email client
			HtmlViewGo htmlview, url
			Return
		ElseIf url.Find("http://") = -1 Then ' check if file exists to prevent htmlview weirdness
			Local fn:String = url
			If fn.Find("file://") = 0 Then
				fn = fn[7..]
				?Win32
				If fn[0]=Asc("/") Then fn = fn[1..]
				?
			EndIf
			fn = fn.Replace("%20", " ")
			Local pos:Int = fn.Find("#")
			If pos <> -1 Then fn = fn[..pos]			
			If FileType(fn) <> FILETYPE_FILE Then
				Notify LanguageString( "error_message/msg_file_not_found", fn), True
				Return
			EndIf
		EndIf		

		HtmlViewGo htmlview,url$
		host.SelectPanel Self
		node=helproot.FindNodeByArgument(RealPath(url))
		If node
			node.Highlight
		Else
'			print "node not found"
		EndIf
		ActivateGadget htmlview
	End Method

	Method Home()
		Go homepage$, True
	End Method

	Function Create:THelpPanel(host:TGUIHost,path$,options:THelpOptions,helproot:TDataNode)
		Local	p:THelpPanel

		p = New THelpPanel
		p.helproot=helproot
		p.homepage = path$ + "/docs/html/index.html"
		p.host=host
		p.options_help=options
		p.name="@@help_panel/title"
		host.addpanel(p)
		p.htmlview=CreateHTMLView(0,0,ClientWidth(p.panel),ClientHeight(p.panel),p.panel,HTMLVIEW_NONAVIGATE)		'HTMLVIEW_NOCONTEXTMENU
		SetGadgetLayout p.htmlview,1,1,1,1
		'p.SyncDocs(path$)
		Return p
	End Function
End Type

Type THelpSource Extends TDataNode
	Field	help_panel:THelpPanel

	Method Invoke:Object(command:Int,argument:Object=Null)
		Select command
		Case TOOLOPEN
			host.OpenSource( argument.ToString(), help_panel.options_help.tempsourcelinks)
		End Select
	End Method

	Function CreateHelpSource:THelpSource(host:TGUIHost,path$,hp:THelpPanel)
		Local	p:THelpSource
		Local	folder:TFolderNode
		Local	dir%,t$

		p=New THelpSource
		p.help_panel = hp
		p.SetName("@@help_tree/source_code")
		p.host = host
		dir% = ReadDir(path + "/mod/")
		If dir Then
			Repeat
				t$ = NextFile(dir)
				If FileType(path + "/mod/" + t$) = FILETYPE_DIR And t$<>"." And t$<>".." Then
					folder=New TFolderNode.CreateFolderNode(p,path+"/mod/"+t$,TFolderNode.DIRECTORYFOLDER)
					folder.Scan
					p.Append folder
				EndIf
			Until t$=""
		EndIf
		folder=New TFolderNode.CreateFolderNode(p,path+"/src",TFolderNode.DIRECTORYFOLDER)
		folder.Scan
		p.Append folder
		Return p
	End Function
End Type

Type THelpSamples Extends TDataNode
	Field	help_panel:THelpPanel

	Method Invoke:Object(command:Int,argument:Object=Null)
		Select command
		Case TOOLOPEN
			host.OpenSource( argument.ToString(), help_panel.options_help.tempsourcelinks)
		End Select
	End Method

	Function CreateHelpSamples:THelpSamples (host:TGUIHost,path$,hp:THelpPanel)
		Local	p:THelpSamples
		Local	folder:TFolderNode
		Local	dir% , file$
		Local	n:TDataNode

		p=New THelpSamples
		p.help_panel = hp
		p.SetName("@@help_tree/samples")
		p.host = host
		n = p.AddNode("@@help_tree/welcome_samples")
		n.SetAction(TOOLACTIVATE,hp,TOOLNAVIGATE,RealPath(path + "/samples/index.html"))

		dir%=ReadDir(path+"/samples")
		If dir Then
			file$=NextFile(dir)
			Repeat
				If FileType(path + "/samples/" + file) = FILETYPE_DIR And file <> "." And file <> ".." Then
					folder=New TFolderNode.CreateFolderNode(p,path+"/samples/"+file,TFolderNode.DIRECTORYFOLDER)
					folder.Scan
					p.Append folder
				End If
				file$=NextFile(dir)
			Until file=""
		EndIf

		Return p
	End Function
End Type

'---------------- About
Incbin "resources/bmax120.png"

Function HelpAbout(dmo:Int)
	Local fnt_link:TGUIFont,fnt_normal:TGUIFont,fnt_bmx:TGUIFont,fnt_ce:TGUIFont
	Local w:TGadget , b_ok:TGadget , p:TGadget , s$ , pic:TPixmap , link:TGadget, link2:TGadget , l:Tgadget

	fnt_normal = LookupGuiFont(GUIFONT_SYSTEM)
	fnt_link = LookupGuiFont(GUIFONT_SYSTEM,0,FONT_UNDERLINE)
	fnt_bmx = LookupGuiFont(GUIFONT_SYSTEM,12,FONT_BOLD)
	fnt_ce = LookupGuiFont(GUIFONT_SYSTEM,0,FONT_UNDERLINE|FONT_BOLD)

	LanguageSetGroup "about_window"

	w = CreateWindow("@@title" , (GadgetWidth(Desktop() ) - 300) / 2 , (GadgetHeight(Desktop() ) - 340) / 2 , 300 , 340 , Null , WINDOW_TITLEBAR | WINDOW_CLIENTCOORDS)
	p = CreatePanel(20 , 20 , 120 , 120 , w)
	pic = LoadPixmap("incbin::resources/bmax120.png")
	SetPanelPixmap p , pic , PANELPIXMAP_CENTER
	l = CreateLabel("BlitzMax IDE" , 160 , 20 , 120 , 20 , w)
	SetGadgetFont l,fnt_bmx
	l=CreateLabel("Community Edition" , 160 , 40 , 120 , 40 , w)
	SetGadgetFont l,fnt_ce
	s$ = "@@blitzmax_version|" + GetBCCVersion()
	If dmo Then s$ = s$ + " " + LanguageString("blitzmax_version_demo")
	l=CreateLabel(s$ , 160 , 80 , 140 , 30 , w)
	SetGadgetFont l,fnt_normal
	l=CreateLabel("@@ide_version|" + VERSION$ , 160 , 110 , 140 , 30 , w)
	SetGadgetFont l,fnt_normal

	l = CreateLabel("@@info_community" , 20 , 160 , 260 , 32 , w , LABEL_CENTER)
	SetGadgetFont l,fnt_normal
	l = CreateLabel("@@info_contribute" , 20 , 200 , 260 , 20 , w , LABEL_CENTER)
	SetGadgetFont l , fnt_normal
	link = CreateHyperlink("http://blitzmax-ide.sourceforge.net/" , 20 , 225 , 260 , 20 , w , LABEL_CENTER)
'	link = CreateLabel("http://blitzmax-ide.sourceforge.net/" , 20 , 220 , 260 , 30 , w , LABEL_CENTER)
'	SetGadgetFont link , fnt_link
'	SetGadgetTextColor link,0,0,255
	l = CreateLabel("@@info_blitzmax" , 20 , 250 , 260 , 20 , w , LABEL_CENTER)
	SetGadgetFont l , fnt_normal
	link2 = CreateHyperlink("http://www.blitzbasic.com/" , 20 , 275 , 260 , 20 , w , LABEL_CENTER)	
'	link2 = CreateLabel("http://www.blitzbasic.com/" , 20 , 270 , 260 , 26 , w , LABEL_CENTER)
'	SetGadgetFont link2 , fnt_link
'	SetGadgetTextColor link2,0,0,255

	b_ok = CreateButton("@@requester/ok" , 100 , 304 , 100 , 24 , w)

	Repeat
		WaitEvent
		If EventSource() = w And EventID() = EVENT_WINDOWCLOSE Then Exit
		If EventSource() = b_ok And EventID() = EVENT_GADGETACTION Then Exit
		ActivateWindow w
	Forever
	FreeGadget w
End Function