'host.bmx
'
' All the different forms of window
'  THost - the main IDE
'  TRequester - a basic ok cancel reQuester
'  TPanelRequester - a reQuester with tabs to select panels
'  TListRequester - a requester with a treelist to select panels (NOT DONE YET!)

SuperStrict

Import "tool.bmx"

Type TGUITabbedPanelGroup Extends TGUIContainer Final
	Field	_navtab:TGUINavTabber
	Field	_panelgroup:TGUIPanelGroup

	Method InitTabbedPanelGroup(host:TGUIContainer) 
		owner=host
		_panelgroup = New TGUIPanelGroup
		_navtab = New TGUINavTabber
		_navtab.Initialize(host , _panelgroup)
		_panelgroup.Initialise(_navtab.gadget(),host) ' specific to tabber navigators?
	End Method

	Method AddPanel(panel:TGUIPanel)
		_navtab.AddNavigation panel.name, panel
		_panelgroup.addPanel panel
	End Method

	Method Invoke:Object(command:Int , argument:Object = Null)
		Local i:Int

		Select command
			Case TOOLNAVIGATE ' navigate to the specified panel
				If TGUIPanel(argument) Then
					_panelgroup.ShowPanel(TGUIPanel(argument) )
				ElseIf argument = "" Then
					_panelgroup.ShowPanel(Null , 0) 'panels[index].Hide()
					_navtab.SetNavigation(0)
				Else
					i=Int(argument.tostring())
					_panelgroup.ShowPanel(Null , i) 'panels[index].Hide()
					_navtab.SetNavigation(i)
				EndIf
			Default
				'NEVER put anything here!
		End Select
	End Method

	Method Query:Int(property:Int)
		' there are no properties to query
	End Method

	Method Refresh()
		_navtab.refresh()
		_panelgroup.Refresh()
	End Method

	Method Poll(e:TEvent)
		_navtab.poll(e)
		_panelgroup.poll(e)
	End Method

	Method PanelGroup:TGUIContainer()
		Return _panelgroup
	End Method

End Type

Type TGUIToolWindow Extends TGUIContainer
	Field _modal:Int		' is this container a modal tool window (i.e. a requester)?
	Field _centered:Int
	Field _ok:TGadget
	Field _cancel:TGadget

	Const TOOLWINDOW_CANCEL:Int = 1
	Const TOOLWINDOW_OK:Int = 2
	Const TOOLWINDOW_CENTER:Int = 4
	Const TOOLWINDOW_MODAL:Int = 8

	Method InitToolWindow:TGadget(o:TGUIContainer,label$,w:Int=280,h:Int=128,flags:Int=15)
		owner=o
		_gadget=CreateWindow(label,0,0,w,h,Null,WINDOW_CLIENTCOORDS|WINDOW_TITLEBAR|WINDOW_HIDDEN|WINDOW_TOOL)
		If flags&TOOLWINDOW_CANCEL Then _cancel=CreateButton("@@requester/cancel",8,ClientHeight(_gadget)-36,110,34,_gadget,BUTTON_CANCEL)
		If flags&TOOLWINDOW_OK Then _ok = CreateButton("@@requester/ok" , ClientWidth(_gadget) - 118 , ClientHeight(_gadget) - 36 , 110 , 34 , _gadget, BUTTON_OK)
		_modal = flags&TOOLWINDOW_MODAL
		_centered = flags & TOOLWINDOW_CENTER
		Return _gadget
	End Method

	Method Show()
		Local	x:Int , y:Int , w:Int , h:Int , win:TGadget
		TGUIHost(owner).DisableHotKeys()
		TGUIHost(owner).ToolWindowAdd Self,_modal
		If _centered Then
			win=owner.Gadget()
			w=GadgetWidth(_gadget)
			h=GadgetHeight(_gadget)
			x=GadgetX(win)+(GadgetWidth(win)-w)/2
			y=GadgetY(win)+(GadgetHeight(win)-h)/2
			SetGadgetShape _gadget,x,y,w,h
		EndIf
		ShowGadget _gadget
		ActivateGadget _gadget
		PollSystem
	End Method

	Method Hide() ' use this if it's a window you expect to hide & show often
		HideGadget _gadget
		TGUIHost(owner).EnableHotKeys()
		TGUIHost(owner).ToolWindowRemove Self
		TGUIHost(owner).Activate()		
	End Method

End Type

Type TGUIHost Extends TGUIContainer
	Field	_modal_tw:TList
	Field	_path_map:TMap
	Field	_name:String
	Field	eventhandlers:TList=New TList
	Field	running:Int
	Field	regions:TGUIContainer[] ' the graphical regions that make up the visible host
	'TODO: not sure currentpanel should exist on the host
	Field	currentpanel:TGUIPanel
	Field	_currentstatustext$

	Method New()
		_modal_tw = New TList
		_path_map = New TMap
	End Method

	Method SetAppTitle(name$)
		_name = name$
		AppTitle = name$
	End Method

	Method AddPathKey(key$ , value$)
		_path_map.Insert key , value
	End Method

	Method FullPath$(path$) 
		If path[0] = Asc("$") Then
			For Local s$=EachIn _path_map.Keys()
				If path[..s.length] = s$ Then path = String(_path_map.ValueForKey(s$)) + path[s.length..]
			Next
		EndIf
		If path[0] = Asc("$") Then Notify "Could not translate path!~n"+path;Return path
		path = RealPath(path) 
?win32
		path=path.Replace("\","/")
?
		Return path
	End Method

	' event handling stuff
	Method Invoke:Object(command:Int , argument:Object = Null)
		Select command
			Case TOOLOPEN
				OpenSource String(argument)
			Case TOOLEXECUTE ' requires the input to be an array of 5 items
				Local arr:Object[]=array(argument)
				Execute String(arr[0]),String(arr[1]),String(arr[2]),Int(String(arr[3])),TBase(arr[4])
		End Select
	End Method

	Method Query:Int(property:Int)
		' there are no properties to query
	End Method

	Method AddHandler(handler:TGUIElement)
		eventhandlers.addlast handler
	End Method

	Method RemoveHandler(handler:TGUIElement)
		eventhandlers.remove handler
	End Method

	Method CheckHandlers(e:TEvent)
		Local	handler:TGUIElement

		If Not _modal_tw.IsEmpty() Then
			Local wnd:TGUIToolWindow = TGUIToolWindow( _modal_tw.last())
			If wnd Then wnd.Poll(e)
			Return
		EndIf

		For handler=EachIn eventhandlers
			handler.Poll(e)
		Next
	End Method

	Method PopupMenu(menu:TGadget,extra:Object=Null)
		PopupWindowMenu _gadget , menu , extra
	End Method

	Method ToolWindowAdd(tw:TGUIToolWindow , modal:Int)
		If modal
			If _modal_tw.count() > 0 Then
				TGUIToolWindow(_modal_tw.Last()).disable
			End If
			_modal_tw.AddLast tw
			DisableGadget _gadget
		Else
			AddHandler tw
		EndIf
	End Method

	Method ToolWindowRemove(tw:TGUIToolWindow)
		If tw=_modal_tw.Last() Then
			_modal_tw.RemoveLast
			If _modal_tw.count() > 0 Then
				' return control to the next modal window in the stack
				TGUIToolWindow(_modal_tw.Last() ).Enable
				TGUIToolWindow(_modal_tw.Last()).Activate
			Else ' return control back to the host gadget
				EnableGadget _gadget
				SelectPanel currentpanel
			EndIf
		Else
			RemoveHandler tw
		EndIf
	End Method

	Method AddRegion(region:TGUIContainer)
		If regions = Null Then
			regions = New TGUIContainer[1]
		Else
			regions = regions[0..regions.length+1]
		EndIf
		regions[regions.length-1] = region
		AddHandler region
	End Method

	'TODO: RemoveRegion - needed for customising your IDE

	'Abstract method stubs
	Method Initialize() Abstract

	'UI Stuff
	Method SetStatus(text$)
		_currentstatustext = text
		SetStatusText _gadget,text
	End Method

'TODO: move menu & hotkey handling code in here
	Method DisableHotKeys() Abstract
	Method EnableHotKeys() Abstract	

'TODO: the following may not be genuine TGUIHost items

	'UI stuff
	Method SetTitle(title$ = "")
		If _name And title Then title=" - "+title
		SetGadgetText _gadget,_name+title
	End Method
	
	Method SelectPanel(panel:TGUIPanel) Abstract
	Method AddPanel(tabpanel:TGUIPanel) Abstract
	Method RemovePanel(tabpanel:TGUIPanel) Abstract
	Method RefreshPanel(tabpanel:TGUIPanel) Abstract
	Method RefreshToolbar() Abstract
	Method RefreshAll() Abstract
	Method UpdateOptions() Abstract
	Method ShowHelp$(subject$="") Abstract

	' executing stuff
	Method GetCommandLine$() Abstract
	Method SetCommandLine(text$) Abstract
	Method Execute(cmd$,mess$="",post$="",home:Int=True,tool:TBase=Null) Abstract

	' file stuff
	Method AddRecent(file$) Abstract
	Method AddOpenFile(file$) Abstract

	' Source code related stuff
	Method OpenSource:Object(path$,temp:Int=False) Abstract
	Method OpenSourceAt:Object(path$,line:Int,column:Int) Abstract
	Method IsSourceOpen(path$) Abstract

	' Debug stuff
	Const EDITMODE:Int=1
	Const DEBUGMODE:Int=2

	Method SetMode(m:Int) Abstract
	Method WritePipe(text$) Abstract
	Method DebugSource(path$,line:Int,column:Int) Abstract

End Type


