' language.bmx - localisation
' Also includes os specific label offset _LABELOFF

'TODO: Add CreatePanel to this!

SuperStrict

Import "../gui_module.bmx"
Import BRL.LinkedList
Import BRL.RamStream ' for Incbin
Import "../util/stringtable.bmx"

?MacOS
Import "language.m"
Private
Extern
	Function LocaliseAppMenu(hide:String, hideothers:String, show:String, quit:String)
EndExtern
Public
?

' different OS have different ways of handling labels.  MaxGUI should do this, but encoded here for now
?macos
Const _LABELOFF:Int = 2
?win32
Const _LABELOFF:Int = 4
?linux
Const _LABELOFF:Int = 0
?

Incbin "language.default.ini"
Incbin "language.error.ini" ' error message definition
'Incbin "language.italiano.ini"
'Incbin "italiano.error.ini"

Private
	Const LANG_VERSION:String = "2.01"

	Global TransTable:TStringTable
	Global ErrorTable:TStringTable
	Global TransDateMonth:String[]
	Global TransTimeAMPM:String[]
	Global UseGadgetList:Int = False
	Global GadgetList:TList = CreateList()
	Global WindowList:TList = CreateList()
	Global LastKey:String
	Global LastInsert:Object
	Global LastKey2:String
	Global LastInsert2:Object
	Global LastIcon:Int
	Global LastFlags:Int
	Global LastExtra:Object

	ErrorTable = TStringTable.CreateFromFile("incbin::language.error.ini")
Public

Function LanguageLoad(url:Object = Null)
	TransTable = TStringTable.CreateFromFile("incbin::language.default.ini")

	If url Then TransTable.Load url

	TransDateMonth = Split(Translate("@@/date_month"))
	If TransDateMonth.Length <> 12 Then TransDateMonth = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]
	TransTimeAMPM = Split(Translate("@@/time_am_pm"))
	If TransTimeAMPM.Length <> 2 Then TransTimeAMPM = ["AM", "PM"]

	LanguageSetGroup ""

	Local FileVersion:String = LanguageString("version")

	If FileVersion.ToLower() <> LANG_VERSION.ToLower() Then
		Notify "@@version_warning" + "|" + FileVersion + "|" + LANG_VERSION + "|" + url.ToString(), True
	EndIf
EndFunction

Function LanguageSetGroup(Key:String)
	If Not TransTable Then Return
	TransTable.SetCurrGroupByKey(Key)
EndFunction


Function LanguageUpdateGadgets()
	For Local ge:TGadgetListEntry = EachIn GadgetList
		Select ge.LocType
			Case LOC_SETGADGETTEXT
				Local text:String = LanguageString(ge.Key, ge.Insert)
				MaxGUI.maxgui.SetGadgetText ge.Gadget, text

			Case LOC_SETMENUTEXT
				Local text:String = LanguageString(ge.Key, ge.Insert)
				MaxGUI.maxgui.SetMenuText ge.Gadget, text

			Case LOC_MODIFYGADGETITEM
				Local text:String = LanguageString(ge.Key, ge.Insert)
				Local tip:String = LanguageString(ge.Key2, ge.Insert2)
				'Print "MGI: " + ge.Key + " [" + ge.Item + "]"
				Local i:Int = SelectedGadgetItem(ge.Gadget)
				MaxGUI.maxgui.ModifyGadgetItem ge.Gadget, ge.Item, text, ge.Flags, ge.Icon, tip, ge.Extra
				SelectGadgetItem(ge.Gadget,i) 'TODO: this is a workaround for an FLTKmaxgui bug

			Case LOC_MODIFYTREEVIEWNODE
				Local text:String = LanguageString(ge.Key, ge.Insert)
' don't believe there's any need to distinguish on fltk...
'				?Win32
				' check if the node has been created internally yet
				If ge.Gadget And Not ge.Gadget.parent Then Continue
'				?Linux
				' check if the node has been created internally yet
'				If TFLNode(ge.Gadget) And Not TFLNode(ge.Gadget).owner Then Continue
'				?
				MaxGUI.maxgui.ModifyTreeViewNode ge.Gadget, text, ge.Icon
		EndSelect
	Next
	PollSystem
	For Local win:TGadget = EachIn WindowList
		?Linux
			Local enabled:Int
			enabled = Not GadgetDisabled(win)
			If Not enabled Then EnableGadget win
		?
		UpdateWindowMenu win
		?Linux
			If Not enabled Then DisableGadget win
		?
	Next
	LanguageLocaliseAppMenu
EndFunction

Function LanguageLocaliseAppMenu()
	?MacOS
	LocaliseAppMenu..
		LanguageString("app_menu/hide", "MaxIDE"), ..
		LanguageString("app_menu/hide_others"), ..
		LanguageString("app_menu/show_all"), ..
		LanguageString("app_menu/quit", "MaxIDE")
	?
EndFunction

Function LanguageEnumFiles(Path:String, LangFile:String[] Var, LangCaption:String[] Var)
	Local count:Int = -1

	FixPath Path

	Local dir:Int = ReadDir(Path)
	If Not dir Then Return

	LangFile = [""] ; LangCaption = [""]

	Repeat
		Local file:String = NextFile(dir)
		If file = "" Then Exit

		If file.ToLower().StartsWith("language.") And file.ToLower().EndsWith(".ini") Then
			file = Path + "/" + file
			Local name:String = "", version:String = ""
			Local tt:TStringTable = TStringTable.CreateFromFile(file)
			If Not tt Then Continue

			tt.SetCurrGroupByKey("")

			Local en:TStringTableEntry = tt.FindEntryByKey("name")
			If en Then name = en.Value Else Continue

			en = tt.FindEntryByKey("version")
			If en Then version = en.Value Else Continue

			If version <> LANG_VERSION Then name:+ " (" + version + ")"

			count:+ 1
			LangFile = LangFile[..count + 1] ; LangFile[count] = file
			LangCaption = LangCaption[..count + 1] ; LangCaption[count] = name
		EndIf
	Forever
	If count = -1 Then LangFile = Null ; LangCaption = Null
EndFunction

' Returns a translated version of a text
'  it also supports replacing wildcards ( "{{1}}", "{{2}}" , ... )
'  using a string, a string array or an int-array
Function LanguageString:String(Key:String, Insert:Object = Null, ForceInsert:Int = True, InsertTranslate:Int = True)
	If Key = "" Then Return Key
	Return Translate("@@" + Key, Insert, ForceInsert, InsertTranslate)
EndFunction

Function Translate:String(Key:String, Insert:Object = Null, ForceInsert:Int = True, InsertTranslate:Int = True)

	LastKey = "" ; LastInsert = Null

	If Key = "" Then Return ""
	If Key[..2] <> "@@" Then Return Key
	Key = Key[2..]

	If Not Insert And Key.Contains("|") Then
		Local temp:String[] = Split(Key)
		Key = temp[0]
		Insert = temp[1..]
	EndIf


	Local TempKey:String = Key
	If Not Key.Contains("/") Then TempKey = TransTable.CurrGroup.Key + "/" + Key
	Local TempInsert:Object = Insert

	Local out:String = "(? " + TempKey + ")"

	Local en:TStringTableEntry
	If TransTable Then
		If Key.Contains("/") Then
			Local group:String = Key[..Key.Find("/")].Trim()
			Key = Key[Key.Find("/")+1..].Trim()
			Local gr:TStringTableGroup = TransTable.FindGroupByKey(group)
			If gr Then en = gr.FindEntryByKey(Key)
		Else
			en = TransTable.FindEntryByKey(Key)
		EndIf
	EndIf

	If en Then out = en.Value

	If Insert Then
		If String(Insert) Then
			Local p:Int = out.Find("{{1}}")
			If p = - 1 And ForceInsert Then
				out:+ "|" + String(Insert)
			Else
				out = out.Replace("{{1}}", String(Insert))
			EndIf

		ElseIf String[](Insert) Then
			Local InsArray:String[] = String[](Insert)
			For Local n:Int = 0 Until InsArray.Length
				If InsertTranslate Then
					InsArray[n] = Translate(InsArray[n])
				EndIf
				Local p:Int = out.Find("{{" + (n + 1) + "}}")
				If p = - 1 And ForceInsert Then
					out:+ "|" + InsArray[n]
				Else
					out = out.Replace("{{" + (n + 1) + "}}", InsArray[n])
				EndIf
			Next

		ElseIf Int[](Insert) Then
			Local InsArray:Int[] = Int[](Insert)
			For Local n:Int = 0 Until InsArray.Length
				Local p:Int = out.Find("{{" + (n + 1) + "}}")
				If p = - 1 And ForceInsert Then
					out:+ "|" + InsArray[n]
				Else
					out = out.Replace("{{" + (n + 1) + "}}", InsArray[n])
				EndIf
			Next
		EndIf
	EndIf

	LastKey = TempKey
	LastInsert = TempInsert

	Return out'.ToUpper()
EndFunction

Function TranslateMessage:String(Message:String)
	Local gr_msg:TStringTableGroup = ErrorTable.FindGroupByKey("msg")
	Local gr_insert:TStringTableGroup = ErrorTable.FindGroupByKey("insert")
	Local FoundParts:String[] = Null
	Message = Message.Trim()

	For Local n:Int = 0 To 26
		Local en:TStringTableEntry = gr_msg.Content[n]
		While en
			If MatchParts(Message, Split(en.Value), FoundParts) Then
				' translate found parts "insert_"
				For Local n:Int = 0 Until FoundParts.Length
					Local en2:TStringTableEntry = gr_insert.FindEntryByValue(FoundParts[n])
					If en2 Then FoundParts[n] = LanguageString("error_message/insert_" + en2.Key)
				Next

				' translate message "msg_"
				Local Mess:String = LanguageString("error_message/msg_"+en.Key, FoundParts, True, False)
				Return Mess[..1].ToUpper() + Mess[1..]
			EndIf
			en = en.Succ
		Wend
	Next

	Return LanguageString("error_message/no_translation|" + Message, , True, False)
EndFunction

Function TranslateExtensions:String( extension:String)
	Local ext:String[], exttrans:String
	ext = Split(extension, ";")

	LanguageSetGroup "file_type"
	For Local n:Int = 0 Until ext.Length
		If ext[n].Contains(":") Then
			exttrans:+ (Translate(ext[n][..ext[n].Find(":")]) + ext[n][ext[n].Find(":")..])
			If exttrans.EndsWith(":") Then exttrans:+ "*"
		Else
			exttrans:+ (LanguageString("file") + ":" + ext[n])
			If exttrans.EndsWith(":") Then exttrans:+ "*"
		EndIf
		If n < ext.Length - 1 Then exttrans:+ ";"
	Next

	Return exttrans
EndFunction


' Functions to create localised gadgets

Function CreateWindow:TGadget(name:String, x:Int, y:Int, w:Int, h:Int, group:TGadget = Null, style:Int = 15)
	name = Translate(name)
	Local g:TGadget = AddToGadgetList(MaxGUI.maxgui.CreateWindow(name, x, y, w, h, group, style), LOC_SETGADGETTEXT)
	WindowList.AddLast g
	Return g
EndFunction

Function CreateButton:TGadget(label:String, x:Int, y:Int, w:Int, h:Int, group:TGadget, style:Int = BUTTON_PUSH)
	label = Translate(label)
	Return AddToGadgetList(MaxGUI.maxgui.CreateButton(label, x, y, w, h, group, style), LOC_SETGADGETTEXT)
EndFunction

Function CreateLabel:TGadget(name:String, x:Int, y:Int, w:Int, h:Int, group:TGadget = Null, style:Int = 0)
	name = Translate(name)
	Return AddToGadgetList(MaxGUI.maxgui.CreateLabel(name, x, y+_LABELOFF, w, h, group, style), LOC_SETGADGETTEXT)
EndFunction

Function CreateMenu:TGadget(text:String, tag:Int, parent:TGadget, hotKey:Int = 0, modifier:Int = 0)
	text = Translate(text)
	Return  AddToGadgetList(MaxGUI.maxgui.CreateMenu(text, tag, parent, hotKey, modifier), LOC_SETMENUTEXT)
EndFunction


Function AddGadgetItem(gadget:TGadget, text:String, flags:Int = 0, icon:Int = - 1, tip:String = "", extra:Object = Null)
	If tip <> "" Then
		tip = Translate(tip)
		LastKey2 = LastKey
		LastInsert2 = LastInsert
	EndIf
	text = Translate(text)
	Local index:Int = CountGadgetItems(gadget)
	LastIcon = icon
	LastFlags = flags
	LastExtra = extra
	MaxGUI.maxgui.AddGadgetItem(gadget, text, flags, icon, tip, extra)
	AddToGadgetList gadget, LOC_MODIFYGADGETITEM, index
	LastKey2 = "" ; LastInsert2 = Null
EndFunction

Function InsertGadgetItem(gadget:TGadget, index:Int, text:String, flags:Int = 0, icon:Int = - 1, tip:String = "", extra:Object = Null)
	text = Translate(text)
	LastIcon = icon
	LastFlags = flags
	LastExtra = extra
	If tip <> "" Then tip = Translate(tip)
	MaxGUI.maxgui.InsertGadgetItem(gadget, index, text, flags, icon, tip, extra)
EndFunction

Function ModifyGadgetItem(gadget:TGadget, index:Int, text:String, flags:Int = 0, icon:Int = - 1, tip:String = "", extra:Object = Null)
	text = Translate(text)
	LastIcon = icon
	LastFlags = flags
	LastExtra = extra
	If tip <> "" Then tip = Translate(tip)
	AddToGadgetList gadget, LOC_MODIFYGADGETITEM, Index
	MaxGUI.maxgui.ModifyGadgetItem(gadget, index, text, flags, icon, tip, extra)
EndFunction

Function RemoveGadgetItem(gadget:TGadget, index:Int)
	RemoveFromGadgetList gadget, index
	MaxGUI.maxgui.RemoveGadgetItem(gadget, index)
EndFunction

Function ClearGadgetItems(gadget:TGadget)
	RemoveFromGadgetList gadget
	MaxGUI.maxgui.ClearGadgetItems gadget
EndFunction

Function SetToolBarTips(toolbar:TGadget,tips:String[])
	If tips.Length = 1 And tips[0][..2] = "@@" Then
		tips = Split(Translate(tips[0]))
	Else
		For Local n:Int = 0 Until tips.Length
			tips[n] = Translate(tips[n])
		Next
	EndIf

	MaxGUI.maxgui.SetToolBarTips(toolbar, tips)
EndFunction

Function AddTreeViewNode:TGadget(text:String, node:TGadget, icon:Int = - 1)
	text = Translate(text)
	LastIcon = icon
	Return AddToGadgetList(MaxGUI.maxgui.AddTreeViewNode(text, node, icon), LOC_MODIFYTREEVIEWNODE)
EndFunction

Function InsertTreeViewNode:TGadget(index:Int, text:String, node:TGadget, icon:Int = - 1)
	text = Translate(text)
	LastIcon = icon
	Return AddToGadgetList(MaxGUI.maxgui.InsertTreeViewNode(index, text, node, icon), LOC_MODIFYTREEVIEWNODE)
EndFunction

Function ModifyTreeViewNode(node:TGadget, text:String, icon:Int = - 1)
	text = Translate(text)
	LastIcon = icon
	MaxGUI.maxgui.ModifyTreeViewNode(node, text, icon)
	AddToGadgetList(node, LOC_MODIFYTREEVIEWNODE)
EndFunction

Function SetMenuText(menu:TGadget, text:String)
	text = Translate(text)
	AddToGadgetList(menu, LOC_SETMENUTEXT)
	MaxGUI.maxgui.SetMenuText(menu, text)
EndFunction

Function SetGadgetText(gadget:TGadget, text:String)
	text = Translate(text)
	AddToGadgetList(gadget, LOC_SETGADGETTEXT)
	MaxGUI.maxgui.SetGadgetText(gadget, text)
EndFunction

' This will remove a gadget and its children from the list of all localised gadgets.
Function FreeGadget(gadget:TGadget)
	RemoveFromGadgetList gadget
	WindowList.Remove(gadget)

	MaxGUI.maxgui.FreeGadget(gadget)
EndFunction


' Simple user interface

Function Notify(text:String, serious:Int = False)
	text = Translate(text)
	BRL.System.Notify text, serious
EndFunction

Function Proceed:Int(text:String, serious:Int = False)
	text = Translate(text)
	Return BRL.System.Proceed(text, serious)
EndFunction

Function Confirm:Int(text:String, serious:Int = False)
	text = Translate(text)
	Return BRL.System.Confirm(text, serious)
EndFunction


Function RequestDir:String(text:String, initial_path:String = "")
	text = Translate(text)
	Return BRL.System.RequestDir(text, initial_path)
EndFunction

Function RequestFile:String(text:String, extension:String = "", save_flag:Int = False, initial_path:String = "")
	Return BRL.System.RequestFile( Translate(text), TranslateExtensions(extension), save_flag, initial_path)
EndFunction


' Time and date Functions

Function CurrentTime:String()
	If Not TransTable Then Return BRL.System.CurrentTime()

	Local time:String = BRL.System.CurrentTime()
	Local hour24:String = time[..2]
	Local hour12:String = ( (hour24.ToInt() + 11) Mod 12) + 1
	If hour12.Length = 1 Then hour12 = "0" + hour12
	Local minute:String = time[3..5]
	Local Second:String = time[6..]
	Local ampm:Int = hour24.ToInt() / 12

	Return Translate("@@/time", [hour24, hour12, minute, Second, TransTimeAMPM[ampm]], False)
EndFunction

Function CurrentDate:String()
	If Not TransTable Then Return BRL.System.CurrentDate()

	Local date:String = BRL.System.CurrentDate()
	Local day:String = date[..2]
	Local Month:String, month_str:String
	Select date[3..6].ToLower()
		Case "jan" ; Month = "01" ; month_str = TransDateMonth[0]
		Case "feb" ; Month = "02" ; month_str = TransDateMonth[1]
		Case "mar" ; Month = "03" ; month_str = TransDateMonth[2]
		Case "apr" ; Month = "04" ; month_str = TransDateMonth[3]
		Case "may" ; Month = "05" ; month_str = TransDateMonth[4]
		Case "jun" ; Month = "06" ; month_str = TransDateMonth[5]
		Case "jul" ; Month = "07" ; month_str = TransDateMonth[6]
		Case "aug" ; Month = "08" ; month_str = TransDateMonth[7]
		Case "sep" ; Month = "09" ; month_str = TransDateMonth[8]
		Case "oct" ; Month = "10" ; month_str = TransDateMonth[9]
		Case "nov" ; Month = "11" ; month_str = TransDateMonth[10]
		Case "dec" ; Month = "12" ; month_str = TransDateMonth[11]
		Default ; Month = date[3..6] ; month_str = Month
	EndSelect
	Local year:String = date[7..]

	Return Translate("@@/date", [day, Month, month_str, year], False)
EndFunction



Private

' GadgetList

Const LOC_SETGADGETTEXT:Int = 1
Const LOC_SETMENUTEXT:Int = 2
Const LOC_MODIFYGADGETITEM:Int = 3
Const LOC_MODIFYTREEVIEWNODE:Int = 4

Function AddToGadgetList:TGadget(Gadget:TGadget, LocType:Int, Item:Int = -1)
	If LastKey = "" Then Return Gadget

'	Print "ADD: " + LastKey + " [" + Item + "]"

	Local ge:TGadgetListEntry = FindItemByGadget(Gadget, Item)
	If Not ge Then
		ge = New TGadgetListEntry
		ge.Gadget = gadget
		ge.LocType = LocType
		GadgetList.AddLast ge
	EndIf

	ge.Item = Item
	ge.Key = LastKey
	ge.Insert = LastInsert
	ge.Key2 = LastKey2
	ge.Insert2 = LastInsert2
	ge.Icon = LastIcon
	ge.Flags = LastFlags
	ge.Extra = LastExtra
	Return Gadget
EndFunction

Function RemoveFromGadgetList(Gadget:TGadget, Item:Int = -1)
	For Local kid:TGadget = EachIn Gadget.kids
		RemoveFromGadgetList kid
	Next

	' remove all matching entries
	Local ge:TGadgetListEntry = FindItemByGadget(Gadget, Item)
	While ge
'		Print "REMOVE: " + ge.Key + " [" + ge.Item + "]"
		GadgetList.Remove ge
		ge = FindItemByGadget(Gadget, Item)
	Wend

	' update affected list items' indices
	For Local ge:TGadgetListEntry = EachIn GadgetList
		If (ge.LocType = LOC_MODIFYGADGETITEM) And (Item <> -1) Then
			If (ge.Gadget = Gadget) And (ge.Item > Item) Then
'				Print "MOVE: " + ge.Key + " [" + ge.Item + "] -> [" + (ge.Item - 1) + "]"
				ge.Item:- 1
			EndIf
		EndIf
	Next
EndFunction

Function FindItemByGadget:TGadgetListEntry(Gadget:TGadget, Item:Int = -1)
	For Local ge:TGadgetListEntry = EachIn GadgetList
		If (ge.LocType = LOC_MODIFYGADGETITEM) And (Item <> -1) Then
			If (ge.Gadget = Gadget) And (ge.Item = Item) Then Return ge
		Else
			If ge.Gadget = Gadget Then Return ge
		EndIf
	Next
EndFunction

Type TGadgetListEntry
	Field Gadget:TGadget
	Field Item:Int
	Field LocType:Int
	Field Key:String
	Field Insert:Object
	Field Key2:String
	Field Insert2:Object
	Field Icon:Int
	Field Flags:Int
	Field Tips:String[]
	Field Extra:Object
EndType

' Helper Functions

Function Split:String[](Text:String, SepChar:String = "|")
	Local ret:String[] = [""]
	Local i:Int = 0
	For Local n:Int = 0 Until Text.Length
		If Text[n..n + 1] = SepChar Then
			i:+ 1
			ret = ret[..i+1]
		Else
			ret[i]:+ Text[n..n + 1]
		EndIf
	Next
	Return ret
EndFunction

Function MatchParts:Int(Match:String, Parts:String[], FoundParts:String[] Var)
	FoundParts = Null

	If Parts.Length = 1 Then Return Match = Parts[0]

	FoundParts = New String[Parts.Length-1]

	' match start
	If Parts[0] <> "" Then
		If Match.StartsWith(Parts[0]) Then
			Match = Match[Parts[0].Length..]
		Else
			Return False
		EndIf
	EndIf

	If Parts[Parts.Length-1] <> "" Then
		If Match.EndsWith(Parts[Parts.Length-1]) Then
			Match = Match[..Match.Length-Parts[Parts.Length-1].Length]
		Else
			Return False
		EndIf
	EndIf

	For Local n:Int = 1 Until Parts.Length - 1
		Local p:Int = Match.Find(Parts[n])
		If p > - 1 Then
			FoundParts[n - 1] = Match[..p]
			Match = Match[p..]
			Match = Match[Parts[n].Length..]
		Else
			Return False
		EndIf
	Next

	FoundParts[Parts.Length - 2] = Match

	Return True
EndFunction