'options.bmx

SuperStrict

Import "style.bmx"
Import "util/miscfuncs.bmx"
Import "projects.bmx"

?MacOS
Const DEFAULTFONT$="Courier"
Const DEFAULTFONTSIZE:Int=12
'Const HOMEPAGE$="/doc/index.html"
?
?Win32
Const DEFAULTFONT$="Courier"
Const DEFAULTFONTSIZE:Int=10
'Const HOMEPAGE$="/doc/index.html"
?
?Linux
Const DEFAULTFONT$="Lucida"
Const DEFAULTFONTSIZE:Int=12
'Const HOMEPAGE$="/doc/home.html"
?

Type TOptions Extends TBase Abstract
	Field	_undo:TBank
	Field	myname$
	Field	_group:TOptionsGroup

	Method Snapshot() Final
		If Not _undo Then _undo=CreateBank(8192)
		Local stream:TStream=CreateBankStream(_undo)
		write stream
		stream.close
	End Method

	Method Restore() Final
		Local stream:TStream
		Local f$,p:Int,n$,v$

		If Not _undo Then Return
		stream:TStream=CreateBankStream(_undo)

		While Not stream.Eof()
			f$=stream.ReadLine()
			p=f.find("=")
			n$=f[..p]
			v$=f[p+1..]
			Read(n,v)
		Wend

		stream.close
	End Method

	Method InitOptions(h:TOptionsGroup)
		_group=h
		_group.AddOptions(Self) 'make sure that the host options group knows about me!
		SetDefaults()
		Snapshot()
	End Method

	Method SetDirty(b:Int = True)
		_group.dirty = True
	End Method

	Method Invoke:Object(command:Int , argument:Object) ; End Method 		'nothing to do - yet

	Method Query:Int(property:Int)
		Select property
			Case QUERY_ISDIRTY ; Return _group.dirty
			Default Return False
		End Select
	End Method

	Function Create:TOptions(host:TOptionsGroup) Abstract

	' option based stuff
	Method Initialise() Abstract				'does any required object set up
	Method Write(stream:TStream) Abstract 		'this allows the options to be saved
	Method Clear() Abstract						'clears any old options to allow new ones to be loaded
	Method Read(key$,value$) Abstract 			'this allows the options to be loaded
	Method SetDefaults() Abstract 				'this sets the default values for the options

End Type

Type TOptionsGroup Extends TBase Final
	Field	_optionsfile$
	Field	optionslist:TList = New TList
	Field	dirty:Int 		= 	False
	Field	dirtyfonts:Int	=	False

	Method Query:Int(property:Int)
		Select property
			Case QUERY_ISDIRTY ; Return dirty
			Default Return False
		End Select
	End Method

	Method AddOptions(opt:TOptions)
		optionslist.AddLast opt
	End Method

	Method ReadConfig(file$)

		Local	stream:TStream
		Local	curroption:TOptions
		Local	f$,p:Int,n$,v$

		' Set defaults first, to be over-ridden with file contents
		SetDefaults()

		' Start reading the options file
		stream=ReadFile(file$)
		_optionsfile$=file$

		' If there is a file, start reading it
		If stream Then
			' loop through the file, line by line
			While Not stream.Eof()
				f$=stream.ReadLine()

				' first check to see if we changed section [SECTION]
				p=f.find("[")
				If p=0 Then
					curroption=Null
					curroption=FindOptions(f[1..f.length-1])
					If curroption<>Null Then curroption.Clear() ' reset it
				End If

				' only if we have a valid section, do we need to worry about the contents of it
				If curroption=Null Then Continue

				' so now we check to see if we've got a name/value pair
				p=f.find("=")
				If p=-1 Continue ' if not, give it up!

				' we have a valid section, and a name/value pair: let's break it up!
				n$=f[..p]
				v$=f[p+1..]

				' and give it to the current option to read
				curroption.Read(n,v)

			Wend
			' close the file
			stream.close
			dirty = False
			dirtyfonts = False
		End If
		' save a snapshot of the options, and we're done!
		Snapshot()
	End Method

	Method FindOptions:TOptions(name$)
		Local o:TOptions
		For o:TOptions = EachIn optionslist
			If o.myname=name Then Return o
		Next
	End Method

	Method SetDefaults()
		Local o:TOptions
		For o:TOptions = EachIn optionslist
			o.SetDefaults
		Next
		dirty = True
		dirtyfonts = True
	End Method

	Method Snapshot()
		Local o:TOptions
		For o:TOptions = EachIn optionslist
			o.Snapshot
		Next
	End Method

	Method Restore()
		Local o:TOptions
		For o:TOptions = EachIn optionslist
			o.Restore
		Next
		dirty = False
		dirtyfonts = False
	End Method

	Method WriteConfig:Int(newfile$="")
		Local	stream:TStream
		Local	o:TOptions

		' open the file for writing to it
		If newfile$<>"" Then _optionsfile$=newfile$
		stream=WriteFile(_optionsfile$)
		If Not stream Return False

		' loop through all options, and save them
		For	o = EachIn optionslist
			stream.WriteLine ""
			stream.WriteLine "["+o.myname+"]"
			o.write(stream)
		Next

		stream.close ' all done
		dirty = False
		dirtyfonts = False
		Return True
	End Method

	Method Invoke:Object(command:Int , argument:Object) ; End Method 		'nothing to do - yet

End Type

Type TEditorOptions Extends TOptions Final
	Field	editstyle:TGadgetStyle
	Field	styles:TTextStyle[6]

	Const NORMAL:Int=	0
	Const COMMENT:Int=	1
	Const QUOTED:Int=	2
	Const KEYWORD:Int=	3
	Const NUMERAL:Int	=4
	Const ATTRIBUTE:Int = 5

	Method Initialise()
		myname$="EditorOptions"
		editstyle = TGadgetStyle.Create()
		styles[NORMAL]=TTextStyle.Create()
		styles[COMMENT]=TTextStyle.Create()
		styles[QUOTED]=TTextStyle.Create()
		styles[KEYWORD]=TTextStyle.Create()
		styles[NUMERAL]=TTextStyle.Create()
		styles[ATTRIBUTE]=TTextStyle.Create()
	End Method

	Method SetDefaults()
		editstyle.set($FFFFFF,$01516b,DEFAULTFONT,DEFAULTFONTSIZE)
		styles[NORMAL].set($ffffff,0)
		styles[COMMENT].set($b1e7eb,0)
		styles[QUOTED].set($00ff66,0)
		styles[KEYWORD].set($ffff00,1)
		styles[NUMERAL].set($15ffb9,0)
		styles[ATTRIBUTE].set($808080,2)
	End Method

	Method Write(stream:TStream)
		stream.WriteLine "edit_style="+editstyle.ToString()
		stream.WriteLine "normal_style="+styles[NORMAL].ToString()
		stream.WriteLine "comment_style="+styles[COMMENT].ToString()
		stream.WriteLine "quote_style="+styles[QUOTED].ToString()
		stream.WriteLine "keyword_style="+styles[KEYWORD].ToString()
		stream.WriteLine "numeral_style="+styles[NUMERAL].ToString()
		stream.WriteLine "attribute_style="+styles[ATTRIBUTE].ToString()
	End Method

	Method Clear()
	End Method

	Method Read(key$,value$)
		Select key$
			Case "edit_style" ; editstyle.FromString(value$)
			Case "normal_style" ; styles[NORMAL].FromString(value$)
			Case "comment_style" ; styles[COMMENT].FromString(value$)
			Case "quote_style" ; styles[QUOTED].FromString(value$)
			Case "keyword_style" ; styles[KEYWORD].FromString(value$)
			Case "numeral_style" ; styles[NUMERAL].FromString(value$)
			Case "attribute_style" ; styles[ATTRIBUTE].FromString(value$)
		End Select
	End Method

	Function Create:TEditorOptions(host:TOptionsGroup)
		Local	o:TEditorOptions
		o=New TEditorOptions
		o.Initialise()
		o.InitOptions(host)
		Return o
	End Function

End Type

Type TCodeOptions Extends TOptions Final
	Field 	autocapitalize:Int
	Field 	syntaxhighlight:Int
	Field 	autoindent:Int
	Field	tabsize:Int
	Field	autocomplete:Int
	Field	dockeysfirst:Int
	Field	autotidy:Int
	Field	autotidymode:Int ' flags: $1 = space before seperator
	Field	undosize:Int	

	Method Initialise()
		myname$="CodeOptions"
	End Method

	Method SetDefaults()
		autocapitalize=True
		syntaxhighlight=True
		autoindent=True
		autocomplete=True
		dockeysfirst=True
		autotidy=True
		autotidymode=$1
		tabsize = 4
		undosize=1000		
	End Method

	Method Write(stream:TStream)
		stream.WriteLine "autocapitalize="+autocapitalize
		stream.WriteLine "syntaxhighlight="+syntaxhighlight
		stream.WriteLine "autoindent="+autoindent
		stream.WriteLine "autocomplete="+autocomplete		
		stream.WriteLine "document_keywords_first="+dockeysfirst
		stream.WriteLine "autotidy="+autotidy
		stream.WriteLine "autotidymode="+autotidymode
		stream.WriteLine "tabsize=" + tabsize
		stream.WriteLine "undosize=" +undosize		
	End Method

	Method Clear()
	End Method

	Method Read(key$,value$)
		Select key$
			Case "autocapitalize" ; autocapitalize=Int(value$)
			Case "syntaxhighlight" ; syntaxhighlight=Int(value$)
			Case "autoindent" ; autoindent=Int(value$)
			Case "autocomplete" ; autocomplete=Int(value$)
			Case "document_keywords_first" dockeysfirst=Int(value)
			Case "autotidy" ; autotidy=Int(value$)
			Case "autotidymode" ; autotidymode=Int(value$)
			Case "tabsize" ; tabsize = Int(value$) 
			Case "undosize" ; undosize = Int(value)			
		End Select
	End Method

	Function Create:TCodeOptions(host:TOptionsGroup)
		Local	o:TCodeOptions
		o=New TCodeOptions
		o.Initialise()
		o.InitOptions(host)
		Return o
	End Function

End Type

Type THostOptions Extends TOptions Final
	Field	showtoolbar:Int
	Field	restoreopenfiles:Int
	Field 	autobackup:Int
	Field	languagefile:String
	Field	singleinstance:Int

	' non-options
	Field	maximised:Int
	Field	winsize:TRect=New TRect
	Field	projlist:TStringList
	Field	openlist:TStringList
	Field	recentlist:TStringList

	Method SetProjectList(list:TStringList)
		projlist = list
		_group.dirty = True
	End Method

	Method SetWindowSize(maxi%,x%=0,y%=0,w%=0,h%=0)
		maximised=maxi
		If Not maxi Then
			winsize.x=x
			winsize.y=y
			winsize.w=w
			winsize.h=h
		End If
		_group.dirty=True
	End Method

	Method Initialise()
		myname$="HostOptions"
		projlist=New TStringList
		openlist=New TStringList
		recentlist = New TStringList
	End Method

	Method Write(stream:TStream)
		Local f$

		stream.WriteLine "showtoolbar="+showtoolbar
		stream.WriteLine "restoreopenfiles="+restoreopenfiles
		stream.WriteLine "autobackup="+autobackup
		stream.WriteLine "languagefile="+languagefile
		stream.WriteLine "win_size="+winsize.ToString()
		stream.WriteLine "win_max=" + maximised
		stream.WriteLine "single_instance=" + singleinstance

		If restoreopenfiles Then
			For f$=EachIn openlist
				stream.WriteLine "file_open="+f$
			Next
		EndIf
		Local n:Int = 0
		For f$=EachIn recentlist
			stream.WriteLine "file_recent=" + f$
			n:+ 1
			If n = 16 Then Exit
		Next
		For f$=EachIn projlist
			stream.WriteLine "proj_node=" + f$
		Next

	End Method

	Method Clear()
		recentlist=New TStringList
		openlist=New TStringList
		projlist=New TStringList
	End Method

	Method Read(key$,value$)
		Select key$
			Case "showtoolbar" ; showtoolbar=Int(value$)
			Case "restoreopenfiles" ; restoreopenfiles=Int(value$)
			Case "autobackup" ; autobackup=Int(value$)
			Case "languagefile" ; languagefile = value$
			Case "win_size" ; winsize.FromString(value$)
			Case "win_max" ; maximised = Int(value$)
			Case "single_instance" ; singleinstance = Int(value)
			Case "file_recent"
				If Not recentlist.contains(value$) And FileType(value) = FILETYPE_FILE Then recentlist.addlast value$
			Case "file_open"
				If Not openlist.contains(value$) And FileType(value) = FILETYPE_FILE Then openlist.addlast value$
			Case "proj_node"
				'BUG: something causes the projlist to duplicate its content
				'  this just prevents the doubled entries from being loaded -- Byteemoz
				For Local proj:Object = EachIn projlist
					If String(proj) = value$ Then projlist.Remove proj
				Next
				projlist.AddLast value$
		End Select
	End Method

	Method SetDefaults()
		showtoolbar=True
		restoreopenfiles=True
		autobackup = True
		languagefile = ""
		Local wh:Int=GadgetHeight(Desktop())-80'32
		Local ww:Int=wh
		Local wx:Int=(GadgetWidth(Desktop())-ww)/2
		Local wy:Int=(GadgetHeight(Desktop())-wh)/2
		winsize.set( wx,wy,ww,wh )
		maximised=True
		singleinstance=False

		recentlist.clear
		openlist.clear
		projlist.clear

	End Method

	Function Create:THostOptions(host:TOptionsGroup)
		Local	o:THostOptions
		o=New THostOptions
		o.Initialise()
		o.InitOptions(host)
		Return o
	End Function

End Type

Type TSyncOptions Extends TOptions Final
	Field	syncstate$
	Field	modules$[]
	Field	module_names$[]

	Method Initialise()
		myname$="SyncOptions"
		syncstate=""
	End Method

	Method Write(stream:TStream)
		stream.WriteLine "sync_state=" + syncstate
		stream.WriteLine "modules=" + ConvertArrayToString(modules , ",")
		stream.WriteLine "module_names=" + ConvertArrayToString(module_names,",")
	End Method

	Method Clear()
	End Method

	Method Read(key$,value$)
		Select key$
			Case "sync_state"	 ; syncstate = value$
			Case "modules"		 ; modules = ConvertStringToArray(value$ , ",")
			Case "module_names"		; module_names=ConvertStringToArray(value$,",")			
		End Select
	End Method

	Method SetDefaults()
		syncstate = ""
		modules = ["BRL" , "PUB" , "AXE" ,"BAH", "MAK" , "GMAN" , "BIRDIE" ]
		module_names=["BlitzMax Core Modules","Public Domain Core Modules","User Contributed Modules","Brucey's Modules","Mark Sibly's","Simon Armstrong's","Garrit Grandberg's","Dave Bird's"]
	End Method

	Function Create:TSyncOptions(host:TOptionsGroup)
		Local	o:TSyncOptions
		o=New TSyncOptions
		o.Initialise
		o.InitOptions(host)
		Return o
	End Function

End Type

Const APPTYPE_CONSOLE:Int = 1
Const APPTYPE_GUI:Int = 2
Const APPTYPE_WXGUI:Int = 3

Const TARGETPLATFORM_WIN32:Int = 1
Const TARGETPLATFORM_LINUX:Int = 2
Const TARGETPLATFORM_MACOS_NATIVE:Int = 3
Const TARGETPLATFORM_MACOS_UNIVERSAL:Int = 4

?win32
Const TARGETPLATFORM_NATIVE:Int = 1
?linux
Const TARGETPLATFORM_NATIVE:Int = 2
?macos
Const TARGETPLATFORM_NATIVE:Int = 3
?

Type TBuildOptions Extends TOptions Final
	Field	outputstyle:TGadgetStyle
	Field 	hideoutput:Int
	Field	quickenabled:Int
	Field	debugenabled:Int
	Field	threadenabled:Int
	
	'Field	guienabled:Int
	
	Field	targetplatform:Int
	Field	apptype:Int
	
	Field	buildfile$
	Field	cmdline$
	Field	wordwrap:Int

	Method Initialise()
		myname$="BuildOptions"
		outputstyle = TGadgetStyle.Create()
	End Method

	Method Write(stream:TStream)
		stream.WriteLine "output_style="+outputstyle.ToString()
		stream.WriteLine "hide_output="+hideoutput
		stream.WriteLine "prg_quick="+quickenabled
		stream.WriteLine "prg_debug="+debugenabled
		'stream.WriteLine "prg_gui="+guienabled
		stream.WriteLine "prg_apptype=" + apptype
		stream.WriteLine "prg_target="+targetplatform		
		stream.WriteLine "prg_thread="+threadenabled
		stream.WriteLine "prg_locked="+buildfile
		stream.WriteLine "cmd_line=" + cmdline
		stream.WriteLine "output_wordwrap="+wordwrap
	End Method

	Method Clear()
	End Method

	Method Read(key$,value$)
		Select key$
			Case "output_style" ; outputstyle.FromString(value$)
			Case "hide_output"	; hideoutput=Int(value$)
			Case "prg_quick"	; quickenabled=Int(value$)
			Case "prg_debug"	; debugenabled=Int(value$)
			Case "prg_gui"
				'guienabled = Int(value$)
				If Int(value$) = 0 Then apptype = APPTYPE_GUI Else apptype = APPTYPE_CONSOLE
			Case "prg_apptype"	 ; apptype = Int(value$)
			Case "prg_target"	 ; targetplatform = Int(value$)
			Case "prg_thread"	; threadenabled=Int(value$)
			Case "prg_locked"	; buildfile=value$
			Case "cmd_line"		 ; cmdline = value$
			Case "output_wordwrap"		; wordwrap=Int(value$)
		End Select

	End Method

	Method SetDefaults()
		outputstyle.set(0,-1,DEFAULTFONT,DEFAULTFONTSIZE)
		hideoutput=False
		quickenabled=False
		debugenabled=True
		'guienabled = True
		apptype = APPTYPE_GUI
		?macos
			targetplatform = TARGETPLATFORM_MACOS_NATIVE
		?win32
			targetplatform = TARGETPLATFORM_WIN32
		?linux
			targetplatform = TARGETPLATFORM_LINUX		
		?
		threadenabled = False
		wordwrap=False
		cmdline=""
	End Method

	Function Create:TBuildOptions(host:TOptionsGroup)
		Local	o:TBuildOptions
		o=New TBuildOptions
		o.Initialise()
		o.InitOptions(host)
		Return o
	End Function

End Type

Type TToolsOptions Extends TOptions Final
	Field	sorted:Int
	Field	navstyle:TGadgetStyle
	Field	navswap:Int
	Field	splitpos:Int

	Method Initialise()
		myname$="ToolsOptions"
		navstyle=TGadgetStyle.Create()
	End Method

	Method Write(stream:TStream)
		stream.WriteLine "nav_style="+navstyle.ToString()
		stream.WriteLine "nav_sorted="+sorted
		stream.WriteLine "nav_swap="+navswap
		stream.WriteLine "nav_pos="+splitpos
	End Method

	Method Clear()
	End Method

	Method Read(key$,value$)
		Select key$
			Case "nav_style" ; navstyle.FromString(value$)
			Case "nav_sorted" ; sorted=Int(value$)
			Case "nav_swap" ; navswap=Int(value$)
			Case "nav_pos" ; splitpos=Int(value$)
		End Select
	End Method

	Method SetDefaults()
		sorted=False
		navstyle.set(0,-1,DEFAULTFONT,DEFAULTFONTSIZE)
		navswap=0
		splitpos=400
	End Method

	Function Create:TToolsOptions (host:TOptionsGroup)
		Local	o:TToolsOptions
		o=New TToolsOptions
		o.Initialise()
		o.InitOptions(host)
		Return o
	End Function

End Type

Type TKeyOptions Extends TOptions Final
	Field	NumKeys%=0
	Field	_keyname$[]
	Field	_keycode%[]
	Field	_keymod%[]
	Field	_keycode_def%[]
	Field	_keymod_def%[]

''TODO: remove these once maxide code adds hotkeys properly
'Const KEY_AUTOC% = 2
'Const KEY_HELP% = 1

	Method Initialise()
		myname$ = "KeyOptions"
		_keyname = [""]
		_keycode = [0]
		_keymod = [0]
		_keycode_def = [0]
		_keymod_def = [0]
	End Method

	Method Write(stream:TStream)
		For Local i% = 1 To NUMKEYS 
			If _keycode[i] = _keycode_def[i] And _keymod[i]=_keymod_def[i] Then Continue
			stream.WriteLine "key_name="+_keyname[i]
			stream.WriteLine "key_code="+_keycode[i]
			stream.WriteLine "key_mod="+_keymod[i]
		Next
	End Method

	Method Clear() 
		For Local i%=1 To NUMKEYS
			_keycode[i] = 0
			_keymod[i]=0			
		Next		
	End Method

	Field	_cur_keyid%
	Method Read(key$ , value$) 
		Select key$
			Case "key_name" ; _cur_keyid=AddHotKey(value$,0,0)
			Case "key_code" ; If _cur_keyid>0 Then _keycode[_cur_keyid]=Int(value$)
			Case "key_mod" ; If _cur_keyid>0 Then _keymod[_cur_keyid]=Int(value$)
		End Select
	End Method

	Method FindKeyID%(s$) 
		For Local i%=1 To NUMKEYS
			If _keyname[i]=s$ Then Return i
		Next
		Return 0
	End Method

	Method SetDefaults() 
		For Local i%=1 To NUMKEYS
			SetKeyToDefault(i)
		Next					
	End Method
	
	Method AddHotKey:Int(name$,keyc:Int,keym:Int,keyc_def%=0,keym_def%=0)
		Local index:Int
		name = Trim(name) 
		'if it already exists, DO NOT use the keycode and mod specified, just store the defaults and return the index
		index = FindKeyID(name) 
		If index > 0 Then
			If keyc_def >0 Then _keycode_def[index]=keyc_def;_keymod_def[index]=keym_def
			Return index
		EndIf
		'otherwise, create a new item for the list and store it
		NumKeys:+ 1
		index=NumKeys
		_keyname = _keyname[..index + 1]
		_keycode = _keycode[..index + 1]
		_keymod = _keymod[..index + 1]
		_keycode_def = _keycode_def[..index + 1]
		_keymod_def = _keymod_def[..index + 1]		
		_keyname[index] = name
		_keycode[index] = keyc
		_keymod[index] = keym
		_keycode_def[index] = keyc_def
		_keymod_def[index] = keym_def
		Return index
	End Method
	
	Method AddHotKeyDef:Int(name$ , keyc:Int , keym:Int)
		'to be used from code to define the hotkey and set the defaults
		Return AddHotKey(name , keyc , keym , keyc , keym)
	End Method
	
	Method KeyName$(i:Int) 
		Return _keyname[i]		
	End Method
	
	Method HotKeyName$(i:Int)
		Return KeyToName(_keycode[i] , _keymod[i])		
	End Method
	
	Method GetKey:Int(i:Int) 
		Return _keycode[i]
	End Method
	
	Method GetKeyMod:Int(i:Int) 
		Return _keymod[i]
	End Method
	
	Method SetKey(index:Int , keyc:Int , keyM:Int)
		_keycode[index] = keyc
		_keymod[index]=keym	
	End Method
	
	Method SetKeyToDefault(index:Int)
		_keycode[index] = _keycode_def[index]
		_keymod[index]=_keymod_def[index]		
	End Method

	Method CheckKey%(id%,key%,mods%)
		If key=_keycode[id] And mods=_keymod[id] Then Return True
		Return False
	End Method

	Function Create:TKeyOptions (host:TOptionsGroup)
		Local	o:TKeyOptions
		o=New TKeyOptions
		o.Initialise()
		o.InitOptions(host)
		Return o
	End Function

End Type

Type THelpOptions Extends TOptions Final
	Field	tempsourcelinks%		'should we prepend a . to source links?
	Field	externalhelp:Int

	Method Initialise()
		myname$="HelpOptions"
	End Method

	Method Write(stream:TStream)
		stream.WriteLine "temp_source_links="+tempsourcelinks
		stream.WriteLine "external_help=" + externalhelp
	End Method

	Method Clear()
	End Method

	Method Read(key$,value$)
		Select key$
			Case "temp_source_links" ; tempsourcelinks=Int(value$)
			Case "external_help" ; externalhelp=Int(value$)
		End Select
	End Method

	Method SetDefaults()
		tempsourcelinks=True
		externalhelp = False
	End Method

	Function Create:THelpOptions (host:TOptionsGroup)
		Local	o:THelpOptions
		o=New THelpOptions
		o.Initialise()
		o.InitOptions(host)
		Return o
	End Function

End Type


'
' AutoComplete has its own section for future additions.
'
Type TAutoCompleteOptions Extends TOptions Final
	' x,y not used but TRect has nice handling allready ;)
	Field	winsize:TRect = New TRect

	Method SetWindowSize( width:Int,height:Int)
		winsize.x = 0
		winsize.y = 0
		winsize.w = width
		winsize.h = height
		_group.dirty = True
	EndMethod

	Method Initialise()
		myname = "AutoCompleteOptions"
	EndMethod

	Method Write(stream:TStream)
		stream.WriteLine "win_size=" + winsize.ToString()
	EndMethod

	Method Clear()
	EndMethod

	Method Read( key:String, value:String)
		Select key
			Case "win_size" winsize.FromString( value)
		EndSelect
	EndMethod

	Method SetDefaults()
		winsize.Set( 0,0, 256,192)
	EndMethod

	Function Create:TAutoCompleteOptions( host:TOptionsGroup)
		Local o:TAutoCompleteOptions = New TAutoCompleteOptions
		o.Initialise()
		o.InitOptions(host)
		Return o
	EndFunction
EndType
