'projects.bmx

SuperStrict

Import "datanodes.bmx"
Import "host.bmx" 
Import "util/miscfuncs.bmx"
Import "tool_svn.bmx"
Import "locale/language.bmx"
Import "util/requesttext.bmx"

Const NODEFILETYPES:String = ..
	"|bmx|bbdoc" + ..
	"|bb" + ..
	"|cpp|cxx|c|m|h|hpp|s|rc|asm" + ..
	"|txt" + ..
	"|html|htm|js|css|xml" + ..
	"|bat|cmd|sh" + ..
	"|ini|cfg|conf|config" + ..
	"|jpg|png|bmp|tga" + ..
	"|wav|ogg|mp3" + ..
	"|pl|doc|plist" + ..
	"|glsl|hlsl" + ..
	"|lua" + ..
	"|py|"
	
Const TEXTFILETYPES:String = ..
	"|bmx|bbdoc" + ..
	"|bb" + ..
	"|cpp|cxx|c|m|h|hpp|s|rc|asm" + ..
	"|txt" + ..
	"|html|htm|js|css|xml" + ..
	"|bat|cmd|sh" + ..
	"|ini|cfg|conf|config" + ..
	"|pl|doc|plist" + ..
	"|glsl|hlsl" + ..
	"|lua" + ..
	"|py|"

Const IMAGEFILETYPES:String =	"|jpg|png|bmp|tga|"

Const SOUNDFILETYPES:String = "|wav|ogg|mp3|"

?Win32
Const EXECFILETYPES:String = "|exe|cmd|bat|"
?Linux
Const EXECFILETYPES:String = "|sh|"
?MacOS
Const EXECFILETYPES:String = "|app|sh|"
?


Type TProjects Extends TDataNode
	Field	addproj:TDataNode
	Field	projects:TProjectFolderNode[]

	Method RemoveProject(index:Int)
		If index >= 0 And index < projects.Length Then
			projects[index].Free()
			projects = projects[..index] + projects[index+1..]
			UpdateNodes()
			host.UpdateOptions()
		EndIf
	EndMethod

	Method MoveProject:Int(index:Int,dir:Int)
		If index >= 0 And index < projects.Length Then
			Local newidx:Int = index + dir
			If (dir < 0 And newidx >= 0) Or (dir > 0 And newidx < projects.Length) Then
				Local tmp:TProjectFolderNode = projects[newidx]
				projects[newidx] = projects[index]
				projects[index] = tmp
				UpdateNodes()
				Return newidx
			EndIf
		EndIf
		Return index
	EndMethod
	
	Method UpdateNodes()
		Hide()
		For Local n:TDataNode = EachIn _kids
			n.Detach()
		Next
		For Local p:TProjectFolderNode = EachIn projects
			Append(p)
		Next
		Append(addproj)
		Refresh()
		Open()
	EndMethod

	Method NewProject()
		Local idx:Int = projects.Length
		projects = projects[..idx+1]
		projects[idx] = TProjectFolderNode.CreateProjectNode( Self, "New Project")
		UpdateNodes()
		host.UpdateOptions()
		
		' show the properties requester
		TGUIProjectPropertiesRequester.CreateProjectPropertiesRequester( host, projects[idx], True)
	EndMethod
		
	Method AddProject(info$)
'		Local folder:TFolderNode

		Local idx:Int = projects.Length
		projects = projects[..idx+1]
		projects[idx] = TProjectFolderNode.CreateProjectNode( Self, info)
		UpdateNodes()

'TODO: this code isnt used, as folder is never set ??
'		' now set up the project itself
'		Local popen:Int
'		Local pversion:Int
'		popen=Int(GetInfo(info))
'		pversion = Int(GetInfo(info) )
'			
'		If folder
'			folder.SetVersion pversion
'			If popen folder.Open
'		EndIf		
	EndMethod

	Method Invoke:Object(command:Int,argument:Object=Null)
		Select command
		Case TOOLNEW
			NewProject
		End Select
	End Method

	Function CreateProjects:TProjects(host:TGUIHost)
		Local	p:TProjects
		p=New TProjects
		p.SetName("@@projects_tree/projects")
		p.host=host
		p.addproj=p.AddNode("@@projects_tree/add_project")
		p.addproj.SetAction TOOLACTIVATE, p, TOOLNEW ,Null
		Return p
	End Function
	
	Method ProjectList:TStringList()
		Local projlist:TStringList =New TStringList
		For Local p:TProjectFolderNode = EachIn projects
			projlist.addlast p.ToString()
		Next
		Return projlist
	End Method	
	
End Type

Type TGUIProjectRequester Extends TGUIToolWindow
	Field	projects:TProjects
	Field	listbox:TGadget
	Field	add:TGadget
	Field	remove:TGadget
	Field	props:TGadget
	Field	moveup:TGadget
	Field	movedown:TGadget
	Field	Current:TProjectFolderNode
	
	Method Invoke:Object(command:Int,arg:Object=Null)
		Select command
		Case TOOLACTIVATE
			Refresh
		End Select
	End Method

	Method SetCurrent(i:Int)
		If i=-1 
			DisableGadget remove
			DisableGadget moveup
			DisableGadget movedown
			DisableGadget props 
			Current=Null
		Else
			Current=TProjectFolderNode(GadgetItemExtra(listbox,i))
			If Current 
				EnableGadget remove
				EnableGadget props 
				If i = 0 Then 
					' first item
					DisableGadget moveup
					EnableGadget movedown
				ElseIf i = CountGadgetItems(listbox) - 1 Then 
					' last item
					EnableGadget moveup
					DisableGadget movedown
				Else
					' middle items
					EnableGadget moveup
					EnableGadget movedown
				EndIf
			EndIf
		EndIf
	End Method

	Method Query:Int(property:Int)
		' there are no properties to query
	End Method

	Method Poll(e:TEvent)
		Local	id:Int,data:Int,index:Int,src:Object
		id=e.id
		data = e.data
		src = e.source
		
		If src = _gadget Then
			Select id
				Case EVENT_WINDOWCLOSE Hide ; Return
				Case EVENT_WINDOWACTIVATE Refresh() ; Return
				Default Return
			End Select
		End If
		If src = listbox And id = EVENT_GADGETSELECT Then SetCurrent SelectedGadgetItem(listbox);Return	
		If id<>EVENT_GADGETACTION Then Return
		Select src
			Case add
				projects.NewProject
				Refresh
			Case remove
				projects.RemoveProject( SelectedGadgetItem(listbox))
				Refresh()
			Case _ok
				projects.host.UpdateOptions()
				Hide
			Case _cancel
				Hide
			Case props
				If Current
					' show the properties
					TGUIProjectPropertiesRequester.CreateProjectPropertiesRequester(owner,Current)
				EndIf
			Case moveup
				index = projects.MoveProject( SelectedGadgetItem(listbox), -1)
				Refresh()
				SetCurrent(index)	
				SelectGadgetItem listbox, index
				
			Case movedown
				index = projects.MoveProject( SelectedGadgetItem(listbox), 1)
				Refresh()
				SetCurrent(index)	
				SelectGadgetItem listbox, index
		End Select
	End Method

	Method Refresh()
		ClearGadgetItems listbox
		For Local p:TProjectFolderNode = EachIn projects.projects
			AddGadgetItem listbox, p._name, 0,-1,"", p
		Next
		SetCurrent -1
	End Method

	Method Open(projnode:TProjects)
		projects=projnode
		Refresh
		Show
	End Method

	Function CreateProjectRequester:TGUIProjectRequester(host:TGUIHost)
		Local x:Int,y:Int
		Local	proj:TGUIProjectRequester
		Local	win:Tgadget
		
		proj=New TGUIProjectRequester

		LanguageSetGroup "project_requester"
		proj.owner=host
		win=proj.InitToolWindow(host,"@@title",400,256)
		proj.listbox=CreateListBox(  8,8,240,180,win  )

		x=proj.width()-136
		proj.add=CreateButton("@@add_project",x,8,128,28,win)
		proj.remove=CreateButton("@@remove_project",x,40,128,28,win)
		proj.moveup=CreateButton("@@move_up",x,72,128,28,win)
		proj.movedown=CreateButton("@@move_down",x,104,128,28,win)
		proj.props=CreateButton("@@properties",x,136,128,28,win)

		DisableGadget proj.remove
		DisableGadget proj.moveup
		DisableGadget proj.movedown
		DisableGadget proj.props
		
		Return proj
	End Function
	
End Type

Type TGUIProjectPropertiesRequester Extends TGUIToolWindow
	Field	proj:TProjectFolderNode
	Field	localname:TGadget
	Field	localpath:TGadget
	Field	pathbutton:TGadget
	Field	vctype:TGadget
	Field	path:TGadget
	Field	user:TGadget
	Field	password:TGadget
	Field	_module:TGadget
	Field	checkout:TGadget
	Field	update:TGadget
	Field	commit:TGadget
	Field	dirty:Int
	Field	_isnew:Int

	Method Invoke:Object(command:Int,arg:Object=Null)
		Select command
		Case TOOLACTIVATE 
			Refresh
		End Select
	End Method
	
	Method Query:Int(property:Int)
		' there are no properties to query
	End Method	
	
	Method Tidy()
		
		If dirty
			Local a$,b$,c$,d$,e$
			Local idx:Int = SelectedGadgetItem(vctype)
			If  idx > 0 Then
				a = GadgetItemText(vctype, idx)
				b = GadgetText(path)
				c = GadgetText(user)
				d = GadgetText(password)
				e$ = GadgetText(_module)
			EndIf
			Local n$=GadgetText(localname)
			Local p$=GadgetText(localpath)
			proj.Set n,p,a,b,c,d,e
			dirty = False
			proj.host.UpdateOptions()
		EndIf
	End Method
	
	Method Poll(e:TEvent)
		Select e.source
			Case localname,localpath,path,user,password,_module
				dirty = True
				
			Case vctype
				If e.id<>EVENT_GADGETACTION Then Return
				dirty=True
				Tidy
				_SetGadgetState
				
			Case pathbutton
				If e.id<>EVENT_GADGETACTION Then Return
				Local dir$=RequestDir("@@project_requester/select_folder")
				If dir
					If dir[dir.length-1..]="/"	'fltk hack
						dir=dir[..dir.length-1]
					EndIf
					SetGadgetText localpath,dir
					If GadgetText(localname)=""
						SetGadgetText localname,StripDir(dir)
					EndIf
					dirty=True
				EndIf
								
			Case checkout
				If e.id<>EVENT_GADGETACTION Then Return
				Tidy
				Hide				
				proj.Checkout		

			Case commit
				If e.id<>EVENT_GADGETACTION Then Return
				Tidy
				Hide				
				proj.Commit	
					
			Case update
				If e.id<>EVENT_GADGETACTION Then Return
				Tidy
				Hide				
				proj.Update		

			Case _ok
				If e.id<>EVENT_GADGETACTION Then Return
				Tidy
				Free
								
			Case _cancel
				If e.id<>EVENT_GADGETACTION Then Return
				If _isnew Then
					'proj.detach()
					proj.Free()
				EndIf
				Free
				
			Case _gadget
				If e.id = EVENT_WINDOWCLOSE
					If _isnew Then
						'proj.detach()
						proj.Free()
					EndIf
					Free
				EndIf
		End Select
	End Method
	
	Method Refresh()
		SetGadgetText localname,proj.name()
		SetGadgetText localpath , proj.path()
		If proj.vc Then
			Local count:Int = 0
			For Local a$ = EachIn VC_List
				count:+1
				If a = proj.vc.vctype Then SelectGadgetItem vctype, count
			Next
			SetGadgetText path,proj.vc._path
			SetGadgetText user,proj.vc._user
			SetGadgetText password , proj.vc._pwd
			SetGadgetText _module, proj.vc._module
		EndIf
		_SetGadgetState
	End Method
	
	Method _SetGadgetState()
		If SelectedGadgetItem(vctype)>0 Then
			EnableGadget path
			If proj.vc And proj.vc.RequiresModuleName() Then EnableGadget _module Else DisableGadget _module
			EnableGadget user
			EnableGadget password
			EnableGadget checkout
			EnableGadget update
			EnableGadget commit
		Else
			SelectGadgetItem vctype, 0
			DisableGadget path
			DisableGadget _module			
			DisableGadget user
			DisableGadget password
			DisableGadget checkout
			DisableGadget update
			DisableGadget commit
		EndIf
	End Method
	
	Function CreateProjectPropertiesRequester:TGUIProjectPropertiesRequester(host:TGUIContainer,projnode:TProjectFolderNode,isnew:Int=False)
		Local	projGUI:TGUIProjectPropertiesRequester
		Local	i:Int , n:Int , y:Int
		Local	win:Tgadget
		Local	title$ = "Project Properties"

		projGUI = New TGUIProjectPropertiesRequester
		projGUI.proj=projnode
		projGUI._isnew = isnew
		If isnew Then title$="New "+title$
		win = projGUI.InitToolWindow(host , title$ , 512 , 380) 
		y=12

		CreateLabel("Name:",8,y,72,24,win)
		projGUI.localname=CreateTextField(88,y,340,20,win)
		y:+30

		CreateLabel("Path:",8,y,72,24,win)
		projGUI.localpath=CreateTextField(88,y,340,20,win)
		projGUI.pathbutton=CreateButton("..",434,y,20,20,win)
		y:+30

		Local svnbox:TGadget
		svnbox=CreatePanel(4,y,500,310-y,win,PANEL_GROUP)
		SetGadgetText svnbox,"Version Control"
		y = 12
		CreateLabel "Software Type:" , 8 , y , 72 , 24 , svnbox
		projGUI.vctype = CreateComboBox(88 , y , 100 , 20 , svnbox) 
		AddGadgetItem projGUI.vctype , "None"
		For Local a$ = EachIn VC_List
			AddGadgetItem projGUI.vctype , a$
		Next
		
		y:+30
		CreateLabel("URL:",8,y,72,24,svnbox)
		projGUI.path=CreateTextField(88,y,380,20,svnbox)
		y:+30
		CreateLabel("Module:",8,y,72,24,svnbox)
		projGUI._module=CreateTextField(88,y,380,20,svnbox)
		y:+30
		CreateLabel("Username:",8,y,72,24,svnbox)
		projGUI.user=CreateTextField(88,y,380,20,svnbox)
		y:+30
		CreateLabel("Password:",8,y,72,24,svnbox)
		projGUI.password=CreateTextField(88,y,380,20,svnbox,TEXTFIELD_PASSWORD)		
		y:+30
		projGUI.checkout=CreateButton("Check Out Project",10,y,150,28,svnbox)
		projGUI.update=CreateButton("Update Project",170,y,150,28,svnbox)
		projGUI.commit=CreateButton("Commit Project",330,y,150,28,svnbox)

		projGUI.Refresh
		projGUI.Show

		Return projGUI
	End Function
End Type

Type TProjectFolderNode Extends TFolderNode
	Field	svnversion:Int
	Field	svnerr$
	Field	vc:TVersionControl

	Method ToString$()
		Local prj$
		prj=_name+"|"+_path
		If vc Then prj:+ "|" + vc.ToString()
		Return prj
	End Method
	
	Method FromString(info$)
		Local n$
		n$=GetInfo(info)
		If n="" Then n="Unknown"
		setname n
		_path = GetInfo(info)
		If Trim(info)<>"" Then vc=TVersionControl.CreateFromString(host,info)
		Scan()
		Local isopen:Int,vers:Int
		isopen=Int(GetInfo(info))
		If isopen
			Open
'			Rescan
'			scanned=False
		EndIf	
		vers=Int(GetInfo(info))
		If vers
			SetVersion vers
		EndIf
'		Notify "svnpass="+svnpass
	End Method

	Method Invoke:Object(command:Int,argument:Object=Null)
		Select command
		
		Case TOOLPROPERTIES
			Highlight()
			TGUIProjectPropertiesRequester.CreateProjectPropertiesRequester(TGUIHost(host) , Self)
			
		Case TOOLUPDATE
			highlight()
			Update() 
				
		Case TOOLCOMMIT
			Highlight()
			Commit()
						
		Case TOOLMENU
			Highlight()
			Local menu:TGadget = CreatePopupMenu()
			TGUIHost(host).popupmenu(menu,Self)
			
		Default
			Return Super.Invoke(command , argument)							
			
		End Select

	End Method

	Method Set(n$ , p$ , vctype$,s$ , user$ , pass$,modul$)
		_path=p		
		setname n
		vc=TVersionControl.CreateVC(host,vctype,s,user,pass,modul$)
		Rescan
	End Method

	Method CreatePopupMenu:TGadget()
		'create menus
		Local projmenu:TGadget=CreateMenu("Projects",0,Null)
		CreateMenu "Refresh",MENUREFRESH,projmenu
		CreateMenu "Explore...",MENUBROWSE,projmenu
		CreateMenu "Shell...",MENUSHELL,projmenu
		CreateMenu "" , 0 , projmenu
		If vc Then
			CreateMenu "Update Version",MENUUPDATE,projmenu
			CreateMenu "Commit Version",MENUCOMMIT,projmenu
			CreateMenu "" , 0 , projmenu
		EndIf
		CreateMenu "Properties..." , MENUPROPS , projmenu
		Return projmenu
	End Method	

	Function CreateProjectNode:TProjectFolderNode(projects:TProjects,info$)
		Local n:TProjectFolderNode
		
		n=New TProjectFolderNode
		n.SetParent(projects)
		n.host = projects.host
		n.foldertype = PROJECTFOLDER
		n.fromstring(info)
		Return n
	End Function
	
	Method Checkout()
		If vc Then vc.Checkout(_path)
	End Method

	Method Update()
		If vc Then vc.Update(_path)
	End Method

	Method Commit() 
		If vc Then 
			'TODO: add support for entering a comment
			host.disable()
			LanguageSetGroup "version_control"
			Local c$ = RequestText(Translate("@@title") , Translate("@@message") , "") 
			host.enable()
			If c<>"" Then vc.Commit(_path,c)
		EndIf
	End Method
	
	Method GetIndex:Int()
		Local	node:TNewDataNode
		Local	i:Int

		If _parent
			For node=EachIn _parent._kids
				If node=Self Then Return i
				If TProjectFolderNode(node) Then i :+ 1
			Next
		EndIf
	End Method
End Type

Type TFileNode Extends TDataNode
	Field _path:String
	Field _project:TProjectFolderNode
	
	Method New()
		_typename = "TFileNode"
	EndMethod
	
	Method CreatePopupMenu:TGadget()
		Local menu:TGadget = CreateMenu( "File Properties", 0, Null)
		CreateMenu "Shell...", MENUSHELL, menu
		If _project Then
'			CreateMenu "" , 0 , menu
'			CreateMenu "Properties..." , MENUPROPS , menu
		EndIf
		Return menu
	EndMethod
	
	Method Invoke:Object( command:Int, argument:Object=Null)
		Select command
			Case TOOLPROPERTIES
				Highlight()
				'TODO: add file properties dialog?
				Notify "File Properties..."
	
			Case TOOLSHELL
				Local ext:String = "|" + ExtractExt(_name).ToLower() + "|"
				If EXECFILETYPES.Find(ext) <> - 1 Then
					Local cd:String = CurrentDir()
					ChangeDir RealPath(_path)
					host.execute _path + "/" + _name, "Executing external application ("+_name+")"
					ChangeDir cd					
				Else
					OpenURL( RealPath(_path + "/" + _name))
				EndIf
				
			Case TOOLMENU
				Highlight()
				TGUIHost(host).PopupMenu( CreatePopupMenu(), Self)
				
			Case TOOLOPEN,TOOLACTIVATE
				Local ext:String = "|" + ExtractExt(_name).ToLower() + "|"
				If IMAGEFILETYPES.Find(ext) <> -1 Then
					'TODO: open image viewer here
					Notify "Image Viewer..."
				ElseIf SOUNDFILETYPES.Find(ext) <> - 1 Then
					'TODO: open sound player here				
					Notify "Sound Player..."
				ElseIf TEXTFILETYPES.Find(ext) <> - 1 Then
					Return Super.Invoke( command, argument)
				Else
					' just the default for now
					Return Super.Invoke( command, argument)
				EndIf
	
			Default
				Return Super.Invoke( command, argument)
		EndSelect
	EndMethod	
EndType

Type TFolderNode Extends TDataNode
	Field	_path$
	Field	scanned:Int
	Field	_projectfolder:TProjectFolderNode

	Field	version:Int
	Field	foldertype:Int
	
	Const PROJECTFOLDER:Int=0
	Const DIRECTORYFOLDER:Int=1
	Const FILEFOLDER:Int=2

	Method Name$() ; Return _name ; End Method
	Method Path$() ; Return _path$ ; End Method

	Method Invoke:Object(command:Int,argument:Object=Null)
		Local p:Int
		Local line$
			
		Select command
		Case TOOLOUTPUT
			line$=String(argument)
			p=line.find(" revision ")
			If p>-1
				SetVersion Int(line[p+10..])
			EndIf
'			If line[..12]="At revision "
'			DebugLog "TOOLOUTPUT:"+line
			
		Case TOOLERROR
			line$=String(argument)
'			DebugLog "TOOLERROR:"+line

		Case TOOLPROPERTIES
			' show the properties
			If _projectfolder Then TGUIProjectPropertiesRequester.CreateProjectPropertiesRequester(host , _projectfolder) 

		Case TOOLREFRESH				
			Rescan
			
		Case TOOLBROWSE				
			OpenURL RealPath(_path)
			
		Case TOOLSHELL				
'			system_ RealPath(path)
			Local cd$=CurrentDir()
			ChangeDir RealPath(_path)
?MacOS
			Local s$="/bin/bash"' --help"
			host.execute s$,"Shell Terminal"
?Linux
			Local s$="/bin/bash"' --help"
			host.execute s$,"Shell Terminal"
?Win32
			host.execute "cmd","Shell Terminal - Type Exit To End"
?
			ChangeDir cd
								
		Case TOOLMENU
			'Highlight	
			If _projectfolder Then _projectfolder.Invoke TOOLMENU,[0]
			
		Default
			'if nothing trapped the event, pass it up the chain
			Return Super.Invoke(command,argument)
		End Select
		
	End Method

	Method FindFolderFromPath:TFolderNode(dir$)
		Local folder:TFolderNode
		Local result:TFolderNode
		If _path=dir Return Self
		For folder=EachIn _kids
			result=folder.FindFolderFromPath(dir)
			If result Return result
		Next
	End Method
	
	Method SetName(n$)
		If version n=n+"("+version+")"
		Super.SetName n
		Refresh
	End Method
		
	Method SetVersion(ver:Int)
		Local n$	
		version=ver
		n=StripDir(_path)
'		If version n=n+"("+version+")"
		SetName n
	End Method
	
'	Method Write(stream:TStream)
'		Local folder:TFolderNode
'		Local isopen:Int
'		If GetState()=OPENSTATE isopen=True
'		If version Or isopen
'			stream.WriteLine "proj_data="+_path+"|"+isopen+"|"+version+"|"
'		EndIf
'		For folder=EachIn _kids
'			folder.Write(stream)
'		Next
'	End Method

	Method Open(view:Int=-1)
		If Not scanned
			ScanKids		'Path name,path
		EndIf
		Super.Open view
	End Method

	Method AddFileNode:TDataNode(file$)
		Local	f:TFileNode
		Local 	ext$

		If NodeFileTypes.find("|"+ExtractExt(file).toLower()+"|")<0 Return Null
		
		f = New TFileNode
		f.SetName( StripDir(file))
		f._path = ExtractDir(file)
		f.host = host
		Append(f)
		
		' check if this file is in a project
		Local prj:TNewDataNode
		prj=f
		While prj
			If TProjectFolderNode(prj) Then f._project = TProjectFolderNode(prj)
			prj=prj._parent
		Wend
	
		f.SetAction(TOOLACTIVATE,host,TOOLOPEN,file)
		f.SetAction(TOOLNAVIGATE,host,TOOLMENU,file)
		ext=ExtractExt(file$).ToLower()
		f._sortname=ext+f._name
		Return f
	End Method

	Method AddFolderNode:TDataNode(path$)
		Local	n:TDataNode
		n = TFolderNode.CreateFolderNode( Self, path , DIRECTORYFOLDER)
		'TODO: whats this space being added for?
		n._sortname=" "+n._name
		Append n
		Return n
	End Method

	Method Scan()
		Local	flist:TList
		Local	dir:Int,f$,p$

		flist=New TList
		dir=ReadDir(_path)
		If dir
			Repeat
				f$=NextFile(dir)
				If f$="" Exit
				If f[..1]="." Continue
				p$=RealPath(_path+"/"+f)
				Select FileType(p$)
				Case FILETYPE_FILE
					AddFileNode p$
				Case FILETYPE_DIR
					AddFolderNode p$
				End Select
			Forever
			CloseDir dir
		EndIf
		SortKids
	End Method

	Method Rescan()
		FreeKids
		Scan
		scanned=False
		ScanKids
		Refresh
	End Method

	Method ScanKids()
		Local	f:TFolderNode
		If scanned Return
		For f=EachIn _kids
			f.Scan
		Next
		scanned=True
	End Method

	Function CreateFolderNode:TFolderNode(p:TDataNode,path$,foldertype:Int)
		Local	n:TFolderNode
		n=New TFolderNode
		n.setname StripDir(path)
		n.setparent p
		n._path = path
		n.foldertype = foldertype
		n.host = p.host
		If n.host = Null Then RuntimeError "Oops, I lost my host!"
		
		' find my parent project folder node.  It will be null if it doesn't exist (samples?)
		Local prj:TNewDataNode
		prj=p
		While prj
			If TProjectFolderNode(prj) Then n._projectfolder = TProjectFolderNode(prj)
			prj=prj._parent
		Wend
		Return n
	End Function

End Type


