SuperStrict

Import "tool.bmx"

Const SPLITWIDTH:Int=4

Type TSplitter Extends TGUIElement
	Const	HORIZONTAL:Int=0
	Const	VERTICAL:Int=1

	Field	axis:Int
	Field	gadget0:TGadget
	Field	gadget1:TGadget
	Field	group:TGadget
	Field	panel:TGadget
	Field	pos:Int,flipped:Int,snap:Int,oldpos:Int,oldsize:Int

	Method  Invoke:Object(command:Int,argument:Object=Null)
		' there is no point in invoking a command on a splitter object at this time, so do nothing!
	End Method

	Method Query:Int(property:Int)
		' there are no properties to query
	End Method

	Method Refresh()
	End Method

	Method Hide()
		HideGadget panel
	End Method

	Method Show()
		HideGadget panel
	End Method

	Method Position(p:Int)
		Move p-pos
	End Method
	
	Method Gadget:TGadget()
		Return panel
	End Method

	Method Move(d:Int)
		Local	x:Int,y:Int,w:Int,h:Int,p:Int

		pos:+d
		w=ClientWidth(group)
		h=ClientHeight(group)
		If pos<0 pos=0
		If axis=VERTICAL
			If pos>w-SPLITWIDTH pos=w-SPLITWIDTH
			y=Min(gadget0.ypos,gadget1.ypos)
			h=Max(gadget0.height+gadget0.ypos,gadget1.height+gadget1.ypos)-y
			If (d<0)~flipped
				SetGadgetShape gadget0,0,y,pos-SPLITWIDTH,h
				SetGadgetShape panel,pos-SPLITWIDTH,0,SPLITWIDTH*2,h
				SetGadgetShape gadget1,pos+SPLITWIDTH,0,w-(pos+SPLITWIDTH),h
			Else
				SetGadgetShape gadget1,pos+SPLITWIDTH,0,w-(pos+SPLITWIDTH),h
				SetGadgetShape panel,pos-SPLITWIDTH,0,SPLITWIDTH*2,h
				SetGadgetShape gadget0,0,0,pos-SPLITWIDTH,h
			EndIf
		Else
			If pos>h-SPLITWIDTH pos=h-SPLITWIDTH
			If (d<0)~flipped
				SetGadgetShape gadget0,0,0,w,pos-SPLITWIDTH
				SetGadgetShape panel,0,pos-SPLITWIDTH,w,SPLITWIDTH*2
				SetGadgetShape gadget1,0,pos+SPLITWIDTH,w,h-(pos+SPLITWIDTH)
			Else
				SetGadgetShape gadget1,0,pos+SPLITWIDTH,w,h-(pos+SPLITWIDTH)
				SetGadgetShape panel,0,pos-SPLITWIDTH,w,SPLITWIDTH*2
				SetGadgetShape gadget0,0,0,w,pos-SPLITWIDTH
			EndIf
		EndIf
	End Method

	Method Resize()
		If axis=VERTICAL
			pos=GadgetX(panel)+SPLITWIDTH
			If pos<0 Or pos>ClientWidth(group) move(0)
		Else
			pos=GadgetY(panel)+SPLITWIDTH
			If pos<0 Or pos>ClientHeight(group) move(0)
		EndIf
	End Method

	Method Poll(e:TEvent)
		If e.Source<>panel Return
		Select e.id
			Case EVENT_MOUSEDOWN
				snap=True
				If axis
					oldpos=e.x
				Else
					oldpos=e.Y
				EndIf
			Case EVENT_MOUSEUP
				snap=False
			Case EVENT_MOUSEMOVE
				If axis
					SetPointer POINTER_SIZEWE
				Else
					SetPointer POINTER_SIZENS
				EndIf
				If snap
					If axis
						Move e.x-oldpos
					Else
						Move e.Y-oldpos
					EndIf
				EndIf
			Case EVENT_MOUSELEAVE
				SetPointer POINTER_DEFAULT				
		End Select
	End Method

	Method SetFlip(f:Int,init:Int=True)
		Local	t:TGadget
		If f<>flipped
			flipped=f
			t=gadget0;gadget0=gadget1;gadget1=t
			If init
				If axis=VERTICAL
					pos=ClientWidth(group)-pos-1
				Else
					pos=ClientHeight(group)-pos-1
				EndIf
			EndIf
		EndIf
		If axis
			If flipped
				SetGadgetLayout gadget0,1,0,1,1
				SetGadgetLayout gadget1,1,1,1,1
				SetGadgetLayout panel,1,0,1,1
			Else
				SetGadgetLayout gadget0,1,1,1,1
				SetGadgetLayout gadget1,0,1,1,1
				SetGadgetLayout panel,0,1,1,1
			EndIf
		Else
			If flipped
				SetGadgetLayout gadget0,1,1,1,0
				SetGadgetLayout gadget1,1,1,1,1
				SetGadgetLayout panel,1,1,1,0
			Else
				SetGadgetLayout gadget0,1,1,1,1
				SetGadgetLayout gadget1,1,1,0,1
				SetGadgetLayout panel,1,1,0,1
			EndIf
		EndIf
		Position pos
	End Method

	Function Create:TSplitter(axis:Int,pos:Int,flipped:Int,gadget0:TGadget,gadget1:TGadget)
		Local	s:TSplitter
		s=New TSplitter
		s.axis=axis
		s.gadget0=gadget0
		s.gadget1=gadget1
		s.group=GadgetGroup(gadget0)
		s.panel=CreatePanel(0,0,0,0,s.group,PANEL_ACTIVE)'|PANEL_BORDER)
		s.pos=pos
		s.setflip flipped,False
		Return s
	End Function
End Type
