' style.bmx
'
' This file contains code relating to defining styles for text and gadgets

SuperStrict

Import "gui_module.bmx"
Import BRL.Max2d
Import BRL.GLMax2D

Type TColor
	Field	red:Int,green:Int,blue:Int

	Method Set(rgb:Int)
		red=(rgb Shr 16)&255
		green=(rgb Shr 8)&255
		blue=rgb&255
	End Method

	Method ToString$()
		Return ""+red+","+green+","+blue
	End Method

	Method FromString(s$)
		Local	p:Int,q:Int
		p=s.Find(",")+1
		If Not p Return
		q=s.Find(",",p)+1
		If Not q Return
		red=Int(s[..p-1])
		green=Int(s[p..q-1])
		blue=Int(s[q..])
	End Method

	Method FromName(s$)
		Select s.ToLower()
			Case "red";red=255;green=0;blue=0
			Case "green";red=0;green=255;blue=0
			Case "blue";red=0;green=0;blue=255						
		End Select
		Return
	End Method

	Method Request:Int()
		If RequestColor(red,green,blue)
			red=RequestedRed()
			green=RequestedGreen()
			blue=RequestedBlue()
			Return True
		EndIf				
	End Method	
End Type

Type TTextStyle
	Field	color:TColor
	Field	flags:Int
	
	Method Set(rgb:Int,bolditalic:Int)
		color.set(rgb)
		flags=bolditalic
	End Method

	Method Format(textarea:TGadget,pos:Int,length:Int)
		FormatTextAreaText textarea,color.red,color.green,color.blue,flags,pos,length
	End Method

	Method ToString$()
		Return ""+color.red+","+color.green+","+color.blue+","+flags
	End Method

	Method FromString(s$)
		Local	p:Int,q:Int,r:Int
		p=s.Find(",")+1;If Not p Return
		q=s.Find(",",p)+1;If Not q Return
		r=s.Find(",",q)+1;If Not r Return
		color.red=Int(s[..p-1])
		color.green=Int(s[p..q-1])
		color.blue=Int(s[q..r-1])
		flags=Int(s[r..])
	End Method

	Function Create:TTextStyle()
		Local	s:TTextStyle
		s=New TTextStyle
		s.color=New TColor
		Return s
	End Function

End Type

Type TGadgetStyle
	Field	font_name$,font_size:Int',font_width:Int
	Field	fg:TColor
	Field	bg:TColor
	Field	font:TGUIFont

	Method Apply(gadget:TGadget)
		SetGadgetFont gadget,font
		SetGadgetColor gadget,bg.red,bg.green,bg.blue,True
		SetGadgetColor gadget,fg.red,fg.green,fg.blue,False
	End Method
	
	Method Set(fg_rgb:Int,bg_rgb:Int,fname$,fsize:Int)
		fg.set(fg_rgb)
		bg.set(bg_rgb)
		font_name=fname
		font_size=fsize
		font = LoadGuiFont(font_name , font_size) 
'		Local f:TImageFont = LoadImageFont(font_name , font_size) 
'		SetImageFont f
'		font_width=TextWidth("W")
'		f=Null
	End Method

	Method ToString$()
		Return font_name+","+font_size+","+fg.ToString()+","+bg.ToString()
	End Method
	
	Function GetArg$(a$ Var)
		Local 	p:Int,r$
		p=a.Find(",")+1
		If p=0 p=Len a$+1
		r$=a$[..p-1]
		a$=a$[p..]
		Return r$	
	End Function
	
	Method FromString(s$)		
		font_name=GetArg(s$)
		font_size=Int(GetArg(s$))
		fg.red=Int(GetArg(s$))
		fg.green=Int(GetArg(s$))
		fg.blue=Int(GetArg(s$))
		bg.red=Int(GetArg(s$))
		bg.green=Int(GetArg(s$))
		bg.blue=Int(GetArg(s$))
		font=LoadGuiFont(font_name,font_size)
	End Method

	Function Create:TGadgetStyle()
		Local	s:TGadgetStyle
		s=New TGadgetStyle
		s.fg=New TColor
		s.bg=New TColor
		Return s
	End Function
End Type
