'
' Test Harness for BlitzMax Integrated Development Environment - Community Edition
' ================================================================================
'
' Requires BRL.MaxGui & BAH.MaxUnit modules to compile
'
' This program provides an automated test suite for the CE IDE, using Brucey's nifty MaxUnit module.

SuperStrict

Framework BRL.Blitz

' Import the relevant BRL / PUB modules
Import brl.standardio
Import brl.retro
Import brl.filesystem
Import brl.system
Import pub.freeprocess
Import brl.pngloader
Import brl.timer
Import brl.maxutil
Import pub.macos ' for ExportMod on MacOS
Import maxgui.proxygadgets
Import bah.maxunit

' Import the generic code used by the IDE
Import "util/bytebuffer.bmx"	' buffers up a stream
Import "splitter.bmx"		' provides the GUI splitter 'gadget'
Import "requestmultifile.bmx"	' Import multiple file requester
Import "singleinstance.bmx"	' Import single instance helper

'Import the MaxIDE specific stuff
Import "style.bmx"			' provides functionality to specify and manage text styles
Import "host.bmx"			' provides base IDE functionality
Import "requesters.bmx"		' provides all of the MaxIDE specific requesters
Import "projects.bmx"		' provides all project based functionality
Import "debugtree.bmx"		' debugging code
Import "help.bmx"			' all help stuff
Import "todo.bmx"			' todo viewing stuff
Import "codetree.bmx"		' code tree stuff
Import "options.bmx"		' code relating to options
Import "optionsgui.bmx"		' GUI for setting options
Import "datanodeview.bmx"	' the navigation bar on the side
Import "version.bmx"		' the versioning file
Import "tool_svn.bmx"		' Subversion support
Import "icons.bmx"			' icons and related constants

Import "diff.bmx"			' difference
Include "autocomplete.bmx"	' autocomplete window
Include "debug.bmx"			' debugging code (output panel)
Include "editor.bmx"		' code editor
Include "tool_buildmods.bmx"	' BlitzMax syncmods tool
Include "bmxhost.bmx"		' main code block for the IDE (host)
' Notes: on dependent Imports:
'  language.bmx has been imported in tool.bmx

'---------------------------------------------------------------

' Start MaxIDE
Global maxide:TBMXHost
maxide = New TBMXHost
maxide.Initialize()

' Run the tests
New TTestSuite.run()

'Simply end to prevent any changes (e.g. to IDE options) from being saved
End

'----------------------------------------------------------------

Function DebugLog(s:String);End Function ' over-ride DebugLog so that the output window doesn't get barfed by debug output

'----------------------------------------------------------------

Type MAXIDE_Basic_Tests Extends TTest

'TODO: Should really set all the IDE Options to expected states (e.g. auto-tidy & auto-complete on, format of highlighted text, etc)

	Method testCloseAll() { test }
		Local cp:TCodePanel
		assertTrue(maxide.CloseAll("",True))
	End Method

End Type

'----------------------------------------------------------------

Type MAXIDE_File_Tests Extends TTest

	Method setup() { beforetype }
		'anything before the test(s)
		maxide.CloseAll("",True)
	End Method

	Method testOpen() { test }
		assertNotNull(maxide.OpenSource("maxide.bmx"))
		assertEqualS("maxide.bmx",maxide.currentpanel.name)		
	End Method

	Method testOpenTemp() { test }
		assertNotNull(maxide.OpenSource(bmxpath+"/src/bmk/bmk.bmx",True))
		assertEqualS(".bmk.bmx",maxide.currentpanel.name)
	End Method

	Method testNew() { test }
		assertNotNull(maxide.OpenSource(""))
		assertEqualS("untitled1.bmx",maxide.currentpanel.name)
		assertNotNull(maxide.OpenSource(""))
		assertEqualS("untitled2.bmx",maxide.currentpanel.name)		
	End Method
	
	Method testCloseAll() { test }
		assertTrue(maxide.CloseAll())
		assertEqualS("@@help_panel/title",maxide.currentpanel.name)
	End Method	
	
	Method finish() { aftertype }
		maxide.CloseAll("",True)
	End Method	

End Type

'----------------------------------------------------------------

Type MAXIDE_Edit_Tests Extends TTest

	Global cp:TCodePanel
	Global expected$

	Method setup() { before }
		'anything before the test(s)
		maxide.CloseAll("",True)
		cp=maxide.OpenSource("")
		expected=cp.text()
	End Method

	Method testKeys() { test }
		assertequalS("",cp.text())
		
		Local test$="0123456789abcefghifghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ '~q(){}[]=+-/\*^<>$%!&_@#?|.,:;~~~t"
		cp.simulatetyping(test)
		assertEqualS(test,cp.text()) 
	End Method

	Method testReturnKey() { test }
		assertequalS("",cp.text())
		
		cp.simulatekeypress(KEY_RETURN)
		assertEqualS("~n",cp.text()) 
		
		cp.simulatetyping("~n")
		assertEqualS("~n~n",cp.text()) 

		cp.simulatetyping(" test ~n")
		assertEqualS("~n~n test~n",cp.text()) 
				
	End Method

	Method testCommandCapitaliseNotTooEarly() { test }
		assertequalS("",cp.text())
			
		cp.simulatetyping("wHiL")
		assertEqualS("wHiL" , cp.text() )
		
		' add an E, should remain capitalised as typed (don't change too early!)
		cp.simulatekeypress(KEY_E , MODIFIER_SHIFT)
		assertEqualS("wHiLE" , cp.text() )
		
		cp.simulatekeypress(KEY_SPACE)
		assertEqualS("While ",cp.text())			
			
		cp.Undo()
		assertEqualS("" , cp.text() )
		
		cp.simulatetyping("MODIFIER")	
		assertEqualS("MODIFIER" , cp.text() )
		
		cp.Undo()
		cp.simulatetyping("F")
		cp.simulatekeypress(KEY_LEFT)
		cp.simulatetyping("id")
		assertEqualS("idF",cp.text()) ' over-aggressive capitalisation yields Idf.  Also had diF as a result once!
		
		cp.Undo()
		cp.Undo()
		cp.simulatekeypress(KEY_D)
		cp.simulatekeypress(KEY_LEFT)
		assertEqualS("d" , cp.text() )		
		
		cp.simulatetyping("mo")
		assertEqualS("mod" , cp.text() )	' may be Mod too soon
		
		cp.simulatekeypress(KEY_RIGHT)
		cp.simulatekeypress(KEY_SPACE)
		assertEqualS("Mod " , cp.text() )		
		
	End Method

	Method testCommandCapitaliseAfterEdit() { test }
		assertequalS("",cp.text())
			
		cp.simulatetyping("moad")
		assertEqualS("moad" , cp.text() )
				
		cp.simulatekeypress(KEY_SPACE)
		assertEqualS("moad " , cp.text() )			
		
		cp.simulatekeypress(KEY_LEFT , , 2)	
		cp.simulatekeypress(KEY_BACKSPACE)		
		assertEqualS("mod " , cp.text() )
		
		cp.simulatekeypress(KEY_RIGHT, , 2)
		assertEqualS("mod " , cp.text() )
		
		cp.simulatekeypress(KEY_SPACE)
		assertEqualS("Mod  " , cp.text() )	
				
	End Method

	Method testCommandHighlighting() { test }
		assertequalS("",cp.text())
		
		cp.simulatetyping("wHiL")
		assertEqualS("wHiL",cp.text())
		
		' should remain as "normal" style
		cp.simulatekeypress(KEY_LEFT,MODIFIER_SHIFT,4)
		AssertEqualsI(0,TextAreaColor(cp._textarea),"Is black? (1)")
		AssertFalse(IsTextAreaStyle(cp._textarea , TEXTFORMAT_BOLD) ,"Is not bold? (1)")
		cp.simulatekeypress(KEY_RIGHT)
		
		' add an E, should now be "keyword" style
		cp.simulatekeypress(KEY_E , MODIFIER_SHIFT)
		'assertEqualS("wHiLE",cp.text()) ' should be this, but no need to explicitly tested as already proven by another test
		cp.simulatekeypress(KEY_LEFT,MODIFIER_SHIFT,5)
		AssertEqualsI(255,TextAreaColor(cp._textarea),"Is red? (1)")
		AssertTrue(IsTextAreaStyle(cp._textarea,TEXTFORMAT_BOLD),"Is bold? (1)")
		cp.simulatekeypress(KEY_RIGHT)	
		
		' add a space
		cp.simulatekeypress(KEY_SPACE)
		cp.simulatekeypress(KEY_LEFT,MODIFIER_SHIFT)
		assertEqualS("While " , cp.text() )
		cp.simulatekeypress(KEY_RIGHT)

		' end char and space should be "normal"
		AssertEqualsI(0,TextAreaColor(cp._textarea),"Is black? (2)")
		AssertFalse(IsTextAreaStyle(cp._textarea,TEXTFORMAT_BOLD),"Is not bold? (2)")

		' end char and space should be "normal"
		AssertEqualsI(0,TextAreaColor(cp._textarea),"Is black? (3)")
		AssertFalse(IsTextAreaStyle(cp._textarea,TEXTFORMAT_BOLD),"Is not bold? (3)")		
		cp.simulatekeypress(KEY_LEFT , MODIFIER_SHIFT)
		AssertEqualsI(0,TextAreaColor(cp._textarea),"Is black? (4)")
		AssertFalse(IsTextAreaStyle(cp._textarea,TEXTFORMAT_BOLD),"Is not bold? (4)")
		cp.simulatekeypress(KEY_LEFT)		
		
		'While should be "keyword"
		cp.simulatekeypress(KEY_LEFT , MODIFIER_SHIFT , 5)
		AssertEqualsI(255,TextAreaColor(cp._textarea),"Is red? (2)")
		AssertTrue(IsTextAreaStyle(cp._textarea,TEXTFORMAT_BOLD),"Is bold? (2)")
		cp.simulatekeypress(KEY_LEFT)	
		
		'remove a char
		cp.simulatekeypress(KEY_RIGHT , , 3)
		cp.simulatekeypress(KEY_BACKSPACE)
		assertEqualS("Whle ",cp.text(),"Backspace")

		cp.simulatekeypress(KEY_RIGHT)
		cp.simulatekeypress(KEY_LEFT,MODIFIER_SHIFT,4)
		AssertEqualsI(0,TextAreaColor(cp._textarea),"Is black? (5)")
		AssertFalse(IsTextAreaStyle(cp._textarea , TEXTFORMAT_BOLD) ,"Is not bold? (5)")

	End Method

	Method testCommandHighlightingWithSpecialKeys() { test }
		assertequalS("",cp.text())
		
		cp.simulatekeypress(KEY_TAB)
		assertEqualS("~t",cp.text())
		
		AssertEqualsI(0,TextAreaColor(cp._textarea))
		AssertFalse(IsTextAreaStyle(cp._textarea,TEXTFORMAT_BOLD))
		
		cp.simulatetyping("while ")
		assertEqualS("~tWhile " , cp.text() , "Before Home")	
		
		cp.simulatekeypress(KEY_LEFT)
		cp.simulatekeypress(KEY_LEFT , MODIFIER_SHIFT , 5)
		AssertEqualsI(255,TextAreaColor(cp._textarea))
		AssertTrue(IsTextAreaStyle(cp._textarea,TEXTFORMAT_BOLD))
				
		cp.simulatekeypress(KEY_HOME)
		assertEqualS("~tWhile ",cp.text(), "After Home")

		cp.simulatekeypress(KEY_RIGHT,MODIFIER_SHIFT,5)
		AssertEqualsI(255,TextAreaColor(cp._textarea))
		AssertTrue(IsTextAreaStyle(cp._textarea , TEXTFORMAT_BOLD) )	
		
		cp.simulatekeypress(KEY_LEFT)
		cp.simulatekeypress(KEY_LEFT , MODIFIER_SHIFT)
		AssertEqualsI(0,TextAreaColor(cp._textarea))
		AssertFalse(IsTextAreaStyle(cp._textarea,TEXTFORMAT_BOLD))		
						
	End Method

	Method testAutoIndent() {test}
		Local t$
		assertequalS("",cp.text())
	
		cp.simulatetyping("While")
		cp.simulatekeypress(KEY_RETURN)
		t$:+"While~n~t"
		assertEqualS(t$,cp.text(),"While")
		
		cp.simulatetyping("For")
		cp.simulatekeypress(KEY_RETURN)
		t$:+"For~n~t~t"
		assertEqualS(t$,cp.text(),"For")

		cp.simulatetyping("Repeat")
		cp.simulatekeypress(KEY_RETURN)
		t$:+"Repeat~n~t~t~t"
		assertEqualS(t$,cp.text(),"Repeat")

		cp.simulatetyping("Type")
		cp.simulatekeypress(KEY_RETURN)
		t$:+"Type~n~t~t~t~t"
		assertEqualS(t$,cp.text(),"Type")

		cp.simulatetyping("Method")
		cp.simulatekeypress(KEY_RETURN)
		t$:+"Method~n~t~t~t~t~t"
		assertEqualS(t$,cp.text(),"Method")
		
		cp.simulatetyping("Function")
		cp.simulatekeypress(KEY_RETURN)
		t$:+"Function~n~t~t~t~t~t~t"
		assertEqualS(t$,cp.text(),"Function")		
		
	End Method

	Method testAutoTidy() {test}
		assertequalS("",cp.text())
		
		cp.simulatetyping("if (a+b)*7.0/1-c(f,7) then a();z(1,2)")
		assertEqualS("If (a+b)*7.0/1-c(f,7) Then a();z(1,2)",cp.text())
		
		cp.simulatekeypress(KEY_RETURN)
		assertEqualS("If (a + b) * 7.0 / 1 - c(f , 7) Then a() ; z(1 , 2)~n",cp.text())
	End Method

	Method finish() { aftertype }
		maxide.CloseAll("",True)
	End Method

End Type

'----------------------------------------------------------------

Type MAXIDE_Undo_Tests Extends TTest

	Global cp:TCodePanel

	Method setup() { before }
		'anything before the test(s)
		maxide.CloseAll("",True)
		cp=maxide.OpenSource("")
	End Method
	
	Method finish() { aftertype }
		maxide.CloseAll("",True)
	End Method	

	Method testSingleKeyUndo() { test }
		assertequalS("",cp.text())
		
		cp.simulatekeypress(KEY_A)
		assertEqualS("a",cp.text()) 
		
		cp.Undo()
		assertEqualS("",cp.text()) 		
		
		cp.Redo()
		assertEqualS("a",cp.text())
	End Method

	Method testSingleKeywordUndo() { test }
		assertequalS("",cp.text())
		
		cp.simulatetyping("wHiLe ")
		assertEqualS("While ",cp.text()) 
		
		cp.Undo()
		assertEqualS("",cp.text()) 		
		
		cp.Redo()
		assertEqualS("While ",cp.text())
	End Method

	Method testSimpleBackspaceUndo() { test }
		assertequalS("",cp.text())
		
		cp.simulatetyping("abcdefg")
		assertEqualS("abcdefg",cp.text(),"Typing 1") 

		cp.simulatekeypress(KEY_BACKSPACE)
		assertEqualS("abcdef",cp.text(),"Backspace") 
				
		cp.Undo()
		assertEqualS("abcdefg",cp.text(),"Undo 1")

		cp.Undo()
		assertEqualS("",cp.text(),"Undo 2")
		
		cp.Redo()
		assertEqualS("abcdefg",cp.text(),"Redo 1")

		cp.Redo()
		assertEqualS("abcdef",cp.text(),"Redo 2")
		
		cp.simulatetyping("ghijklmnop")
		assertEqualS("abcdefghijklmnop",cp.text(),"Typing 2")		
	End Method

	Method testBackspaceMidTypingUndo() { test }
		assertequalS("",cp.text())
		
		cp.simulatetyping("abcdefgg")
		assertEqualS("abcdefgg",cp.text(),"Typing 1") 

		cp.simulatekeypress(KEY_BACKSPACE)
		assertEqualS("abcdefg",cp.text(),"Backspace") 

		cp.simulatetyping("hijklmnop")
		assertEqualS("abcdefghijklmnop",cp.text(),"Typing 2") 
				
		cp.Undo()
		assertEqualS("abcdefg",cp.text(),"Undo 1")
		
		cp.Undo()
		assertEqualS("abcdefgg",cp.text(),"Undo 2")

		cp.Undo()
		assertEqualS("",cp.text(),"Undo 3")

		cp.Redo()
		assertEqualS("abcdefgg",cp.text(),"Redo 1")

		cp.Redo()
		assertEqualS("abcdefg",cp.text(),"Redo 2")
				
		cp.Redo()
		assertEqualS("abcdefghijklmnop",cp.text(),"Redo 3")
		
		cp.simulatetyping("qrst")
		assertEqualS("abcdefghijklmnopqrst",cp.text(),"Typing 3") 
	End Method
	
	Method testSingleLineUndo() { test }
		assertequalS("",cp.text())
		
		Local test$="If a Then b"
		cp.simulatetyping(test)
		assertEqualS(test,cp.text()) 
		
		cp.Undo()
		assertEqualS("",cp.text()) 		
		
		cp.Redo()
		assertEqualS(test,cp.text())
	End Method
	
	Method testSingleLineUndoWithAutoTidy() { test }
		assertequalS("",cp.text())
		
		Local test1$="If a(1,2)/2 = True Then b();c(1,1*2)"
		cp.simulatetyping(test1)
		assertEqualS(test1,cp.text()) 
		
		cp.Undo()
		assertEqualS("",cp.text()) 		
		
		cp.Redo()
		assertEqualS(test1,cp.text()) 	
		
		cp.simulatekeypress(KEY_RETURN)
		assertEqualS("If a(1 , 2) / 2 = True Then b() ; c(1 , 1 * 2)~n",cp.text()) 	
				
		cp.Undo()
		assertEqualS(test1,cp.text()) 		

		cp.Undo()
		assertEqualS("",cp.text())
		
		cp.Redo()
		assertEqualS(test1,cp.text()) 		
		
		cp.Redo()
		assertEqualS("If a(1 , 2) / 2 = True Then b() ; c(1 , 1 * 2)~n",cp.text()) 				
	End Method	
	
End Type

'--------------------------------------------

Type MAXIDE_ComplexUndo_Tests Extends TTest

	Global cp:TCodePanel

	Method setup() { before }
		'anything before the test(s)
		maxide.CloseAll("",True)
		cp=maxide.OpenSource("")
	End Method
	
	Method finish() { aftertype }
		maxide.CloseAll("",True)
	End Method	
		
	Method testArrowsMidTypingUndo() { test }
		assertequalS("",cp.text())
		
		cp.simulatetyping("Fild")
		assertEqualS("Fild",cp.text(),"Typing 1") 

		cp.simulatekeypress(KEY_LEFT)
		cp.simulatekeypress(KEY_LEFT)		
		assertEqualS("Fild",cp.text(),"2xLeft") 

		cp.simulatetyping("e")
		assertEqualS("Field",cp.text(),"Insert e") 
				
		cp.Undo()
		assertEqualS("Fild",cp.text(),"Undo 1")

		cp.Undo()
		assertEqualS("",cp.text(),"Undo 2")
		
		cp.Redo()
		assertEqualS("Fild",cp.text(),"Redo 1")
		
		cp.Redo()
		assertEqualS("Field",cp.text(),"Redo 2")

		cp.simulatekeypress(KEY_S)
		assertEqualS("Fiesld",cp.text(),"Typing 2")
		
		cp.simulatekeypress(KEY_RIGHT)
		cp.simulatekeypress(KEY_RIGHT)		
		assertEqualS("Fiesld",cp.text(),"2xRight")
		
		cp.simulatekeypress(KEY_S)
		assertEqualS("Fieslds",cp.text(),"Typing 3")				
				
	End Method

	Method testBackspaceMidTypingUndo() { test }
		assertequalS("",cp.text())
		
		cp.simulatetyping("testing")
		assertEqualS("testing",cp.text(),"Typing 1") 

		cp.simulatekeypress(KEY_LEFT,,4)
		assertEqualS("testing",cp.text(),"4xLeft") 

		cp.simulatekeypress(KEY_SPACE)
		assertEqualS("tes ting",cp.text(),"Insert SPACE") 

		cp.simulatekeypress(KEY_RIGHT,,3)
		assertEqualS("tes ting",cp.text(),"3xRight") 

		cp.simulatekeypress(KEY_BACKSPACE)
		assertEqualS("tes tig",cp.text(),"Backspace 1") 

		cp.simulatekeypress(KEY_BACKSPACE)
		assertEqualS("tes tg",cp.text(),"Backspace 2") 
				
		cp.Undo()
		assertEqualS("tes ting",cp.text(),"Undo 1")

		cp.Undo()
		assertEqualS("testing" , cp.text() , "Undo 2")
		
		cp.Undo()
		assertEqualS("",cp.text(),"Undo 3")		
		
		cp.Redo()
		assertEqualS("testing",cp.text(),"Redo 1")
		
		cp.Redo()
		assertEqualS("tes ting" , cp.text() , "Redo 2")
		
		cp.Redo()
		assertEqualS("tes tg",cp.text(),"Redo 3")		
				
	End Method

'TODO: test Delete as well as Backspace

	Method testBlockBackspaceAndUndo() { test }
		assertequalS("",cp.text())
		
		cp.simulatetyping("testing")
		assertEqualS("testing",cp.text(),"Typing 1") 

		cp.simulatekeypress(KEY_LEFT,,4)
		assertEqualS("testing",cp.text(),"4xLeft") 

		cp.simulatekeypress(KEY_SPACE)
		assertEqualS("tes ting",cp.text(),"Insert SPACE 1") 

		cp.simulatekeypress(KEY_RIGHT,,3)
		assertEqualS("tes ting",cp.text(),"3xRight") 

		cp.simulatekeypress(KEY_LEFT,MODIFIER_SHIFT,3)			
		assertEqualS("tes ting",cp.text(),"3xShift+Left") 

		cp.simulatekeypress(KEY_BACKSPACE)
		assertEqualS("tes g",cp.text(),"Backspace 1") 

		cp.simulatekeypress(KEY_LEFT,,3) 
		assertEqualS("tes g" , cp.text() , "3xLeft")
		
		cp.simulatekeypress(KEY_SPACE)
		assertEqualS("t es g",cp.text(),"Insert SPACE 2") 		
		
		cp.Undo()
		assertEqualS("tes g",cp.text(),"Undo 1")

		cp.Undo()
		assertEqualS("tes ting" , cp.text() , "Undo 2")
		
		cp.Undo()
		assertEqualS("testing",cp.text(),"Undo 3")		
		
		cp.Redo()
		assertEqualS("tes ting",cp.text(),"Redo 1")
		
		cp.Redo()
		assertEqualS("tes g" , cp.text() , "Redo 2")
		
		cp.Redo()
		assertEqualS("t es g",cp.text(),"Redo 3")		
				
	End Method

	Method testBlockOvertypeAndUndo() { test }
		assertequalS("",cp.text())
		
		cp.simulatetyping("testing")
		assertEqualS("testing",cp.text(),"Typing 1") 

		cp.simulatekeypress(KEY_LEFT,,4)
		assertEqualS("testing",cp.text(),"4xLeft") 

		cp.simulatekeypress(KEY_SPACE)
		assertEqualS("tes ting",cp.text(),"Insert SPACE 1") 

		cp.simulatekeypress(KEY_RIGHT,MODIFIER_SHIFT,3)			
		assertEqualS("tes ting",cp.text(),"3xShift+Right") 

		cp.simulatekeypress(KEY_X,MODIFIER_SHIFT)
		assertEqualS("tes Xg",cp.text(),"Overtype X") 

		cp.simulatekeypress(KEY_LEFT,,3) 
		assertEqualS("tes Xg" , cp.text() , "3xLeft")
		
		cp.simulatekeypress(KEY_SPACE)
		assertEqualS("te s Xg",cp.text(),"Insert SPACE 2") 		
		
		cp.Undo()
		assertEqualS("tes Xg",cp.text(),"Undo 1")

		cp.Undo()
		assertEqualS("tes ting" , cp.text() , "Undo 2")
		
		cp.Undo()
		assertEqualS("testing",cp.text(),"Undo 3")		
		
		cp.Redo()
		assertEqualS("tes ting",cp.text(),"Redo 1")
		
		cp.Redo()
		assertEqualS("tes Xg" , cp.text() , "Redo 2")
		
		cp.Redo()
		assertEqualS("te s Xg",cp.text(),"Redo 3")		
				
	End Method

End Type

'----------------------------------------------------------------

Type MAXIDE_Help_Tests Extends TTest

	Global cp:TCodePanel
	Global expected$

	Method setup() { beforetype }
		'anything before the test(s)
		maxide.CloseAll("",True)
		cp=maxide.OpenSource("")
		expected=cp.text()
	End Method
	
	Method test() { test }
		cp.simulatekeypress(KEY_I)
		cp.simulatekeypress(KEY_F)
		cp.invoke TOOLHELP		
		assertSame(cp,maxide.currentpanel)		
		assertEqualS("If : ",Left(maxide._currentstatustext,5))
		cp.invoke TOOLHELP
		assertEqualS("@@help_panel/title",maxide.currentpanel.name)
		maxide.selectpanel(cp)
		assertSame(cp,maxide.currentpanel)	
	End Method

	Method finish() { aftertype }
		maxide.CloseAll("",True)
	End Method

End Type

'------------------------------------------------------------------
?win32
Function GetTextAreaFormat:CHARFORMAT(textarea:TGadget)
	Local fmt:CHARFORMAT=New CHARFORMAT
	fmt.cbSize = SizeOf(fmt)
	Local i:Int=SendMessageW(textarea.query(QUERY_HWND),EM_GETCHARFORMAT,SCF_SELECTION,Int Byte Ptr fmt)
	Return fmt
End Function
?

Function TextAreaColor:Int(textarea:TGadget)
?win32
	Return GetTextAreaFormat(textarea).crTextColor
?Not win32
	Return True				'TODO: Equivalent functionality needed on gtk,fltk,macos
?
End Function

Function IsTextAreaStyle:Int(textarea:TGadget , style:Int)
?win32
	Local c:Int
	If style = TEXTFORMAT_BOLD Then c = c | CFM_BOLD
	If style = TEXTFORMAT_ITALIC Then c = c | CFM_ITALIC
	Return (GetTextAreaFormat(textarea).dwEffects & c)
?Not win32
	Return True				'TODO: Equivalent functionality needed on gtk,fltk,macos
?
End Function

