' Provides SVN & CVS integration
SuperStrict

Import BRL.FileSystem

Import "util/miscfuncs.bmx"
Import "tool.bmx"

Const VC_SVN$ = "SVN"
Const VC_CVS$ = "CVS"
Global VC_List:String[] = [VC_SVN,VC_CVS] ' why cant const's be based on consts!

Type TVersionControl Extends TBase
	Field	vctype$ 'what type of version control is it?
	Field	_path$
	Field	_user$
	Field	_pwd$
	Field	_module$
	Field	_host:TBase 'who should I invoke when I need something doing?
	
	Function CreateFromString:TVersionControl(host:TBase,str:String Var)
		Local vct$ = getinfo(str)
		Local path$ = getinfo(str)
		Local user$ = getinfo(str)
		Local pwd$ = crypt(getinfo(str) ) 
		Local modul$ = getinfo(str)
		Return CreateVC(host,vct,path,user,pwd,modul)
	End Function
	
	Function CreateVC:TVersionControl(host:TBase , VCType$, path$, user$, pwd$, modul$)
		Local vc:TVersionControl
		Select VCType
			Case VC_SVN ; vc = New TVC_SVN
			Case VC_CVS ; vc = New TVC_CVS
			Case "" ; Return Null
			Default ; RuntimeError "oops - version control type not recognised!"
		End Select
		vc._host = host
		vc._path = path
		vc._user = user
		vc._pwd = pwd
		vc._module = modul
		Return vc
	End Function
	
	Function ErrorBox( msg:String)
		Local oldapptitle$ = AppTitle
		AppTitle = LanguageString("error_message/vc_error_title")
		Notify msg
		AppTitle = oldapptitle$
	EndFunction	

	Method ToString:String()
		Return vctype + "|" + _path + "|" + _user + "|" + crypt(_pwd) + "|" + _module
	End Method

	Method Run(cmd$ , about$ , refresh:Int)
'		DebugLog "CUR-DIR:: " + CurrentDir()
'		DebugLog "VC-RUN:: " + cmd
		Local arr:Object[5]
		arr[0] = cmd
		arr[1] = about
		arr[4] = Self
		If refresh
			arr[2] = String(MENUREFRESH)
			arr[3] = String(True)
			_host.invoke TOOLEXECUTE,arr
		Else
			_host.invoke TOOLEXECUTE,arr
		EndIf
	End Method
	
	Method Invoke:Object(command:Int , argument:Object = Null)
		Select command		
			Case TOOLPARSEERROR
				Return String.FromInt( ParseError( argument.ToString()))
		EndSelect
	EndMethod
	
	Method Query:Int(property:Int)
	EndMethod	
	
	Method Update(path$) Abstract		' bring down all changes
	Method Commit(path$,comment$) Abstract		' send your changes up
	Method Checkout(path$) Abstract		' creates a working copy based on content of server
	Method RequiresModuleName:Int() Abstract
	Method ParseError:Int( err:String) Abstract ' parses error output of command
	
End Type

Type TVC_SVN Extends TVersionControl
	
?Macos
Const SVNCMD$="/usr/local/bin/svn"
?win32
Const SVNCMD$="svn"
?linux
Const SVNCMD$="/usr/bin/svn"
?	
	
	Method New()
		vctype = VC_SVN
	End Method

	Method AddLogon$(cmd$)
		If _user 
			cmd:+" --username "+_user
			If _pwd cmd:+" --password "+_pwd
		EndIf
		Return cmd$
	End Method

	Method Update(path$)
		Local cmd$
		cmd$ = svncmd + " update"
		cmd=addlogon(cmd)
		cmd:+" "+quote(path)
		Run cmd,"Updating "+path,True
	End Method
	
	Method Commit(path$,comment$)
		Local cmd$
		cmd = svncmd + " commit"
		cmd=addlogon(cmd)
		cmd:+" -m "+quote(comment)
		cmd:+" "+quote(path)
		Run cmd,"Committing "+path,False
	End Method	

	Method Checkout(path$)	'to do - needs to move old version to temp?
		Local cmd$
		cmd$ = svncmd + " checkout"
		cmd$=addlogon(cmd)
		cmd:+" "+quote(_path)
		cmd:+" "+quote(path)
		Run cmd,"Checking out "+_path+" to "+path,True
	End Method
	
	Method ParseError:Int( err:String)
		Local lines:String[] = err.Split("~n")
		If lines.Length <= 0 Then Return False
		
		For Local error:String = EachIn lines
			Local i:Int = error.Find(":")
			If i > 0 Then
				ErrorBox( error[i..].Trim())
				Return True
			EndIf
		Next
		
		Return False
	EndMethod
	
	Method RequiresModuleName:Int(); Return False;End Method

End Type

Type TVC_CVS Extends TVersionControl

?Macos
Const SVNCMD$="/usr/local/bin/cvs"
?win32
Const SVNCMD$="cvs"
?linux
Const SVNCMD$="/usr/bin/cvs"
?		
	
	Method New()
		vctype = VC_CVS
	End Method
	
	Method AddLogon$(cmd$)
		If _pwd<>"" Or _user.tolower()="anonymous" Then
			cmd = ":pserver:" + _user + ":" + _pwd + "@" + cmd$
		Else
			cmd = ":ext:" + _user + "@" + cmd$
		EndIf
		Return cmd$
	End Method	
		
	Method Update(path$)
		Local cmd$ , prevdir$
		prevdir=CurrentDir()
		ChangeDir(path)
		cmd:+ svncmd
		cmd:+ " update"
		Run cmd , "Updating " + path , True	
		ChangeDir prevdir	
	End Method
	
	Method Commit(path$ , comment$)
		Local cmd$ , prevdir$
		prevdir=CurrentDir()
		ChangeDir path
		cmd = svncmd
		cmd:+" commit"
		cmd:+" -m "+quote(comment)
		Run cmd , "Committing " + path , False		
		ChangeDir prevdir
	End Method		
	
	Method Checkout(path$)
		' cvs gets confused if the working directory already exists, so delete it.
		If FileType(path) <> FILETYPE_NONE Then
			Local files:String[] = LoadDir(path)
			If files.Length > 0 Then
				If Not Confirm( "@@version_control/delete_all_files") Then Return
			EndIf
			DeleteDir(path,True)
		EndIf
		Local cmd$
		cmd$ = svncmd
		cmd:+ " -d" + addlogon(_path) ' repository
		cmd:+ " checkout"
		cmd:+ " -d " + quote(path) ' project root
		cmd:+ " " + _module
		Run cmd,"Checking out "+_path+" to "+path,True	
	End Method
	
	Method ParseError:Int( err:String)		
		Local lines:String[] = err.Split("~n")
		If lines.Length = 0 Then Return False
		
		Local noerrcount:Int = 0
		For Local error:String = EachIn lines
			' it seems CVS spits out sub-directories its updating on stderr, try to detect these and skip them
			If error.Find( "cvs checkout: Updating") = 0 Or error.Find( "cvs update: Updating") = 0 Then
				noerrcount :+ 1
				Continue
			EndIf
			
			' server errors
			If error.Find( "cvs server:") = 0 Then
				ErrorBox( error[11..].Trim())
				Return True
			Else
				' other errors
				Local start:Int = error.Find("[")
				Local stop:Int = error.Find("]", start)
				If (start <> -1) And (stop <> -1) And (start < stop) Then
					ErrorBox( error[stop+2..].Replace(":","~r~n"))
					Return True
				EndIf
			EndIf
		Next
		
		' dont allow any further error parsing if there are no real errors
		If noerrcount >= lines.Length Then Return True
		
		Return False		
	EndMethod	

	Method RequiresModuleName:Int(); Return True;End Method

End Type

