' diff.bmx - string comparison and undo/redo functionality

SuperStrict

Import brl.linkedlist

Type TDiff

	Field CurrText:String

	Field UndoList:TList = CreateList()
	Field RedoList:TList = CreateList()

	Method Init(Text:String)
		CurrText = Text
		UndoList.Clear()
		RedoList.Clear()

	EndMethod

	Method CalcDiff:Int(NewText:String Var, Action:Int = 0 , cursorpos:Int=-1 , cursorlen:Int=-1)
		Local dummy1:Int, dummy2:Int, dummy3:String
		Local FirstValid:Int = 0
		Local LastValid:Int = 0

		If Not DiffString(CurrText, NewText, FirstValid, LastValid, True)
			Return False ' the strings are equal
		EndIf

		Local OldPart:String = CurrText[FirstValid..CurrText.length - LastValid]

		Local dd:TDiffData = New TDiffData

		dd.FirstValid = FirstValid
		dd.LastValid = LastValid
		dd.ChangedPart = OldPart
		dd.Action = Action
		dd.CursorPos = cursorpos
		dd.CursorLen = cursorlen

		UndoList.AddLast dd
		RedoList.Clear()


		CurrText = NewText

?debug
		Print "CalcDiff"
		Print "FirstValid = " + dd.FirstValid
		Print "LastValid =  " + dd.LastValid
		Print "ChangedPart =    " + dd.ChangedPart
		Print "Data =       " + UndoList.Count()
?
		Return True
	EndMethod

	Method CanUndo:Int()
		Return UndoList.Count() > 0
	EndMethod

	Method CanRedo:Int()
		Return RedoList.Count() > 0
	EndMethod

	Method Undo:TDiffData()
		If Not CanUndo() Then Return Null

		Local dd:TDiffData = TDiffData(UndoList.RemoveLast())

		Local OldPart:String = CurrText[dd.FirstValid..CurrText.length - dd.LastValid]
		CurrText = CurrText[..dd.FirstValid] + dd.ChangedPart + CurrText[CurrText.length - dd.LastValid..]

		Local ndd:TDiffData = dd.CreateCopy()
		ndd.ChangedPart = OldPart
		RedoList.AddLast ndd

		Return dd
	EndMethod

	Method Redo:TDiffData()
		If Not CanRedo() Then Return Null

		Local dd:TDiffData = TDiffData(RedoList.RemoveLast())

		Local OldPart:String = CurrText[dd.FirstValid..CurrText.length - dd.LastValid]
		CurrText = CurrText[..dd.FirstValid] + dd.ChangedPart + CurrText[CurrText.length - dd.LastValid..]

		Local ndd:TDiffData = dd.CreateCopy()
		ndd.ChangedPart = OldPart
		UndoList.AddLast ndd

		Return dd
	EndMethod

EndType

Type TDiffData
	Field FirstValid:Int
	Field LastValid:Int
	Field ChangedPart:String
	Field Action:Int
	Field CursorPos:Int
	Field CursorLen:Int

	Method CreateCopy:TDiffData()
		Local t:TDiffData = New TDiffData
		t.FirstValid = self.FirstValid
		t.LastValid= self.LastValid
		t.ChangedPart= self.ChangedPart
		t.Action= self.Action
		t.CursorPos= self.CursorPos
		t.CursorLen= self.CursorLen
		Return t
	End Method
EndType

Function DiffString:Int(Str1:String, Str2:String, FirstValid:Int Var, LastValid:Int Var, MinimalDiff:Int = True)
	FirstValid = 0
	LastValid = 0

	' calculate difference between Str1 and Str2
	If Str1.length = Str2.length And Str1 = Str2 Then Return False

	Local MinStrLen:Int = Min(Str1.length , Str2.length)
	Local MaxStrLen:Int = Max(Str1.length , Str2.length)

	' find first equal characters
	For Local n:Int = 0 Until MinStrLen
		If Str1[n] = Str2[n] Then
			FirstValid = n + 1
		Else
			Exit
		EndIf
	Next

	' find last equal characters
	For Local n:Int = 0 Until MinStrLen
		Local n1:Int = Str1.length - 1 - n
		Local n2:Int = Str2.length - 1 - n
		If Str1[n1] = Str2[n2] Then
			LastValid = n + 1
		Else
			Exit
		EndIf
	Next

	' make sure that LastValid and FirstValid don't overlap if the change happend
	' within equal sequences of characters ("12 34 56" <-> "12 34 34 34 56")
	'
	' LastValid <= NewTextLength - AddedCharcters - FirstValid
	Local Delta:Int = Str2.length - Str1.length

	If MinStrLen < FirstValid + LastValid
		If MinimalDiff Then
			LastValid = MaxStrLen - Abs(Delta) - FirstValid
		Else
			Local TempLV:Int = MaxStrLen - Abs(Delta) - FirstValid
			Local TempFV:Int = MaxStrLen - Abs(Delta) - LastValid
			FirstValid = TempFV ; LastValid = TempLV
		EndIf
	EndIf

'	If Str1.Length <> FirstValid
'	LastValid = Min(NewText.length - Max(0 , NewText.length - CurrText.length) - FirstValid, LastValid)
'	LastValid = Min(NewText.length - Max(0 , NewText.length - CurrText.length) - FirstValid, LastValid)


	Return True
EndFunction