' miscfuncs.bmx
SuperStrict
Import pub.freeprocess
Import brl.retro
Import "../gui_module.bmx"

' Provides basic 'rectangle definition / overlap' functionality
' useful for remembering gadget sizes and checking for overlap
Type TRect
	Field		x:Int,y:Int,w:Int,h:Int

	Method Set(xpos:Int,ypos:Int,width:Int,height:Int)
		x=xpos;y=ypos;w=width;h=height
	End Method

	Method ToString$()
		Return ""+x+","+y+","+w+","+h
	End Method

	Method FromString(s$)
		Local	p:Int,q:Int,r:Int
		p=s.Find(",")+1;If Not p Return
		q=s.Find(",",p)+1;If Not q Return
		r=s.Find(",",q)+1;If Not r Return
		x=Int(s[..p-1])
		y=Int(s[p..q-1])
		w=Int(s[q..r-1])
		h=Int(s[r..])
	End Method

	Method IsOutside:Int(tx:Int,ty:Int,tw:Int,th:Int)
		If x+w<=tx Or y+h<=ty Return True
		If x>=tx+tw Or y>=ty+th Return True
	End Method

End Type

' special string list
Type TStringList Extends TList
	Method ToString$()
		Local s$,o$
		For o$=EachIn Self
			s$=s$+o$+","
		Next
		Return s[..Len(s)-2] 'chop the last char
	End Method
End Type

'Since we can't do a list of Ints without a type
Type IntStack
	Field a:Int[]
	Field maxsize:Int

	Method NotEmpty:Int()
		If Len(a) > 0 Return True Else Return False
	End Method

	Method Size:Int()
		Return Len(a)
	End Method

	Method Push(i:Int) 
		If maxsize > 0 And Len(a) = maxsize Then
			a = a[1..Len(a) + 1]
		Else
			a = a[0..Len(a) + 1]
		EndIf
		a[Len(a) - 1] = i
	End Method

	Method Pop:Int()
		Assert Len(a)>0," Tried to Pop and empty IntStack"

		Local i:Int = a[Len(a)-1]
		a = a[..Len(a)-1]
		Return i
	End Method

	Method Clear()
		a = a[..0]
	End Method

	Method Dump:String()
		Local s:String
		For Local i:Int = EachIn a
			s = s + i + ", "
		Next
		Return s
	End Method
End Type

'used to represent a point on screen, e.g. for passing to api calls
Type POINT
	Field x%
	Field y%
End Type

'add quotes to arg if spaces found, e.g. for filenames
Function Quote$(a$)
	Local	p:Int
	If a.length=0 Return ""
	If a[0]=34 Return a	'already quoted
	p=a.find(" ")
	If p=-1 Return a	'no spaces
	Return Chr(34)+a+Chr(34)
End Function

'removes whitespace from any string that occurs between printable characters and EOL on all lines
Function StripWhiteSpace$(s$)
	' takes the input string and finds the EOL (Chr10) characters
	' and strips any whitespace immediately before the EOL.
	' Whitespace includes tabs & spaces.
	Local p:Int , t$ , r$ , p2:Int , op2:Int
	p = s.find(Chr(10))
	While p > -1
		t = s[..p]
		s = s[p + 1..]
		op2 = -2
		p2 = Len(t)-1
		While p2<>op2
			op2=p2
			If t[p2] = 9 Then t = t[..p2] ; p2=Len(t)-1
			If t[p2] = 32 Then t = t[..p2] ; p2=Len(t)-1
		Wend
		r=r+t+Chr(10)
		p=s.find(Chr(10))
	Wend
	r=r+s
	Return r$
End Function

'get the version of blitzmax
Function GetBCCVersion$()
	Const	BCCVERSIONTEXT$ = "BlitzMax Release Version "
	Local	bytes:Byte[]
	Local	line$

	' I think calling bcc everytime we need the version-number is worse than using a global
	'  to store the result of the first call.
	Global	version$
	If version <> "" And version <>  "N/A" Then Return version
	version = "N/A"

	Local	bccpath$ = getenv_("BMXPATH") + "/bin/bcc"
	?win32
	bccpath:+ ".exe"
	?
'	Print bccpath
	Local process:TProcess=CreateProcess("~q"+bccpath+"~q",HIDECONSOLE)
	If Not process Then Return version
	If Not process.status() Then Return version

	While process.status()
		PollSystem
		bytes=process.pipe.ReadPipe()
		If bytes
			line$=String.FromBytes(bytes,Len bytes)
			line=line.Replace$(Chr(13),"")
			line=line.Replace$(Chr(0),"")
			If Left(line,Len(BCCVERSIONTEXT)) = BCCVERSIONTEXT Then
				version$=Trim(Mid(line,Len(BCCVERSIONTEXT)))
			End If
		EndIf
	Wend
	process.terminate
	process.close
	process = Null
	Return version$
End Function

Function UTF8BytesToString$(bytes:Byte[])
	Local pos:Int
	Local b1:Byte , b2:Byte , b3:Byte
	Local s$

	For pos = 0 To Len(bytes)-1
		b1 = bytes[pos]
		If b1 < 128 Then ' one byte long
			s$ = s$ + Chr(b1)

		ElseIf b1 < 224 Then 'two bytes long
			pos = pos + 1
			b2 = bytes[pos]
			s$ = s$ + Chr( ( (b1 - 192) * 64) + (b2 - 128))

		ElseIf b1 < 240 Then 'three bytes long
			pos=pos+1
			b2 = bytes[pos]
			pos=pos+1
			b3 = bytes[pos]
			s$ = s$ + Chr( ( (b1 - 224) * 4096) + ( (b2 - 128) * 64) + (b3 - 128) )

		EndIf
	Next
	Return s$
End Function

Function OutsideDesktop:Int(winrect:TRect)
	Local desk:TGadget,x:Int,y:Int,w:Int,h:Int
	desk=Desktop()
	x=GadgetX(desk)
	y=GadgetY(desk)
	w=GadgetWidth(desk)
	h=GadgetHeight(desk)
	Return winrect.IsOutside(x,y,w,h)
End Function

Function ConvertArrayToString$(a:Object[], sep$=",")
	Local s$,i:Int
	For i = 0 To Len(a)-1
		s = s + a[i].tostring() + sep
	Next
	Return s[..Len(s)-1]
End Function

Function ConvertStringToArray:String[](s$,sep$=",")
	Local a$[], pos:Int
	
	pos = s.find(sep)
	While pos > - 1
		a=a[..a.length+1]
		a[a.length-1] = s[..pos]
		s = s[pos+1..]
		pos=s.find(sep)
	Wend	
	If s.length > 0 Then
		a=a[..a.length+1]
		a[a.length-1] = s
	EndIf
	Return a
End Function

Function GetInfo$(a$ Var, sep$="|")
	' extracts the next item up to sep$, returning the value and removing it from the original string
	Local 	p:Int,r$
	p=a.Find("|")+1
	If p=0 p=Len a$+1
	r$=a$[..p-1]
	a$=a$[p..]
	Return r$	
End Function

Function Crypt$(a$)
	Local b$,c:Int,i:Int
	For i=0 Until Len a
		c=a[i]
		If c>31 c:~((i*-5)&31)
		b:+Chr(c&255)
	Next
	Return b
End Function

Function KeyToName$(keycode:Int , keymod:Int) 

		Local	m$
		If keycode>=KEY_0 And keycode<=KEY_9
			m$=Chr(keycode)
		ElseIf keycode>=KEY_A And keycode<=KEY_Z
			m$=Chr(keycode)
		ElseIf keycode>=KEY_F1 And keycode<=KEY_F12
			m$="F"+(keycode+1-KEY_F1)
		ElseIf keycode>=KEY_NUM0 And keycode<=KEY_NUM9
			m$="Num "+(keycode+1-KEY_NUM0)
		Else
			Select keycode
				Case KEY_LEFT m="Left"
				Case KEY_UP m="Up"
				Case KEY_RIGHT m="Right"
				Case KEY_DOWN m="Down"
				Case KEY_BACKSPACE m="Backspace"
				Case KEY_TAB m="Tab"
				Case KEY_CLEAR m="Clear"
				Case KEY_RETURN m="Enter"
				Case KEY_ESCAPE m="Escape"
				Case KEY_SPACE m="Space"
				Case KEY_PAGEUP m="PageUp"
				Case KEY_PAGEDOWN m="PageDown"
				Case KEY_END m="End"
				Case KEY_HOME m="Home"
				Case KEY_SELECT m="Select"
				Case KEY_PRINT m="Print"
				Case KEY_EXECUTE m="Exec"
'				Case KEY_SCREEN m="
				Case KEY_INSERT m="Insert"
				Case KEY_DELETE m="Delete"
				Case KEY_TILDE m="~~"
				Case KEY_MINUS m="-"
				Case KEY_EQUALS m="="
				Case KEY_OPENBRACKET m="["
				Case KEY_CLOSEBRACKET m="]"
				Case KEY_BACKSLASH m="\"
				Case KEY_SEMICOLON m=";"
				Case KEY_QUOTES m="'"
				Case KEY_COMMA m=","
				Case KEY_PERIOD m="."
				Case KEY_SLASH m="/"
				Case 220 m="\"
				Case 19,3 m="Break"	'Const KEY_PAUSE=19
'				Default
'					DebugLog "key="+keycode
			End Select
		EndIf				
		If m
			If keymod&1 m$="Shift+"+m$
			If keymod&8 m$="Apple"+m$	'apple key
			If keymod&4 m$="Alt+"+m$
			If keymod&2 m$="Ctrl+"+m$
		EndIf
'		If m="" m="undefined"
		Return m

End Function

Function KeyToChar$(keycode:Int, keymod:Int)
		Local	m$
		
		If keycode>=KEY_0 And keycode<=KEY_9
			If keymod=MODIFIER_SHIFT Then
				Select keycode
					Case KEY_1;m$="!"
					Case KEY_2;m$="~q"
					Case KEY_3;m$=""
					Case KEY_4;m$="$"
					Case KEY_5;m$="%"
					Case KEY_6;m$="^"
					Case KEY_7;m$="&"
					Case KEY_8;m$="*"
					Case KEY_9;m$="("
					Case KEY_0;m$=")"
				End Select
			Else
				m$=Chr(keycode)
			EndIf
		ElseIf keycode>=KEY_A And keycode<=KEY_Z Then
			If keymod=MODIFIER_SHIFT 'Or keymod=MODIFIER_CAPSLOCK Then
				m$=Chr(keycode)			
			Else
				m$=Chr(keycode+32)			
			EndIf
		ElseIf keycode>=KEY_NUM0 And keycode<=KEY_NUM9
			m$=keycode+1-KEY_NUM0
		ElseIf keymod=0
			Select keycode		
'				Case KEY_LEFT m="Left"
'				Case KEY_UP m="Up"
'				Case KEY_RIGHT m="Right"
'				Case KEY_DOWN m="Down"
'				Case KEY_BACKSPACE m="Backspace"
				Case KEY_TAB m="~t"
'				Case KEY_CLEAR m="Clear"
				Case KEY_RETURN m="~n"
'				Case KEY_ESCAPE m="Escape"
				Case KEY_SPACE m=" "
'				Case KEY_PAGEUP m="PageUp"
'				Case KEY_PAGEDOWN m="PageDown"
'				Case KEY_END m="End"
'				Case KEY_HOME m="Home"
'				Case KEY_SELECT m="Select"
'				Case KEY_PRINT m="Print"
'				Case KEY_EXECUTE m="Exec"
'				Case KEY_SCREEN m="
'				Case KEY_INSERT m="Insert"
'				Case KEY_DELETE m="Delete"
				Case KEY_TILDE m="'"
				Case KEY_MINUS m="-"
				Case KEY_EQUALS m="="
				Case KEY_OPENBRACKET m="["
				Case KEY_CLOSEBRACKET m="]"
				Case KEY_BACKSLASH m="\"
				Case KEY_SEMICOLON m=";"
				Case KEY_COMMA m=","
				Case KEY_PERIOD m="."
				Case KEY_SLASH m="/"
				Case 220 m="\"
				Case KEY_QUOTES m="#"
'				Case 19,3 m="Break"	'Const KEY_PAUSE=19
'				Default
'					DebugLog "key="+keycode
			End Select
		ElseIf keymod=MODIFIER_SHIFT Then
			Select keycode		
				Case KEY_MINUS m="_"
				Case KEY_EQUALS m="+"
				Case KEY_OPENBRACKET m="{"
				Case KEY_CLOSEBRACKET m="}"
				Case KEY_BACKSLASH m="|"
				Case KEY_SEMICOLON m=":"
				Case KEY_COMMA m="<"
				Case KEY_PERIOD m=">"
				Case KEY_SLASH m="?"
				Case 220 m="|"
				Case KEY_QUOTES m="~~"
				Case KEY_TILDE m="@"
			End Select
		EndIf				

		Return m

End Function
