' stringtable.bmx

SuperStrict

Import BRL.System
Import BRL.LinkedList
Import BRL.TextStream

Type TStringTable
	Field CaseSensitive:Int
	Field IndexChar:Int

	Field Encoding:Int = ENCODING_LATIN1
	Field Escape:Int = 0

	Field Groups:TList
	Field CurrGroup:TStringTableGroup

	Const AUTO_ID_START:Int = 10000
	Global AutoID:Int = AUTO_ID_START

	Const ENCODING_LATIN1:Int = TTextStream.LATIN1
	Const ENCODING_UTF8:Int = TTextStream.UTF8

	Const ESCAPE_CSTYLE:Int = 1
	Const ESCAPE_BLITZ:Int = 2

	Function Create:TStringTable(DefaultCaseSensitive:Int = False, DefaultIndexChar:Int = 0)
		Local st:TStringTable = New TStringTable

		st.CaseSensitive = DefaultCaseSensitive
		st.IndexChar = DefaultIndexChar

		st.Groups = CreateList()
		st.CurrGroup = st.CreateGroup("")

		Return st
	EndFunction

	Function CreateFromFile:TStringTable(url:Object, Groups:Object = Null)
		Local st:TStringTable = Create()
		st.Load url, Groups

		Return st
	EndFunction

	Method CreateGroup:TStringTableGroup(Key:String = "", ID:Int = 0, CaseSensitive:Int = -1, IndexChar:Int = -1)
		If CaseSensitive = -1 Then CaseSensitive = Self.CaseSensitive
		If IndexChar = -1 Then IndexChar = Self.IndexChar
		Key = Key.Trim()

		Local gr:TStringTableGroup

		If ID > 0 Then gr = FindGroupByID(ID)
		If gr Then CurrGroup = gr ; Return gr

		If ID <= 0 And Key <> "" Then gr = FindGroupByKey(Key)
		If gr Then CurrGroup = gr ; Return gr

		If ID <= 0 Then ID = AutoID ; AutoID:+ 1

		gr = New TStringTableGroup
		gr.ID = ID
		gr.Key = Key
		gr.Content = New TStringTableEntry[27]
		gr.CaseSensitive = CaseSensitive
		gr.IndexChar = IndexChar

		Groups.AddLast gr
		CurrGroup = gr

		Return gr
	EndMethod


	Method AddEntry(Value:String, ID:Int = 0, Key:String = "")
		CurrGroup.AddEntry Value, ID, Key
	EndMethod

	Method AddEntries(Entries:String[])
		For Local Entry:String = EachIn Entries
			CurrGroup.AddEntry Entry
		Next
	EndMethod


	Method FindEntryByID:TStringTableEntry(ID:Int)
		Return CurrGroup.FindEntryByID(ID)
	EndMethod

	Method FindEntryByKey:TStringTableEntry(Key:String)
		Return CurrGroup.FindEntryByKey(Key)
	EndMethod

	Method FindEntryByIDEx:TStringTableEntry(ID:Int)
		Local en:TStringTableEntry
		For Local gr:TStringTableGroup = EachIn Groups
			en = gr.FindEntryByID(ID)
			If en Then Return en
		Next
	EndMethod

	Method FindEntryByKeyEx:TStringTableEntry(Key:String)
		Local en:TStringTableEntry
		For Local gr:TStringTableGroup = EachIn Groups
			en = gr.FindEntryByKey(Key)
'			If en Print Key + " -- " + en.ToString()
			If en Then Return en
		Next
	EndMethod


	Method FindGroupByID:TStringTableGroup(ID:Int)
		For Local gr:TStringTableGroup = EachIn Groups
			If gr.ID = ID Then Return gr
		Next
	EndMethod

	Method FindGroupByKey:TStringTableGroup(Key:String)
		For Local gr:TStringTableGroup = EachIn Groups
			If gr.Key = Key Then Return gr
		Next
	EndMethod


	Method SetCurrGroupByID(ID:Int)
		CurrGroup = FindGroupByID(ID)
		If Not CurrGroup Then CurrGroup = TStringTableGroup(Groups.First())
	EndMethod

	Method SetCurrGroupByKey(Key:String)
		CurrGroup = FindGroupByKey(Key)
		If Not CurrGroup Then CurrGroup = TStringTableGroup(Groups.First())
	EndMethod


	Method Save(url:Object, Intro:String = "???", NewEncoding:Int = -1, NewEscape:Int = -1)
		Local File:TStream = WriteStream(url)
		If Not File Then Return

		If NewEncoding <> -1 Then Encoding = NewEncoding
		If NewEscape <> -1 Then Escape = NewEscape

		' write utf8 tag
		If Encoding = ENCODING_UTF8 Then
			File.WriteByte $EF
			File.WriteByte $BB
			File.WriteByte $BF
		EndIf

		' write intro
		If Intro = "???" Then
			Intro = "Saved " + CurrentDate() + " " + CurrentTime() + ..
			"~n  Items: " + Count()
			If String(url) Then Intro = String(url) + "~n  " + Intro
		EndIf

		Intro = "# " + Intro.Replace("~n", "~n# ")

		File.WriteLine Intro

		File.WriteLine ""

		Select Encoding
			Case ENCODING_UTF8 ; File.WriteLine "$encoding=utf8"
			Case ENCODING_LATIN1 ; File.WriteLine "$encoding=latin1"
		EndSelect

		Select Escape
			Case ESCAPE_CSTYLE ; File.WriteLine "$escape=cstyle"
			Case ESCAPE_BLITZ ; File.WriteLine "$escape=blitz"
		EndSelect

		File.WriteLine "$casesensitive=" + CaseSensitive
		File.WriteLine "$indexchar=" + IndexChar

		File = TTextStream.Create(File, Encoding)

		For Local gr:TStringTableGroup = EachIn Groups
			If gr.ID >= AUTO_ID_START Then
				If gr.Key <> "" Then
					File.WriteLine ""
					File.WriteLine "[" + gr.Key + "]"
				EndIf
			Else
				If gr.Key = "" Then
					File.WriteLine ""
					File.WriteLine "[" + gr.ID + "]"
				Else
					File.WriteLine ""
					File.WriteLine "[" + gr.ID + "," + gr.Key + "]"
				EndIf
			EndIf

			If gr.CaseSensitive <> CaseSensitive Then
				File.WriteLine "$casesensitive=" + gr.CaseSensitive
			EndIf
			If gr.IndexChar <> IndexChar Then
				File.WriteLine "$indexchar=" + gr.IndexChar
			EndIf

			For Local n:Int = 0 To 26
				Local en:TStringTableEntry = gr.Content[n]
				While en
					Local value:String = en.Value.Trim()
					Select Escape ' escape unicode chars when the encoding can't handle them
						Case ESCAPE_CSTYLE ; value = CStyleEscape(value, Encoding = ENCODING_LATIN1)
						Case ESCAPE_BLITZ ; value = BlitzEscape(value, Encoding = ENCODING_LATIN1)
					EndSelect
					If en.ID >= AUTO_ID_START Then
						' don't save automatically created keys
						If en.Key = ValueToKey(value, CaseSensitive) Then
							If value.StartsWith("[") Or value.StartsWith("#") Or value.Contains("=") Then
								File.WriteLine "=" + value
							Else
								File.WriteLine value
							EndIf
						Else
							File.WriteLine en.Key + "=" + value
						EndIf
					Else
						' don't save empty or automatically created keys
						If (en.Key = "") ..
						Or en.Key = ValueToKey(value, CaseSensitive) Then
							File.WriteLine en.ID + "=" + value
						Else
							File.WriteLine gr.ID + "," + gr.Key + "=" + value
						EndIf
					EndIf

					en = en.Succ
				Wend
			Next
		Next

		File.Close
	EndMethod

	Method Load(url:Object, Groups:Object = Null)
		Local File:TStream = ReadStream(url)
		If Not File Then Return

		Encoding = 0
		Escape = 0

		SetCurrGroupByKey ""
		Local LoadGroup:Int = True

		While Not File.Eof()
			Local Line:String = File.ReadLine().Trim()

			If Line = "" Then Continue
			If Line[0] = $EF And Line[1] = $BB And Line[2] = $BF Then
				Line = Line[3..] ' remove utf8 tag
			EndIf

			If Line = "" Then Continue
			If Line[..1] = "#" Then Continue

			If Escape = ESCAPE_CSTYLE And (Line[Line.Length - 2..] = " \") Then
				While Not File.Eof() And (Line[Line.Length - 2..] = " \")
					Line = Line[..Line.Length - 2].Trim()
					Line:+ File.ReadLine().Trim()
				Wend
			ElseIf Escape = ESCAPE_BLITZ And (Line[Line.Length - 3..] = " ..") Then
				While Not File.Eof() And (Line[Line.Length - 3..] = " ..")
					Line = Line[..Line.Length - 3].Trim()
					Line:+ " " + File.ReadLine().Trim()
				Wend
			EndIf

			If (Line[..1] = "[") And (Line[Line.Length-1..] = "]")
				Local key:String = Line[1..Line.Length - 1].Trim()
				Local id:Int = 0
				If key.Contains(",") Then
					id = key[..key.Find(",")].ToInt()
					key = key[key.Find(",") + 1..].Trim()
				Else
					If String(key.ToInt()) = key Then
						id = key.ToInt()
						key = ""
					EndIf
				EndIf

				If String[](Groups) Then
					LoadGroup = False
					For Local k:String = EachIn String[](Groups)
						If key = k Then LoadGroup = True
					Next
				ElseIf Int[](Groups) Then
					LoadGroup = False
					For Local i:Int = EachIn Int[](Groups)
						If id = i Then LoadGroup = True
					Next
				Else
					LoadGroup = True
				EndIf

				If LoadGroup Then
					CreateGroup key, id
				EndIf

			ElseIf LoadGroup Then
				Local value:String
				Local id:Int = 0, key:String = ""

				value = line

				If value.Contains("=") Then
					key = value[..value.Find("=")].Trim()
					value = value[value.Find("=") + 1..].Trim()
				EndIf
				If key.Contains(",") Then
					id = key[..key.Find(",")].ToInt()
					key = key[key.Find(",") + 1..].Trim()
				Else
					If String(key.ToInt()) = key Then
						id = key.ToInt()
						key = ""
					EndIf
				EndIf

				If key[..1] = "$" Then
					Select key[1..].ToLower()
						Case "encoding"
							If Encoding = 0 Then
								Select value.ToLower()
									Case "latin1"
										Encoding = ENCODING_LATIN1
									Case "utf8"
										Encoding = ENCODING_UTF8
									Default
										Encoding = - 1
								EndSelect
								If Encoding > 0 Then File = TTextStream.Create(File, Encoding)
							EndIf
						Case "escape"
							If Escape = 0 Then
								Select value.ToLower()
									Case "cstyle"
										Escape = ESCAPE_CSTYLE
									Case "blitz"
										Escape = ESCAPE_BLITZ
									Default
										Escape = - 1
								EndSelect
							EndIf
						Case "casesensitive"
							If CurrGroup.Key = "" Then
								CaseSensitive = value.ToInt()
							EndIf
							CurrGroup.SetCaseSensitive value.ToInt()
						Case "indexchar"
							If CurrGroup.Key = "" Then
								IndexChar = value.ToInt()
							EndIf
							CurrGroup.SetIndexChar value.ToInt()
					EndSelect
				Else
					Select Escape
						Case ESCAPE_CSTYLE ; value = CStyleUnescape(value)
						Case ESCAPE_BLITZ ; value = BlitzUnescape(value)
					EndSelect
					AddEntry value, id, key
				EndIf

			EndIf
		Wend
	EndMethod

	Method Count:Int()
		Local c:Int
		For Local gr:TStringTableGroup = EachIn Groups
			c:+ gr.Count()
		Next
		Return c
	EndMethod
EndType


Type TStringTableGroup
	Field ID:Int
	Field Key:String

	Field Content:TStringTableEntry[27]

	Field CaseSensitive:Int
	Field IndexChar:Int

	Method AddEntry(Value:String, ID:Int = 0, Key:String = "")
		If Key = "" Then Key = ValueToKey(Value, CaseSensitive)
		If Not CaseSensitive Then Key = Key.ToLower()

		Local Pos:Int
		If Key = "" Then
			Pos = 26
		Else
			Pos = Key[Min(IndexChar, Key.Length-1)]
			If (Pos >= $41) And (Pos <= $5A) Then ' A-Z
				Pos:- $41
			ElseIf (Pos >= $61) And (Pos <= $7A) Then ' a-z
				Pos:- $61
			Else ' everything else
				Pos = 26
			EndIf
		EndIf

		Local en:TStringTableEntry
		If Not en Then
			en = FindEntryByID(ID)
		EndIf
		If Not en Then
			en = FindEntryByKey(Key)
		EndIf
		If Not en Then
			If ID <= 0 Then ID = TStringTable.AutoID ; TStringTable.AutoID:+ 1
			en = New TStringTableEntry
			en.ID = ID
			en.Key = Key
			en.Group = Self.ID
			If Not Content[Pos] Then
				Content[Pos] = en
			Else
				Local en2:TStringTableEntry = Content[Pos]
				While en2.Succ
					en2 = en2.Succ
				Wend
				en2.Succ = en
			EndIf
			en.Succ = Null
		EndIf

		en.Value = Value
	EndMethod

	Method FindEntryByID:TStringTableEntry(ID:Int)
		If id = 0 Then Return Null
		For Local n:Int = 0 To 26
			Local en:TStringTableEntry = Content[n]
			While en
				If en.ID = ID Then Return en
				en = en.Succ
			Wend
		Next
	EndMethod

	Method FindEntryByKey:TStringTableEntry(Key:String)
		If Not CaseSensitive Then Key = Key.ToLower()
		If Key = "" Then Return Null

		Local Pos:Int
		If Key = "" Then
			Pos = 26
		Else
			Pos = Key[Min(IndexChar, Key.Length-1)]
			If (Pos >= $41) And (Pos <= $5A) Then ' A-Z
				Pos:- $41
			ElseIf (Pos >= $61) And (Pos <= $7A) Then ' a-z
				Pos:- $61
			Else ' everything else
				Pos = 26
			EndIf
		EndIf

		Local en:TStringTableEntry = Content[Pos]
		While en
			If en.Key = Key Then Return en
			en = en.Succ
		Wend
	EndMethod

	Method FindEntryByValue:TStringTableEntry(Value:String)
		For Local n:Int = 0 To 26
			Local en:TStringTableEntry = Content[n]
			While en
				If en.Value = Value Then Return en
				en = en.Succ
			Wend
		Next
	EndMethod



	Method SetCaseSensitive(CaseSensitive:Int)
		If Count() = 0 Then
			Self.CaseSensitive = CaseSensitive
		EndIf
	EndMethod

	Method SetIndexChar(IndexChar:Int)
		If Count() = 0 Then
			Self.IndexChar = IndexChar
		EndIf
	EndMethod

	Method Count:Int()
		Local c:Int = 0
		For Local n:Int = 0 To 26
			Local en:TStringTableEntry = Content[n]
			While en
				c:+ 1
				en = en.Succ
			Wend
		Next
		Return c
	EndMethod
EndType


Type TStringTableEntry
	Field ID:Int
	Field Key:String
	Field Value:String

	Field Group:Int

	Field Succ:TStringTableEntry

	Method ToString:String()
		Return "ID: " + ID + " Key: " + Key + " Group: " + Group + " Value: " + Value
	EndMethod
EndType

Private


Function BlitzEscape:String(Data:String, EscapeUnicode:Int = True)
	Local out:String

	For Local a:Int = 0 Until Data.length
		Local ch:Short = Data[a]

		Select ch
			Case 0 ; out :+ "~~0"
			Case 9 ; out :+ "~~t"
			Case 13 ; out :+ "~~r"
			Case 10 ; out :+ "~~n"
			Case 34 ; out :+ "~~q"
			Case 126 ; out :+ "~~~~"
			Default
				If ch < 32 Then
					out :+ "~~" + ch + "~~"
				ElseIf (ch > 255) And EscapeUnicode Then
					out :+ "~~" + ch + "~~"
				Else
					out :+ Chr$(ch)
				EndIf
		EndSelect
	Next

	Return out
EndFunction

Function BlitzUnescape:String(Data:String)
	Local out:String

	For Local a:Int = 0 Until Data.length
		Local ch:Short = Data[a]

		If ch = 126 Then
			Select Data[a..a+2]
				Case "~~0" ; out :+ "~0" ; a :+ 1
				Case "~~t" ; out :+ "~t" ; a :+ 1
				Case "~~r" ; out :+ "~r" ; a :+ 1
				Case "~~n" ; out :+ "~n" ; a :+ 1
				Case "~~q" ; out :+ "~q" ; a :+ 1
				Case "~~~~" ; out :+ "~~" ; a :+ 1
				Case "~~1", "~~2", "~~3", "~~4", "~~5", "~~6", "~~7", "~~8", "~~9"
					If Data.Find("~~", a) = -1 Then Continue
					out :+ Chr$(Data[a+1..Data.Find("~~", a)].ToLong() )
					a = Data.Find("~~", a)
				Default

					out :+ Chr$(ch)
			EndSelect
		Else
			out :+ Chr$(ch)
		EndIf
	Next

	Return out
EndFunction


Function CStyleEscape:String(Data:String, EscapeUnicode:Int = True)
	Local out:String

	For Local a:Int = 0 Until Data.length
		Local ch:Short = Data[a]

		Select ch
			Case 0 ; out :+ "\0"
			Case 9 ; out :+ "\t"
			Case 13 ; out :+ "\r"
			Case 10 ; out :+ "\n"
			Case 34 ; out :+ "\~q"
			Case 92 ; out :+ "\\"
			Default
				If ch < 32 Then
					out :+ "\u" + ShortHex(ch)
				ElseIf (ch > 255) And EscapeUnicode Then
					out :+ "\u" + ShortHex(ch)
				Else
					out :+ Chr$(ch)
				EndIf
		EndSelect
	Next

	Return out
EndFunction

Function CStyleUnescape:String(Data:String)
	Local out:String

	For Local a:Int = 0 Until Data.length
		Local ch:Short = Data[a]

		If ch = 92 Then
			Select Data[a..a+2]
				Case "\0" ; out :+ "~0" ; a :+ 1
				Case "\t" ; out :+ "~t" ; a :+ 1
				Case "\r" ; out :+ "~r" ; a :+ 1
				Case "\n" ; out :+ "~n" ; a :+ 1
				Case "\~q" ; out :+ "~q" ; a :+ 1
				Case "\\" ; out :+ "\" ; a :+ 1
				Case "\u"
					out :+ Chr$(FromShortHex(Data[a+2..a+6]))
					a:+5
				Default
					out :+ Chr$(ch)
			EndSelect
		Else
			out :+ Chr$(ch)
		EndIf
	Next

	Return out
EndFunction

Function ShortHex:String(val:Int)
	Local buf:Short[4]
	For Local k:Int = 3 To 0 Step - 1
		buf[k] = (val & 15)
		If buf[k] > 9 Then buf[k]:+ $41 - 10 Else buf[k]:+ $30
		val:Shr 4
	Next
	Return String.FromShorts( buf,4 )
End Function

Function FromShortHex:Int(val:String)
	Local ret:Int = 0
	For Local k:Int = 0 Until Min(val.length, 4)
		ret:Shl 4
		If (val[k] >= $30) And (val[k] <= $39) Then ' 0-9
			ret:+ (val[k] - $30)
		ElseIf (val[k] >= $61) And (val[k] <= $66) Then ' a-f
			ret:+ (val[k] - $61 + 10)
		ElseIf (val[k] >= $41) And (val[k] <= $46) Then ' A-F
			ret:+ (val[k] - $41 + 10)
		Else
			Return ret
		EndIf
	Next
	Return ret
End Function

Function ValueToKey:String(Value:String, CaseSensitive:Int = False)
	Local out:String, f:Int = True
	For Local n:Int = 0 Until Value.Length
		If (Value[n] >= $30) And (Value[n] <= $39) Then ' 0-9
			out:+ Chr$(Value[n])
			f = False
		ElseIf (Value[n] >= $61) And (Value[n] <= $7A) Then ' a-z
			out:+ Chr$(Value[n])
			f = False
		ElseIf (Value[n] >= $41) And (Value[n] <= $5A) Then ' A-Z
			If CaseSensitive Then
				out:+ Chr$(Value[n])
			Else
				out:+ Chr$(Value[n] + $20)
			EndIf
			f = False
		ElseIf "_@".Contains(Value[n..n+1]) Then
			out:+ Chr$(Value[n])
			f = False
		Else
			If Not f Then out:+ "_"
			f = True
		EndIf
	Next
	out = out.Replace("__", "_")
	If out[out.Length-1..] = "_" Then out = out[..out.Length-1]
	Return out
EndFunction